/* Waving Flag Articulation
   using OpenGL in Windows GUI
   by Dudley J. Benton
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include "..\\..\\OpenGL\\gl.h"
#include "..\\..\\utilities\\keys.h"
#include "flag.h"

/*******************************************************************************
  WINDOWS OBJECTS AND PROCEDURES - NOT RELATED TO OPENGL
*******************************************************************************/

DWORD BLACK =0x000000;
DWORD GRAY  =0x808080;
DWORD LTGRAY=0xC0C0C0;
DWORD WHITE =0xFFFFFF;

#define StockBrush(stock_color) ((HBRUSH)(stock_color+1))

HINSTANCE hInst;
HWND hMain;
HWND hAbou;
HWND hAnim;
HWND hCapt;
HWND hExit;
HWND hFast;
HWND hLogo;
HWND hPlot;
HWND hRoll;
HWND hSlow;
HWND hSpin;
HWND hStat;
HWND hWind;
HWND hWire;

BOOL auto_capture;
BOOL wind_capture;
BOOL animate=FALSE;
BOOL back=TRUE;
BOOL fill=TRUE;
int ticks=1;
int waits;
double flap;

BITMAPINFOHEADER*bLogo;
DWORD tLogo;

HDC   pDC; /* plot window device context */
HGLRC rDC; /* OpenGL rendering context */

struct{float a,b,c;}Camera={3,5,10};
struct{int a,b;}Light={0,0};
struct{int a,b,c;float s,x,y,z;}Objective={0,0,0,1,0,0,-0.5F};

void Abort(int code,char*format,...)
  {
  char bufr[1024],head[32];
  va_list arg_marker;
  KillTimer(hMain,1);
  va_start(arg_marker,format);
  vsprintf(bufr,format,arg_marker);
  sprintf(head,"Fatal Error Code %i",code);
  MessageBox(NULL,bufr,head,MB_ICONSTOP|MB_OK);
  _flushall();
  _fcloseall();
  ExitProcess(0);
  }

void WindowText(HWND hWnd,char*format,...)
  {
  char bufr[128];
  va_list arg_marker;
  va_start(arg_marker,format);
  if(vsprintf(bufr,format,arg_marker)>=sizeof(bufr))
    FatalAppExit(0,"WindowText(): buffer overflow");
  SetWindowText(hWnd,bufr);
  }

void CenterWindowOnScreen(HWND hWnd)
  {
  int high,left,top,wide;
  RECT wr;
  GetWindowRect(hWnd,&wr);
  wide=wr.right-wr.left;
  high=wr.bottom-wr.top;
  left=(GetSystemMetrics(SM_CXSCREEN)-wide)/2;
  top=(GetSystemMetrics(SM_CYSCREEN)-high)/2;
  MoveWindow(hWnd,left,top,wide,high,TRUE);
  }

void PaintBitmap(HDC hDC,RECT*rc,BITMAPINFOHEADER*bHead,int stretch)
  {
  int colors,h,H,l,t,w,W;
  BYTE*bImage;
  l=rc->left;
  t=rc->top;
  w=rc->right-rc->left;
  h=rc->bottom-rc->top;
  W=bHead->biWidth;
  H=bHead->biHeight;
  if(w<8||h<8||W<8||H<8)
    return;
  if(bHead->biBitCount==24)
    colors=0;
  else
    {
    colors=bHead->biClrUsed;
    if(colors==0)
      colors=1<<bHead->biBitCount;
    }
  bImage=((BYTE*)bHead)+sizeof(BITMAPINFOHEADER)+colors*sizeof(DWORD);
  if(stretch<0)
    SetDIBitsToDevice(hDC,l,t,W,H,0,0,0,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS);
  else if(stretch>0)
    StretchDIBits(hDC,l,t,w,h,0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  else if(w*abs(H)<abs(W)*h)
    StretchDIBits(hDC,l,t+(h-w*abs(H)/abs(W))/2,w,w*abs(H)/abs(W),0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  else
    StretchDIBits(hDC,l+(w-h*abs(W)/abs(H))/2,t,h*abs(W)/abs(H),h,0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  }

void CaptureScreen(BOOL main,char*fname)
  {
  int h,high,i,j,w,wide,width;
  BITMAP bm;
  BITMAPINFO bi;
  BITMAPFILEHEADER bf;
  BYTE a,b,*bits,g,r;
  FILE*fp;
  HBITMAP hBM,hOld;
  HDC cDC,wDC;
  RECT rw;
  if(main)
    {
    wDC=GetWindowDC(hMain);
    GetWindowRect(hMain,&rw);
    }
  else
    {
    wDC=GetWindowDC(hPlot);
    GetWindowRect(hPlot,&rw);
    }
  high=rw.bottom-rw.top;
  wide=rw.right-rw.left;
  if(wDC==NULL)
    Abort(__LINE__,"can't get window device context\nWindows error code %i",GetLastError());
  if(!(hBM=CreateCompatibleBitmap(wDC,wide,high)))
    Abort(__LINE__,"can't create compatible bitmap\nWindows error code %i",GetLastError());
  if(GetObject(hBM,sizeof(BITMAP),&bm)!=sizeof(BITMAP))
    Abort(__LINE__,"can't get bitmap\nWindows error code %i",GetLastError());
  if(!(bm.bmBitsPixel==24||bm.bmBitsPixel==32))
    Abort(__LINE__,"compatible bitmap not 24 or 32 bits/pixel (%i)",bm.bmBitsPixel);
  if(!(cDC=CreateCompatibleDC(wDC)))
    Abort(__LINE__,"can't create compatible device context\nWindows error code %i",GetLastError());
  if((hOld=SelectObject(cDC,hBM))==NULL)
    Abort(__LINE__,"can't select bitmap into device context\nWindows error code %i",GetLastError());
  if(!BitBlt(cDC,0,0,wide,high,wDC,0,0,SRCCOPY))
    Abort(__LINE__,"BitBlt failed\nWindows error code %i",GetLastError());
  if((bits=calloc(bm.bmWidthBytes*bm.bmHeight,sizeof(BYTE)))==NULL)
    Abort(__LINE__,"can't allocate memory for bitmap");
  memset(&bi,0,sizeof(BITMAPINFO));
  bi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  bi.bmiHeader.biWidth=wide;
  bi.bmiHeader.biHeight=high;
  bi.bmiHeader.biPlanes=1;
  bi.bmiHeader.biSizeImage=bm.bmWidthBytes*bm.bmHeight;
  bi.bmiHeader.biBitCount=bm.bmBitsPixel;
  if(GetDIBits(cDC,hBM,0,high,bits,&bi,DIB_RGB_COLORS)!=high)
    Abort(__LINE__,"can't get bitiap bits\nWindows error code %i",GetLastError());
  memset(&bf,0,sizeof(BITMAPFILEHEADER));
  bf.bfType=0x4D42;
  bf.bfOffBits=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);
  bf.bfSize=bf.bfOffBits+bi.bmiHeader.biSizeImage;
  if((fp=fopen(fname,"wb"))==NULL)
    Abort(__LINE__,"can't create capture file %s",fname);
  if(fwrite(&bf,sizeof(BITMAPFILEHEADER),1,fp)!=1)
    Abort(__LINE__,"can't write bitmap file header");
  width=wide*3;
  while(width%4)
    width++;
  bi.bmiHeader.biSizeImage=width*high;
  bi.bmiHeader.biBitCount=24;
  if(fwrite(&bi.bmiHeader,sizeof(BITMAPINFOHEADER),1,fp)!=1)
    Abort(__LINE__,"can't write bitmap info header");
  if(bm.bmBitsPixel==24)
    {
    if(fwrite(bits,sizeof(BYTE),width*high,fp)!=width*high)
      Abort(__LINE__,"can't write bitmap bits");
    }
  else
    {
    BYTE*row;
    if((row=calloc(width,sizeof(BYTE)))==NULL)
      Abort(__LINE__,"can't allocate memory for row");
    for(h=0;h<high;h++)
      {
      i=bm.bmWidthBytes*h;
      for(w=j=0;w<wide;w++)
        {
        b=bits[i++];
        g=bits[i++];
        r=bits[i++];
        a=bits[i++];
        row[j++]=b;
        row[j++]=g;
        row[j++]=r;
        }
      if(fwrite(row,sizeof(BYTE),width,fp)!=width)
        Abort(__LINE__,"can't write bitmap row");
      }
    free(row);
    }
  fclose(fp);
  free(bits);
  if(SelectObject(cDC,hOld)!=hBM)
    Abort(__LINE__,"can't deselect bitmap from device context\nWindows error code %i",GetLastError());
  DeleteObject(hBM);
  DeleteDC(cDC);
  if(main)
    ReleaseDC(hMain,wDC);
  else
    ReleaseDC(hPlot,wDC);
  }

LRESULT WINAPI LogoProc(HWND hWnd,UINT wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_DESTROY)
    return(FALSE);
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    RECT rc;
    if((hDC=BeginPaint(hWnd,&ps))==NULL)
      Abort(__LINE__,"can't get device context to begin painting\nerror code %i",GetLastError());
    GetClientRect(hWnd,&rc);
    if(bLogo)
      PaintBitmap(hDC,&rc,bLogo,TRUE);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

/*******************************************************************************
  OBJECTS AND PROCEDURES RELATED TO THE MESH
  NOT NECESSARILY OPENGL OR WINDOWS SPECIFIC
*******************************************************************************/

typedef struct{double u,v,x,y,z;}NODE;NODE*Node;
typedef struct{int i,j,k;}ELEM;ELEM*Elem;
int nodes,elems;

void CreateModel()
  {
  int e,n,nx=21,ny=21,x,y;
  nodes=nx*ny;
  if((Node=calloc(nodes,sizeof(NODE)))==NULL)
    Abort(__LINE__,"can't allocate memory for nodes");
  for(n=y=0;y<ny;y++)
    {
    for(x=0;x<nx;x++)
      {
      Node[n].u=x/(nx-1.);
      Node[n].v=y/(ny-1.);
      Node[n].x=-0.50+x*1.0/(nx-1);
      Node[n].y=-0.25+y*0.5/(ny-1);
      Node[n++].z=0.;
      }
    }
  elems=(nx-1)*(ny-1)*2;
  if((Elem=calloc(elems,sizeof(ELEM)))==NULL)
    Abort(__LINE__,"can't allocate memory for elements");
  for(e=y=0;y<ny-1;y++)
    {
    for(x=0;x<nx-1;x++,n++)
      {
      Elem[e].i=x  + y   *nx;
      Elem[e].j=x+1+ y   *nx;
      Elem[e].k=x+1+(y+1)*nx;
      e++;
      Elem[e].i=x  + y   *nx;
      Elem[e].j=x+1+(y+1)*nx;
      Elem[e].k=x  +(y+1)*nx;
      e++;
      }
    }
  }

/*******************************************************************************
  OPENGL OBJECTS AND PROCEDURES - NOT RELATED TO WINDOWS
*******************************************************************************/

WINGDIAPI void APIENTRY glGenTextures(GLsizei n,GLuint*textures);
WINGDIAPI void APIENTRY glBindTexture(GLenum target,GLuint texture);

#define GL_BGR_EXT 0x80E0

DWORD InitializeTexture(BITMAPINFOHEADER*bm)
  {
  BYTE*bitmap;
  int i;
  DWORD t;
  if(bm->biBitCount!=24)
    {
    WindowText(hStat,"texture depth must be 24 bits/pixel (%li)",bm->biBitCount);
    return(0);
    }
  if(bm->biWidth>1024)
    {
    WindowText(hStat,"texture width must not exceed 1024 (%li)",bm->biWidth);
    return(0);
    }
  for(i=2;i<=1024;i*=2)
    if(i==bm->biWidth)
      break;
  if(i>1024)
    {
    WindowText(hStat,"texture width must be 2^n (%li)",bm->biWidth);
    return(0);
    }
  if(bm->biHeight>1024)
    {
    WindowText(hStat,"texture height must not exceed 1024 (%li)",bm->biHeight);
    return(0);
    }
  for(i=2;i<=1024;i*=2)
    if(i==bm->biHeight)
      break;
  if(i>1024)
    {
    WindowText(hStat,"texture height must be 2^n (%li)",bm->biHeight);
    return(0);
    }
  bitmap=((BYTE*)bm)+sizeof(BITMAPINFOHEADER);
  t=0;
  glGenTextures(1,&t);
  if(t==0)
    {
    WindowText(hStat,"can't generate texture: OpenGL error code=0x%lX",glGetError());
    return(0);
    }
  glBindTexture(GL_TEXTURE_2D,t);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexImage2D(GL_TEXTURE_2D,0,3,bm->biWidth,bm->biHeight,0,GL_BGR_EXT,GL_UNSIGNED_BYTE,bitmap);
  glDisable(GL_TEXTURE_2D);
  return(t);
  }

float*floatColor(DWORD c)
  {
  static float f[4];
  union{BYTE b[4];DWORD d;}u;
  u.d=c;
  f[0]=(float)(u.b[2]/255.);
  f[1]=(float)(u.b[1]/255.);
  f[2]=(float)(u.b[0]/255.);
  return(f);
  }

void glColor(DWORD color)
  {
  glMaterialfv(back?GL_FRONT_AND_BACK:GL_FRONT,GL_AMBIENT_AND_DIFFUSE,floatColor(color));
  }

typedef struct{float x,y,z;}VECTOR;

VECTOR Normalize(VECTOR V)
  {
  static VECTOR N;
  float r=(float)sqrt(V.x*V.x+V.y*V.y+V.z*V.z);
  if(r>FLT_EPSILON)
    {
    N.x=V.x/r;
    N.y=V.y/r;
    N.z=V.z/r;
    }
  else
    {
    N.x=0;
    N.y=0;
    N.z=1;
    }
  return(N);
  }

VECTOR CrossProduct(VECTOR P,VECTOR Q)
  {
  static VECTOR R;
  R.x=P.y*Q.z-P.z*Q.y;
  R.y=P.z*Q.x-P.x*Q.z;
  R.z=P.x*Q.y-P.y*Q.x;
  return(R);
  }

void RenderSphere(double X,double Y,double Z,double R,int L,int M)
  {
  int i,j,l,m;
  double A,B,*cosA,*cosB,*sinA,*sinB,dL=L,dM=M;
  VECTOR*N,*V;
  if(R<=0.)
    return;
  if(L<3)
    return;
  if(M<3)
    return;
  cosA=calloc(L,sizeof(double));
  sinA=calloc(L,sizeof(double));
  for(l=0;l<L;l++)
    {
    A=l*M_PI/L-M_PI/2.;
    cosA[l]=cos(A);
    sinA[l]=sin(A);
    }
  cosB=calloc(M,sizeof(double));
  sinB=calloc(M,sizeof(double));
  for(m=0;m<M;m++)
    {
    B=m*2.*M_PI/M;
    cosB[m]=cos(B);
    sinB[m]=sin(B);
    }
  N=calloc(L*M,sizeof(VECTOR));
  V=calloc(L*M,sizeof(VECTOR));
  V[0].x=(float)X;
  V[0].z=(float)Z;
  V[0].y=(float)(Y-R);
  N[0].x=0.F;
  N[0].z=0.F;
  N[0].y=-1.F;
  V[1].x=(float)X;
  V[1].z=(float)Z;
  V[1].y=(float)(Y+R);
  N[1].x=0.F;
  N[1].z=0.F;
  N[1].y=1.F;
  for(l=1;l<L;l++)
    {
    for(m=0;m<M;m++)
      {
      i=M*l+m;
      N[i].x=(float)(cosA[l]*sinB[m]);
      N[i].z=(float)(cosA[l]*cosB[m]);
      N[i].y=(float) sinA[l];
      V[i].x=(float)(X+R*N[i].x);
      V[i].z=(float)(Z+R*N[i].z);
      V[i].y=(float)(Y+R*N[i].y);
      }
    }
  glBegin(GL_TRIANGLES);
  for(m=0;m<M;m++)
    {
    i=M+(m+1)%M;
    j=M+m;
    glNormal3d(N[0].x,N[0].y,N[0].z);
    glVertex3d(V[0].x,V[0].y,V[0].z);
    glNormal3d(N[i].x,N[i].y,N[i].z);
    glVertex3d(V[i].x,V[i].y,V[i].z);
    glNormal3d(N[j].x,N[j].y,N[j].z);
    glVertex3d(V[j].x,V[j].y,V[j].z);
    }
  glEnd();
  for(l=1;l<L-2;l++)
    {
    glBegin(GL_QUAD_STRIP);
    for(m=0;m<=M;m++)
      {
      i=M*(l+1)+m%M;
      j=M*l+m%M;
      glNormal3d(N[i].x,N[i].y,N[i].z);
      glVertex3d(V[i].x,V[i].y,V[i].z);
      glNormal3d(N[j].x,N[j].y,N[j].z);
      glVertex3d(V[j].x,V[j].y,V[j].z);
      }
    glEnd();
    }
  glBegin(GL_TRIANGLES);
  for(m=0;m<M;m++)
    {
    i=M*(L-2)+m;
    j=M*(L-2)+(m+1)%M;
    glNormal3d(N[1].x,N[1].y,N[1].z);
    glVertex3d(V[1].x,V[1].y,V[1].z);
    glNormal3d(N[i].x,N[i].y,N[i].z);
    glVertex3d(V[i].x,V[i].y,V[i].z);
    glNormal3d(N[j].x,N[j].y,N[j].z);
    glVertex3d(V[j].x,V[j].y,V[j].z);
    }
  glEnd();
  free(V);
  free(N);
  free(cosA);
  free(sinA);
  free(cosB);
  free(sinB);
  }

void RenderCylinder(double X,double Y,double Z,double R,double H,int M)
  {
  int m;
  double B,*cosB,*sinB;
  VECTOR N1,N2,V1,V2;
  if(R<=0.)
    return;
  if(H<=0.)
    return;
  if(M<3)
    return;
  cosB=calloc(M+1,sizeof(double));
  sinB=calloc(M+1,sizeof(double));
  for(m=0;m<=M;m++)
    {
    B=m*2.*M_PI/M;
    cosB[m]=cos(B);
    sinB[m]=sin(B);
    }
  glBegin(GL_QUAD_STRIP);
  N1.y=N2.y=0.F;
  V1.y=(float)(Y-H/2.);
  V2.y=(float)(Y+H/2.);
  for(m=0;m<=M;m++)
    {
    N1.x=N2.x=(float)cosB[m];
    N1.z=N2.z=(float)sinB[m];
    V1.x=V2.x=(float)(X+R*N1.x);
    V1.z=V2.z=(float)(Z+R*N1.z);
    glNormal3d(N1.x,N1.y,N1.z);
    glVertex3d(V1.x,V1.y,V1.z);
    glNormal3d(N2.x,N2.y,N2.z);
    glVertex3d(V2.x,V2.y,V2.z);
    }
  glEnd();
  free(cosB);
  free(sinB);
  }

void glRepaint()
  {
  int e,i,j,k;
  float p[4],x1,x2,x3,y1,y2,y3,z1,z2,z3;
  double h=1.3,r=0.02,s,x=-0.52,y=-0.4,z=0.;
  NODE n1,n2,n3;
  VECTOR N,P,Q;
  WindowText(hStat,"Camera=%.0hf,%.0hf,%.0hf, Light=%i,%i, Objective=%i,%i,%i,%hG,%hG,%hG, Zoom=%.3hf",
    Camera.a,Camera.b,Camera.c,Light.a,Light.b,Objective.a,Objective.b,Objective.c,Objective.x,Objective.y,Objective.z,Objective.s);
  if(!pDC)
    return;
  if(!tLogo)
    return;
  glClearDepth(1);
  glClearColor(0.F,0.F,1.F,0.F);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glFrontFace(GL_CCW);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_NORMALIZE);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,FALSE);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,FALSE);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,floatColor(0x000000));
  glLightfv(GL_LIGHT0,GL_AMBIENT ,floatColor(0x333333));
  glLightfv(GL_LIGHT0,GL_DIFFUSE ,floatColor(0xCCCCCC));
  glLightfv(GL_LIGHT0,GL_SPECULAR,floatColor(0x000000));
    p[0]=(float)(sin(Light.a*M_PI/180)*cos(Light.b*M_PI/180));
    p[1]=(float) sin(Light.b*M_PI/180);
    p[2]=(float)(cos(Light.a*M_PI/180)*cos(Light.b*M_PI/180));
    p[3]=0;/* infinite distance to light source */
  glLightfv(GL_LIGHT0,GL_POSITION,p);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT1,GL_AMBIENT ,floatColor(0xFFFFFF));
  glLightfv(GL_LIGHT1,GL_DIFFUSE ,floatColor(0x000000));
  glLightfv(GL_LIGHT1,GL_SPECULAR,floatColor(0x000000));
  glLightfv(GL_LIGHT1,GL_POSITION,p);
  glDisable(GL_LIGHT1);
  glEnable(GL_LIGHTING);
  glTranslatef(Objective.x,Objective.y,Objective.z);
  glRotatef((float)Objective.a,1.F,0.F,0.F);
  glRotatef((float)Objective.b,0.F,1.F,0.F);
  glRotatef((float)Objective.c,0.F,0.F,1.F);
  glScalef(Objective.s,Objective.s,Objective.s);
  glScalef(2.5F,2.5F,2.5F);
  glPolygonMode(back?GL_FRONT_AND_BACK:GL_FRONT,fill?GL_FILL:GL_LINE);
  glColor(0xFF0000);
  RenderSphere(x,y+h/2.+3.*r/2.,z,2.*r,20,20);
  RenderCylinder(x,y,z,r,h,20);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,tLogo);
  glBegin(GL_TRIANGLES);
  glColor(0xFFFFFF);
  s=sin(flap*M_PI/180.);
  for(e=0;e<elems;e++)
    {
    i=Elem[e].i;
    j=Elem[e].j;
    k=Elem[e].k;
    n1=Node[i];
    n2=Node[j];
    n3=Node[k];
    n1.z+=s*sin(3.5*(n1.x+0.5)*M_PI)/ 20.;
    n2.z+=s*sin(3.5*(n2.x+0.5)*M_PI)/ 20.;
    n3.z+=s*sin(3.5*(n3.x+0.5)*M_PI)/ 20.;
    n1.y+=s*sin(3.5*(n1.x+0.5)*M_PI)/100.;
    n2.y+=s*sin(3.5*(n2.x+0.5)*M_PI)/100.;
    n3.y+=s*sin(3.5*(n3.x+0.5)*M_PI)/100.;
    P.x=(float)(n2.x-n1.x);
    P.y=(float)(n2.y-n1.y);
    P.z=(float)(n2.z-n1.z);
    Q.x=(float)(n3.x-n1.x);
    Q.y=(float)(n3.y-n1.y);
    Q.z=(float)(n3.z-n1.z);
    N=Normalize(CrossProduct(P,Q));
    x1=(float)n1.x;
    y1=(float)n1.y;
    z1=(float)n1.z;
    x2=(float)n2.x;
    y2=(float)n2.y;
    z2=(float)n2.z;
    x3=(float)n3.x;
    y3=(float)n3.y;
    z3=(float)n3.z;
    glNormal3f(N.x,N.y,N.z);
    glTexCoord2d(n1.u,n1.v);
    glVertex3f(x1,y1,z1);
    glTexCoord2d(n2.u,n2.v);
    glVertex3f(x2,y2,z2);
    glTexCoord2d(n3.u,n3.v);
    glVertex3f(x3,y3,z3);
    }
  glEnd();
  glFinish();
  SwapBuffers(pDC);
  if(auto_capture)
    {
    char fname[14];
    static int frame;
    sprintf(fname,"flag%05i.BMP",frame++);
    CaptureScreen(wind_capture,fname);
    }
  }

void QuickSortIndices(int*M,int*I,int n)
  {
  int i,j,k,l,m;
  while(n>1)
    {
    k=I[0];
    I[0]=I[n/2];
    I[n/2]=k;
    i=0;
    j=n;
    while(1)
      {
      m=M[I[0]];
      while(j>0)
        {
        j--;
        if(M[I[j]]<=m)
          break;
        }
      while(1)
        {
        i++;
        if(i>=j)
          break;
        if(M[I[i]]>=m)
          break;
        }
      if(i>=j)
        break;
      k=I[i];
      I[i]=I[j];
      I[j]=k;
      }
    k=I[0];
    I[0]=I[j];
    I[j]=k;
    k=j;
    j++;
    l=n-j;
    if(k<l)
      {
      QuickSortIndices(M,I,k);
      I+=j;
      n=l;
      }
    else
      {
      QuickSortIndices(M,I+j,l);
      n=k;
      }
    }
  }

HDC GetBestPixelFormat(HWND hWnd)
  {
  int adjacent_color_bits,bits_per_pixel,color_planes,pf,pixel_formats;
  int*index,*score,pFS;
  HDC hDC;
  PIXELFORMATDESCRIPTOR*pFD;
  if(sizeof(PIXELFORMATDESCRIPTOR)!=40)
    Abort(__LINE__,"sizeof(PIXELFORMATDESCRIPTOR)<?>40");
  if((hDC=GetDC(hWnd))==NULL)
    Abort(__LINE__,"can't get device context\nerror code %i",GetLastError());
  color_planes=max(1,GetDeviceCaps(hDC,PLANES));
  adjacent_color_bits=GetDeviceCaps(hDC,BITSPIXEL);
  bits_per_pixel=color_planes*adjacent_color_bits;
  pixel_formats=DescribePixelFormat(hDC,1,0,NULL);
  if(pixel_formats<1)
    Abort(__LINE__,"device doesn't support pixel formats");
  pFD=calloc(pixel_formats+1,sizeof(PIXELFORMATDESCRIPTOR));
  index=calloc(pixel_formats+1,sizeof(int));
  score=calloc(pixel_formats+1,sizeof(int));
  for(pFS=1;pFS<=pixel_formats;pFS++)
    {
    index[pFS]=pFS;
    DescribePixelFormat(hDC,pFS,sizeof(PIXELFORMATDESCRIPTOR),pFD+pFS);
    if(!(pFD[pFS].dwFlags&PFD_SUPPORT_OPENGL))
      continue;
    if(!(pFD[pFS].dwFlags&PFD_DOUBLEBUFFER))
      continue;
    if(pFD[pFS].iPixelType!=PFD_TYPE_RGBA)
      continue;
    if(pFD[pFS].dwFlags&PFD_NEED_PALETTE)
      continue;
    if(pFD[pFS].dwFlags&PFD_NEED_SYSTEM_PALETTE)
      continue;
    if(pFD[pFS].cColorBits<8)
      continue;
    if(pFD[pFS].cDepthBits<16)
      continue;
    score[pFS]=pFD[pFS].cDepthBits/16;
    if(pFD[pFS].dwFlags&PFD_GENERIC_FORMAT)
      score[pFS]+=8;
    if(pFD[pFS].dwFlags&PFD_GENERIC_ACCELERATED)
      score[pFS]+=16;
    if(pFD[pFS].cColorBits==GetDeviceCaps(hDC,BITSPIXEL))
      score[pFS]+=32;
    }
  QuickSortIndices(score,index,pixel_formats+1);
  for(pf=pixel_formats;pf>=0;pf--)
    {
    pFS=index[pf];
    if(score[pFS])
      if(SetPixelFormat(hDC,pFS,pFD+pFS))
        break;
    }
  if(pf<0)
    Abort(__LINE__,"device doesn't support OpenGL features critical to this application");
  if(!SetPixelFormat(hDC,pFS,pFD+pFS))
    Abort(__LINE__,"can't set pixel format\nerror code %i",GetLastError());
  free(score);
  free(index);
  free(pFD);
  return(hDC);
  }

/*******************************************************************************
  MORE WINDOWS OBJECTS AND PROCEDURES - NOT RELATED TO OPENGL
*******************************************************************************/

void WINAPI Animate(HWND hWnd,WPARAM wParam,DWORD idEvent,int uTime)
  {
  if(!animate)
    return;
  if(waits)
    {
    waits--;
    return;
    }
  flap+=5.;
  if(flap>359.)
    flap=0.;
  waits=ticks;
  glRepaint();
  }

void PositionWindows()
  {
  int h,t,w,x,y;
  RECT rc;
  if(IsIconic(hMain))
    return;
  if(!hPlot)
    return;
  GetClientRect(hMain,&rc);
  t=GetSystemMetrics(SM_CYCAPTION);
  w=bLogo->biWidth;
  h=bLogo->biHeight;
  x=rc.right-1-w;
  y=1;
  MoveWindow(hLogo,x,y,w,h,TRUE);y+=h+2;
  MoveWindow(hSpin,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hRoll,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hFast,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hSlow,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hAnim,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWire,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hCapt,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWind,x,y,w,t,TRUE);y+=t+2;
  y=rc.bottom-t;
  y-=t+1;MoveWindow(hExit,x,y,w,t,TRUE);
  y-=t+1;MoveWindow(hAbou,x,y,w,t,TRUE);
  MoveWindow(hStat,0,rc.bottom-t,rc.right,t,TRUE);
  w=rc.right-w-2;
  h=rc.bottom-t;
  MoveWindow(hPlot,2,2,w-4,h-4,TRUE);
  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(w>h)
    {
    float a=(float)(w/(h+0.01));
    glOrtho(-a,a,-1,1,Camera.a,Camera.c);
    }
  else
    {
    float a=(float)(h/(w+0.01));
    glOrtho(-1,1,-a,a,Camera.a,Camera.c);
    }
  glTranslated(0.,0.,-Camera.b);
  }

LRESULT WINAPI AboutProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_COMMAND)
    {
    if(HIWORD(wParam)==BN_CLICKED||HIWORD(wParam)==BN_DOUBLECLICKED)
      return(EndDialog(hWnd,LOWORD(wParam)));
    else
      return(FALSE);
    }
  if(wMsg==WM_CTLCOLORDLG)
    return((DWORD)StockBrush(COLOR_BTNFACE));
  if(wMsg==WM_CTLCOLORSTATIC)
    {
    SetBkMode((HDC)wParam,TRANSPARENT);
    SetTextColor((HDC)wParam,BLACK);
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_DESTROY)
    return(TRUE);
  if(wMsg==WM_INITDIALOG)
    {
    CenterWindowOnScreen(hWnd);
    BringWindowToTop(hWnd);
    return(TRUE);
    }
  return(FALSE);
  }

int IsCommand(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_EXIT||wParam==PUSH_ALT_F4)
    {
    PostQuitMessage(0);
    return(TRUE);
    }
  if(wParam==PUSH_ABOUT||wParam==PUSH_F1)
    {
    DialogBox(hInst,"ABOUT",hMain,(DLGPROC)AboutProc);
    return(TRUE);
    }
  if(wParam==PUSH_ANIM)
    {
    animate=1-animate;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    return(TRUE);
    }
  if(wParam==PUSH_CAPT)
    {
    auto_capture=1-auto_capture;
    SendMessage(hCapt,BM_SETCHECK,auto_capture,0);
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_FAST)
    {
    ticks--;
    if(ticks<0)
      ticks=0;
    return(TRUE);
    }
  if(wParam==PUSH_ROLL)
    {
    int a,i;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    a=Objective.a;
    for(i=0;i<360;i+=5)
      {
      Objective.a=i;
      glRepaint();
      Sleep(25);
      }
    Objective.a=a;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SLOW)
    {
    ticks++;
    return(TRUE);
    }
  if(wParam==PUSH_SPIN)
    {
    int b,i;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    b=Objective.b;
    for(i=0;i<360;i+=5)
      {
      Objective.b=i;
      glRepaint();
      Sleep(25);
      }
    Objective.b=b;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_WIND)
    {
    wind_capture=1-wind_capture;
    SendMessage(hWind,BM_SETCHECK,!wind_capture,0);
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_WIRE)
    {
    fill=1-fill;
    SendMessage(hWire,BM_SETCHECK,!fill,0);
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI PlotProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_DESTROY)
    return(FALSE);
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    if(pDC)
      {
      if(hDC!=pDC)
        Abort(__LINE__,"OpenGL device context not preserved");
      glRepaint();
      }
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  if(wMsg==WM_SIZE)
    {
    PositionWindows();
    glRepaint();
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

int IsOrientation(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_ALT_UP)
    {
    Objective.y-=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_DOWN)
    {
    Objective.y+=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_LEFT)
    {
    Objective.x+=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_RIGHT)
    {
    Objective.x-=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_HOME)
    {
    Objective.z-=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_END)
    {
    Objective.z+=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_PGUP)
    {
    Objective.s*=1.025F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_PGDN)
    {
    Objective.s/=1.025F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_UP)
    {
    Objective.a-=5;
    if(Objective.a<0)
      Objective.a=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_DOWN)
    {
    Objective.a+=5;
    if(Objective.a>355)
      Objective.a=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_LEFT)
    {
    Objective.c+=5;
    if(Objective.c>355)
      Objective.c=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_RIGHT)
    {
    Objective.c-=5;
    if(Objective.c<0)
      Objective.c=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_HOME)
    {
    Objective.b+=5;
    if(Objective.b>355)
      Objective.b=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_END)
    {
    Objective.b-=5;
    if(Objective.b<0)
      Objective.b=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_UP)
    {
    Light.b+=5;
    if(Light.b>355)
      Light.b=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_DOWN)
    {
    Light.b-=5;
    if(Light.b<0)
      Light.b=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_LEFT)
    {
    Light.a-=5;
    if(Light.a<0)
      Light.a=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_RIGHT)
    {
    Light.a+=5;
    if(Light.a>355)
      Light.a=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI MainProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    {
    PostQuitMessage(0);
    return(FALSE);
    }
  if(wMsg==WM_COMMAND)
    {
    if(IsCommand(wParam&0xFFFF,lParam))
      {
      SetFocus(hMain);
      return(FALSE);
      }
    if(IsOrientation(wParam&0xFFFF,lParam))
      return(FALSE);
    return(FALSE);
    }
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_CTLCOLORBTN)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,LTGRAY);
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_CTLCOLOREDIT)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,WHITE);
    return((DWORD)GetStockObject(WHITE_BRUSH));
    }
  if(wMsg==WM_CTLCOLORSTATIC)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,LTGRAY);
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_GETMINMAXINFO)
    {
    MINMAXINFO*mInfo=(MINMAXINFO*)lParam;
    mInfo->ptMinTrackSize.x=560;
    mInfo->ptMinTrackSize.y=480;
    return(FALSE);
    }
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  if(wMsg==WM_SIZE)
    {
    if(wParam==SIZE_MAXIMIZED||wParam==SIZE_RESTORED)
      PositionWindows();
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

HWND CreatePushButton(DWORD ID,char*text)
  {
  static HWND hWnd;
  if((hWnd=CreateWindow("BUTTON",text,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|BS_PUSHBUTTON,
      0,0,0,0,hMain,(void*)ID,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create pushbutton \"%s\"\nerror code %i",text,GetLastError());
  SendMessage(hWnd,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  EnableWindow(hWnd,ID!=0&&text!=NULL);
  return(hWnd);
  }

HWND CreateRadioButton(DWORD ID,char*text,int check)
  {
  static HWND hWnd;
  if((hWnd=CreateWindow("BUTTON",text,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|BS_RADIOBUTTON,
      0,0,0,0,hMain,(void*)ID,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create radio button \"%s\"\nerror code %i",text,GetLastError());
  SendMessage(hWnd,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  SendMessage(hWnd,BM_SETCHECK,check,0);
  return(hWnd);
  }

void CreateWindows()
  {
  int h,high,w,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  w=2*wide/3;
  h=3*high/4;
  if((hMain=CreateWindow("MAIN","use directional keys + alt/ctl/shift to view",WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN|WS_VISIBLE,
      (wide-w)/2,(high-h)/2,w,h,NULL,LoadMenu(hInst,"MENU"),hInst,NULL))==NULL)
    Abort(__LINE__,"can't create main window\nerror code %i",GetLastError());
  if((hStat=CreateWindow(STATUSCLASSNAME,NULL,
      WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create status control\nerror code %i",GetLastError());
  SendMessage(hStat,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  if((hLogo=CreateWindow("LOGO",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_WHITERECT,
      0,0,0,0,hMain,NULL,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create logo window\nerror code %i",GetLastError());
  if((hPlot=CreateWindow("PLOT",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|WS_BORDER,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nerror code %i",GetLastError());
  hExit=CreatePushButton(PUSH_EXIT ,"Exit");
  hAbou=CreatePushButton(PUSH_ABOUT,"About");
  hFast=CreatePushButton(PUSH_FAST ,"Faster");
  hRoll=CreatePushButton(PUSH_ROLL ,"Roll");
  hSlow=CreatePushButton(PUSH_SLOW ,"Slower");
  hSpin=CreatePushButton(PUSH_SPIN ,"Spin");
  hAnim=CreateRadioButton(PUSH_ANIM,"Animate",animate);
  hCapt=CreateRadioButton(PUSH_CAPT,"auto capture",auto_capture);
  hWind=CreateRadioButton(PUSH_WIND,"window/all",wind_capture);
  hWire=CreateRadioButton(PUSH_WIRE,"Solid/Wire",!fill);
  }

void Register(char*cName,void*wProc,HBRUSH hBrush,char*dCursor,DWORD style)
  {
  WNDCLASS wc;
  memset(&wc,0,sizeof(WNDCLASS));
  wc.hInstance    =hInst;
  wc.hIcon        =LoadIcon(hInst,"ICON");
  wc.style        =CS_HREDRAW|CS_VREDRAW|style;
  wc.lpszClassName=cName;
  wc.lpfnWndProc  =(WNDPROC)wProc;
  wc.hbrBackground=hBrush;
  wc.hCursor      =LoadCursor(NULL,dCursor);
  if(!RegisterClass(&wc))
    Abort(__LINE__,"can't register %s class\nerror code %i",cName,GetLastError());
  }

void RegisterClasses()
  {
  Register("MAIN",MainProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0);
  Register("LOGO",LogoProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0);
  Register("PLOT",PlotProc,GetStockObject(NULL_BRUSH),IDC_CROSS,CS_OWNDC);
  }

void*GetResource(char*rname,char*type)
  {
  void*rLock;
  HGLOBAL rLoad;
  HRSRC rFind;
  if((rFind=FindResource(hInst,rname,type))==NULL)
    Abort(__LINE__,"can't find resource %s\nerror code %i",rname,GetLastError());
  if((rLoad=LoadResource(hInst,rFind))==NULL)
    Abort(__LINE__,"can't load resource %s\nerror code %i",rname,GetLastError());
  if((rLock=LockResource(rLoad))==NULL)
    Abort(__LINE__,"can't lock resource %s\nerror code %i",rname,GetLastError());
  return(rLock);
  }

BITMAPINFOHEADER*LoadBitmapResource(char*rname)
  {
  return((BITMAPINFOHEADER*)((BYTE*)GetResource(rname,RT_BITMAP)));
  }

void LoadBitmaps()
  {
  bLogo=LoadBitmapResource("LOGO");
  }

int WINAPI WinMain(HINSTANCE hCurrent,HINSTANCE hPrevious,char*lCommand,int nShow)
  {
  HACCEL acc;
  MSG msg;
  hInst=hCurrent;
  InitCommonControls();
  if((acc=LoadAccelerators(hInst,"FAST"))==NULL)
    Abort(__LINE__,"can't load accelerator table\nerror code %i",GetLastError());
  LoadBitmaps();
  CreateModel();
  RegisterClasses();
  CreateWindows();
  if((pDC=GetBestPixelFormat(hPlot))==0)
    Abort(__LINE__,"can't find best pixel context\nerror code %i",GetLastError());
  if((rDC=wglCreateContext(pDC))==0)
    Abort(__LINE__,"can't create OpenGL context\nerror code %i",GetLastError());
  if(!wglMakeCurrent(pDC,rDC))
    Abort(__LINE__,"can't make OpenGL context current\nerror code %i",GetLastError());
  tLogo=InitializeTexture(bLogo);
  PositionWindows();
  SetTimer(hMain,1,10,(TIMERPROC)Animate);
  while(GetMessage(&msg,NULL,0,0))
    if(!TranslateAccelerator(msg.hwnd,acc,&msg))
      if(!TranslateMessage(&msg))
        DispatchMessage(&msg);
  return(msg.wParam);
  }
