/* A Few Discrete Hairs
   using OpenGL in Windows GUI
   by Dudley J. Benton
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include "..\\..\\OpenGL\\gl.h"
#include "..\\..\\utilities\\keys.h"
#include "..\\..\\utilities\\colors.h"
#include "hair.h"

#define thick_lines    1
#define simple_tubes   2
#define complex_tubes  3
#define rendering      3
int delay_count      = 0;
int downswept_hairs  = 1;
int hairs_per_element=63;
int randomize_gravity= 1;
int randomize_lengths= 1;
int style_option     = 1;

/*******************************************************************************
  WINDOWS OBJECTS AND PROCEDURES - NOT RELATED TO OPENGL
*******************************************************************************/

#define StockBrush(stock_color) ((HBRUSH)(stock_color+1))

HINSTANCE hInst;
HWND hMain;
HWND hAbou;
HWND hAnim;
HWND hCapt;
HWND hDlay;
HWND hExit;
HWND hFast;
HWND hLogo;
HWND hPlot;
HWND hRoll;
HWND hSlow;
HWND hSpin;
HWND hStat;
HWND hWind;
HWND hWire;

BOOL animate         =FALSE;
BOOL auto_capture    =FALSE;
BOOL back            =TRUE;
BOOL busy            =FALSE;
#ifdef __TRACE__
BOOL delayed_painting=FALSE;
#else
BOOL delayed_painting=TRUE;
#endif
BOOL fill            =TRUE;
BOOL partial_paint   =TRUE;
BOOL pending_paint   =FALSE;
BOOL wind_capture    =FALSE;
int ticks=1;
int waits;
double flap;

BITMAPINFOHEADER*bLogo;

HDC   pDC; /* plot window device context */
HGLRC rDC; /* OpenGL rendering context */

struct{float a,b,c;}Camera={3,5,10};
struct{int a,b;}Light={0,0};
struct{int a,b,c;float s,x,y,z;}Objective={0,0,0,1.F,0.F,0.F,0.F};

void Abort(int code,char*format,...)
  {
  char bufr[1024],head[32];
  va_list arg_marker;
  KillTimer(hMain,1);
  va_start(arg_marker,format);
  vsprintf(bufr,format,arg_marker);
  sprintf(head,"Fatal Error Code %i",code);
  MessageBox(NULL,bufr,head,MB_ICONSTOP|MB_OK);
  _flushall();
  _fcloseall();
  ExitProcess(0);
  }

void WindowText(HWND hWnd,char*format,...)
  {
  char bufr[128];
  va_list arg_marker;
  va_start(arg_marker,format);
  if(vsprintf(bufr,format,arg_marker)>=sizeof(bufr))
    FatalAppExit(0,"WindowText(): buffer overflow");
  SetWindowText(hWnd,bufr);
  }

void CenterWindowOnScreen(HWND hWnd)
  {
  int high,left,top,wide;
  RECT wr;
  GetWindowRect(hWnd,&wr);
  wide=wr.right-wr.left;
  high=wr.bottom-wr.top;
  left=(GetSystemMetrics(SM_CXSCREEN)-wide)/2;
  top=(GetSystemMetrics(SM_CYSCREEN)-high)/2;
  MoveWindow(hWnd,left,top,wide,high,TRUE);
  }

void PaintBitmap(HDC hDC,RECT*rc,BITMAPINFOHEADER*bHead,int stretch)
  {
  int colors,h,H,l,t,w,W;
  BYTE*bImage;
  l=rc->left;
  t=rc->top;
  w=rc->right-rc->left;
  h=rc->bottom-rc->top;
  W=bHead->biWidth;
  H=bHead->biHeight;
  if(w<8||h<8||W<8||H<8)
    return;
  if(bHead->biBitCount==24)
    colors=0;
  else
    {
    colors=bHead->biClrUsed;
    if(colors==0)
      colors=1<<bHead->biBitCount;
    }
  bImage=((BYTE*)bHead)+sizeof(BITMAPINFOHEADER)+colors*sizeof(DWORD);
  if(stretch<0)
    SetDIBitsToDevice(hDC,l,t,W,H,0,0,0,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS);
  else if(stretch>0)
    StretchDIBits(hDC,l,t,w,h,0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  else if(w*abs(H)<abs(W)*h)
    StretchDIBits(hDC,l,t+(h-w*abs(H)/abs(W))/2,w,w*abs(H)/abs(W),0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  else
    StretchDIBits(hDC,l+(w-h*abs(W)/abs(H))/2,t,h*abs(W)/abs(H),h,0,0,W,H,bImage,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  }

void CaptureScreen(BOOL main,char*fname)
  {
  int h,high,i,j,w,wide,width;
  BITMAP bm;
  BITMAPINFO bi;
  BITMAPFILEHEADER bf;
  BYTE a,b,*bits,g,r;
  FILE*fp;
  HBITMAP hBM,hOld;
  HDC cDC,wDC;
  RECT rw;
  if(main)
    {
    wDC=GetWindowDC(hMain);
    GetWindowRect(hMain,&rw);
    }
  else
    {
    wDC=GetWindowDC(hPlot);
    GetWindowRect(hPlot,&rw);
    }
  high=rw.bottom-rw.top;
  wide=rw.right-rw.left;
  if(wDC==NULL)
    Abort(__LINE__,"can't get window device context\nWindows error code %i",GetLastError());
  if((hBM=CreateCompatibleBitmap(wDC,wide,high))==NULL)
    Abort(__LINE__,"can't create compatible bitmap\nWindows error code %i",GetLastError());
  if(GetObject(hBM,sizeof(BITMAP),&bm)!=sizeof(BITMAP))
    Abort(__LINE__,"can't get bitmap\nWindows error code %i",GetLastError());
  if(!(bm.bmBitsPixel==24||bm.bmBitsPixel==32))
    Abort(__LINE__,"compatible bitmap not 24 or 32 bits/pixel (%i)",bm.bmBitsPixel);
  if((cDC=CreateCompatibleDC(wDC))==NULL)
    Abort(__LINE__,"can't create compatible device context\nWindows error code %i",GetLastError());
  if((hOld=SelectObject(cDC,hBM))==NULL)
    Abort(__LINE__,"can't select bitmap into device context\nWindows error code %i",GetLastError());
  if(!BitBlt(cDC,0,0,wide,high,wDC,0,0,SRCCOPY))
    Abort(__LINE__,"BitBlt failed\nWindows error code %i",GetLastError());
  if((bits=calloc(bm.bmWidthBytes*bm.bmHeight,sizeof(BYTE)))==NULL)
    Abort(__LINE__,"can't allocate memory for bitmap");
  memset(&bi,0,sizeof(BITMAPINFO));
  bi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  bi.bmiHeader.biWidth=wide;
  bi.bmiHeader.biHeight=high;
  bi.bmiHeader.biPlanes=1;
  bi.bmiHeader.biSizeImage=bm.bmWidthBytes*bm.bmHeight;
  bi.bmiHeader.biBitCount=bm.bmBitsPixel;
  if(GetDIBits(cDC,hBM,0,high,bits,&bi,DIB_RGB_COLORS)!=high)
    Abort(__LINE__,"can't get bitiap bits\nWindows error code %i",GetLastError());
  memset(&bf,0,sizeof(BITMAPFILEHEADER));
  bf.bfType=0x4D42;
  bf.bfOffBits=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);
  bf.bfSize=bf.bfOffBits+bi.bmiHeader.biSizeImage;
  if((fp=fopen(fname,"wb"))==NULL)
    Abort(__LINE__,"can't create capture file %s",fname);
  if(fwrite(&bf,sizeof(BITMAPFILEHEADER),1,fp)!=1)
    Abort(__LINE__,"can't write bitmap file header");
  width=wide*3;
  while(width%4)
    width++;
  bi.bmiHeader.biSizeImage=width*high;
  bi.bmiHeader.biBitCount=24;
  if(fwrite(&bi.bmiHeader,sizeof(BITMAPINFOHEADER),1,fp)!=1)
    Abort(__LINE__,"can't write bitmap info header");
  if(bm.bmBitsPixel==24)
    {
    if(fwrite(bits,sizeof(BYTE),width*high,fp)!=width*high)
      Abort(__LINE__,"can't write bitmap bits");
    }
  else
    {
    BYTE*row;
    if((row=calloc(width,sizeof(BYTE)))==NULL)
      Abort(__LINE__,"can't allocate memory for row");
    for(h=0;h<high;h++)
      {
      i=bm.bmWidthBytes*h;
      for(w=j=0;w<wide;w++)
        {
        b=bits[i++];
        g=bits[i++];
        r=bits[i++];
        a=bits[i++];
        row[j++]=b;
        row[j++]=g;
        row[j++]=r;
        }
      if(fwrite(row,sizeof(BYTE),width,fp)!=width)
        Abort(__LINE__,"can't write bitmap row");
      }
    free(row);
    }
  fclose(fp);
  free(bits);
  if(SelectObject(cDC,hOld)!=hBM)
    Abort(__LINE__,"can't deselect bitmap from device context\nWindows error code %i",GetLastError());
  DeleteObject(hBM);
  DeleteDC(cDC);
  if(main)
    ReleaseDC(hMain,wDC);
  else
    ReleaseDC(hPlot,wDC);
  }

LRESULT WINAPI LogoProc(HWND hWnd,UINT wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_DESTROY)
    return(FALSE);
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    RECT rc;
    if((hDC=BeginPaint(hWnd,&ps))==NULL)
      Abort(__LINE__,"can't get device context to begin painting\nerror code %i",GetLastError());
    GetClientRect(hWnd,&rc);
    if(bLogo)
      PaintBitmap(hDC,&rc,bLogo,TRUE);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

/*******************************************************************************
  OBJECTS AND PROCEDURES RELATED TO THE MESH
  NOT NECESSARILY OPENGL OR WINDOWS SPECIFIC
*******************************************************************************/

#ifndef PFD_GENERIC_ACCELERATED
#define PFD_GENERIC_ACCELERATED 0x00001000
#endif

typedef struct{double x,y,z;}NODE;NODE*Node;
typedef struct{int i,j,k;DWORD rgb;}ELEM;ELEM*Elem;
int nodes,elems;

void ReadModel(char*fname)
  {
  char bufr[128],*ptr;
  int e,i,j,k,n;
  double S,X,Xc,Xm,Xx,Y,Yc,Ym,Yx,Z,Zc,Zm,Zx;
  FILE*fp;
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open model file: %s",fname);
  if(!fgets(bufr,sizeof(bufr),fp))
    Abort(__LINE__,"unexpected end of file");
  if(sscanf(bufr,"%i",&nodes)!=1)
    Abort(__LINE__,"scan error on number of nodes");
  if(nodes<100)
    Abort(__LINE__,"expected at least 100 nodes but found %i",nodes);
  if((Node=calloc(nodes,sizeof(NODE)))==NULL)
    Abort(__LINE__,"can't allocate memory for nodes");
  for(i=0;i<nodes;i++)
    {
    if(!fgets(bufr,sizeof(bufr),fp))
      Abort(__LINE__,"unexpected end of file");
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf",&X,&Y,&Z)!=3)
      Abort(__LINE__,"scan error on X,Y,Z");
    Node[i].x=X;
    Node[i].y=Y;
    Node[i].z=Z;
    if(i)
      {
      if(X<Xm)
        Xm=X;
      if(X>Xx)
        Xx=X;
      if(Y<Ym)
        Ym=Y;
      if(Y>Yx)
        Yx=Y;
      if(Z<Zm)
        Zm=Z;
      if(Z>Zx)
        Zx=Z;
      }
    else
      {
      Xm=Xx=X;
      Ym=Yx=Y;
      Zm=Zx=Z;
      }
    }
  S=sqrt((Xx-Xm)*(Xx-Xm)+(Yx-Ym)*(Yx-Ym)+(Zx-Zm)*(Zx-Zm));
  Xc=(Xm+Xx)/2.;
  Yc=(Ym+Yx)/2.;
  Zc=(Zm+Zx)/2.;
  for(i=0;i<nodes;i++)
    {
    Node[i].x=(Node[i].x-Xc)*2.685/S;
    Node[i].y=(Node[i].y-Yc)*2.685/S;
    Node[i].z=(Node[i].z-Zc)*2.685/S;
    }
  if(!fgets(bufr,sizeof(bufr),fp))
    Abort(__LINE__,"unexpected end of file");
  if(sscanf(bufr,"%i",&elems)!=1)
    Abort(__LINE__,"scan error on number of elements");
  if(elems<100)
    Abort(__LINE__,"expected at least 100 elements but found %i",elems);
  if((Elem=calloc(elems,sizeof(ELEM)))==NULL)
    Abort(__LINE__,"can't allocate memory for elements");
  for(e=0;e<elems;e++)
    {
    if(!fgets(bufr,sizeof(bufr),fp))
      Abort(__LINE__,"unexpected end of file");
    if(sscanf(bufr,"%i%*[ ,\t]%i%*[ ,\t]%i%n",&i,&j,&k,&n)!=3)
      Abort(__LINE__,"scan error on i,j,k");
    if(i<1||i>nodes)
      Abort(__LINE__,"no such node %i",i);
    if(j<1||j>nodes)
      Abort(__LINE__,"no such node %i",j);
    if(k<1||k>nodes)
      Abort(__LINE__,"no such node %i",k);
    Elem[e].i=i-1;
    Elem[e].j=j-1;
    Elem[e].k=k-1;
    ptr=bufr+n;
    while(strchr(" \t,",*ptr))
      ptr++;
    if(_memicmp(ptr,"rgb=",4)==0)
      ptr+=4;
    if(ptr[0]!='0'||ptr[1]!='x')
      Abort(__LINE__,"scan error on 0xRRGGBB following i,j,k");
    sscanf(ptr,"%i",&Elem[e].rgb);
    }
  fclose(fp);
  }

int randbetween(int lo,int hi)
  {
  return(lo+rand()%(hi-lo+1));
  }

double drand()
  {
  return(rand()/32767.);
  }

double xrand()
  {
  return((rand()-16484)/16384.);
  }

/*******************************************************************************
  OPENGL OBJECTS AND PROCEDURES - NOT RELATED TO WINDOWS
*******************************************************************************/

float*floatColor(DWORD c)
  {
  static float f[4];
  union{BYTE b[4];DWORD d;}u;
  u.d=c;
  f[0]=(float)(u.b[2]/255.);
  f[1]=(float)(u.b[1]/255.);
  f[2]=(float)(u.b[0]/255.);
  return(f);
  }

void glColor(DWORD color)
  {
  glMaterialfv(back?GL_FRONT_AND_BACK:GL_FRONT,GL_AMBIENT_AND_DIFFUSE,floatColor(color));
  }

typedef struct{float x,y,z;}VECTOR;

VECTOR Normalize(VECTOR V)
  {
  static VECTOR N;
  float r=(float)sqrt(V.x*V.x+V.y*V.y+V.z*V.z);
  if(r>FLT_EPSILON)
    {
    N.x=V.x/r;
    N.y=V.y/r;
    N.z=V.z/r;
    }
  else
    {
    N.x=0;
    N.y=0;
    N.z=1;
    }
  return(N);
  }

VECTOR CrossProduct(VECTOR P,VECTOR Q)
  {
  static VECTOR R;
  R.x=P.y*Q.z-P.z*Q.y;
  R.y=P.z*Q.x-P.x*Q.z;
  R.z=P.x*Q.y-P.y*Q.x;
  return(R);
  }

VECTOR vNormal(VECTOR V1,VECTOR V2,VECTOR V3)
  {
  double a1,a2,a3,b1,b2,b3,r;
  static VECTOR N;
  a1=V2.x-V1.x;
  a2=V2.y-V1.y;
  a3=V2.z-V1.z;
  b1=V3.x-V1.x;
  b2=V3.y-V1.y;
  b3=V3.z-V1.z;
  N.x=(float)(a2*b3-a3*b2);
  N.y=(float)(a3*b1-a1*b3);
  N.z=(float)(a1*b2-a2*b1);
  r=sqrt(N.x*N.x+N.y*N.y+N.z*N.z);
  if(r>DBL_EPSILON)
    {
    N.x/=(float)r;
    N.y/=(float)r;
    N.z/=(float)r;
    }
  else
    {
    N.x=0.F;
    N.y=0.F;
    N.z=1.F;
    }
  return(N);
  }

#if(rendering==thick_lines)

void RenderHair(double u,double v,double w,double ignore,double s,double x,double y,double z)
  {
  int i,n=11;
  double l,r;
  l=0.5*s;
  r=hypot(u,w)*l/2.;
  glLineWidth(5.F);
  glBegin(GL_LINE_STRIP);
  for(i=0;i<11;i++)
    glVertex3d(
      (float)(x+i*u*l/(n-1)),
      (float)(y+i*v*l/(n-1)+i*i*r/(n-1)/(n-1)),
      (float)(z+i*w*l/(n-1)));
  glEnd();
  }

#elif(rendering==simple_tubes)

typedef struct{double x,y,r;}XYR;

void RenderTube(XYR*xyr,int L,int M)
  {
  int i,j,l,m;
  double A,B,*cosA,*cosB,*sinA,*sinB;
  VECTOR*N,*V;
  if(L<3||M<3)
    return;
  cosA=calloc(L,sizeof(double));
  sinA=calloc(L,sizeof(double));
  for(l=0;l<L;l++)
    {
    i=max(0,l-1);
    j=min(L-1,l+1);
    A=atan2(xyr[i].x-xyr[j].x,xyr[j].y-xyr[i].y);
    cosA[l]=cos(A);
    sinA[l]=sin(A);
    }
  cosB=calloc(M,sizeof(double));
  sinB=calloc(M,sizeof(double));
  for(m=0;m<M;m++)
    {
    B=m*2.*M_PI/M;
    cosB[m]=cos(B);
    sinB[m]=sin(B);
    }
  N=calloc(L*M,sizeof(VECTOR));
  V=calloc(L*M,sizeof(VECTOR));
  for(i=l=0;l<L;l++)
    {
    for(m=0;m<M;m++,i++)
      {
      N[i].x=(float)(cosB[m]*cosA[l]);
      N[i].y=(float)(cosB[m]*sinA[l]);
      N[i].z=(float)(sinB[m]);
      V[i].x=(float)(xyr[l].x+xyr[l].r*N[i].x);
      V[i].y=(float)(xyr[l].y+xyr[l].r*N[i].y);
      V[i].z=(float)(     xyr[l].r*N[i].z);
      }
    }
  for(l=0;l<L-1;l++)
    {
    glBegin(GL_QUAD_STRIP);
    for(m=0;m<=M;m++)
      {
      i=M*l+m%M;
      j=M*(l+1)+m%M;
      glNormal3d(N[i].x,N[i].y,N[i].z);
      glVertex3d(V[i].x,V[i].y,V[i].z);
      glNormal3d(N[j].x,N[j].y,N[j].z);
      glVertex3d(V[j].x,V[j].y,V[j].z);
      }
    glEnd();
    }
  free(V);
  free(N);
  free(cosA);
  free(sinA);
  free(cosB);
  free(sinB);
  }

double arctan(double x,double y)
  {
  if(hypot(x,y)<FLT_EPSILON)
    return(0.);
  return(atan2(y,x)*180./M_PI);
  }

void RenderHair(double u,double v,double w,double ignore,double s,double x,double y,double z)
  {
  float a,c;
  static XYR xyr[]={
    {0.0000,0.00,0.01000},
    {0.0031,0.05,0.00794},
    {0.0122,0.10,0.00631},
    {0.0272,0.15,0.00501},
    {0.0477,0.20,0.00398},
    {0.0732,0.25,0.00316},
    {0.1031,0.30,0.00251},
    {0.1365,0.35,0.00200},
    {0.1727,0.40,0.00159},
    {0.2109,0.45,0.00126},
    {0.2500,0.50,0.00100}};
  glPushMatrix();
  glScaled(s,s,s);
  glTranslated(x,y,z);
  a=(float)arctan(v,w);
  c=(float)(-arctan(v,u));
  glRotatef(a,1.F,0.F,0.F);
  glRotatef(c,0.F,0.F,1.F);
  RenderTube(xyr,sizeof(xyr)/sizeof(XYR),16);
  glPopMatrix();
  }

#elif(rendering==complex_tubes)

typedef struct{float r,t,x,y,z;}RTXYZ;

void RenderComplexTube(RTXYZ*rtxyz,int n,int m)
  {
  int i,i1,i2,i3,j,k;
  float a,c,dxdt,dydt,dzdt,phi,s,theta;
  VECTOR p,u,v,*w;
  w=calloc(n*m,sizeof(VECTOR));
  for(k=i=0;i<n;i++)
    {
    if(i==0)
      {
      i1=i;
      i2=i+1;
      }
    else if(i==n-1)
      {
      i1=n-2;
      i2=n-1;
      }
    else
      {
      i1=i-1;
      i2=i+1;
      }
    dxdt=(rtxyz[i2].x-rtxyz[i1].x)/(rtxyz[i2].t-rtxyz[i1].t);
    dydt=(rtxyz[i2].y-rtxyz[i1].y)/(rtxyz[i2].t-rtxyz[i1].t);
    dzdt=(rtxyz[i2].z-rtxyz[i1].z)/(rtxyz[i2].t-rtxyz[i1].t);
    theta=(float)(-atan2(dydt,dxdt));
    phi=(float)atan2(hypot(dxdt,dydt),dzdt);
    u.x=(float)sin(theta);
    u.y=(float)cos(theta);
    u.z=0.F;
    if(fabs(cos(theta))>fabs(sin(theta)))
      v.z=(float)(dxdt/cos(theta));
    else
      v.z=(float)(-dydt/sin(theta));
    v.x=(float)(-dzdt*cos(theta));
    v.y=(float)(dzdt*sin(theta));
    a=(float)sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
    v.x/=a;
    v.y/=a;
    v.z/=a;
    for(j=0;j<m;j++,k++)
      {
      c=(float)cos(j*2.*M_PI/m);
      s=(float)sin(j*2.*M_PI/m);
      w[k].x=(float)(rtxyz[i].x+rtxyz[i].r*(s*u.x+c*v.x));
      w[k].y=(float)(rtxyz[i].y+rtxyz[i].r*(s*u.y+c*v.y));
      w[k].z=(float)(rtxyz[i].z+rtxyz[i].r*(s*u.z+c*v.z));
      }
    }
  for(i=0;i<n-1;i++)
    {
    glBegin(GL_QUAD_STRIP);
    for(j=0;j<=m;j++)
      {
      i1=m*i+(j%m);
      i2=m*(i+1)+(j%m);
      i3=m*(i+1)+((j+1)%m);
      u.x=w[i2].x-w[i1].x;
      u.y=w[i2].y-w[i1].y;
      u.z=w[i2].z-w[i1].z;
      v.x=w[i3].x-w[i1].x;
      v.y=w[i3].y-w[i1].y;
      v.z=w[i3].z-w[i1].z;
      p=Normalize(CrossProduct(u,v));
      glNormal3d(p.x,p.y,p.z);
      glVertex3f(w[i1].x,w[i1].y,w[i1].z);
      glVertex3f(w[i2].x,w[i2].y,w[i2].z);
      }
    glEnd();
    }
  free(w);
  }

NODE hoop[]={
  {-0.6116,-0.7, 0.5134},
  {-0.6720,-0.7, 0.3949},
  {-0.7111,-0.7, 0.2666},
  {-0.7360,-0.7, 0.1404},
  {-0.7422,-0.7,-0.0018},
  {-0.7320,-0.7,-0.1374},
  {-0.7088,-0.7,-0.2857},
  {-0.6752,-0.7,-0.3974},
  {-0.6051,-0.7,-0.5134},
  {-0.5134,-0.7,-0.6116},
  {-0.3992,-0.7,-0.6914},
  {-0.2730,-0.7,-0.7505},
  {-0.1389,-0.7,-0.7864},
  { 0.0000,-0.7,-0.7984},
  { 0.1390,-0.7,-0.7864},
  { 0.2731,-0.7,-0.7505},
  { 0.3993,-0.7,-0.6914},
  { 0.5134,-0.7,-0.6116},
  { 0.6051,-0.7,-0.5134},
  { 0.6752,-0.7,-0.3974},
  { 0.7089,-0.7,-0.2857},
  { 0.7321,-0.7,-0.1374},
  { 0.7423,-0.7,-0.0018},
  { 0.7361,-0.7, 0.1404},
  { 0.7112,-0.7, 0.2666},
  { 0.6720,-0.7, 0.3949},
  { 0.6116,-0.7, 0.5134}};

double sign(double x,double y)
  {
  if(y<0.)
    return(-fabs(x));
  return(fabs(x));
  }

double hypot3d(double x,double y,double z)
  {
  return(sqrt(x*x+y*y+z*z));
  }

double AreaTriangle(double x1,double y1,double z1,double x2,double y2,double z2,double x3,double y3,double z3)
  { /* Heron's formula */
  double a,b,c,r,s;
  a=hypot3d(x1-x3,y1-y3,z1-z3);
  b=hypot3d(x2-x1,y2-y1,z2-z1);
  c=hypot3d(x3-x2,y3-y2,z3-z2);
  s=(a+b+c)/2.;
  r=s*(s-a)*(s-b)*(s-c);
  if(r>0.)
    return(sqrt(r));
  return(0.);
  }

void RenderHair(double u,double v,double w,double r,double s,double x,double y,double z)
  {
  int i,j,m,n;
  double l,q,rm,rr,t,tt;
  RTXYZ rtxyz[11];
  VECTOR g,h;
  m=sizeof(hoop)/sizeof(NODE);
  n=sizeof(rtxyz)/sizeof(RTXYZ);
  if(style_option==2)
    {
    l=0.25*s;
    g.x=g.y=g.z=0.F;
    u+=xrand()/2.;
    v+=xrand()/2.;
    w+=xrand()/2.;
    }
  else if(downswept_hairs)
    {
    l=0.25*(1.+1.5*y)*s;
    if(x>0.)
      g.x=(float)(2.*(0.5-x*5./3.)*l);
    else
      g.x=(float)(-2.*(0.5+x*5./3.)*l);
    if(z>0.)
      g.x*=(float)(1.+5.*z);
    g.y=(float)(-3.*l);
    if(z>0.)
      g.z=(float)(-z/1.5);
    else
      g.z=(float)(-z/2.5);
    g.x=(float)(g.x*(1.+xrand()/5.));
    g.y=(float)(g.y*(1.+xrand()/5.));
    g.z=(float)(g.z*(1.+xrand()/5.));
    if(hypot(x,z)<0.1)
      {
      u+=xrand()/10.;
      v+=xrand()/10.;
      w+=xrand()/10.;
      }
    }
  else
    {
    if(randomize_lengths)
      l=0.5*(1.+0.25*drand())*s;
    else
      l=0.5*s;
    g.y=(float)(hypot(u,w)*l/2.);
    if(randomize_gravity)
      {
      g.x=(float)(g.y*xrand());
      g.z=(float)(g.y*xrand());
      }
    else
      g.x=g.z=0.F;
    }
  if(style_option==1)
    {
    for(j=i=0;i<m;i++)
      {
      rr=hypot3d(x-hoop[i].x,y-hoop[i].y,z-hoop[i].z);
      if(i==0)
        rm=rr;
      else if(rr<rm)
        {
        rm=rr;
        j=i;
        }
      }
    if(j>0)
      i=j-1;
    else
      i=1;
    q=drand();
    h.x=(float)(q*hoop[i].x+(1.-q)*hoop[j].x);
    h.y=(float)(q*hoop[i].y+(1.-q)*hoop[j].y);
    h.z=(float)(q*hoop[i].z+(1.-q)*hoop[j].z);
    h.x+=(float)(xrand()/50.);
    h.y+=(float)(xrand()/50.);
    h.z+=(float)(xrand()/50.);
    for(i=0;i<n;i++)
      {
      t=i/(n-1.);
      rtxyz[i].t=(float)t;
      rtxyz[i].r=(float)r;
      if(i<n/2)
        {
        rtxyz[i].x=(float)(x+i*u*l/(n-1));
        rtxyz[i].y=(float)(y+i*v*l/(n-1));
        rtxyz[i].z=(float)(z+i*w*l/(n-1));
        }
      else
        {
        q=(n-1-i)*2./n;
        rtxyz[i].x=(float)(q*rtxyz[n/2-1].x+(1.-q)*h.x);
        rtxyz[i].y=(float)(q*rtxyz[n/2-1].y+(1.-q)*h.y);
        rtxyz[i].z=(float)(q*rtxyz[n/2-1].z+(1.-q)*h.z);
        }
      r*=0.9;
      }
    for(i=1;i<n;i++)
      if(rtxyz[i].z>=0.4)
        if(rtxyz[i].y<=0.45)
          if(fabs(rtxyz[i].x)<0.5)
            rtxyz[i].x=(float)sign(0.5,rtxyz[i].x);
    for(j=0;j<3;j++)
      {
      for(i=1;i<n-1;i++)
        {
        rtxyz[i].x=(rtxyz[i-1].x+rtxyz[i+1].x)/2.F;
        rtxyz[i].y=(rtxyz[i-1].y+rtxyz[i+1].y)/2.F;
        rtxyz[i].z=(rtxyz[i-1].z+rtxyz[i+1].z)/2.F;
        }
      }
    }
  else
    {
    for(i=0;i<n;i++)
      {
      t=i/(n-1.);
      tt=pow(t,3.5);
      rtxyz[i].t=(float)t;
      rtxyz[i].r=(float)r;
      rtxyz[i].x=(float)(x+u*l*t+g.x*tt);
      rtxyz[i].y=(float)(y+v*l*t+g.y*tt);
      rtxyz[i].z=(float)(z+w*l*t+g.z*tt);
      r*=0.9;
      g.x*=0.95F;
      }
    }
  if(partial_paint)
    {
    glBegin(GL_LINE_STRIP);
    for(i=0;i<n;i++)
      glVertex3d(rtxyz[i].x,rtxyz[i].y,rtxyz[i].z);
    glEnd();
    }
  else
    RenderComplexTube(rtxyz,n,16);
  }

#else
#error unexpected rendering option
#endif

void RenderHead()
  {
  int e,i,j,k,l,n;
  float x1,x2,x3,xc,y1,y2,y3,yc,z1,z2,z3,zc;
  double area;
  NODE n1,n2,n3;
  VECTOR N,P,Q;
  for(e=0;e<elems;e++)
    {
    i=Elem[e].i;
    j=Elem[e].j;
    k=Elem[e].k;
    n1=Node[i];
    n2=Node[j];
    n3=Node[k];
    P.x=(float)(n2.x-n1.x);
    P.y=(float)(n2.y-n1.y);
    P.z=(float)(n2.z-n1.z);
    Q.x=(float)(n3.x-n1.x);
    Q.y=(float)(n3.y-n1.y);
    Q.z=(float)(n3.z-n1.z);
    N=Normalize(CrossProduct(P,Q));
    x1=(float)n1.x;
    y1=(float)n1.y;
    z1=(float)n1.z;
    x2=(float)n2.x;
    y2=(float)n2.y;
    z2=(float)n2.z;
    x3=(float)n3.x;
    y3=(float)n3.y;
    z3=(float)n3.z;
    if(Elem[e].rgb==brown)
      {
      if(style_option==1)
        glColor(dark_red);
      else if(style_option==2)
        glColor(black);
      else
        glColor(brown);
      glPushMatrix();
      if(hairs_per_element==1||partial_paint)
        {
        xc=(x1+x2+x3)/3.F;
        yc=(y1+y2+y3)/3.F;
        zc=(z1+z2+z3)/3.F;
        RenderHair(N.x,N.y,N.z,0.01,1.,xc,yc,zc);
        }
      else if(hairs_per_element==3)
        {
        xc=(3.F*x1+x2+x3)/5.F;
        yc=(3.F*y1+y2+y3)/5.F;
        zc=(3.F*z1+z2+z3)/5.F;
        RenderHair(N.x,N.y,N.z,0.005,1.,xc,yc,zc);
        xc=(x1+3.F*x2+x3)/5.F;
        yc=(y1+3.F*y2+y3)/5.F;
        zc=(z1+3.F*z2+z3)/5.F;
        RenderHair(N.x,N.y,N.z,0.005,1.,xc,yc,zc);
        xc=(x1+x2+3.F*x3)/5.F;
        yc=(y1+y2+3.F*y3)/5.F;
        zc=(z1+z2+3.F*z3)/5.F;
        RenderHair(N.x,N.y,N.z,0.005,1.,xc,yc,zc);
        }
      else if(hairs_per_element==7)
        {
        xc=(x1+x2+x3)/3.F;
        yc=(y1+y2+y3)/3.F;
        zc=(z1+z2+z3)/3.F;
        RenderHair(N.x,N.y,N.z,0.00333333,1.,xc,yc,zc);
        xc=(3.F*x1+x2+x3)/5.F;
        yc=(3.F*y1+y2+y3)/5.F;
        zc=(3.F*z1+z2+z3)/5.F;
        RenderHair(N.x,N.y,N.z,0.00333333,1.,xc,yc,zc);
        xc=(x1+3.F*x2+x3)/5.F;
        yc=(y1+3.F*y2+y3)/5.F;
        zc=(z1+3.F*z2+z3)/5.F;
        RenderHair(N.x,N.y,N.z,0.00333333,1.,xc,yc,zc);
        xc=(x1+x2+3.F*x3)/5.F;
        yc=(y1+y2+3.F*y3)/5.F;
        zc=(z1+z2+3.F*z3)/5.F;
        RenderHair(N.x,N.y,N.z,0.00333333,1.,xc,yc,zc);
        xc=(6.F*x1+x2+x3)/8.F;
        yc=(6.F*y1+y2+y3)/8.F;
        zc=(6.F*z1+z2+z3)/8.F;
        RenderHair(N.x,N.y,N.z,0.00333333,1.,xc,yc,zc);
        xc=(x1+6.F*x2+x3)/8.F;
        yc=(y1+6.F*y2+y3)/8.F;
        zc=(z1+6.F*z2+z3)/8.F;
        RenderHair(N.x,N.y,N.z,0.00333333,1.,xc,yc,zc);
        xc=(x1+x2+6.F*x3)/8.F;
        yc=(y1+y2+6.F*y3)/8.F;
        zc=(z1+z2+6.F*z3)/8.F;
        RenderHair(N.x,N.y,N.z,0.00333333,1.,xc,yc,zc);
        }
      else
        {
        area=AreaTriangle(n1.x,n1.y,n1.z,n2.x,n2.y,n2.z,n3.x,n3.y,n3.z);
        n=(int)(area*hairs_per_element/0.00643);
        if(n<1)
          n=1;
        for(l=0;l<n;l++)
          {
          i=rand();
          j=rand();
          k=rand();
          xc=(i*x1+j*x2+k*x3)/(i+j+k);
          yc=(i*y1+j*y2+k*y3)/(i+j+k);
          zc=(i*z1+j*z2+k*z3)/(i+j+k);
          RenderHair(N.x,N.y,N.z,0.00333333,1.,xc,yc,zc);
          }
        }
      glPopMatrix();
      if(style_option==1)
        glColor(0xFFDFBF);
      else if(style_option==2)
        glColor(brown);
      else
        glColor(pink);
      }
    else if(style_option==1&&Elem[e].rgb==pink)
      glColor(0xFFDFBF);
    else if(style_option==1&&Elem[e].rgb==green)
      glColor(blue);
    else if(style_option==2&&(Elem[e].rgb==pink||Elem[e].rgb==green))
      glColor(brown);
    else
      glColor(Elem[e].rgb);
    glBegin(GL_POLYGON);
    glNormal3f(N.x,N.y,N.z);
    glVertex3f(x1,y1,z1);
    glVertex3f(x2,y2,z2);
    glVertex3f(x3,y3,z3);
    glEnd();
    }
  }

void glRepaint(BOOL partial)
  {
  float p[4];
  WindowText(hStat,"Camera=%.0hf,%.0hf,%.0hf, Light=%i,%i, Objective=%i,%i,%i,%hG,%hG,%hG, Zoom=%.3hf",
    Camera.a,Camera.b,Camera.c,Light.a,Light.b,Objective.a,Objective.b,Objective.c,Objective.x,Objective.y,Objective.z,Objective.s);
  if(!pDC)
    return;
  if(busy)
    return;
  busy=TRUE;
  partial_paint=delayed_painting&&partial;
  glClearDepth(1);
  glClearColor(0.685F,0.735F,0.925F,0.0F);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glFrontFace(GL_CCW);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_NORMALIZE);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,FALSE);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,FALSE);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,floatColor(0x000000));
  glLightfv(GL_LIGHT0,GL_AMBIENT ,floatColor(0x333333));
  glLightfv(GL_LIGHT0,GL_DIFFUSE ,floatColor(0xCCCCCC));
  glLightfv(GL_LIGHT0,GL_SPECULAR,floatColor(0x000000));
    p[0]=(float)(sin(Light.a*M_PI/180)*cos(Light.b*M_PI/180));
    p[1]=(float) sin(Light.b*M_PI/180);
    p[2]=(float)(cos(Light.a*M_PI/180)*cos(Light.b*M_PI/180));
    p[3]=0;/* infinite distance to light source */
  glLightfv(GL_LIGHT0,GL_POSITION,p);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT1,GL_AMBIENT ,floatColor(0xFFFFFF));
  glLightfv(GL_LIGHT1,GL_DIFFUSE ,floatColor(0x000000));
  glLightfv(GL_LIGHT1,GL_SPECULAR,floatColor(0x000000));
  glLightfv(GL_LIGHT1,GL_POSITION,p);
  glDisable(GL_LIGHT1);
  glEnable(GL_LIGHTING);
  glTranslatef(Objective.x,Objective.y,Objective.z);
  glRotatef((float)Objective.a,1.F,0.F,0.F);
  glRotatef((float)Objective.b,0.F,1.F,0.F);
  glRotatef((float)Objective.c,0.F,0.F,1.F);
  glScalef(Objective.s,Objective.s,Objective.s);
  if(partial_paint)
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  else
    glPolygonMode(back?GL_FRONT_AND_BACK:GL_FRONT,fill?GL_FILL:GL_LINE);
  RenderHead();
  glFinish();
  SwapBuffers(pDC);
  if(auto_capture)
    {
    char fname[14];
    static int frame;
    sprintf(fname,"hair%05i.BMP",frame++);
    CaptureScreen(wind_capture,fname);
    }
  pending_paint=partial_paint;
  if(pending_paint)
    delay_count=0;
  busy=FALSE;
#ifdef __TRACE__
  PostQuitMessage(0);
#endif
  }

void WINAPI DelayedPaint(HWND hWnd,WPARAM wParam,DWORD idEvent,int uTime)
  {
  if(!delayed_painting)
    return;
  if(busy)
    return;
  if(!pending_paint)
    return;
  if(++delay_count>3)
    glRepaint(FALSE);
  }

void QuickSortIndices(int*M,int*I,int n)
  {
  int i,j,k,l,m;
  while(n>1)
    {
    k=I[0];
    I[0]=I[n/2];
    I[n/2]=k;
    i=0;
    j=n;
    while(1)
      {
      m=M[I[0]];
      while(j>0)
        {
        j--;
        if(M[I[j]]<=m)
          break;
        }
      while(1)
        {
        i++;
        if(i>=j)
          break;
        if(M[I[i]]>=m)
          break;
        }
      if(i>=j)
        break;
      k=I[i];
      I[i]=I[j];
      I[j]=k;
      }
    k=I[0];
    I[0]=I[j];
    I[j]=k;
    k=j;
    j++;
    l=n-j;
    if(k<l)
      {
      QuickSortIndices(M,I,k);
      I+=j;
      n=l;
      }
    else
      {
      QuickSortIndices(M,I+j,l);
      n=k;
      }
    }
  }

HDC GetBestPixelFormat(HWND hWnd)
  {
  int adjacent_color_bits,bits_per_pixel,color_planes,pf,pixel_formats;
  int*index,*score,pFS;
  HDC hDC;
  PIXELFORMATDESCRIPTOR*pFD;
  if(sizeof(PIXELFORMATDESCRIPTOR)!=40)
    Abort(__LINE__,"sizeof(PIXELFORMATDESCRIPTOR)<?>40");
  if((hDC=GetDC(hWnd))==NULL)
    Abort(__LINE__,"can't get device context\nerror code %i",GetLastError());
  color_planes=max(1,GetDeviceCaps(hDC,PLANES));
  adjacent_color_bits=GetDeviceCaps(hDC,BITSPIXEL);
  bits_per_pixel=color_planes*adjacent_color_bits;
  pixel_formats=DescribePixelFormat(hDC,1,0,NULL);
  if(pixel_formats<1)
    Abort(__LINE__,"device doesn't support pixel formats");
  pFD=calloc(pixel_formats+1,sizeof(PIXELFORMATDESCRIPTOR));
  index=calloc(pixel_formats+1,sizeof(int));
  score=calloc(pixel_formats+1,sizeof(int));
  for(pFS=1;pFS<=pixel_formats;pFS++)
    {
    index[pFS]=pFS;
    DescribePixelFormat(hDC,pFS,sizeof(PIXELFORMATDESCRIPTOR),pFD+pFS);
    if(!(pFD[pFS].dwFlags&PFD_SUPPORT_OPENGL))
      continue;
    if(!(pFD[pFS].dwFlags&PFD_DOUBLEBUFFER))
      continue;
    if(pFD[pFS].iPixelType!=PFD_TYPE_RGBA)
      continue;
    if(pFD[pFS].dwFlags&PFD_NEED_PALETTE)
      continue;
    if(pFD[pFS].dwFlags&PFD_NEED_SYSTEM_PALETTE)
      continue;
    if(pFD[pFS].cColorBits<8)
      continue;
    if(pFD[pFS].cDepthBits<16)
      continue;
    score[pFS]=pFD[pFS].cDepthBits/16;
    if(pFD[pFS].dwFlags&PFD_GENERIC_FORMAT)
      score[pFS]+=8;
    if(pFD[pFS].dwFlags&PFD_GENERIC_ACCELERATED)
      score[pFS]+=16;
    if(pFD[pFS].cColorBits==GetDeviceCaps(hDC,BITSPIXEL))
      score[pFS]+=32;
    }
  QuickSortIndices(score,index,pixel_formats+1);
  for(pf=pixel_formats;pf>=0;pf--)
    {
    pFS=index[pf];
    if(score[pFS])
      if(SetPixelFormat(hDC,pFS,pFD+pFS))
        break;
    }
  if(pf<0)
    Abort(__LINE__,"device doesn't support OpenGL features critical to this application");
  if(!SetPixelFormat(hDC,pFS,pFD+pFS))
    Abort(__LINE__,"can't set pixel format\nerror code %i",GetLastError());
  free(score);
  free(index);
  free(pFD);
  return(hDC);
  }

/*******************************************************************************
  MORE WINDOWS OBJECTS AND PROCEDURES - NOT RELATED TO OPENGL
*******************************************************************************/

void WINAPI Animate(HWND hWnd,WPARAM wParam,DWORD idEvent,int uTime)
  {
  if(!animate)
    return;
  if(waits)
    {
    waits--;
    return;
    }
  flap+=5.;
  if(flap>359.)
    flap=0.;
  waits=ticks;
  glRepaint(TRUE);
  }

void PositionWindows()
  {
  int h,t,w,x,y;
  RECT rc;
  if(IsIconic(hMain))
    return;
  if(!hPlot)
    return;
  GetClientRect(hMain,&rc);
  t=GetSystemMetrics(SM_CYCAPTION);
  w=bLogo->biWidth;
  h=bLogo->biHeight;
  x=rc.right-1-w;
  y=1;
  MoveWindow(hLogo,x,y,w,h,TRUE);y+=h+2;
  MoveWindow(hSpin,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hRoll,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hFast,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hSlow,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hDlay,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWire,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hAnim,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hCapt,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWind,x,y,w,t,TRUE);y+=t+2;
  y=rc.bottom-t;
  y-=t+1;MoveWindow(hExit,x,y,w,t,TRUE);
  y-=t+1;MoveWindow(hAbou,x,y,w,t,TRUE);
  MoveWindow(hStat,0,rc.bottom-t,rc.right,t,TRUE);
  w=rc.right-w-2;
  h=rc.bottom-t;
  MoveWindow(hPlot,2,2,w-4,h-4,TRUE);
  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(w>h)
    {
    float a=(float)(w/(h+0.01));
    glOrtho(-a,a,-1,1,Camera.a,Camera.c);
    }
  else
    {
    float a=(float)(h/(w+0.01));
    glOrtho(-1,1,-a,a,Camera.a,Camera.c);
    }
  glTranslated(0.,0.,-Camera.b);
  }

LRESULT WINAPI AboutProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_COMMAND)
    {
    if(HIWORD(wParam)==BN_CLICKED||HIWORD(wParam)==BN_DOUBLECLICKED)
      return(EndDialog(hWnd,LOWORD(wParam)));
    else
      return(FALSE);
    }
  if(wMsg==WM_CTLCOLORDLG)
    return((DWORD)StockBrush(COLOR_BTNFACE));
  if(wMsg==WM_CTLCOLORSTATIC)
    {
    SetBkMode((HDC)wParam,TRANSPARENT);
    SetTextColor((HDC)wParam,BLACK);
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_DESTROY)
    return(TRUE);
  if(wMsg==WM_INITDIALOG)
    {
    CenterWindowOnScreen(hWnd);
    BringWindowToTop(hWnd);
    return(TRUE);
    }
  return(FALSE);
  }

int IsCommand(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_EXIT||wParam==PUSH_ALT_F4)
    {
    PostQuitMessage(0);
    return(TRUE);
    }
  if(wParam==PUSH_ABOUT||wParam==PUSH_F1)
    {
    DialogBox(hInst,"ABOUT",hMain,(DLGPROC)AboutProc);
    return(TRUE);
    }
  if(wParam==PUSH_ANIM)
    {
    animate=1-animate;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    return(TRUE);
    }
  if(wParam==PUSH_CAPT)
    {
    auto_capture=1-auto_capture;
    SendMessage(hCapt,BM_SETCHECK,auto_capture,0);
    glRepaint(TRUE);
    return(TRUE);
    }
  if(wParam==PUSH_DELAY)
    {
    delayed_painting=1-delayed_painting;
    SendMessage(hDlay,BM_SETCHECK,delayed_painting,0);
    glRepaint(TRUE);
    return(TRUE);
    }
  if(wParam==PUSH_FAST)
    {
    ticks--;
    if(ticks<0)
      ticks=0;
    return(TRUE);
    }
  if(wParam==PUSH_ROLL)
    {
    int a,i;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    a=Objective.a;
    for(i=0;i<360;i+=5)
      {
      Objective.a=i;
      glRepaint(TRUE);
      Sleep(25);
      }
    Objective.a=a;
    glRepaint(TRUE);
    return(TRUE);
    }
  if(wParam==PUSH_SLOW)
    {
    ticks++;
    return(TRUE);
    }
  if(wParam==PUSH_SPIN)
    {
    int b,i;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    b=Objective.b;
    for(i=0;i<360;i+=5)
      {
      Objective.b=i;
      glRepaint(TRUE);
      Sleep(25);
      }
    Objective.b=b;
    glRepaint(TRUE);
    return(TRUE);
    }
  if(wParam==PUSH_WIND)
    {
    wind_capture=1-wind_capture;
    SendMessage(hWind,BM_SETCHECK,!wind_capture,0);
    glRepaint(TRUE);
    return(TRUE);
    }
  if(wParam==PUSH_WIRE)
    {
    fill=1-fill;
    SendMessage(hWire,BM_SETCHECK,!fill,0);
    glRepaint(TRUE);
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI PlotProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_DESTROY)
    return(FALSE);
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    if(pDC)
      {
      if(hDC!=pDC)
        Abort(__LINE__,"OpenGL device context not preserved");
      glRepaint(TRUE);
      }
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  if(wMsg==WM_SIZE)
    {
    PositionWindows();
    glRepaint(TRUE);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

int IsOrientation(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_ALT_UP)
    {
    Objective.y-=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_DOWN)
    {
    Objective.y+=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_LEFT)
    {
    Objective.x+=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_RIGHT)
    {
    Objective.x-=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_HOME)
    {
    Objective.z-=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_ALT_END)
    {
    Objective.z+=0.05F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_PGUP)
    {
    Objective.s*=1.025F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_PGDN)
    {
    Objective.s/=1.025F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_UP)
    {
    Objective.a-=5;
    if(Objective.a<0)
      Objective.a=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_DOWN)
    {
    Objective.a+=5;
    if(Objective.a>355)
      Objective.a=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_LEFT)
    {
    Objective.c+=5;
    if(Objective.c>355)
      Objective.c=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_RIGHT)
    {
    Objective.c-=5;
    if(Objective.c<0)
      Objective.c=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_HOME)
    {
    Objective.b+=5;
    if(Objective.b>355)
      Objective.b=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_END)
    {
    Objective.b-=5;
    if(Objective.b<0)
      Objective.b=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_UP)
    {
    Light.b+=5;
    if(Light.b>355)
      Light.b=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_DOWN)
    {
    Light.b-=5;
    if(Light.b<0)
      Light.b=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_LEFT)
    {
    Light.a-=5;
    if(Light.a<0)
      Light.a=355;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_RIGHT)
    {
    Light.a+=5;
    if(Light.a>355)
      Light.a=0;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  if(wParam==PUSH_CTL_Z)
    {
    Objective.a=Objective.b=Objective.c=0;
    Objective.s=1.F;
    Objective.x=Objective.y=Objective.y=0.F;
    InvalidateRect(hPlot,NULL,FALSE);
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI MainProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    {
    PostQuitMessage(0);
    return(FALSE);
    }
  if(wMsg==WM_COMMAND)
    {
    if(IsCommand(wParam&0xFFFF,lParam))
      {
      SetFocus(hMain);
      return(FALSE);
      }
    if(IsOrientation(wParam&0xFFFF,lParam))
      return(FALSE);
    return(FALSE);
    }
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_CTLCOLORBTN)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,LTGRAY);
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
//if(wMsg==WM_CTLCOLOREDIT)
//  {
//  SetTextColor((HDC)wParam,BLACK);
//  SetBkColor((HDC)wParam,WHITE);
//  return((DWORD)GetStockObject(WHITE_BRUSH));
//  }
//if(wMsg==WM_CTLCOLORSTATIC)
//  {
//  SetTextColor((HDC)wParam,BLACK);
//  SetBkColor((HDC)wParam,LTGRAY);
//  return((DWORD)StockBrush(COLOR_BTNFACE));
//  }
  if(wMsg==WM_GETMINMAXINFO)
    {
    MINMAXINFO*mInfo=(MINMAXINFO*)lParam;
    mInfo->ptMinTrackSize.x=560;
    mInfo->ptMinTrackSize.y=480;
    return(FALSE);
    }
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  if(wMsg==WM_SIZE)
    {
    if(wParam==SIZE_MAXIMIZED||wParam==SIZE_RESTORED)
      PositionWindows();
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

HWND CreatePushButton(DWORD ID,char*text)
  {
  static HWND hWnd;
  if((hWnd=CreateWindow("BUTTON",text,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|BS_PUSHBUTTON,
      0,0,0,0,hMain,(void*)ID,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create pushbutton \"%s\"\nerror code %i",text,GetLastError());
  SendMessage(hWnd,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  EnableWindow(hWnd,ID!=0&&text!=NULL);
  return(hWnd);
  }

HWND CreateRadioButton(DWORD ID,char*text,int check)
  {
  static HWND hWnd;
  if((hWnd=CreateWindow("BUTTON",text,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|BS_RADIOBUTTON,
      0,0,0,0,hMain,(void*)ID,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create radio button \"%s\"\nerror code %i",text,GetLastError());
  SendMessage(hWnd,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  SendMessage(hWnd,BM_SETCHECK,check,0);
  return(hWnd);
  }

void CreateWindows()
  {
  int h,high,w,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  w=2*wide/3;
  h=3*high/4;
  if((hMain=CreateWindow("MAIN","use directional keys + alt/ctl/shift to view",WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN|WS_VISIBLE,
      (wide-w)/2,(high-h)/2,w,h,NULL,LoadMenu(hInst,"MENU"),hInst,NULL))==NULL)
    Abort(__LINE__,"can't create main window\nerror code %i",GetLastError());
  if((hStat=CreateWindow(STATUSCLASSNAME,NULL,
      WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create status control\nerror code %i",GetLastError());
  SendMessage(hStat,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  if((hLogo=CreateWindow("LOGO",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_WHITERECT,
      0,0,0,0,hMain,NULL,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create logo window\nerror code %i",GetLastError());
  if((hPlot=CreateWindow("PLOT",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|WS_BORDER,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nerror code %i",GetLastError());
  hAbou=CreatePushButton(PUSH_ABOUT ,"About");
  hExit=CreatePushButton(PUSH_EXIT  ,"Exit");
  hFast=CreatePushButton(PUSH_FAST  ,"Faster");
  hRoll=CreatePushButton(PUSH_ROLL  ,"Roll");
  hSlow=CreatePushButton(PUSH_SLOW  ,"Slower");
  hSpin=CreatePushButton(PUSH_SPIN  ,"Spin");
  hAnim=CreateRadioButton(PUSH_ANIM ,"animate"         ,animate);
  hCapt=CreateRadioButton(PUSH_CAPT ,"auto capture"    ,auto_capture);
  hDlay=CreateRadioButton(PUSH_DELAY,"delayed painting",delayed_painting);
  hWind=CreateRadioButton(PUSH_WIND ,"window/all"      ,wind_capture);
  hWire=CreateRadioButton(PUSH_WIRE ,"solid/wire"      ,!fill);
  }

void Register(char*cName,void*wProc,HBRUSH hBrush,char*dCursor,DWORD style)
  {
  WNDCLASS wc;
  memset(&wc,0,sizeof(WNDCLASS));
  wc.hInstance    =hInst;
  wc.hIcon        =LoadIcon(hInst,"ICON");
  wc.style        =CS_HREDRAW|CS_VREDRAW|style;
  wc.lpszClassName=cName;
  wc.lpfnWndProc  =(WNDPROC)wProc;
  wc.hbrBackground=hBrush;
  wc.hCursor      =LoadCursor(NULL,dCursor);
  if(!RegisterClass(&wc))
    Abort(__LINE__,"can't register %s class\nerror code %i",cName,GetLastError());
  }

void RegisterClasses()
  {
  Register("MAIN",MainProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0);
  Register("LOGO",LogoProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0);
  Register("PLOT",PlotProc,GetStockObject(NULL_BRUSH),IDC_CROSS,CS_OWNDC);
  }

void*GetResource(char*rname,char*type)
  {
  void*rLock;
  HGLOBAL rLoad;
  HRSRC rFind;
  if((rFind=FindResource(hInst,rname,type))==NULL)
    Abort(__LINE__,"can't find resource %s\nerror code %i",rname,GetLastError());
  if((rLoad=LoadResource(hInst,rFind))==NULL)
    Abort(__LINE__,"can't load resource %s\nerror code %i",rname,GetLastError());
  if((rLock=LockResource(rLoad))==NULL)
    Abort(__LINE__,"can't lock resource %s\nerror code %i",rname,GetLastError());
  return(rLock);
  }

BITMAPINFOHEADER*LoadBitmapResource(char*rname)
  {
  return((BITMAPINFOHEADER*)((BYTE*)GetResource(rname,RT_BITMAP)));
  }

void LoadBitmaps()
  {
  bLogo=LoadBitmapResource("LOGO");
  }

HACCEL acc;
MSG msg;

#ifdef __TRACE__
int main(int argc,char**argv,char**envp)
  {
  hInst=GetModuleHandle(NULL);
#else
int WINAPI WinMain(HINSTANCE hCurrent,HINSTANCE hPrevious,char*lCommand,int nShow)
  {
  hInst=hCurrent;
#endif
  InitCommonControls();
  if((acc=LoadAccelerators(hInst,"FAST"))==NULL)
    Abort(__LINE__,"can't load accelerator table\nerror code %i",GetLastError());
  ReadModel("head.3dv");
  LoadBitmaps();
  RegisterClasses();
  CreateWindows();
  if((pDC=GetBestPixelFormat(hPlot))==0)
    Abort(__LINE__,"can't find best pixel context\nerror code %i",GetLastError());
  if((rDC=wglCreateContext(pDC))==0)
    Abort(__LINE__,"can't create OpenGL context\nerror code %i",GetLastError());
  if(!wglMakeCurrent(pDC,rDC))
    Abort(__LINE__,"can't make OpenGL context current\nerror code %i",GetLastError());
  PositionWindows();
  SetTimer(hMain,1, 50,(TIMERPROC)Animate);
  SetTimer(hMain,2,250,(TIMERPROC)DelayedPaint);
  while(GetMessage(&msg,NULL,0,0))
    if(!TranslateAccelerator(msg.hwnd,acc,&msg))
      if(!TranslateMessage(&msg))
        DispatchMessage(&msg);
  return(msg.wParam);
  }
