/* Blue Pony based on the program written by Brian Paul.
   Modified to function within a standard Windows GUI by Dudley J. Benton. */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <io.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include "..\\..\\OpenGL\\gl.h"
#include "..\\..\\utilities\\keys.h"
#include "..\\..\\utilities\\WinApp.c"
#include "..\\..\\utilities\\OpenGL.c"
#include "pony.h"

HWND hAbou;
HWND hAnim;
HWND hExit;
HWND hFast;
HWND hRoll;
HWND hSlow;
HWND hSpin;
HWND hStat;
HWND hWire;

int ticks=1;
int waits=0;
double zoom=1.;

float ambientLight0[ 4]={0.2F,0.2F,0.2F,0.0F};
float diffuseLight0[ 4]={0.3F,0.3F,0.3F,0.0F};
float positionLight0[4]={0.0F,0.0F,5.0F,0.0F};

struct{double a,b,c;}Camera={15,0,0};
struct{double a,b;}Light={0,45};
struct{double a,b,c,x,y,z;}Objective={345,130,0,0,0,0};

float*body,*mane,*fleg,*rleg;

void ReadElements(char*fname)
  {
  char bufr[128],*ptr;
  int b,e,f,i,j,k,l,m,n,p,r;
  float*x,*y,*z;
  double xi,yi,zi;
  FILE*fp;
  if((fp=fopen(fname,"rt"))==NULL)
    exit(1);
  if(!fgets(bufr,sizeof(bufr),fp))
    exit(1);
  if(sscanf(bufr,"%i",&n)!=1)
    exit(1);
  if(n<3)
    exit(1);
  for(i=0;i<n;i++)
    if(!fgets(bufr,sizeof(bufr),fp))
      exit(1);
  if(!fgets(bufr,sizeof(bufr),fp))
    exit(1);
  if(sscanf(bufr,"%i",&e)!=1)
    exit(1);
  if(e<4)
    exit(1);
  b=f=m=r=0;
  for(l=0;l<e;l++)
    {
    if(!fgets(bufr,sizeof(bufr),fp))
      exit(1);
    if(sscanf(bufr,"%i %i %i%n",&i,&j,&k,&p)!=3)
      exit(1);
    if(i<1||i>n||j<1||j>n||k<1||k>n)
      exit(1);
    ptr=bufr+p;
    while(*ptr==' ')
      ptr++;
    if(*ptr=='b')
      b++;
    else if(*ptr=='r')
      m++;
    else if(*ptr=='g')
      f++;
    else if(*ptr=='y')
      r++;
    else
      exit(1);
    }
  if(b<1||m<1||f<1||r<1)
    exit(1);
  if((x=calloc(n*3,sizeof(float)))==NULL)
    exit(1);
  y=x+n;
  z=y+n;
  if((body=calloc(9*b+1+9*m+1+9*f+1+9*r+1,sizeof(float)))==NULL)
    exit(1);
  mane=body+9*b+1;
  fleg=mane+9*m+1;
  rleg=fleg+9*f+1;
  body[9*b]=mane[9*m]=fleg[9*f]=rleg[9*r]=FLT_MAX;
  rewind(fp);
  if(!fgets(bufr,sizeof(bufr),fp))
    exit(1);
  for(i=0;i<n;i++)
    {
    if(!fgets(bufr,sizeof(bufr),fp))
      exit(1);
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf",&xi,&yi,&zi)!=3)
      exit(1);
    x[i]=(float)xi;
    y[i]=(float)yi;
    z[i]=(float)zi;
    }
  if(!fgets(bufr,sizeof(bufr),fp))
    exit(1);
  b=f=m=r=0;
  for(l=0;l<e;l++)
    {
    if(!fgets(bufr,sizeof(bufr),fp))
      exit(1);
    if(sscanf(bufr,"%i %i %i%n",&i,&j,&k,&p)!=3)
      exit(1);
    ptr=bufr+p;
    while(*ptr==' ')
      ptr++;
    if(*ptr=='b')
      {
      body[b++]=x[i-1];
      body[b++]=y[i-1];
      body[b++]=z[i-1];
      body[b++]=x[j-1];
      body[b++]=y[j-1];
      body[b++]=z[j-1];
      body[b++]=x[k-1];
      body[b++]=y[k-1];
      body[b++]=z[k-1];
      }
    else if(*ptr=='r')
      {
      mane[m++]=x[i-1];
      mane[m++]=y[i-1];
      mane[m++]=z[i-1];
      mane[m++]=x[j-1];
      mane[m++]=y[j-1];
      mane[m++]=z[j-1];
      mane[m++]=x[k-1];
      mane[m++]=y[k-1];
      mane[m++]=z[k-1];
      }
    else if(*ptr=='g')
      {
      fleg[f++]=x[i-1];
      fleg[f++]=y[i-1];
      fleg[f++]=z[i-1];
      fleg[f++]=x[j-1];
      fleg[f++]=y[j-1];
      fleg[f++]=z[j-1];
      fleg[f++]=x[k-1];
      fleg[f++]=y[k-1];
      fleg[f++]=z[k-1];
      }
    else if(*ptr=='y')
      {
      rleg[r++]=x[i-1];
      rleg[r++]=y[i-1];
      rleg[r++]=z[i-1];
      rleg[r++]=x[j-1];
      rleg[r++]=y[j-1];
      rleg[r++]=z[j-1];
      rleg[r++]=x[k-1];
      rleg[r++]=y[k-1];
      rleg[r++]=z[k-1];
      }
    else
      exit(1);
    }
  fclose(fp);
  free(x);
  }

float DeltaWalkAngle=0.225F;
float LegAngle=0.F;
float LegAngleStep=0.75F;
float LegDeltaAngle=0.5F;
float LegMaxAngle=15.F;
float Speed=2.F;
float WalkAngle=30.F;
float WalkRadius=4.F;
float Xrot=0.F;
float Yrot=30.F;
float FrontLegPos[3]={-0.36F,-0.324F,0.108F};
float BackLegPos[3]={0.684F,-0.324F,0.108F};

void RenderPony()
  {
  glColor(0.1F,0.1F,1.0F);
  glTriangles(body);
  glColor(1.0F,0.5F,0.5F);
  glTriangles(mane);
  glColor(0.1F,0.1F,1.0F);
  glPushMatrix();
  glTranslatef(FrontLegPos[0],FrontLegPos[1],FrontLegPos[2]);
  glRotatef(LegAngle,0.0,0.0,1.0);
  glTriangles(fleg);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(FrontLegPos[0],FrontLegPos[1],-FrontLegPos[2]);
  glRotatef(-LegAngle,0.0,0.0,1.0);
  glTriangles(fleg);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(BackLegPos[0],BackLegPos[1],BackLegPos[2]);
  glRotatef(-LegAngle,0.0,0.0,1.0);
  glTriangles(rleg);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(BackLegPos[0],BackLegPos[1],-BackLegPos[2]);
  glRotatef(LegAngle,0.0,0.0,1.0);
  glTriangles(rleg);
  glPopMatrix();
  }

void DrawGround(void)
  {
  int i,j,rows=20,columns=20;
  float sizeA=1.25F,sizeB=0.2F,x,z;
  float mat[2][4]={
    {0.0F,0.6F,0.0F,1.0F},
    {0.1F,0.8F,0.1F,1.0F}};
  glNormal3f(0.F,1.F,0.F);
  x=-(columns*(sizeA+sizeB))/4;
  for(i=0;i<=rows;i++)
    {
    float size=(i&1)?sizeA:sizeB;
    z=-(rows*(sizeA+sizeB))/4;
    glBegin(GL_QUAD_STRIP);
    for(j=0;j<=columns;j++)
      {
      glColor4fv(mat[(i+j)%2]);
      glVertex3f(x+size,0.F,z);
      glVertex3f(x,0.F,z);
      if(j&1)
        z+=sizeA;
      else
        z+=sizeB;
      }
    glEnd();
    x+=size;
    }
  }

void DrawBillboard(void)
  {
  glColor3f(1.F,0.F,0.F);
  glBegin(GL_POLYGON);
  glVertex2f(-1.0F,-0.5F);
  glVertex2f(1.0F,-0.5F);
  glVertex2f(1.0F,0.5F);
  glVertex2f(-1.0F,0.5F);
  glEnd();
  }

void RenderScene()
  {
  float xPos,zPos;
  glRotatef(Xrot,1.F,0.F,0.F);
  glRotatef(Yrot,0.F,1.F,0.F);
  glDisable(GL_LIGHTING);
  glPushMatrix();
  glTranslatef(0.0F,-1.6F,0.0F);
  DrawGround();
  glPopMatrix();
  glPushMatrix();
  glTranslatef(0.0F,-0.2F,0.0F);
  glScalef(2.5F,2.5F,2.5F);
  DrawBillboard();
  glPopMatrix();
  xPos=(float)(WalkRadius*cos(WalkAngle*M_PI/180.));
  zPos=(float)(WalkRadius*sin(WalkAngle*M_PI/180.));
  glEnable(GL_LIGHTING);
  glPushMatrix();
  glTranslatef(xPos,0.F,zPos);
  glRotatef(-WalkAngle+90.F,0.F,1.F,0.F);
  RenderPony();
  glPopMatrix();
  }

void glRepaint()
  {
  if(!pDC)
    return;
  glClearColor(0.F,0.F,0.F,0.F);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glShadeModel(GL_SMOOTH);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,TRUE);
  glDisable(GL_AUTO_NORMAL);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0,GL_AMBIENT ,ambientLight0);
  glLightfv(GL_LIGHT0,GL_DIFFUSE ,diffuseLight0);
    positionLight0[0]=(float)(sin(M_PI*Light.a/180)*cos(M_PI*Light.b/180));
    positionLight0[1]=(float)(sin(M_PI*Light.b/180));
    positionLight0[2]=(float)(cos(M_PI*Light.a/180)*cos(M_PI*Light.b/180));
    positionLight0[3]=0;/* infinite distance to light source */
  glLightfv(GL_LIGHT0,GL_POSITION,positionLight0);
  glEnable(GL_LIGHT0);
  glDisable(GL_TEXTURE_2D);
  glTranslated(Objective.x,Objective.y,Objective.z);
  glRotated(Objective.a,1,0,0);
  glRotated(Objective.b,0,1,0);
  glRotated(Objective.c,0,0,1);
  WindowText(hStat,"Camera=%lG,%lG,%lG, Light=%lG,%lG, Objective=%lG,%lG,%lG,%lG,%lG,%lG",
    Camera.a,Camera.b,Camera.c,Light.a,Light.b,Objective.a,Objective.b,
    Objective.c,Objective.x,Objective.y,Objective.z);
  glPolygonMode(back?GL_FRONT_AND_BACK:GL_FRONT,fill?GL_FILL:GL_LINE);
  glScaled(zoom/5.,zoom/5.,zoom/5.);
  RenderScene();
  glFinish();
  SwapBuffers(pDC);
  }

void PositionWindows()
  {
  int h,t,w,x,y;
  RECT rc;
  if(IsIconic(hMain))
    return;
  if(!hPlot)
    return;
  GetClientRect(hMain,&rc);
  t=GetSystemMetrics(SM_CYCAPTION);
  w=bLogo->biWidth;
  h=bLogo->biHeight;
  x=rc.right-1-w;
  y=1;
  MoveWindow(hLogo,x,y,w,h,TRUE);y+=h+2;
  MoveWindow(hSpin,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hRoll,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hFast,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hSlow,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hAnim,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWire,x,y,w,t,TRUE);
  y=rc.bottom-t;
  y-=t+1;MoveWindow(hExit,x,y,w,t,TRUE);
  y-=t+1;MoveWindow(hAbou,x,y,w,t,TRUE);
  MoveWindow(hStat,0,rc.bottom-t,rc.right,t,TRUE);
  w=rc.right-w-2;
  h=rc.bottom-t;
  MoveWindow(hPlot,2,2,w-4,h-4,TRUE);
  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(w>h)
    {
    float a=(float)(w/(h+0.01));
    glFrustum(-a,a,-1,1,Camera.a,Camera.c);
    }
  else
    {
    float a=(float)(h/(w+0.01));
    glFrustum(-1,1,-a,a,Camera.a,Camera.c);
    }
  glTranslated(0.,0.,-Camera.b);
  }

int IsCommand(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_EXIT||wParam==PUSH_ALT_F4)
    {
    PostQuitMessage(0);
    return(TRUE);
    }
  if(wParam==PUSH_ABOUT)
    {
    DialogBox(hInst,"ABOUT",hMain,(DLGPROC)AboutProc);
    return(TRUE);
    }
  if(wParam==PUSH_ANIM)
    {
    animate=1-animate;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    return(TRUE);
    }
  if(wParam==PUSH_FAST)
    {
    ticks--;
    if(ticks<0)
      ticks=0;
    return(TRUE);
    }
  if(wParam==PUSH_ROLL)
    {
    int i;
    double a;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    a=Objective.a;
    for(i=0;i<360;i+=5)
      {
      Objective.a=i;
      glRepaint();
      Sleep(25);
      }
    Objective.a=a;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SLOW)
    {
    ticks++;
    return(TRUE);
    }
  if(wParam==PUSH_SPIN)
    {
    int i;
    double b;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    b=Objective.b;
    for(i=0;i<360;i+=5)
      {
      Objective.b=i;
      glRepaint();
      Sleep(25);
      }
    Objective.b=b;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_WIRE)
    {
    fill=1-fill;
    SendMessage(hWire,BM_SETCHECK,!fill,0);
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

void ResetView()
  {
  Light.a= 0.;
  Light.b=45.;
  Objective.a=345.;
  Objective.b=130.;
  Objective.c=0.;
  Objective.x=0.;
  Objective.y=0.;
  Objective.z=0.;
  zoom=1.;
  }

int IsOrientation(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_ALT_UP)
    {
    Objective.y+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_RIGHT)
    {
    Objective.x+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_LEFT)
    {
    Objective.x-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_DOWN)
    {
    Objective.y-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_DOWN)
    {
    Objective.a-=5;
    if(Objective.a<0)
      Objective.a=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_RIGHT)
    {
    Objective.c+=5;
    if(Objective.c>359)
      Objective.c=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_LEFT)
    {
    Objective.c-=5;
    if(Objective.c<0)
      Objective.c=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_UP)
    {
    Objective.a+=5;
    if(Objective.a>359)
      Objective.a=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_DOWN)
    {
    Light.b-=5;
    if(Light.b<0)
      Light.b=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_LEFT)
    {
    Light.a-=5;
    if(Light.a<0)
      Light.a=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_RIGHT)
    {
    Light.a+=5;
    if(Light.a>359)
      Light.a=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_UP)
    {
    Light.b+=5;
    if(Light.b>359)
      Light.b=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_Z)
    {
    ResetView();
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_DOWN)
    {
    Objective.z-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_RIGHT)
    {
    Objective.b+=5;
    if(Objective.b>359)
      Objective.b=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_LEFT)
    {
    Objective.b-=5;
    if(Objective.b<0)
      Objective.b=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_UP)
    {
    Objective.z+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_PGDN)
    {
    zoom*=0.95;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_PGUP)
    {
    zoom/=0.95;
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI MainProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    {
    PostQuitMessage(0);
    return(FALSE);
    }
  if(wMsg==WM_COMMAND)
    {
    if(IsCommand(wParam&0xFFFF,lParam))
      {
      SetFocus(hMain);
      return(FALSE);
      }
    if(IsOrientation(wParam&0xFFFF,lParam))
      return(FALSE);
    return(FALSE);
    }
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_CTLCOLORBTN)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_GETMINMAXINFO)
    {
    MINMAXINFO*mInfo=(MINMAXINFO*)lParam;
    mInfo->ptMinTrackSize.x=640;
    mInfo->ptMinTrackSize.y=480;
    return(FALSE);
    }
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  if(wMsg==WM_SIZE)
    {
    PositionWindows();
    glRepaint();
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

void WINAPI Animate(HWND hWnd,WPARAM wParam,DWORD idEvent,int uTime)
  {
  if(!animate)
    return;
  if(waits)
    {
    waits--;
    return;
    }
  LegAngle+=LegDeltaAngle*Speed;
  if(LegAngle>LegMaxAngle)
    LegDeltaAngle=-LegAngleStep;
  else if(LegAngle<-LegMaxAngle)
    LegDeltaAngle=LegAngleStep;
  WalkAngle+=DeltaWalkAngle*Speed;
  waits=ticks;
  glRepaint();
  }

void CreateWindows()
  {
  int h,high,w,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  h=7*high/8;
  w=h+bLogo->biWidth-55;
  if((hMain=CreateWindow("MAIN","use directional keys + alt/ctl/shift to view",
      WS_OVERLAPPEDWINDOW|WS_VISIBLE,(wide-w)/2,(high-h)/2,w,h,
      NULL,LoadMenu(hInst,"MENU"),hInst,NULL))==NULL)
    Abort(__LINE__,"can't create main window\nWindows error code %i",GetLastError());
  if((hStat=CreateWindow(STATUSCLASSNAME,NULL,
      WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create status control\nerror code %i",GetLastError());
  SendMessage(hStat,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  if((hLogo=CreateWindow("LOGO",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWindows error code %i",GetLastError());
  if((hPlot=CreateWindow("PLOT",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_BLACKRECT,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWindows error code %i",GetLastError());
  hAbou=CreatePushButton(PUSH_ABOUT,"About");
  hExit=CreatePushButton(PUSH_EXIT ,"Exit");
  hFast=CreatePushButton(PUSH_FAST ,"Faster");
  hRoll=CreatePushButton(PUSH_ROLL ,"Roll");
  hSlow=CreatePushButton(PUSH_SLOW ,"Slower");
  hSpin=CreatePushButton(PUSH_SPIN ,"Spin");
  hAnim=CreateRadioButton(PUSH_ANIM,"Animate",animate);
  hWire=CreateRadioButton(PUSH_WIRE,"Solid/Wire",!fill);
  }

void LoadResources()
  {
  bLogo=GetBitmap("logo");
  }

void RegisterClasses()
  {
  Register("MAIN",MainProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0,0);
  Register("LOGO",LogoProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0,0);
  Register("PLOT",PlotProc,GetStockObject(NULL_BRUSH),IDC_CROSS,0,CS_OWNDC);
  }

int WINAPI WinMain(HINSTANCE hCurrent,HINSTANCE hPrevious,char*lCommand,int nShow)
  {
  HACCEL acc;
  MSG msg;
  hInst=hCurrent;
  InitCommonControls();
  acc=LoadAccelerators(hInst,"FAST");
  LoadResources();
  RegisterClasses();
  CreateWindows();
  PositionWindows();
  ReadElements("pony.3dv");
  OpenGL();
  SetTimer(hMain,1,10,(TIMERPROC)Animate);
  while(GetMessage(&msg,NULL,0,0))
    if(!TranslateAccelerator(msg.hwnd,acc,&msg))
      if(!TranslateMessage(&msg))
        DispatchMessage(&msg);
  return(0);
  }
