/* Rigid Articulation of a Human Skeleton
   (based on Thomas Baier's VRML model)
   using OpenGL in a Windows GUI
   by Dudley J. Benton
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <io.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include "..\\..\\OpenGL\\gl.h"
#include "..\\..\\utilities\\keys.h"
#include "..\\..\\utilities\\WinApp.c"
#include "..\\..\\utilities\\OpenGL.c"
#include "skeleton.h"

int group=-1;
enum{skull,mandible,cervneck,thorax,lclavicl,lscapula,lhumerus,lulna,lhand,
  rclavicl,rscapula,rhumerus,rulna,rhand,pelvis,saclumbr,lfemur,lpatella,
  lfibia,lfoot,ltoes,rfemur,rpatella,rfibia,rfoot,rtoes};

HWND hAbou;
HWND hAnim;
HWND hCapt;
HWND hColo;
HWND hExit;
HWND hFast;
HWND hRoll;
HWND hSlow;
HWND hSpin;
HWND hStat;
HWND hWind;
HWND hWire;

BOOL auto_capture;
BOOL wind_capture;
BOOL colors=TRUE;
int ticks=1;
int waits=0;
double zoom=1.;

float ambientLight0[ 4]={0.2F,0.2F,0.2F,0.0F};
float diffuseLight0[ 4]={0.3F,0.3F,0.3F,0.0F};
float positionLight0[4]={0.0F,0.0F,5.0F,0.0F};

struct{double a,b,c;}Camera={15.,0.,0.};
struct{double a,b;}Light={0.,45.};
struct{double a,b,c,x,y,z;}Objective={0.,0.,0.,0.,0.,-1.5};

float Mandible;
float lAnkle,rAnkle;
float lElbow,rElbow;
float lHip,rHip;
float lKnee,rKnee;
float lShoulder,rShoulder;
float lWristF,rWristF;
float lWristR,rWristR;
float Twalk;

typedef struct{float x,y,z;}XYZ;
typedef struct{XYZ c,n,p,q,r;}TRI;
struct{int*i,n;TRI*t;}model;

void ReadTriangles(char*fname)
  {
  char bufr[1024],*ptr;
  int l,m,n;
  BYTE b,g,r;
  DWORD rgb1,rgb2;
  VECTOR N,V1,V2,V3;
  TRI t;
  FILE*fp;
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open model %s",fname);
  m=0;
  while(fgets(bufr,sizeof(bufr),fp))
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%n",&V1.x,&V1.y,&V1.z,&V2.x,&V2.y,&V2.z,&V3.x,&V3.y,&V3.z,&l)==9)
      m++;
  if(m<3)
    Abort(__LINE__,"less than 3 elements found");
  if((model.i=calloc(rtoes+2,sizeof(int)))==NULL)
    Abort(__LINE__,"can't allocate memory");
  if((model.t=calloc(m,sizeof(TRI)))==NULL)
    Abort(__LINE__,"can't allocate memory");
  rewind(fp);
  model.n=n=0;
  rgb2=-1;
  while(fgets(bufr,sizeof(bufr),fp))
    {
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%n",&V1.x,&V1.y,&V1.z,&V2.x,&V2.y,&V2.z,&V3.x,&V3.y,&V3.z,&l)==9)
      {
      ptr=bufr+l;
      if(strchr(ptr,'=')==NULL)
        continue;
      while(*ptr!='=')
        ptr++;
      if(ptr[1]!='0')
        continue;
      if(ptr[2]!='x')
        continue;
      rgb1=rgb2;
      if(sscanf(ptr+1,"%i",&rgb2)!=1)
        Abort(__LINE__,"scan error on rgb=0xRRGGBB");
      N=vNormal(V1,V2,V3);
      t.p.x=(float)V1.x;
      t.p.y=(float)V1.y;
      t.p.z=(float)V1.z;
      t.q.x=(float)V2.x;
      t.q.y=(float)V2.y;
      t.q.z=(float)V2.z;
      t.r.x=(float)V3.x;
      t.r.y=(float)V3.y;
      t.r.z=(float)V3.z;
      t.n.x=(float)N.x;
      t.n.y=(float)N.y;
      t.n.z=(float)N.z;
      r=(BYTE)((rgb2&0xFF0000)>>16);
      g=(BYTE)((rgb2&0x00FF00)>> 8);
      b=(BYTE) (rgb2&0x0000FF);
      t.c.x=(float)(r/255.);
      t.c.y=(float)(g/255.);
      t.c.z=(float)(b/255.);
      if(rgb2!=rgb1)
        model.i[n++]=model.n;
      model.t[model.n++]=t;
      }
    }
  if(model.n!=m)
    Abort(__LINE__,"some elements were not loaded due to rgb=0xRRGGBB errors");
  if(n!=rtoes+1)
    Abort(__LINE__,"some groups were not added %i?%i",n,rtoes+1);
  model.i[n]=model.n;
  fclose(fp);
  }

double sina(double angle,double reference)
  {
  return(sin((angle-reference)*M_PI/180.)-sin(-reference*M_PI/180.));
  }

double cosa(double angle,double reference)
  {
  return(cos((angle-reference)*M_PI/180.)-cos(-reference*M_PI/180.));
  }

void RenderSkeleton()
  {
  int g,n;
  double A,R;
  float dX,dY,dZ;
  float bone[ 3]={0.965F,0.839F,0.647F};
  float green[3]={0.000F,1.000F,0.000F};
  float red[  3]={1.000F,0.000F,0.000F};
  TRI t;
  glColor(bone[0],bone[1],bone[2]);
  for(g=0;g<=rtoes;g++)
    {
    glPushMatrix();
    if(g==mandible)
      {
      R=33.6;
      A=90.;
      dY=(float)( R*sina(Mandible,A));
      dZ=(float)(-R*cosa(Mandible,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(Mandible,1.F,0.F,0.F);
      }
    if(g==lhumerus||g==lulna||g==lhand)
      {
      R=25.35;
      A=91.;
      dY=(float)( R*sina(lShoulder,A));
      dZ=(float)(-R*cosa(lShoulder,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(lShoulder,1.F,0.F,0.F);
      }
    if(g==rhumerus||g==rulna||g==rhand)
      {
      R=25.35;
      A=91.;
      dY=(float)( R*sina(rShoulder,A));
      dZ=(float)(-R*cosa(rShoulder,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(rShoulder,1.F,0.F,0.F);
      }
    if(g==lulna||g==lhand)
      {
      R=11.8;
      A=93.;
      dY=(float)( R*sina(lElbow,A));
      dZ=(float)(-R*cosa(lElbow,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(lElbow,1.F,0.F,0.F);
      }
    if(g==rulna||g==rhand)
      {
      R=11.8;
      A=93.;
      dY=(float)( R*sina(rElbow,A));
      dZ=(float)(-R*cosa(rElbow,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(rElbow,1.F,0.F,0.F);
      }
    if(g==lhand)
      {
      R=1.2;
      A=0.;
      dY=(float)(-R*sina(lWristF,A));
      dZ=(float)( R*cosa(lWristF,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(lWristF,1.F,0.F,0.F);
      R=10.5;
      A=85.;
      dX=(float)(R*sina(lWristR,A));
      dZ=(float)(R*cosa(lWristR,A));
      glTranslatef(dX,0.F,dZ);
      glRotatef(lWristR,0.F,1.F,0.F);
      }
    if(g==rhand)
      {
      R=1.2;
      A=0.;
      dY=(float)(-R*sina(rWristF,A));
      dZ=(float)( R*cosa(rWristF,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(rWristF,1.F,0.F,0.F);
      R=10.5;
      A=-85.;
      dX=(float)(R*sina(-rWristR,A));
      dZ=(float)(R*cosa(-rWristR,A));
      glTranslatef(dX,0.F,dZ);
      glRotatef(-rWristR,0.F,1.F,0.F);
      }
    if(g==lfemur||g==lpatella||g==lfibia||g==lfoot||g==ltoes)
      {
      R=3.6;
      A=120.;
      dY=(float)( R*sina(lHip,A));
      dZ=(float)(-R*cosa(lHip,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(lHip,1.F,0.F,0.F);
      }
    if(g==rfemur||g==rpatella||g==rfibia||g==rfoot||g==rtoes)
      {
      R=3.6;
      A=120.;
      dY=(float)( R*sina(rHip,A));
      dZ=(float)(-R*cosa(rHip,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(rHip,1.F,0.F,0.F);
      }
    if(g==lfibia||g==lfoot||g==ltoes)
      {
      R=20.;
      A=100.;
      dY=(float)(-R*sina(-lKnee,A));
      dZ=(float)( R*cosa(-lKnee,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(-lKnee,1.F,0.F,0.F);
      }
    if(g==rfibia||g==rfoot||g==rtoes)
      {
      R=20.;
      A=100.;
      dY=(float)(-R*sina(-rKnee,A));
      dZ=(float)( R*cosa(-rKnee,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(-rKnee,1.F,0.F,0.F);
      }
    if(g==lfoot||g==ltoes)
      {
      R=38.;
      A=97.;
      dY=(float)(-R*sina(lAnkle,A));
      dZ=(float)( R*cosa(lAnkle,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(lAnkle,1.F,0.F,0.F);
      }
    if(g==rfoot||g==rtoes)
      {
      R=38.;
      A=97.;
      dY=(float)(-R*sina(rAnkle,A));
      dZ=(float)( R*cosa(rAnkle,A));
      glTranslatef(0.F,dY,dZ);
      glRotatef(rAnkle,1.F,0.F,0.F);
      }
    for(n=model.i[g];n<model.i[g+1];n++)
      {
      t=model.t[n];
      if(colors)
        {
        if(group<0)
          glColor(t.c.x,t.c.y,t.c.z);
        else if(g==group)
          glColor(red[0],red[1],red[2]);
        else
          glColor(green[0],green[1],green[2]);
        }
      glBegin(GL_POLYGON);
      glNormal3f(t.n.x,t.n.y,t.n.z);
      glVertex3f(t.p.x,t.p.y,t.p.z);
      glVertex3f(t.q.x,t.q.y,t.q.z);
      glVertex3f(t.r.x,t.r.y,t.r.z);
      glEnd();
      }
    glPopMatrix();
    }
  }

void glRepaint()
  {
  char fname[FILENAME_MAX];
  static int frame;
  if(!pDC)
    return;
  glClearColor(0.F,0.F,0.F,0.F);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glShadeModel(GL_SMOOTH);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,TRUE);
  glDisable(GL_AUTO_NORMAL);
  glEnable(GL_NORMALIZE);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0,GL_AMBIENT ,ambientLight0);
  glLightfv(GL_LIGHT0,GL_DIFFUSE ,diffuseLight0);
    positionLight0[0]=(float)(sin(M_PI*Light.a/180)*cos(M_PI*Light.b/180));
    positionLight0[1]=(float)(sin(M_PI*Light.b/180));
    positionLight0[2]=(float)(cos(M_PI*Light.a/180)*cos(M_PI*Light.b/180));
    positionLight0[3]=0;/* infinite distance to light source */
  glLightfv(GL_LIGHT0,GL_POSITION,positionLight0);
  glEnable(GL_LIGHT0);
  glDisable(GL_TEXTURE_2D);
  glTranslated(Objective.x,Objective.y,Objective.z);
  glRotated(Objective.a,1,0,0);
  glRotated(Objective.b,0,1,0);
  glRotated(Objective.c,0,0,1);
  WindowText(hStat,"Camera=%lG,%lG,%lG, Light=%lG,%lG, Objective=%lG,%lG,%lG,%lG,%lG,%lG, Rot=%.0hf,%.0hf,%.0hf,%.0hf,%.0hf,%.0hf,%.0hf,%.0hf",
    Camera.a,Camera.b,Camera.c,Light.a,Light.b,Objective.a,Objective.b,Objective.c,Objective.x,Objective.y,Objective.z,Mandible,lShoulder,lElbow,lWristF,lWristR,lHip,lKnee,lAnkle);
  glPolygonMode(back?GL_FRONT_AND_BACK:GL_FRONT,fill?GL_FILL:GL_LINE);
  glScaled(zoom/42.,zoom/42.,zoom/42.);
  RenderSkeleton();
  glFinish();
  SwapBuffers(pDC);
  if(auto_capture)
    {
    sprintf(fname,"SKELETON%05i.BMP",frame++);
    CaptureScreen(wind_capture,fname);
    }
  }

void PositionWindows()
  {
  int h,t,w,x,y;
  RECT rc;
  if(IsIconic(hMain))
    return;
  if(!hPlot)
    return;
  GetClientRect(hMain,&rc);
  t=GetSystemMetrics(SM_CYCAPTION);
  w=bLogo->biWidth;
  h=bLogo->biHeight;
  x=rc.right-1-w;
  y=1;
  MoveWindow(hLogo,x,y,w,h,TRUE);y+=h+2;
  MoveWindow(hSpin,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hRoll,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hFast,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hSlow,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hAnim,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWire,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hCapt,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWind,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hColo,x,y,w,t,TRUE);
  y=rc.bottom-t;
  y-=t+1;MoveWindow(hExit,x,y,w,t,TRUE);
  y-=t+1;MoveWindow(hAbou,x,y,w,t,TRUE);
  MoveWindow(hStat,0,rc.bottom-t,rc.right,t,TRUE);
  w=rc.right-w-2;
  h=rc.bottom-t;
  MoveWindow(hPlot,2,2,w-4,h-4,TRUE);
  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(w>h)
    {
    float a=(float)(w/(h+0.01));
    glOrtho(-a,a,-1,1,Camera.a,Camera.c);
    }
  else
    {
    float a=(float)(h/(w+0.01));
    glOrtho(-1,1,-a,a,Camera.a,Camera.c);
    }
  glTranslated(0.,0.,-Camera.b);
  }

int IsCommand(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_EXIT||wParam==PUSH_ALT_F4)
    {
    PostQuitMessage(0);
    return(TRUE);
    }
  if(wParam==PUSH_ABOUT)
    {
    DialogBox(hInst,"ABOUT",hMain,(DLGPROC)AboutProc);
    return(TRUE);
    }
  if(wParam==PUSH_ANIM)
    {
    animate=1-animate;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    return(TRUE);
    }
  if(wParam==PUSH_CAPT)
    {
    auto_capture=1-auto_capture;
    SendMessage(hCapt,BM_SETCHECK,auto_capture,0);
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_COLORS)
    {
    colors=1-colors;
    SendMessage(hColo,BM_SETCHECK,colors,0);
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_FAST)
    {
    ticks--;
    if(ticks<0)
      ticks=0;
    return(TRUE);
    }
  if(wParam==PUSH_ROLL)
    {
    int i;
    double a;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    a=Objective.a;
    for(i=0;i<360;i+=5)
      {
      Objective.a=i;
      glRepaint();
      Sleep(25);
      }
    Objective.a=a;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SLOW)
    {
    ticks++;
    return(TRUE);
    }
  if(wParam==PUSH_SPIN)
    {
    int i;
    double b;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    b=Objective.b;
    for(i=0;i<360;i+=5)
      {
      Objective.b=i;
      glRepaint();
      Sleep(25);
      }
    Objective.b=b;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_WIND)
    {
    wind_capture=1-wind_capture;
    SendMessage(hWind,BM_SETCHECK,!wind_capture,0);
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_WIRE)
    {
    fill=1-fill;
    SendMessage(hWire,BM_SETCHECK,!fill,0);
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

void ResetView()
  {
  Light.a= 0.;
  Light.b=45.;
  Objective.a=0.;
  Objective.b=0.;
  Objective.c=0.;
  Objective.x=0.;
  Objective.y=0.;
  Objective.z=-1.5;
  zoom=1.;
  }

int IsOrientation(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_DOWN)
    {
    Objective.a-=5;
    if(Objective.a<0)
      Objective.a=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_LEFT)
    {
    Objective.b-=5;
    if(Objective.b<0)
      Objective.b=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_RIGHT)
    {
    Objective.b+=5;
    if(Objective.b>359)
      Objective.b=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_PGDN)
    {
    zoom*=0.95;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_PGUP)
    {
    zoom/=0.95;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_UP)
    {
    Objective.a+=5;
    if(Objective.a>359)
      Objective.a=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_DOWN)
    {
    Objective.z-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_LEFT)
    {
    Objective.c-=5;
    if(Objective.c<0)
      Objective.c=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_RIGHT)
    {
    Objective.c+=5;
    if(Objective.c>359)
      Objective.c=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_UP)
    {
    Objective.z+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_A)
    {
    lAnkle+=5.F;
    if(lAnkle>41.F)
      lAnkle=-60.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_DOWN)
    {
    Objective.y-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_E)
    {
    lElbow+=5.F;
    if(lElbow>159.F)
      lElbow=0.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_F)
    {
    lWristF+=5.F;
    if(lWristF>91.F)
      lWristF=-90.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_H)
    {
    lHip+=5.F;
    if(lHip>121.F)
      lHip=-70.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_J)
    {
    Mandible-=5.F;
    if(Mandible<-44.F)
      Mandible=0.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_K)
    {
    lKnee+=5.F;
    if(lKnee>121.F)
      lKnee=0.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_LEFT)
    {
    Objective.x-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_R)
    {
    lWristR+=5.F;
    if(lWristR>181.F)
      lWristR=-180.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_RIGHT)
    {
    Objective.x+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_S)
    {
    lShoulder+=5.F;
    if(lShoulder>359.F)
      lShoulder=0.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_SPACE)
    {
    group--;
    if(group<-1)
      group=rtoes;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_UP)
    {
    Objective.y+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_A)
    {
    rAnkle+=5.F;
    if(rAnkle>41.F)
      rAnkle=-60.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_E)
    {
    rElbow+=5.F;
    if(rElbow>159.F)
      rElbow=0.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_F)
    {
    rWristF+=5.F;
    if(rWristF>91.F)
      rWristF=-90.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_H)
    {
    rHip+=5.F;
    if(rHip>121.F)
      rHip=-70.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_K)
    {
    rKnee+=5.F;
    if(rKnee>121.F)
      rKnee=0.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_R)
    {
    rWristR+=5.F;
    if(rWristR>181.F)
      rWristR=-180.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_S)
    {
    rShoulder+=5.F;
    if(rShoulder>359.F)
      rShoulder=0.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_SPACE)
    {
    group++;
    if(group>rtoes)
      group=-1;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_Z)
    {
    ResetView();
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI MainProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    {
    PostQuitMessage(0);
    return(FALSE);
    }
  if(wMsg==WM_COMMAND)
    {
    if(IsCommand(wParam&0xFFFF,lParam))
      {
      SetFocus(hMain);
      return(FALSE);
      }
    if(IsOrientation(wParam&0xFFFF,lParam))
      return(FALSE);
    return(FALSE);
    }
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_CTLCOLORBTN)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_GETMINMAXINFO)
    {
    MINMAXINFO*mInfo=(MINMAXINFO*)lParam;
    mInfo->ptMinTrackSize.x=640;
    mInfo->ptMinTrackSize.y=480;
    return(FALSE);
    }
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  if(wMsg==WM_SIZE)
    {
    PositionWindows();
    glRepaint();
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

float walk(double mn,double mx,double a)
  {
  return((float)(mn+(mx-mn)*(sin((Twalk-a)*M_PI/180.)+1.)/2.));
  }

void WINAPI Animate(HWND hWnd,WPARAM wParam,DWORD idEvent,int uTime)
  {
  if(!animate)
    return;
  if(waits)
    {
    waits--;
    return;
    }
  Twalk+=5.;
  lShoulder=walk( -20., 20., 90.); /* -180 to 180 */
  rShoulder=walk( -20., 20.,270.); /* -180 to 180 */
  lElbow   =walk(   0., 30.,270.); /*    0 to 160 */
  rElbow   =walk(   0., 30., 90.); /*    0 to 160 */
  lWristF  =walk( -15., 15., 90.); /*  -90 to  90 */
  rWristF  =walk( -15., 15.,270.); /*  -90 to  90 */
  lHip     =walk( -15., 15.,270.); /*  -70 to 120 */
  rHip     =walk( -15., 15., 90.); /*  -70 to 120 */
  lKnee    =walk(   0., 15., 90.); /*    0 to 120 */
  rKnee    =walk(   0., 15.,270.); /*    0 to 120 */
  lAnkle   =walk( -10., 10.,270.); /*  -60 to  40 */
  rAnkle   =walk( -10., 10., 90.); /*  -60 to  40 */
  lWristR=rWristR=180.;
  waits=ticks;
  glRepaint();
  }

void CreateWindows()
  {
  int h,high,w,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  w=2*wide/3;
  h=3*high/4;
  if((hMain=CreateWindow("MAIN","use directional keys + alt/ctl/shift to view",
      WS_OVERLAPPEDWINDOW|WS_VISIBLE,(wide-w)/2,(high-h)/2,w,h,
      NULL,LoadMenu(hInst,"MENU"),hInst,NULL))==NULL)
    Abort(__LINE__,"can't create main window\nWindows error code %i",GetLastError());
  if((hStat=CreateWindow(STATUSCLASSNAME,NULL,
      WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create status control\nerror code %i",GetLastError());
  SendMessage(hStat,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  if((hLogo=CreateWindow("LOGO",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWindows error code %i",GetLastError());
  if((hPlot=CreateWindow("PLOT",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_BLACKRECT,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWindows error code %i",GetLastError());
  hAbou=CreatePushButton(PUSH_ABOUT,"About");
  hExit=CreatePushButton(PUSH_EXIT ,"Exit");
  hFast=CreatePushButton(PUSH_FAST ,"Faster");
  hRoll=CreatePushButton(PUSH_ROLL ,"Roll");
  hSlow=CreatePushButton(PUSH_SLOW ,"Slower");
  hSpin=CreatePushButton(PUSH_SPIN ,"Spin");
  hAnim=CreateRadioButton(PUSH_ANIM  ,"Animate",animate);
  hCapt=CreateRadioButton(PUSH_CAPT  ,"auto capture",auto_capture);
  hColo=CreateRadioButton(PUSH_COLORS,"colors/white",colors);
  hWind=CreateRadioButton(PUSH_WIND  ,"window/all",wind_capture);
  hWire=CreateRadioButton(PUSH_WIRE  ,"Solid/Wire",!fill);
  }

void LoadResources()
  {
  bLogo=GetBitmap("logo");
  }

void RegisterClasses()
  {
  Register("MAIN",MainProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0,0);
  Register("LOGO",LogoProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0,0);
  Register("PLOT",PlotProc,GetStockObject(NULL_BRUSH),IDC_CROSS,0,CS_OWNDC);
  }

int WINAPI WinMain(HINSTANCE hCurrent,HINSTANCE hPrevious,char*lCommand,int nShow)
  {
  HACCEL acc;
  MSG msg;
  hInst=hCurrent;
  InitCommonControls();
  acc=LoadAccelerators(hInst,"FAST");
  LoadResources();
  RegisterClasses();
  CreateWindows();
  ReadTriangles("skeleton.tri");
  OpenGL();
  PositionWindows();
  SetTimer(hMain,1,10,(TIMERPROC)Animate);
  while(GetMessage(&msg,NULL,0,0))
    if(!TranslateAccelerator(msg.hwnd,acc,&msg))
      if(!TranslateMessage(&msg))
        DispatchMessage(&msg);
  return(0);
  }
