/* Miniature Steam Engine based on the program written by Troy Robinette.
   Modified to function within a standard Windows GUI by Dudley J. Benton. */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <io.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include "..\\..\\OpenGL\\gl.h"
#include "..\\..\\utilities\\keys.h"
#include "..\\..\\utilities\\WinApp.c"
#include "..\\..\\utilities\\OpenGL.c"
#include "steam.h"

HWND hAbou;
HWND hAnim;
HWND hExit;
HWND hFast;
HWND hRoll;
HWND hSlow;
HWND hSpin;
HWND hStat;
HWND hWire;

int ticks=5;
int waits=0;
double zoom=1.;

float ambientLight0[ 4]={0.2F,0.2F,0.2F,0.0F};
float diffuseLight0[ 4]={0.3F,0.3F,0.3F,0.0F};
float positionLight0[4]={0.0F,0.0F,5.0F,0.0F};

struct{double a,b,c;}Camera={15,0,0};
struct{double a,b;}Light={0,45};
struct{double a,b,c,x,y,z;}Objective={0,0,0,0,0,0};

double angle_step=10.;
double view_h=270.,view_v=0.,head_angle=0.;
int crank_angle=0;
int crank_step=5;

void glBox(double x,double y,double z)
  {
  glPushMatrix();
  glScaled(x,y,z);
  glHexahedron(1.,1.,1.);
  glPopMatrix();
  }

void glShaft(double outerRadius,double innerRadius,double length)
  {
  glPushMatrix();
  glPushMatrix();
  glRotated(90.,1.0,0.0,0.0);
  glTranslated(0.,length/2.,0.);
  glCylinder(outerRadius,length,20);
  glPopMatrix();
  glPushMatrix();
  glRotated(180.,0.0,1.0,0.0);
  glRing(outerRadius,innerRadius,20);
  glPopMatrix();
  glTranslated(0.0,0.0,length);
  glRing(outerRadius,innerRadius,20);
  glPopMatrix();
  }

void DrawPiston()
  {
  glPushMatrix();
  glColor(0.3F,0.6F,0.9F);
  glPushMatrix();
  glRotated(90.,0.0,1.0,0.0);
  glTranslated(0.0,0.0,-0.07);
  glShaft(0.125,0.06,0.12);
  glPopMatrix();
  glRotated(-90.,1.0,0.0,0.0);
  glTranslated(0.0,0.0,0.05);
  glShaft(0.06,0.0,0.6);
  glTranslated(0.0,0.0,0.6);
  glShaft(0.2,0.0,0.5);
  glPopMatrix();
  }

void DrawEnginePole()
  {
  glPushMatrix();
  glColor(0.9F,0.9F,0.9F);
  glBox(0.5,3.0,0.5);
  glColor(0.5F,0.1F,0.5F);
  glRotated(90.,0.0,1.0,0.0);
  glTranslated(0.0,0.9,-0.4);
  glShaft(0.1,0.0,2);
  glPopMatrix();
  }

void DrawCylinderHead()
  {
  glPushMatrix();
  glColor(0.5F,1.0F,0.5F);
  glRotated(90.,1.0,0.0,0.0);
  glTranslated(0,0.0,0.4);
  glRotated(head_angle,1,0,0);
  glTranslated(0,0.0,-0.4);
  glShaft(0.23,0.21,1.6);
  glRotated(180.,1.0,0.0,0.0);
  glRing(0.,0.23,20);
  glPopMatrix();
  }

void DrawFlywheel()
  {
  glPushMatrix();
  glColor(0.5F,0.5F,1.0F);
  glRotated(90.,0.0,1.0,0.0);
  glShaft(0.625,0.08,0.5);
  glPopMatrix();
  }

void DrawCrankbell()
  {
  glPushMatrix();
  glColor(1.0F,0.5F,0.5F);
  glRotated(90.,0.0,1.0,0.0);
  glShaft(0.3,0.08,0.12);
  glColor(0.5F,0.1F,0.5F);
  glTranslated(0.0,0.2,0.0);
  glShaft(0.06,0.0,0.34);
  glTranslated(0.0,0.0,0.22);
  glRotated(90.,0.0,1.0,0.0);
  glRotated(crank_angle-head_angle,1.0,0.0,0.0);
  DrawPiston();
  glPopMatrix();
  }

void DrawCrank()
  {
  glPushMatrix();
  glRotated(crank_angle,1.0,0.0,0.0);
  glPushMatrix();
  glRotated(90.,0.0,1.0,0.0);
  glTranslated(0.0,0.0,-1.0);
  glShaft(0.08,0.0,1.4);
  glPopMatrix();
  glPushMatrix();
  glTranslated(0.28,0.0,0.0);
  DrawCrankbell();
  glPopMatrix();
  glPushMatrix();
  glTranslated(-0.77,0.0,0.0);
  DrawFlywheel();
  glPopMatrix();
  glPopMatrix();
  }

double HeadAngle(double deg)
  {
  double theta;
  theta=270.112-deg/58.;
  return(120.*atan((0.15*sin(theta))/
              ((2.7-0.15*cos(theta)))));
  }

void RenderEngine()
  {
  head_angle=HeadAngle(crank_angle);
  DrawEnginePole();
  glPushMatrix();
  glTranslated(0.5,1.4,0.0);
  DrawCylinderHead();
  glPopMatrix();
  glPushMatrix();
  glTranslated(0.0,-0.8,0.0);
  DrawCrank();
  glPopMatrix();
  }

void glRepaint()
  {
  if(!pDC)
    return;
  glClearColor(0.F,0.F,0.F,0.F);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glShadeModel(GL_SMOOTH);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,TRUE);
  glDisable(GL_AUTO_NORMAL);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0,GL_AMBIENT ,ambientLight0);
  glLightfv(GL_LIGHT0,GL_DIFFUSE ,diffuseLight0);
    positionLight0[0]=(float)(sin(M_PI*Light.a/180)*cos(M_PI*Light.b/180));
    positionLight0[1]=(float)(sin(M_PI*Light.b/180));
    positionLight0[2]=(float)(cos(M_PI*Light.a/180)*cos(M_PI*Light.b/180));
    positionLight0[3]=0;/* infinite distance to light source */
  glLightfv(GL_LIGHT0,GL_POSITION,positionLight0);
  glEnable(GL_LIGHT0);
  glDisable(GL_TEXTURE_2D);
  glTranslated(Objective.x,Objective.y,Objective.z);
  glRotated(Objective.a,1,0,0);
  glRotated(Objective.b,0,1,0);
  glRotated(Objective.c,0,0,1);
  WindowText(hStat,"Camera=%lG,%lG,%lG, Light=%lG,%lG, Objective=%lG,%lG,%lG,%lG,%lG,%lG",
    Camera.a,Camera.b,Camera.c,Light.a,Light.b,Objective.a,Objective.b,
    Objective.c,Objective.x,Objective.y,Objective.z);
  glPolygonMode(back?GL_FRONT_AND_BACK:GL_FRONT,fill?GL_FILL:GL_LINE);
  glScaled(zoom/2.,zoom/2.,zoom/2.);
  RenderEngine();
  glFinish();
  SwapBuffers(pDC);
  }

void PositionWindows()
  {
  int h,t,w,x,y;
  RECT rc;
  if(IsIconic(hMain))
    return;
  if(!hPlot)
    return;
  GetClientRect(hMain,&rc);
  t=GetSystemMetrics(SM_CYCAPTION);
  w=bLogo->biWidth;
  h=bLogo->biHeight;
  x=rc.right-1-w;
  y=1;
  MoveWindow(hLogo,x,y,w,h,TRUE);y+=h+2;
  MoveWindow(hSpin,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hRoll,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hFast,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hSlow,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hAnim,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWire,x,y,w,t,TRUE);
  y=rc.bottom-t;
  y-=t+1;MoveWindow(hExit,x,y,w,t,TRUE);
  y-=t+1;MoveWindow(hAbou,x,y,w,t,TRUE);
  MoveWindow(hStat,0,rc.bottom-t,rc.right,t,TRUE);
  w=rc.right-w-2;
  h=rc.bottom-t;
  MoveWindow(hPlot,2,2,w-4,h-4,TRUE);
  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(w>h)
    {
    float a=(float)(w/(h+0.01));
    glFrustum(-a,a,-1,1,Camera.a,Camera.c);
    }
  else
    {
    float a=(float)(h/(w+0.01));
    glFrustum(-1,1,-a,a,Camera.a,Camera.c);
    }
  glTranslated(0.,0.,-Camera.b);
  }

int IsCommand(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_EXIT||wParam==PUSH_ALT_F4)
    {
    PostQuitMessage(0);
    return(TRUE);
    }
  if(wParam==PUSH_ABOUT)
    {
    DialogBox(hInst,"ABOUT",hMain,(DLGPROC)AboutProc);
    return(TRUE);
    }
  if(wParam==PUSH_ANIM)
    {
    animate=1-animate;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    return(TRUE);
    }
  if(wParam==PUSH_FAST)
    {
    ticks--;
    if(ticks<0)
      ticks=0;
    return(TRUE);
    }
  if(wParam==PUSH_ROLL)
    {
    int i;
    double a;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    a=Objective.a;
    for(i=0;i<360;i+=5)
      {
      Objective.a=i;
      glRepaint();
      Sleep(25);
      }
    Objective.a=a;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SLOW)
    {
    ticks++;
    return(TRUE);
    }
  if(wParam==PUSH_SPIN)
    {
    int i;
    double b;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    b=Objective.b;
    for(i=0;i<360;i+=5)
      {
      Objective.b=i;
      glRepaint();
      Sleep(25);
      }
    Objective.b=b;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_WIRE)
    {
    fill=1-fill;
    SendMessage(hWire,BM_SETCHECK,!fill,0);
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

void ResetView()
  {
  Light.a= 0.;
  Light.b=45.;
  Objective.a=0.;
  Objective.b=0.;
  Objective.c=0.;
  Objective.x=0.;
  Objective.y=0.;
  Objective.z=0.;
  zoom=1.;
  }

int IsOrientation(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_ALT_UP)
    {
    Objective.y+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_RIGHT)
    {
    Objective.x+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_LEFT)
    {
    Objective.x-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_DOWN)
    {
    Objective.y-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_DOWN)
    {
    Objective.a-=5;
    if(Objective.a<0)
      Objective.a=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_RIGHT)
    {
    Objective.c+=5;
    if(Objective.c>359)
      Objective.c=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_LEFT)
    {
    Objective.c-=5;
    if(Objective.c<0)
      Objective.c=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_UP)
    {
    Objective.a+=5;
    if(Objective.a>359)
      Objective.a=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_DOWN)
    {
    Light.b-=5;
    if(Light.b<0)
      Light.b=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_LEFT)
    {
    Light.a-=5;
    if(Light.a<0)
      Light.a=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_RIGHT)
    {
    Light.a+=5;
    if(Light.a>359)
      Light.a=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_UP)
    {
    Light.b+=5;
    if(Light.b>359)
      Light.b=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_Z)
    {
    ResetView();
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_DOWN)
    {
    Objective.z-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_RIGHT)
    {
    Objective.b+=5;
    if(Objective.b>359)
      Objective.b=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_LEFT)
    {
    Objective.b-=5;
    if(Objective.b<0)
      Objective.b=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_UP)
    {
    Objective.z+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_PGDN)
    {
    zoom*=0.95;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_PGUP)
    {
    zoom/=0.95;
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI MainProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    {
    PostQuitMessage(0);
    return(FALSE);
    }
  if(wMsg==WM_COMMAND)
    {
    if(IsCommand(wParam&0xFFFF,lParam))
      {
      SetFocus(hMain);
      return(FALSE);
      }
    if(IsOrientation(wParam&0xFFFF,lParam))
      return(FALSE);
    return(FALSE);
    }
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_CTLCOLORBTN)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_GETMINMAXINFO)
    {
    MINMAXINFO*mInfo=(MINMAXINFO*)lParam;
    mInfo->ptMinTrackSize.x=640;
    mInfo->ptMinTrackSize.y=480;
    return(FALSE);
    }
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  if(wMsg==WM_SIZE)
    {
    PositionWindows();
    glRepaint();
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

void WINAPI Animate(HWND hWnd,WPARAM wParam,DWORD idEvent,int uTime)
  {
  if(!animate)
    return;
  if(waits)
    {
    waits--;
    return;
    }
  crank_angle+=crank_step;
  if(crank_angle>360)
    crank_angle=0;
  waits=ticks;
  glRepaint();
  }

void CreateWindows()
  {
  int h,high,w,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  h=7*high/8;
  w=h+bLogo->biWidth-55;
  if((hMain=CreateWindow("MAIN","use directional keys + alt/ctl/shift to view",
      WS_OVERLAPPEDWINDOW|WS_VISIBLE,(wide-w)/2,(high-h)/2,w,h,
      NULL,LoadMenu(hInst,"MENU"),hInst,NULL))==NULL)
    Abort(__LINE__,"can't create main window\nWindows error code %i",GetLastError());
  if((hStat=CreateWindow(STATUSCLASSNAME,NULL,
      WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create status control\nerror code %i",GetLastError());
  SendMessage(hStat,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  if((hLogo=CreateWindow("LOGO",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWindows error code %i",GetLastError());
  if((hPlot=CreateWindow("PLOT",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_BLACKRECT,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWindows error code %i",GetLastError());
  hAbou=CreatePushButton(PUSH_ABOUT,"About");
  hExit=CreatePushButton(PUSH_EXIT ,"Exit");
  hFast=CreatePushButton(PUSH_FAST ,"Faster");
  hRoll=CreatePushButton(PUSH_ROLL ,"Roll");
  hSlow=CreatePushButton(PUSH_SLOW ,"Slower");
  hSpin=CreatePushButton(PUSH_SPIN ,"Spin");
  hAnim=CreateRadioButton(PUSH_ANIM,"Animate",animate);
  hWire=CreateRadioButton(PUSH_WIRE,"Solid/Wire",!fill);
  }

void LoadResources()
  {
  bLogo=GetBitmap("logo");
  }

void RegisterClasses()
  {
  Register("MAIN",MainProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0,0);
  Register("LOGO",LogoProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0,0);
  Register("PLOT",PlotProc,GetStockObject(NULL_BRUSH),IDC_CROSS,0,CS_OWNDC);
  }

int WINAPI WinMain(HINSTANCE hCurrent,HINSTANCE hPrevious,char*lCommand,int nShow)
  {
  HACCEL acc;
  MSG msg;
  hInst=hCurrent;
  InitCommonControls();
  acc=LoadAccelerators(hInst,"FAST");
  LoadResources();
  RegisterClasses();
  CreateWindows();
  PositionWindows();
  OpenGL();
  SetTimer(hMain,1,10,(TIMERPROC)Animate);
  while(GetMessage(&msg,NULL,0,0))
    if(!TranslateAccelerator(msg.hwnd,acc,&msg))
      if(!TranslateMessage(&msg))
        DispatchMessage(&msg);
  return(0);
  }
