/* TWAVE Demo from the Original OpenGL SDK
 * (c) Copyright 1993,Silicon Graphics,Inc.
 *               1993,1994 Microsoft Corporation
 * ALL RIGHTS RESERVED
 */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <gl\glaux.h>

int checkerSize=2;
int colorIndexes1[3];
int colorIndexes2[3];
int contouring=0;
int frames=10,curFrame=0,nextFrame=0;
int widthX=10,widthY=10;
int clearMask=GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT;
int depth=GL_TRUE;
int lighting=GL_TRUE;
int rgb=GL_TRUE,doubleBuffer=GL_TRUE;
int smooth=GL_TRUE;
int spinMode=GL_FALSE;
int stepMode=GL_FALSE;
float height=0.2F;

struct facet{float color[3];float normal[3];};
struct coord{float vertex[3];float normal[3];};
struct mesh{int widthX,widthY,numFacets,numCoords,frames;
  struct coord*coords;struct facet*facets;}theMesh;

BYTE contourTexture1[]={255,255,255,255,255,255,255,255,255,255,255,255,127,127,127,127};
BYTE contourTexture2[]={255,255,255,255,255,127,127,127,255,127,127,127,255,127,127,127};

#define GETCOORD(frame,x,y)(&(theMesh.coords[frame*theMesh.numCoords+(x)+(y)*(theMesh.widthX+1)]))
#define GETFACET(frame,x,y)(&(theMesh.facets[frame*theMesh.numFacets+(x)+(y)*theMesh.widthX]))

void CALLBACK Animate(void)
  {
  struct coord*coord;
  struct facet*facet;
  float*lastColor;
  float*thisColor;
  int i,j;
  glClear(clearMask);
  if(nextFrame||!stepMode)
    curFrame++;
  if(curFrame>=theMesh.frames)
    curFrame=0;
  if((nextFrame||!stepMode)&&spinMode)
    glRotatef(5.0,0.0,0.0,1.0);
  nextFrame=0;
  for(i=0;i<theMesh.widthX;i++)
    {
    glBegin(GL_QUAD_STRIP);
    lastColor=NULL;
    for(j=0;j<theMesh.widthY;j++)
      {
      facet=GETFACET(curFrame,i,j);
      if(!smooth&&lighting)
        glNormal3fv(facet->normal);
      if(lighting)
        {
        if(rgb)
          {
          thisColor=facet->color;
          glColor3fv(facet->color);
          }
        else
          {
          thisColor=facet->color;
          glMaterialfv(GL_FRONT_AND_BACK,GL_COLOR_INDEXES,facet->color);
          }
        }
      else
        {
        if(rgb)
          {
          thisColor=facet->color;
          glColor3fv(facet->color);
          }
        else
          {
          thisColor=facet->color;
          glIndexf(facet->color[1]);
          }
        }
      if(!lastColor||(thisColor[0]!=lastColor[0]&&smooth))
        {
        if(lastColor)
          {
          glEnd();
          glBegin(GL_QUAD_STRIP);
          }
        coord=GETCOORD(curFrame,i,j);
        if(smooth&&lighting)
          glNormal3fv(coord->normal);
        glVertex3fv(coord->vertex);
        coord=GETCOORD(curFrame,i+1,j);
        if(smooth&&lighting)
          glNormal3fv(coord->normal);
        glVertex3fv(coord->vertex);
        }
      coord=GETCOORD(curFrame,i,j+1);
      if(smooth&&lighting)
        glNormal3fv(coord->normal);
      glVertex3fv(coord->vertex);
      coord=GETCOORD(curFrame,i+1,j+1);
      if(smooth&&lighting)
        glNormal3fv(coord->normal);
      glVertex3fv(coord->vertex);
      lastColor=thisColor;
      }
    glEnd();
    }
  glFlush();
  if(doubleBuffer)
    auxSwapBuffers();
  }

void SetColorMap(void)
  {
  static float green[3]={0.2F,1.0F,0.2F};
  static float red[3]={1.0F,0.2F,0.2F};
  float*color,percent;
  int*indexes,entries,i,j;
  entries=auxGetColorMapSize();
  colorIndexes1[0]=1;
  colorIndexes1[1]=1+(int)((entries-1)*0.3);
  colorIndexes1[2]=(int)((entries-1)*0.5);
  colorIndexes2[0]=1+(int)((entries-1)*0.5);
  colorIndexes2[1]=1+(int)((entries-1)*0.8);
  colorIndexes2[2]=entries-1;
  for(i=0;i<2;i++)
    {
    if(i==0)
      {
      color=green;
      indexes=colorIndexes1;
      }
    else
      {
      color=red;
      indexes=colorIndexes2;
      }
    for(j=indexes[0];j<indexes[1];j++)
      {
      percent=(float)(0.2+(j-indexes[0])*0.8/(indexes[1]-indexes[0]));
      auxSetOneColor(j,percent*color[0],percent*color[1],percent*color[2]);
      }
    for(j=indexes[1];j<=indexes[2];j++)
      {
      percent=(float)((j-indexes[1])/(float)(indexes[2]-indexes[1]));
      auxSetOneColor(j,percent*(1-color[0])+color[0],percent*(1-color[1])+color[1],percent*(1-color[2])+color[2]);
      }
    }
  }

void InitMesh(void)
  {
  struct coord*coord;
  struct facet*facet;
  float dp1[3],dp2[3];
  float*pt1,*pt2,*pt3;
  float angle,d,x,y;
  int numFacets,numCoords,frameNum,i,j;
  theMesh.widthX=widthX;
  theMesh.widthY=widthY;
  theMesh.frames=frames;
  numFacets=widthX*widthY;
  numCoords=(widthX+1)*(widthY+1);
  theMesh.numCoords=numCoords;
  theMesh.numFacets=numFacets;
  theMesh.coords=(struct coord*)malloc(frames*numCoords*sizeof(struct coord));
  theMesh.facets=(struct facet*)malloc(frames*numFacets*sizeof(struct facet));
  if(theMesh.coords==NULL||theMesh.facets==NULL)
    auxQuit();
  for(frameNum=0;frameNum<frames;frameNum++)
    {
    for(i=0;i<=widthX;i++)
      {
      x=i/(float)widthX;
      for(j=0;j<=widthY;j++)
        {
        y=j/(float)widthY;
        d=(float)sqrt(x*x+y*y);
        if(fabs(d)<0.0001F)
          d=0.0001F;
        angle=(float)(2.*M_PI*d+(2.*M_PI/frames*frameNum));
        coord=GETCOORD(frameNum,i,j);
        coord->vertex[0]=(float)(x-0.5);
        coord->vertex[1]=(float)(y-0.5);
        coord->vertex[2]=(float)((height-height*d)*cos(angle));
        coord->normal[0]=(float)(-(height/d)*x*((1.-d)*2.*M_PI*sin(angle)+cos(angle)));
        coord->normal[1]=(float)(-(height/d)*y*((1.-d)*2.*M_PI*sin(angle)+cos(angle)));
        coord->normal[2]=-1;
        d=(float)(1./sqrt(coord->normal[0]*coord->normal[0]+coord->normal[1]*coord->normal[1]+1));
        coord->normal[0]*=d;
        coord->normal[1]*=d;
        coord->normal[2]*=d;
        }
      }
    for(i=0;i<widthX;i++)
      {
      for(j=0;j<widthY;j++)
        {
        facet=GETFACET(frameNum,i,j);
        if(((i/checkerSize)%2)^(j/checkerSize)%2)
          {
          if(rgb)
            {
            facet->color[0]=1.0F;
            facet->color[1]=0.2F;
            facet->color[2]=0.2F;
            }
          else
            {
            facet->color[0]=(float)colorIndexes1[0];
            facet->color[1]=(float)colorIndexes1[1];
            facet->color[2]=(float)colorIndexes1[2];
            }
          }
        else
          {
          if(rgb)
            {
            facet->color[0]=0.2F;
            facet->color[1]=1.0F;
            facet->color[2]=0.2F;
            }
          else
            {
            facet->color[0]=(float)colorIndexes2[0];
            facet->color[1]=(float)colorIndexes2[1];
            facet->color[2]=(float)colorIndexes2[2];
            }
          }
        pt1=GETCOORD(frameNum,i,j)->vertex;
        pt2=GETCOORD(frameNum,i,j+1)->vertex;
        pt3=GETCOORD(frameNum,i+1,j+1)->vertex;
        dp1[0]=pt2[0]-pt1[0];
        dp1[1]=pt2[1]-pt1[1];
        dp1[2]=pt2[2]-pt1[2];
        dp2[0]=pt3[0]-pt2[0];
        dp2[1]=pt3[1]-pt2[1];
        dp2[2]=pt3[2]-pt2[2];
        facet->normal[0]=dp1[1]*dp2[2]-dp1[2]*dp2[1];
        facet->normal[1]=dp1[2]*dp2[0]-dp1[0]*dp2[2];
        facet->normal[2]=dp1[0]*dp2[1]-dp1[1]*dp2[0];
        d=(float)(1.0/sqrt(facet->normal[0]*facet->normal[0]+facet->normal[1]*facet->normal[1]+facet->normal[2]*facet->normal[2]));
        facet->normal[0]*=d;
        facet->normal[1]*=d;
        facet->normal[2]*=d;
        }
      }
    }
  }

void InitMaterials(void)
  {
  static float ambient[]={0.1F,0.1F,0.1F,1.0F};
  static float diffuse[]={0.5F,1.0F,1.0F,1.0F};
  static float position[]={90.0F,90.0F,150.0F,0.0F};
  static float front_mat_shininess[]={60.0F};
  static float front_mat_specular[]={0.2F,0.2F,0.2F,1.0F};
  static float front_mat_diffuse[]={0.5F,0.28F,0.38F,1.0F};
  static float back_mat_shininess[]={60.0F};
  static float back_mat_specular[]={0.5F,0.5F,0.2F,1.0F};
  static float back_mat_diffuse[]={1.0F,1.0F,0.2F,1.0F};
  static float lmodel_ambient[]={1.0F,1.0F,1.0F,1.0F};
  static float lmodel_twoside[]={GL_TRUE};
  glMatrixMode(GL_PROJECTION);
  gluPerspective(450,1.0F,0.5F,10.0F);
  glLightfv(GL_LIGHT0,GL_AMBIENT,ambient);
  glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuse);
  glLightfv(GL_LIGHT0,GL_POSITION,position);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,lmodel_ambient);
  glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE,lmodel_twoside);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glMaterialfv(GL_FRONT,GL_SHININESS,front_mat_shininess);
  glMaterialfv(GL_FRONT,GL_SPECULAR,front_mat_specular);
  glMaterialfv(GL_FRONT,GL_DIFFUSE,front_mat_diffuse);
  glMaterialfv(GL_BACK,GL_SHININESS,back_mat_shininess);
  glMaterialfv(GL_BACK,GL_SPECULAR,back_mat_specular);
  glMaterialfv(GL_BACK,GL_DIFFUSE,back_mat_diffuse);
  if(rgb)
    glColorMaterial(GL_FRONT_AND_BACK,GL_DIFFUSE);
  if(rgb)
    glEnable(GL_COLOR_MATERIAL);
  else
    SetColorMap();
  }

void InitTexture(void)
  {
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
  }

void Init(void)
  {
  glClearColor(0.0,0.0,0.0,0.0);
  glShadeModel(GL_SMOOTH);
  glFrontFace(GL_CW);
  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);
  InitMaterials();
  InitTexture();
  InitMesh();
  glMatrixMode(GL_MODELVIEW);
  glTranslatef(0.0F,0.4F,-1.8F);
  glScalef(2.0F,2.0F,2.0F);
  glRotatef(-35.0F,1.0F,0.0F,0.0F);
  glRotatef(35.0F,0.0F,0.0F,1.0F);
  }

void CALLBACK Reshape(int width,int height)
  {
  glViewport(0,0,width,height);
  }

void CALLBACK Key_c(void)
  {
  contouring++;
  if(contouring==1)
    {
    static float map[4]={0,0,20,0};
    glTexImage2D(GL_TEXTURE_2D,0,3,4,4,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,(GLvoid*)contourTexture1);
    glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
    glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
    glTexGenfv(GL_S,GL_OBJECT_PLANE,map);
    glTexGenfv(GL_T,GL_OBJECT_PLANE,map);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    }
  else if(contouring==2)
    {
    static float map[4]={0,0,20,0};
    glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
    glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTexGenfv(GL_S,GL_EYE_PLANE,map);
    glTexGenfv(GL_T,GL_EYE_PLANE,map);
    glPopMatrix();
    }
  else
    {
    contouring=0;
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_2D);
    }
  }

void CALLBACK Key_s(void)
  {
  smooth=!smooth;
  glShadeModel(smooth?GL_SMOOTH:GL_FLAT);
  }

void CALLBACK Key_l(void)
  {
  lighting=!lighting;
  if(lighting)
    {
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    if(rgb)
      glEnable(GL_COLOR_MATERIAL);
    }
  else
    {
    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    if(rgb)
      glDisable(GL_COLOR_MATERIAL);
    }
  }

void CALLBACK Key_d(void)
  {
  depth=!depth;
  if(depth)
    {
    glEnable(GL_DEPTH_TEST);
    clearMask|=GL_DEPTH_BUFFER_BIT;
    }
  else
    {
    glDisable(GL_DEPTH_TEST);
    clearMask&=~GL_DEPTH_BUFFER_BIT;
    }
  }

void CALLBACK Key_SPACE(void)
  {
  stepMode=!stepMode;
  if(stepMode)
    auxIdleFunc(0);
  else
    auxIdleFunc(Animate);
  }

void CALLBACK Key_n(void)
  {
  if(stepMode)
    nextFrame=1;
  }

void CALLBACK Key_a(void)
  {
  spinMode=!spinMode;
  }

int main(int argc,char**argv,char**envp)
  {
  int h=300,high,type,w=300,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  auxInitPosition((wide-w)/2,(high-h)/2,w,h);
  type=AUX_DEPTH16;
  type|=(rgb)?AUX_RGB:AUX_INDEX;
  type|=(doubleBuffer)?AUX_DOUBLE:AUX_SINGLE;
  auxInitDisplayMode(type);
  if(auxInitWindow("Wave Demo keys=acdlns+space+Esc")==GL_FALSE)
    auxQuit();
  Init();
  auxExposeFunc((AUXEXPOSEPROC)Reshape);
  auxReshapeFunc((AUXRESHAPEPROC)Reshape);
  auxKeyFunc(AUX_c,Key_c);
  auxKeyFunc(AUX_s,Key_s);
  auxKeyFunc(AUX_l,Key_l);
  auxKeyFunc(AUX_d,Key_d);
  auxKeyFunc(AUX_SPACE,Key_SPACE);
  auxKeyFunc(AUX_n,Key_n);
  auxKeyFunc(AUX_a,Key_a);
  auxIdleFunc(Animate);
  auxMainLoop(Animate);
  return(0);
  }
