/* Walker model with standard Windows GUI
   by Dudley J. Benton
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <io.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include "..\\..\\OpenGL\\gl.h"
#include "..\\..\\utilities\\keys.h"
#include "..\\..\\utilities\\WinApp.c"
#include "..\\..\\utilities\\OpenGL.c"
#include "walker.h"

HWND hAbou;
HWND hAnim;
HWND hCapt;
HWND hExit;
HWND hFast;
HWND hRoll;
HWND hSlow;
HWND hSpin;
HWND hStat;
HWND hWind;
HWND hWire;

BOOL auto_capture;
BOOL wind_capture;
int ticks=1;
int waits=0;
double zoom=1.;

float ambientLight0[ 4]={0.2F,0.2F,0.2F,0.0F};
float diffuseLight0[ 4]={0.3F,0.3F,0.3F,0.0F};
float positionLight0[4]={0.0F,0.0F,5.0F,0.0F};

struct{double a,b,c;}Camera={15.,0.,0.};
struct{double a,b;}Light={0.,45.};
struct{double a,b,c,x,y,z;}Objective={0.,0.,0.,0.,0.,-1.5};

float*body,*hip,*thigh,*calf,*foot;

void ReadTriangles(char*fname)
  {
  char bufr[1024],*ptr;
  int b,c,f,h,l,t;
  VECTOR N,V1,V2,V3;
  FILE*fp;
  if((fp=fopen(fname,"rt"))==NULL)
    exit(1);
  b=h=t=c=f=0;
  while(fgets(bufr,sizeof(bufr),fp))
    {
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%n",&V1.x,&V1.y,&V1.z,&V2.x,&V2.y,&V2.z,&V3.x,&V3.y,&V3.z,&l)==9)
      {
      ptr=bufr+l;
      while(*ptr==' ')
        ptr++;
      if(*ptr=='r') /* body: red */
        b++;
      else if(*ptr=='c') /* hip: cyan */
        h++;
      else if(*ptr=='b') /* thigh: blue */
        t++;
      else if(*ptr=='y') /* calf: yellow */
        c++;
      else if(*ptr=='g') /* foot: green */
        f++;
      }
    }
  if(b<1||h<1||t<1||c<1||f<1)
    exit(1);
  if((body=calloc(12*(b+h+t+c+f)+5,sizeof(float)))==NULL)
    exit(1);
  hip=body+12*b+1;
  thigh=hip+12*h+1;
  calf=thigh+12*t+1;
  foot=calf+12*c+1;
  body[12*b]=hip[12*h]=thigh[12*t]=calf[12*c]=foot[12*f]=FLT_MAX;
  rewind(fp);
  b=h=t=c=f=0;
  while(fgets(bufr,sizeof(bufr),fp))
    {
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%n",&V1.x,&V1.y,&V1.z,&V2.x,&V2.y,&V2.z,&V3.x,&V3.y,&V3.z,&l)==9)
      {
      ptr=bufr+l;
      while(*ptr==' ')
        ptr++;
      if(*ptr=='r') /* body: red */
        {
        body[b++]=(float)V1.x;
        body[b++]=(float)V1.y;
        body[b++]=(float)V1.z;
        body[b++]=(float)V2.x;
        body[b++]=(float)V2.y;
        body[b++]=(float)V2.z;
        body[b++]=(float)V3.x;
        body[b++]=(float)V3.y;
        body[b++]=(float)V3.z;
        N=vNormal(V1,V2,V3);
        body[b++]=(float)N.x;
        body[b++]=(float)N.y;
        body[b++]=(float)N.z;
        }
      else if(*ptr=='c') /* hip: cyan */
        {
        hip[h++]=(float)V1.x;
        hip[h++]=(float)V1.y;
        hip[h++]=(float)V1.z;
        hip[h++]=(float)V2.x;
        hip[h++]=(float)V2.y;
        hip[h++]=(float)V2.z;
        hip[h++]=(float)V3.x;
        hip[h++]=(float)V3.y;
        hip[h++]=(float)V3.z;
        N=vNormal(V1,V2,V3);
        hip[h++]=(float)N.x;
        hip[h++]=(float)N.y;
        hip[h++]=(float)N.z;
        }
      else if(*ptr=='b') /* thigh: blue */
        {
        thigh[t++]=(float)V1.x;
        thigh[t++]=(float)V1.y;
        thigh[t++]=(float)V1.z;
        thigh[t++]=(float)V2.x;
        thigh[t++]=(float)V2.y;
        thigh[t++]=(float)V2.z;
        thigh[t++]=(float)V3.x;
        thigh[t++]=(float)V3.y;
        thigh[t++]=(float)V3.z;
        N=vNormal(V1,V2,V3);
        thigh[t++]=(float)N.x;
        thigh[t++]=(float)N.y;
        thigh[t++]=(float)N.z;
        }
      else if(*ptr=='y') /* calf: yellow */
        {
        calf[c++]=(float)V1.x;
        calf[c++]=(float)V1.y;
        calf[c++]=(float)V1.z;
        calf[c++]=(float)V2.x;
        calf[c++]=(float)V2.y;
        calf[c++]=(float)V2.z;
        calf[c++]=(float)V3.x;
        calf[c++]=(float)V3.y;
        calf[c++]=(float)V3.z;
        N=vNormal(V1,V2,V3);
        calf[c++]=(float)N.x;
        calf[c++]=(float)N.y;
        calf[c++]=(float)N.z;
        }
      else if(*ptr=='g') /* foot: green */
        {
        foot[f++]=(float)V1.x;
        foot[f++]=(float)V1.y;
        foot[f++]=(float)V1.z;
        foot[f++]=(float)V2.x;
        foot[f++]=(float)V2.y;
        foot[f++]=(float)V2.z;
        foot[f++]=(float)V3.x;
        foot[f++]=(float)V3.y;
        foot[f++]=(float)V3.z;
        N=vNormal(V1,V2,V3);
        foot[f++]=(float)N.x;
        foot[f++]=(float)N.y;
        foot[f++]=(float)N.z;
        }
      }
    }
  fclose(fp);
  }

float Twalk;
float Ahip[4];
float Athi[4];
float Acal[4];
float Afot[4];

void Hip(int i)
  {
  float hX,hY;
  hX=(float)(0.465*(cos((Ahip[i]- 25.)*M_PI/180.)-cos( -25.*M_PI/180.)));
  hY=(float)(0.465*(cos((Ahip[i]-115.)*M_PI/180.)-cos(-115.*M_PI/180.)));
  glTranslatef(hX,hY,0.F);
  glRotatef(Ahip[i],0.F,0.F,1.F);
  glColor(0.F,1.F,1.F);
  glNormTriangles(hip);
  }

void Thigh(int i)
  {
  float tX,tY;
  tX=(float)( 0.728*(sin((Athi[i]-286.)*M_PI/180.)-sin(-286.*M_PI/180.)));
  tY=(float)(-0.728*(cos((Athi[i]-286.)*M_PI/180.)-cos(-286.*M_PI/180.)));
  glTranslatef(tX,tY,0.F);
  glRotatef(Athi[i],0.F,0.F,1.F);
  glColor(0.F,1.F,0.F);
  glNormTriangles(thigh);
  }

void Calf(int i)
  {
  float cX,cY;
  cX=(float)( 0.753*(sin((Acal[i]-249.)*M_PI/180.)-sin(-249.*M_PI/180.)));
  cY=(float)(-0.753*(cos((Acal[i]-249.)*M_PI/180.)-cos(-249.*M_PI/180.)));
  glTranslatef(cX,cY,0.F);
  glRotatef(Acal[i],0.F,0.F,1.F);
  glColor(1.F,1.F,0.F);
  glNormTriangles(calf);
  }

void Foot(int i)
  {
  float fX,fY;
  fX=(float)( 0.928*(sin((Afot[i]-229.)*M_PI/180.)-sin(-229.*M_PI/180.)));
  fY=(float)(-0.928*(cos((Afot[i]-229.)*M_PI/180.)-cos(-229.*M_PI/180.)));
  glTranslatef(fX,fY,0.F);
  glRotatef(Afot[i],0.F,0.F,1.F);
  glColor(0.F,0.F,1.F);
  glNormTriangles(foot);
  }

void RenderWalker()
  {
  int i;
  float pX,pY;
  float dX[4]={0.F,0.000F,-0.473F,-0.473F};
  float dZ[4]={0.F,0.309F, 0.000F,-0.309F};
  float dW[4]={0.F,0.423F, 0.000F,-0.423F};
  float yA[4]={0.F,0.F,180.F,180.F};
  glColor(1.F,0.F,0.F);
  glNormTriangles(body);
  for(i=0;i<4;i++)
    {
    glPushMatrix();
    glTranslatef(dX[i],0.F,dZ[i]);
    glRotatef(yA[i],0.F,1.F,0.F);
    Hip(i);
    glPopMatrix();
    pX=(float)( 0.275*(cos(Ahip[i]*M_PI/180.)-1.));if(i<2)pX=-pX;
    pY=(float)(-0.275* sin(Ahip[i]*M_PI/180.)    );
    glPushMatrix();
    glTranslatef(dX[i]+pX,pY,dW[i]);
    glRotatef(yA[i],0.F,1.F,0.F);
    Thigh(i);
    Calf(i);
    Foot(i);
    glPopMatrix();
    }
  }

void glRepaint()
  {
  char fname[FILENAME_MAX];
  static int frame;
  if(!pDC)
    return;
  glClearColor(0.F,0.F,0.F,0.F);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glShadeModel(GL_SMOOTH);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,TRUE);
  glDisable(GL_AUTO_NORMAL);
  glEnable(GL_NORMALIZE);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0,GL_AMBIENT ,ambientLight0);
  glLightfv(GL_LIGHT0,GL_DIFFUSE ,diffuseLight0);
    positionLight0[0]=(float)(sin(M_PI*Light.a/180)*cos(M_PI*Light.b/180));
    positionLight0[1]=(float)(sin(M_PI*Light.b/180));
    positionLight0[2]=(float)(cos(M_PI*Light.a/180)*cos(M_PI*Light.b/180));
    positionLight0[3]=0;/* infinite distance to light source */
  glLightfv(GL_LIGHT0,GL_POSITION,positionLight0);
  glEnable(GL_LIGHT0);
  glDisable(GL_TEXTURE_2D);
  glTranslated(Objective.x,Objective.y,Objective.z);
  glRotated(Objective.a,1,0,0);
  glRotated(Objective.b,0,1,0);
  glRotated(Objective.c,0,0,1);
  WindowText(hStat,"Camera=%lG,%lG,%lG, Light=%lG,%lG, Objective=%lG,%lG,%lG,%lG,%lG,%lG, Rot=%.0hf,%.0hf,%.0hf,%.0hf",
    Camera.a,Camera.b,Camera.c,Light.a,Light.b,Objective.a,Objective.b,Objective.c,Objective.x,Objective.y,Objective.z,Ahip[0],Athi[0],Acal[0],Afot[0]);
  glPolygonMode(back?GL_FRONT_AND_BACK:GL_FRONT,fill?GL_FILL:GL_LINE);
  glScaled(zoom,zoom,zoom);
  RenderWalker();
  glFinish();
  SwapBuffers(pDC);
  if(auto_capture)
    {
    sprintf(fname,"WALKER%05i.BMP",frame++);
    CaptureScreen(wind_capture,fname);
    }
  }

void PositionWindows()
  {
  int h,t,w,x,y;
  RECT rc;
  if(IsIconic(hMain))
    return;
  if(!hPlot)
    return;
  GetClientRect(hMain,&rc);
  t=GetSystemMetrics(SM_CYCAPTION);
  w=bLogo->biWidth;
  h=bLogo->biHeight;
  x=rc.right-1-w;
  y=1;
  MoveWindow(hLogo,x,y,w,h,TRUE);y+=h+2;
  MoveWindow(hSpin,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hRoll,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hFast,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hSlow,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hAnim,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWire,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hCapt,x,y,w,t,TRUE);y+=t+2;
  MoveWindow(hWind,x,y,w,t,TRUE);
  y=rc.bottom-t;
  y-=t+1;MoveWindow(hExit,x,y,w,t,TRUE);
  y-=t+1;MoveWindow(hAbou,x,y,w,t,TRUE);
  MoveWindow(hStat,0,rc.bottom-t,rc.right,t,TRUE);
  w=rc.right-w-2;
  h=rc.bottom-t;
  MoveWindow(hPlot,2,2,w-4,h-4,TRUE);
  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if(w>h)
    {
    float a=(float)(w/(h+0.01));
    glOrtho(-a,a,-1,1,Camera.a,Camera.c);
    }
  else
    {
    float a=(float)(h/(w+0.01));
    glOrtho(-1,1,-a,a,Camera.a,Camera.c);
    }
  glTranslated(0.,0.,-Camera.b);
  }

int IsCommand(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_EXIT||wParam==PUSH_ALT_F4)
    {
    PostQuitMessage(0);
    return(TRUE);
    }
  if(wParam==PUSH_ABOUT)
    {
    DialogBox(hInst,"ABOUT",hMain,(DLGPROC)AboutProc);
    return(TRUE);
    }
  if(wParam==PUSH_ANIM)
    {
    animate=1-animate;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    return(TRUE);
    }
  if(wParam==PUSH_CAPT)
    {
    auto_capture=1-auto_capture;
    SendMessage(hCapt,BM_SETCHECK,auto_capture,0);
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_FAST)
    {
    ticks--;
    if(ticks<0)
      ticks=0;
    return(TRUE);
    }
  if(wParam==PUSH_ROLL)
    {
    int i;
    double a;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    a=Objective.a;
    for(i=0;i<360;i+=5)
      {
      Objective.a=i;
      glRepaint();
      Sleep(25);
      }
    Objective.a=a;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SLOW)
    {
    ticks++;
    return(TRUE);
    }
  if(wParam==PUSH_SPIN)
    {
    int i;
    double b;
    animate=FALSE;
    SendMessage(hAnim,BM_SETCHECK,animate,0);
    b=Objective.b;
    for(i=0;i<360;i+=5)
      {
      Objective.b=i;
      glRepaint();
      Sleep(25);
      }
    Objective.b=b;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_WIND)
    {
    wind_capture=1-wind_capture;
    SendMessage(hWind,BM_SETCHECK,!wind_capture,0);
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_WIRE)
    {
    fill=1-fill;
    SendMessage(hWire,BM_SETCHECK,!fill,0);
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

void ResetView()
  {
  Light.a= 0.;
  Light.b=45.;
  Objective.a=345.;
  Objective.b=130.;
  Objective.c=0.;
  Objective.x=0.;
  Objective.y=0.;
  Objective.z=0.;
  zoom=1.;
  }

int IsOrientation(WPARAM wParam,LPARAM lParam)
  {
  if(wParam==PUSH_UP)
    {
    Objective.a+=5;
    if(Objective.a>359)
      Objective.a=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_DOWN)
    {
    Objective.a-=5;
    if(Objective.a<0)
      Objective.a=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_LEFT)
    {
    Objective.b-=5;
    if(Objective.b<0)
      Objective.b=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_RIGHT)
    {
    Objective.b+=5;
    if(Objective.b>359)
      Objective.b=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_PGUP)
    {
    zoom/=0.95;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_PGDN)
    {
    zoom*=0.95;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_UP)
    {
    Objective.z+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_DOWN)
    {
    Objective.z-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_LEFT)
    {
    Objective.c-=5;
    if(Objective.c<0)
      Objective.c=355;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_SHIFT_RIGHT)
    {
    Objective.c+=5;
    if(Objective.c>359)
      Objective.c=0;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_UP)
    {
    Objective.y+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_DOWN)
    {
    Objective.y-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_LEFT)
    {
    Objective.x-=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_RIGHT)
    {
    Objective.x+=0.05;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_PGUP)
    {
    Acal[0]-=5.F;
    Acal[1]-=5.F;
    Acal[2]-=5.F;
    Acal[3]-=5.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_ALT_PGDN)
    {
    Acal[0]+=5.F;
    Acal[1]+=5.F;
    Acal[2]+=5.F;
    Acal[3]+=5.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_UP)
    {
    Ahip[0]-=5.F;
    Ahip[1]-=5.F;
    Ahip[2]-=5.F;
    Ahip[3]-=5.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_DOWN)
    {
    Ahip[0]+=5.F;
    Ahip[1]+=5.F;
    Ahip[2]+=5.F;
    Ahip[3]+=5.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_LEFT)
    {
    Athi[0]-=5.F;
    Athi[1]-=5.F;
    Athi[2]-=5.F;
    Athi[3]-=5.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_RIGHT)
    {
    Athi[0]+=5.F;
    Athi[1]+=5.F;
    Athi[2]+=5.F;
    Athi[3]+=5.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_PGUP)
    {
    Afot[0]-=5.F;
    Afot[1]-=5.F;
    Afot[2]-=5.F;
    Afot[3]-=5.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_PGDN)
    {
    Afot[0]+=5.F;
    Afot[1]+=5.F;
    Afot[2]+=5.F;
    Afot[3]+=5.F;
    glRepaint();
    return(TRUE);
    }
  if(wParam==PUSH_CTL_Z)
    {
    ResetView();
    glRepaint();
    return(TRUE);
    }
  return(FALSE);
  }

LRESULT WINAPI MainProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    {
    PostQuitMessage(0);
    return(FALSE);
    }
  if(wMsg==WM_COMMAND)
    {
    if(IsCommand(wParam&0xFFFF,lParam))
      {
      SetFocus(hMain);
      return(FALSE);
      }
    if(IsOrientation(wParam&0xFFFF,lParam))
      return(FALSE);
    return(FALSE);
    }
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_CTLCOLORBTN)
    {
    SetTextColor((HDC)wParam,BLACK);
    SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_GETMINMAXINFO)
    {
    MINMAXINFO*mInfo=(MINMAXINFO*)lParam;
    mInfo->ptMinTrackSize.x=640;
    mInfo->ptMinTrackSize.y=480;
    return(FALSE);
    }
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  if(wMsg==WM_SIZE)
    {
    PositionWindows();
    glRepaint();
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

void WINAPI Animate(HWND hWnd,WPARAM wParam,DWORD idEvent,int uTime)
  {
  int i;
  if(!animate)
    return;
  if(waits)
    {
    waits--;
    return;
    }
  Twalk+=5.;
  for(i=0;i<4;i++)
    {
    Ahip[i]=-(float)(10.* sin((Twalk-i*90.)*M_PI/180.)-5. );
    Athi[i]= (float)(20.*(cos((Twalk-i*90.)*M_PI/180.)-1.));
    Acal[i]= (float)(20.* cos((Twalk-i*90.)*M_PI/180.)    );
    Afot[i]=-Athi[i]-Acal[i];
    }
  waits=ticks;
  glRepaint();
  }

void CreateWindows()
  {
  int h,high,w,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  h=7*high/8;
  w=7*wide/8;
  if((hMain=CreateWindow("MAIN","use directional keys + alt/ctl/shift to view",
      WS_OVERLAPPEDWINDOW|WS_VISIBLE,(wide-w)/2,(high-h)/2,w,h,
      NULL,LoadMenu(hInst,"MENU"),hInst,NULL))==NULL)
    Abort(__LINE__,"can't create main window\nWindows error code %i",GetLastError());
  if((hStat=CreateWindow(STATUSCLASSNAME,NULL,
      WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create status control\nerror code %i",GetLastError());
  SendMessage(hStat,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  if((hLogo=CreateWindow("LOGO",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWindows error code %i",GetLastError());
  if((hPlot=CreateWindow("PLOT",NULL,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|SS_BLACKRECT,
      0,0,0,0,hMain,0,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create plotting window\nWindows error code %i",GetLastError());
  hAbou=CreatePushButton(PUSH_ABOUT,"About");
  hExit=CreatePushButton(PUSH_EXIT ,"Exit");
  hFast=CreatePushButton(PUSH_FAST ,"Faster");
  hRoll=CreatePushButton(PUSH_ROLL ,"Roll");
  hSlow=CreatePushButton(PUSH_SLOW ,"Slower");
  hSpin=CreatePushButton(PUSH_SPIN ,"Spin");
  hAnim=CreateRadioButton(PUSH_ANIM,"Animate",animate);
  hWire=CreateRadioButton(PUSH_WIRE,"Solid/Wire",!fill);
  hCapt=CreateRadioButton(PUSH_CAPT,"auto capture",auto_capture);
  hWind=CreateRadioButton(PUSH_WIND,"window/all",wind_capture);
  }

void LoadResources()
  {
  bLogo=GetBitmap("logo");
  }

void RegisterClasses()
  {
  Register("MAIN",MainProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0,0);
  Register("LOGO",LogoProc,StockBrush(COLOR_BTNFACE) ,IDC_ARROW,0,0);
  Register("PLOT",PlotProc,GetStockObject(NULL_BRUSH),IDC_CROSS,0,CS_OWNDC);
  }

int WINAPI WinMain(HINSTANCE hCurrent,HINSTANCE hPrevious,char*lCommand,int nShow)
  {
  HACCEL acc;
  MSG msg;
  hInst=hCurrent;
  InitCommonControls();
  acc=LoadAccelerators(hInst,"FAST");
  LoadResources();
  RegisterClasses();
  CreateWindows();
  ReadTriangles("walker.tri");
  OpenGL();
  PositionWindows();
  SetTimer(hMain,1,10,(TIMERPROC)Animate);
  while(GetMessage(&msg,NULL,0,0))
    if(!TranslateAccelerator(msg.hwnd,acc,&msg))
      if(!TranslateMessage(&msg))
        DispatchMessage(&msg);
  return(0);
  }
