#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "colors.h"
#include "model.h"

void Abort(int code,char*format,...)
  {
  va_list arg_marker;
  if(GetConsoleWindow())
    {
    fprintf(stderr,"\nFatal Error Code: %i\n",code);
    va_start(arg_marker,format);
    vfprintf(stderr,format,arg_marker);
    fprintf(stderr,"\n");
    exit(0);
    }
  else
    {
    char bufr[1024],head[32];
    va_start(arg_marker,format);
    vsprintf(bufr,format,arg_marker);
    sprintf(head,"Fatal Error Code %i",code);
    MessageBox(NULL,bufr,head,MB_ICONSTOP|MB_OK);
    ExitProcess(0);
    }
  }

void*allocate(int line,size_t count,size_t size)
  {
  void*ptr;
  if((ptr=(void*)calloc(count,size))==NULL)
    Abort(line,"can't allocate memory %ux%u=%u bytes",count,size,count*size);
  return(ptr);
  }

void*reallocate(int line,void*old_ptr,size_t old_count,size_t new_count,size_t size)
  {
  void*new_ptr;
  if((new_ptr=calloc(new_count,size))==NULL)
    Abort(line,"can't reallocate memory %ux%u=%u -> %lux%u=%lu bytes",old_count,size,old_count*size,new_count,size,new_count*size);
  if(old_ptr&&old_count)
    {
    memcpy(new_ptr,old_ptr,min(old_count,new_count)*size);
    free(old_ptr);
    }
  return(new_ptr);
  }

void*_free(void*ptr)
  {
  if(ptr)
    free(ptr);
  return(NULL);
  }

int nint(float d)
  {
  if(d<0.)
    return((int)(d-0.5));
  if(d>0.)
    return((int)(d+0.5));
  return(0);
  }

char*stristr(char*str1,char*str2)
  {
  int l,l1,l2;
  l1=(int)strlen(str1);
  l2=(int)strlen(str2);
  if(l2<1||l1<l1)
    return(NULL);
  for(l=0;l<=l1-l2;l++)
    if(_memicmp(str1+l,str2,l2)==0)
      return(str1+l);
  return(NULL);
  }

BOOL match(char*txt1,char*txt2)
  {
  if(_memicmp(txt1,txt2,strlen(txt2))==0)
    return(TRUE);
  return(FALSE);
  }

DWORD Text2Color(char*txt)
  {
  DWORD color;
  if(_memicmp(txt,"RGB=0x",6)==0)
    {
    sscanf(txt+4,"%i",&color);
    return(color);
    }
  if(txt[0]=='0'&&txt[1]=='x')
    {
    sscanf(txt,"%i",&color);
    return(color);
    }
  if(match(txt,"almond"))
    return(almond);
  if(match(txt,"antique white"))
    return(antique_white);
  if(match(txt,"aquamarine"))
    return(aquamarine);
  if(match(txt,"azure"))
    return(azure);
  if(match(txt,"beige"))
    return(beige);
  if(match(txt,"black"))
    return(black);
  if(match(txt,"blue"))
    return(blue);
  if(match(txt,"brass"))
    return(brass);
  if(match(txt,"brick red"))
    return(brick_red);
  if(match(txt,"bronze"))
    return(bronze);
  if(match(txt,"brown"))
    return(brown);
  if(match(txt,"burgundy"))
    return(burgundy);
  if(match(txt,"cadet blue"))
    return(cadet_blue);
  if(match(txt,"chartreuse"))
    return(chartreuse);
  if(match(txt,"copper"))
    return(copper);
  if(match(txt,"coral"))
    return(coral);
  if(match(txt,"cyan"))
    return(cyan);
  if(match(txt,"dark brown"))
    return(dark_brown);
  if(match(txt,"dark cyan"))
    return(dark_cyan);
  if(match(txt,"dark gray"))
    return(dark_gray);
  if(match(txt,"dark green"))
    return(dark_green);
  if(match(txt,"dark lavender"))
    return(dark_lavender);
  if(match(txt,"dark magenta"))
    return(dark_magenta);
  if(match(txt,"dark orange"))
    return(dark_orange);
  if(match(txt,"dark red"))
    return(dark_red);
  if(match(txt,"dark yellow"))
    return(dark_yellow);
  if(match(txt,"dodger blue"))
    return(dodger_blue);
  if(match(txt,"forest green"))
    return(forest_green);
  if(match(txt,"gold"))
    return(gold);
  if(match(txt,"grass green"))
    return(grass_green);
  if(match(txt,"gray"))
    return(gray);
  if(match(txt,"green"))
    return(green);
  if(match(txt,"ivory"))
    return(ivory);
  if(match(txt,"khaki"))
    return(khaki);
  if(match(txt,"light blue"))
    return(light_blue);
  if(match(txt,"light brown"))
    return(light_brown);
  if(match(txt,"light cyan"))
    return(light_cyan);
  if(match(txt,"light gray"))
    return(light_gray);
  if(match(txt,"light lavender"))
    return(light_lavender);
  if(match(txt,"light magenta"))
    return(light_magenta);
  if(match(txt,"light orange"))
    return(light_orange);
  if(match(txt,"light red"))
    return(light_red);
  if(match(txt,"light yellow"))
    return(light_yellow);
  if(match(txt,"magenta"))
    return(magenta);
  if(match(txt,"medium gray"))
    return(medium_gray);
  if(match(txt,"medium yellow"))
    return(medium_yellow);
  if(match(txt,"midnight blue"))
    return(midnight_blue);
  if(match(txt,"navy blue"))
    return(navy_blue);
  if(match(txt,"olive green"))
    return(olive_green);
  if(match(txt,"orange"))
    return(orange);
  if(match(txt,"orchid"))
    return(orchid);
  if(match(txt,"pale green"))
    return(pale_green);
  if(match(txt,"peach"))
    return(peach);
  if(match(txt,"powder blue"))
    return(powder_blue);
  if(match(txt,"purple"))
    return(purple);
  if(match(txt,"red"))
    return(red);
  if(match(txt,"royal blue"))
    return(royal_blue);
  if(match(txt,"salmon"))
    return(salmon);
  if(match(txt,"sea green"))
    return(sea_green);
  if(match(txt,"silver"))
    return(silver);
  if(match(txt,"sky blue"))
    return(sky_blue);
  if(match(txt,"slate blue"))
    return(slate_blue);
  if(match(txt,"spring green"))
    return(spring_green);
  if(match(txt,"steel"))
    return(steel);
  if(match(txt,"steel blue"))
    return(steel_blue);
  if(match(txt,"suede"))
    return(suede);
  if(match(txt,"tan"))
    return(tan);
  if(match(txt,"tomato red"))
    return(tomato_red);
  if(match(txt,"turquoise"))
    return(turquoise);
  if(match(txt,"violet"))
    return(violet);
  if(match(txt,"white"))
    return(white);
  if(match(txt,"yellow"))
    return(yellow);
  return(white);
  }

char*Color2Text(DWORD color)
  {
  static char txt[11];
  if(color==almond)
    return("almond");
  if(color==antique_white)
    return("antique white");
  if(color==aquamarine)
    return("aquamarine");
  if(color==azure)
    return("azure");
  if(color==beige)
    return("beige");
  if(color==black)
    return("black");
  if(color==blue)
    return("blue");
  if(color==brass)
    return("brass");
  if(color==brick_red)
    return("brick red");
  if(color==bronze)
    return("bronze");
  if(color==brown)
    return("brown");
  if(color==burgundy)
    return("burgundy");
  if(color==cadet_blue)
    return("cadet blue");
  if(color==chartreuse)
    return("chartreuse");
  if(color==copper)
    return("copper");
  if(color==coral)
    return("coral");
  if(color==cyan)
    return("cyan");
  if(color==dark_brown)
    return("dark brown");
  if(color==dark_cyan)
    return("dark cyan");
  if(color==dark_gray)
    return("dark gray");
  if(color==dark_green)
    return("dark green");
  if(color==dark_lavender)
    return("dark lavender");
  if(color==dark_magenta)
    return("dark magenta");
  if(color==dark_orange)
    return("dark orange");
  if(color==dark_red)
    return("dark red");
  if(color==dark_yellow)
    return("dark yellow");
  if(color==dodger_blue)
    return("dodger blue");
  if(color==forest_green)
    return("forest green");
  if(color==gold)
    return("gold");
  if(color==grass_green)
    return("grass green");
  if(color==gray)
    return("gray");
  if(color==green)
    return("green");
  if(color==ivory)
    return("ivory");
  if(color==khaki)
    return("khaki");
  if(color==light_blue)
    return("light blue");
  if(color==light_brown)
    return("light brown");
  if(color==light_cyan)
    return("light cyan");
  if(color==light_gray)
    return("light gray");
  if(color==light_lavender)
    return("light lavender");
  if(color==light_magenta)
    return("light magenta");
  if(color==light_orange)
    return("light orange");
  if(color==light_red)
    return("light red");
  if(color==light_yellow)
    return("light yellow");
  if(color==magenta)
    return("magenta");
  if(color==medium_gray)
    return("medium gray");
  if(color==medium_yellow)
    return("medium yellow");
  if(color==midnight_blue)
    return("midnight blue");
  if(color==navy_blue)
    return("navy blue");
  if(color==olive_green)
    return("olive green");
  if(color==orange)
    return("orange");
  if(color==orchid)
    return("orchid");
  if(color==pale_green)
    return("pale green");
  if(color==peach)
    return("peach");
  if(color==powder_blue)
    return("powder blue");
  if(color==purple)
    return("purple");
  if(color==red)
    return("red");
  if(color==royal_blue)
    return("royal blue");
  if(color==salmon)
    return("salmon");
  if(color==sea_green)
    return("sea green");
  if(color==silver)
    return("silver");
  if(color==sky_blue)
    return("sky blue");
  if(color==slate_blue)
    return("slate blue");
  if(color==spring_green)
    return("spring green");
  if(color==steel)
    return("steel");
  if(color==steel_blue)
    return("steel blue");
  if(color==suede)
    return("suede");
  if(color==tan)
    return("tan");
  if(color==tomato_red)
    return("tomato red");
  if(color==turquoise)
    return("turquoise");
  if(color==violet)
    return("violet");
  if(color==white)
    return("white");
  if(color==yellow)
    return("yellow");
  sprintf(txt,"RGB=0x%06X",color);
  return(txt);
  }

BYTE Lightness(DWORD rgb)
  {
  BYTE b,g,r;
  DWORD d;
  r=(BYTE)((rgb&0xFF0000)>>16);
  g=(BYTE)((rgb&0x00FF00)>> 8);
  b=(BYTE) (rgb&0x0000FF)     ;
  d=r;
  d+=g;
  d+=b;
  d/=3;
  b=(BYTE)d;
  return(b);
  }

void MoreNodes(MODEL*model,int Nn)
  {
  NODE*old;
  if(model->max_nodes==0)
    {
    model->max_nodes=max(1000,Nn);
    model->node=(NODE*)calloc(model->max_nodes,sizeof(NODE));
    }
  else if(model->nodes+Nn>model->max_nodes)
    {
    old=model->node;
    model->max_nodes=model->nodes+max(1000,Nn);
    model->node=(NODE*)calloc(model->max_nodes,sizeof(NODE));
    memcpy(model->node,old,model->nodes*sizeof(NODE));
    }
  }

void MoreElems(MODEL*model,int Ne)
  {
  ELEM*old;
  if(model->max_elems==0)
    {
    model->max_elems=max(1000,Ne);
    model->elem=(ELEM*)calloc(model->max_elems,sizeof(ELEM));
    }
  else if(model->elems+Ne>model->max_elems)
    {
    old=model->elem;
    model->max_elems=model->elems+max(1000,Ne);
    model->elem=(ELEM*)calloc(model->max_elems,sizeof(ELEM));
    memcpy(model->elem,old,model->elems*sizeof(ELEM));
    }
  }

void MoreMatls(MODEL*model)
  {
  MATL*old;
  if(model->max_matls==0)
    {
    model->max_matls=10;
    model->matl=(MATL*)calloc(model->max_matls,sizeof(MATL));
    }
  else if(model->matls>=model->max_matls)
    {
    old=model->matl;
    model->max_matls=model->matls+10;
    model->matl=(MATL*)calloc(model->max_matls,sizeof(MATL));
    memcpy(model->matl,old,model->matls*sizeof(MATL));
    }
  }

void FreeModel(MODEL*model)
  {
  int t;
  if(model->textures>0)
    {
    for(t=0;t<model->textures;t++)
      free(model->bitmaps[t]);
    free(model->bitmaps);
    model->bitmaps=NULL;
    model->textures=0;
    }
  if(model->nodes>0)
    {
    free(model->elem);
    free(model->matl);
    free(model->node);
    }
  }

int MaterialIndex(MODEL*model,char*material_name)
  {
  int i;
  if(model->matls<1)
    {
    MoreMatls(model);
    strcpy(model->matl[model->matls].name,material_name);
    model->matls+=1;
    return(0);
    }
  for(i=0;i<model->matls;i++)
    if(strcmp(model->matl[i].name,material_name)==0)
      return(i);
  MoreMatls(model);
  strcpy(model->matl[model->matls].name,material_name);
  model->matls+=1;
  return(model->matls-1);
  }
