/*******************************************************************************
  OPENGL OBJECTS AND PROCEDURES
*******************************************************************************/
#include "opengl.h"

void glColor(float r,float g,float b)
  {
  float c[3];
  c[0]=r;
  c[1]=g;
  c[2]=b;
  glMaterialfv(back?GL_FRONT_AND_BACK:GL_FRONT,GL_AMBIENT_AND_DIFFUSE,c);
  }

void glRGB(DWORD rgb)
  {
  float c[3];
  BYTE b,g,r;
  r=(BYTE)((rgb&0xFF0000)>>16);
  g=(BYTE)((rgb&0x00FF00)>> 8);
  b=(BYTE) (rgb&0x0000FF)     ;
  c[0]=(float)(r/255.);
  c[1]=(float)(g/255.);
  c[2]=(float)(g/255.);
  glMaterialfv(back?GL_FRONT_AND_BACK:GL_FRONT,GL_AMBIENT_AND_DIFFUSE,c);
  }

VECTOR Vector(double x,double y,double z)
  {
  static VECTOR V;
  V.x=x;
  V.y=y;
  V.z=z;
  return(V);
  }

VECTOR vNormal(VECTOR V1,VECTOR V2,VECTOR V3)
  {
  double a1,a2,a3,b1,b2,b3,r;
  static VECTOR N;
  a1=V2.x-V1.x;
  a2=V2.y-V1.y;
  a3=V2.z-V1.z;
  b1=V3.x-V1.x;
  b2=V3.y-V1.y;
  b3=V3.z-V1.z;
  N.x=a2*b3-a3*b2;
  N.y=a3*b1-a1*b3;
  N.z=a1*b2-a2*b1;
  r=sqrt(N.x*N.x+N.y*N.y+N.z*N.z);
  if(r>DBL_EPSILON)
    {
    N.x/=r;
    N.y/=r;
    N.z/=r;
    }
  else
    {
    N.x=0.;
    N.y=0.;
    N.z=1.;
    }
  return(N);
  }

void glTriangles(float*tri)
  {
  int i;
  VECTOR N,V1,V2,V3;
  for(i=0;tri[i]!=FLT_MAX;)
    {
    V1.x=tri[i++];
    V1.y=tri[i++];
    V1.z=tri[i++];
    V2.x=tri[i++];
    V2.y=tri[i++];
    V2.z=tri[i++];
    V3.x=tri[i++];
    V3.y=tri[i++];
    V3.z=tri[i++];
    N=vNormal(V1,V2,V3);
    glBegin(GL_POLYGON);
    glNormal3d(N.x,N.y,N.z);
    glVertex3d(V1.x,V1.y,V1.z);
    glVertex3d(V2.x,V2.y,V2.z);
    glVertex3d(V3.x,V3.y,V3.z);
    glEnd();
    }
  }

void glNormTriangles(float*tri)
  {
  int i;
  float x1,x2,x3,xn,y1,y2,y3,yn,z1,z2,z3,zn;
  for(i=0;tri[i]!=FLT_MAX;)
    {
    x1=tri[i++];
    y1=tri[i++];
    z1=tri[i++];
    x2=tri[i++];
    y2=tri[i++];
    z2=tri[i++];
    x3=tri[i++];
    y3=tri[i++];
    z3=tri[i++];
    xn=tri[i++];
    yn=tri[i++];
    zn=tri[i++];
    glBegin(GL_POLYGON);
    glNormal3f(xn,yn,zn);
    glVertex3f(x1,y1,z1);
    glVertex3f(x2,y2,z2);
    glVertex3f(x3,y3,z3);
    glEnd();
    }
  }

void glDisk(double R,int M)
  {
  int m;
  double B;
  VECTOR N,V1,V2,V3;
  if(R<=0.)
    return;
  if(M<3)
    return;
  V1.z=V2.z=V3.z=0.;
  V1.x=0.;
  V1.y=0.;
  N.x=0.;
  N.y=0.;
  N.z=1.;
  glBegin(GL_TRIANGLES);
  glNormal3d(N.x,N.y,N.z);
  for(m=0;m<M;m++)
    {
    B=m*2.*M_PI/M;
    V2.x=R*cos(B);
    V2.y=R*sin(B);
    B=(m+1)*2.*M_PI/M;
    V3.x=R*cos(B);
    V3.y=R*sin(B);
    glVertex3d(V1.x,V1.y,V1.z);
    glVertex3d(V2.x,V2.y,V2.z);
    glVertex3d(V3.x,V3.y,V3.z);
    }
  glEnd();
  }

void glRing(double R,double r,int M)
  {
  int m;
  double B;
  VECTOR N,V1,V2;
  if(R<=0.)
    return;
  if(r<0.)
    return;
  if(r>=R)
    return;
  if(M<3)
    return;
  if(r<=0.)
    {
    glDisk(R,M);
    return;
    }
  V1.z=V2.z=0.;
  N.x=0;
  N.y=0;
  N.z=1;
  glBegin(GL_QUAD_STRIP);
  glNormal3d(N.x,N.y,N.z);
  for(m=0;m<=M;m++)
    {
    B=m*2.*M_PI/M;
    V1.x=r*cos(B);
    V1.y=r*sin(B);
    V2.x=R*cos(B);
    V2.y=R*sin(B);
    glVertex3d(V1.x,V1.y,V1.z);
    glVertex3d(V2.x,V2.y,V2.z);
    }
  glEnd();
  }

void glPolyhedron(VECTOR*V,int*I)
  {
  int i,j,k,l,m;
  VECTOR N;
  m=l=0;
  while(I[l]>=0)
    {
    i=I[l++];
    j=I[l++];
    k=I[l++];
    glBegin(GL_POLYGON);
    N=vNormal(V[i],V[j],V[k]);
    glNormal3d(N.x,N.y,N.z);
    glVertex3d(V[i].x,V[i].y,V[i].z);
    glVertex3d(V[j].x,V[j].y,V[j].z);
    glVertex3d(V[k].x,V[k].y,V[k].z);
    while(I[l]>=0)
      {
      k=I[l++];
      glVertex3d(V[k].x,V[k].y,V[k].z);
      }
    glEnd();
    l++;
    }
  }

void glHexahedron(double W,double H,double D)
  {
  static int Index[]={
    0,1,5,4,-1,1,2,6,5,-1,7,6,2,3,-1,
    4,7,3,0,-1,3,2,1,0,-1,4,5,6,7,-1,-1};
  VECTOR V[8];
  V[0]=Vector(-W/2.,-H/2.,-D/2.);
  V[1]=Vector( W/2.,-H/2.,-D/2.);
  V[2]=Vector( W/2., H/2.,-D/2.);
  V[3]=Vector(-W/2., H/2.,-D/2.);
  V[4]=Vector(-W/2.,-H/2., D/2.);
  V[5]=Vector( W/2.,-H/2., D/2.);
  V[6]=Vector( W/2., H/2., D/2.);
  V[7]=Vector(-W/2., H/2., D/2.);
  glPolyhedron(V,Index);
  }

void glCylinder(double R,double H,int M)
  {
  int m;
  double B;
  VECTOR N1,N2,V1,V2;
  if(R<=0.)
    return;
  if(H<=0.)
    return;
  if(M<3)
    return;
  glBegin(GL_QUAD_STRIP);
  N1.y=N2.y=0.;
  V1.y=-H/2.;
  V2.y= H/2.;
  for(m=0;m<=M;m++)
    {
    B=m*2.*M_PI/M;
    N1.x=N2.x=cos(B);
    N1.z=N2.z=sin(B);
    V1.x=V2.x=R*N1.x;
    V1.z=V2.z=R*N1.z;
    glNormal3d(N1.x,N1.y,N1.z);
    glVertex3d(V1.x,V1.y,V1.z);
    glNormal3d(N2.x,N2.y,N2.z);
    glVertex3d(V2.x,V2.y,V2.z);
    }
  glEnd();
  }

HDC GetBestPixelFormat(HWND hWnd)
  {
  int adjacent_color_bits,bits_per_pixel,color_planes,i,pixel_formats;
  HDC hDC;
  PIXELFORMATDESCRIPTOR pfd;
  struct{int index;DWORD score;}best,this;
  if(sizeof(PIXELFORMATDESCRIPTOR)!=40)
    Abort(__LINE__,"sizeof(PIXELFORMATDESCRIPTOR)<?>40");
  if((hDC=GetDC(hWnd))==NULL)
    Abort(__LINE__,"can't get device context\nWindows error code %i",GetLastError());
  color_planes=max(1,GetDeviceCaps(hDC,PLANES));
  adjacent_color_bits=GetDeviceCaps(hDC,BITSPIXEL);
  bits_per_pixel=color_planes*adjacent_color_bits;
  pixel_formats=DescribePixelFormat(hDC,1,0,NULL);
  if(pixel_formats<1)
    Abort(__LINE__,"device does not support pixel formats");
  for(i=0;i<pixel_formats;i++)
    {
    DescribePixelFormat(hDC,i+1,sizeof(PIXELFORMATDESCRIPTOR),&pfd);
    this.score=0;
    if(pfd.dwFlags&PFD_SUPPORT_OPENGL)
      this.score|=0x1000000;
    if(pfd.iPixelType==PFD_TYPE_RGBA)
      this.score|=0x200000;
    if(!(pfd.dwFlags&(PFD_NEED_PALETTE|PFD_NEED_SYSTEM_PALETTE)))
      this.score|=0x100000;
    if(pfd.dwFlags&PFD_GENERIC_ACCELERATED)
      this.score|=0x20000;
    if(pfd.dwFlags&PFD_GENERIC_FORMAT)
      this.score|=0x10000;
    if(pfd.dwFlags&PFD_DOUBLEBUFFER)
      this.score|=0x1000;
    if(pfd.cDepthBits==64)
      this.score|=0x800;
    else if(pfd.cDepthBits==32)
      this.score|=0x400;
    else if(pfd.cDepthBits==16)
      this.score|=0x200;
    else if(pfd.cDepthBits==8)
      this.score|=0x100;
    if(pfd.cColorBits==32)
      this.score|=0x80;
    else if(pfd.cColorBits==24)
      this.score|=0x40;
    else if(pfd.cColorBits==16)
      this.score|=0x20;
    else if(pfd.cColorBits==8)
      this.score|=0x10;
    if(pfd.cStencilBits==64)
      this.score|=0x8;
    else if(pfd.cStencilBits==32)
      this.score|=0x4;
    else if(pfd.cStencilBits==16)
      this.score|=0x2;
    else if(pfd.cStencilBits==8)
      this.score|=0x1;
    this.index=i+1;
    if(i==0)
      best=this;
    else if(this.score>best.score)
      best=this;
    }
  if(best.score<0x1000000)
    Abort(__LINE__,"GDI does not support OpenGL");
  if(!(best.score&0x1000))
    Abort(__LINE__,"display does not fully support OpenGL");
  if(best.score&PFD_NEED_SYSTEM_PALETTE)
    Abort(__LINE__,"display has insufficient color depth");
  if(!SetPixelFormat(hDC,best.index,&pfd))
    Abort(__LINE__,"can't set pixel format\nWindows error code %i",GetLastError());
  return(hDC);
  }

void OpenGL()
  {
  HGLRC rDC;
  if((pDC=GetBestPixelFormat(hPlot))==0)
    Abort(__LINE__,"can't find best pixel context\nWindows error code %i",GetLastError());
  if((rDC=wglCreateContext(pDC))==0)
    Abort(__LINE__,"can't create OpenGL context\nWindows error code %i",GetLastError());
  if(!wglMakeCurrent(pDC,rDC))
    Abort(__LINE__,"can't make OpenGL context current\nWindows error code %i",GetLastError());
  glRepaint();
  }
