/* read (or write) an AutoCAD 3D Studio file
   written by Dudley J. Benton
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <io.h>
#include <malloc.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "..\\utilities\\model.h"
#include "..\\utilities\\read3ds.h"

MODEL Read3DS(char*fname)
  {
  float B,G,R;
  BYTE b,g,r;
  COLORREF color;
  WORD f,id,l,m,n,ne,Ne,nn,Nn;
  DWORD l1,l2,l3,l4,m0,m1,m2,m3,m4;
  MATL matl;
  FILE*fp;
  struct{float x,y,z;}No;
  struct{WORD i,j,k,l;}Fa;
  struct{float u,v;}Vc;
  static MODEL model;
  memset(&model,0,sizeof(MODEL));
  if(!(fp=fopen(fname,"rb")))
    {
    MessageBox(NULL,"can't open model",fname,MB_ICONSTOP|MB_OK);
    return(model);
    }
  if(fread(&id,sizeof(WORD),1,fp)!=1)
    goto the_end;
  if(id!=_3DS_M3DMAGIC)
    goto the_end;
  if(fread(&m0,sizeof(DWORD),1,fp)!=1)
    goto the_end;
  m0=min(m0,(DWORD)_filelength(_fileno(fp)));
  if(m0<128)
    goto the_end;
  while((m1=ftell(fp))<m0)                                     /* read file */
    {
    if(fread(&id,sizeof(WORD),1,fp)!=1)
      goto the_end;
    if(fread(&l1,sizeof(DWORD),1,fp)!=1)
      goto the_end;
    if(id==_3DS_MDATA)                                   /* read mesh objects */
      {
      while((m2=ftell(fp))<m1+l1)
        {
        if(fread(&id,sizeof(WORD),1,fp)!=1)
          goto the_end;
        if(fread(&l2,sizeof(DWORD),1,fp)!=1)
          goto the_end;
        if(id==_3DS_NAMED_OBJECT)                       /* read block objects */
          {
          memset(&matl,0,sizeof(matl));
          for(l=0;l<sizeof(matl);l++)
            {
            if(fread(&b,1,1,fp)!=1)
              goto the_end;
            matl.name[l]=b;
            if(b==0)
              {
              char*ptr;
              while((ptr=strchr(matl.name,' '))!=NULL)
                strcpy(ptr,ptr+1);
              break;
              }
            }
          while((m3=ftell(fp))<m2+l2)
            {
            if(fread(&id,sizeof(WORD),1,fp)!=1)
              goto the_end;
            if(fread(&l3,sizeof(DWORD),1,fp)!=1)
              goto the_end;
            if(id==_3DS_N_TRI_OBJECT)         /* read triangular mesh objects */
              {
              Nn=Ne=0;
              nn=model.nodes;
              ne=model.elems;
              while((m4=ftell(fp))<m3+l3)
                {
                if(fread(&id,sizeof(WORD),1,fp)!=1)
                  goto the_end;
                if(fread(&l4,sizeof(DWORD),1,fp)!=1)
                  goto the_end;
                if(id==_3DS_POINT_ARRAY)                     /* read vertices */
                  {
                  if(fread(&Nn,sizeof(WORD),1,fp)!=1)
                    goto the_end;
                  MoreNodes(&model,Nn);
                  for(l=0;l<Nn;l++)
                    {
                    if(fread(&No,sizeof(No),1,fp)!=1)
                      goto the_end;
                    model.node[model.nodes].x=-No.y;
                    model.node[model.nodes].y= No.z;
                    model.node[model.nodes].z =No.x;
                    model.nodes++;
                    }
                  }
                else if(id==_3DS_FACE_ARRAY)             /* read face indices */
                  {
                  if(fread(&Ne,sizeof(WORD),1,fp)!=1)
                    goto the_end;
                  MoreElems(&model,Ne);
                  for(l=0;l<Ne;l++)
                    {
                    if(fread(&Fa,sizeof(Fa),1,fp)!=1)
                      goto the_end;
                    model.elem[model.elems].i=nn+Fa.k;
                    model.elem[model.elems].j=nn+Fa.j;
                    model.elem[model.elems].k=nn+Fa.i;
                    model.elem[model.elems].color=color;
                    model.elems++;
                    }
                  }
                else if(id==_3DS_MSH_MAT_GROUP)         /* read face material */
                  {
                  memset(&matl,0,sizeof(matl));
                  for(l=0;l<sizeof(matl);l++)
                    {
                    if(fread(&b,1,1,fp)!=1)
                      goto the_end;
                    matl.name[l]=b;
                    if(b==0)
                      {
                      char*ptr;
                      while((ptr=strchr(matl.name,' '))!=NULL)
                        strcpy(ptr,ptr+1);
                      break;
                      }
                    }
                  m=MaterialIndex(&model,matl.name);
                  if(fread(&n,sizeof(WORD),1,fp)!=1)
                    goto the_end;
                  for(l=0;l<n;l++)
                    {
                    if(fread(&f,sizeof(WORD),1,fp)!=1)
                      goto the_end;
                    model.elem[ne+f].m=m;
                    }
                  }
                else if(id==_3DS_TEX_VERTS)       /* read mapping coordinates */
                  {
                  if(fread(&n,sizeof(WORD),1,fp)!=1)
                    goto the_end;
                  for(l=0;l<n;l++)
                    {
                    if(fread(&Vc,sizeof(Vc),1,fp)!=1)
                      goto the_end;
                    model.node[nn+l].u=Vc.u;
                    model.node[nn+l].v=Vc.v;
                    }
                  }
                else if(fseek(fp,m4+l4,SEEK_SET))
                  goto the_end;
                }
              }
            else if(fseek(fp,m3+l3,SEEK_SET))
              goto the_end;
            }
          }
        else if(id==_3DS_MAP_ENTRY)               /* read material properties */
          {
          while((m3=ftell(fp))<m2+l2)
            {
            if(fread(&id,sizeof(WORD),1,fp)!=1)
              goto the_end;
            if(fread(&l3,sizeof(DWORD),1,fp)!=1)
              goto the_end;
            if(id==_3DS_MAT_NAME)                       /* read material name */
              {
              memset(&matl,0,sizeof(matl));
              for(l=0;l<sizeof(matl);l++)
                {
                if(fread(&b,1,1,fp)!=1)
                  goto the_end;
                matl.name[l]=b;
                if(b==0)
                  {
                  char*ptr;
                  while((ptr=strchr(matl.name,' '))!=NULL)
                    strcpy(ptr,ptr+1);
                  break;
                  }
                }
              }
            else if(id==_3DS_MAT_DIFFUSE)      /* read material diffuse color */
              {
              if(fread(&id,sizeof(WORD),1,fp)!=1)
                goto the_end;
              if(id==_3DS_COLOR_F)        /* floating-point color description */
                {
                if(fread(&R,sizeof(float),1,fp)!=1)
                  goto the_end;
                if(fread(&G,sizeof(float),1,fp)!=1)
                  goto the_end;
                if(fread(&B,sizeof(float),1,fp)!=1)
                  goto the_end;
                r=(BYTE)max(0,min(255,255*R));
                g=(BYTE)max(0,min(255,255*G));
                b=(BYTE)max(0,min(255,255*B));
                color=RGB(b,g,r);
                }
              else if(id==_3DS_COLOR_24)          /* 24-bit color description */
                {
                if(fread(&r,sizeof(BYTE),1,fp)!=1)
                  goto the_end;
                if(fread(&g,sizeof(BYTE),1,fp)!=1)
                  goto the_end;
                if(fread(&b,sizeof(BYTE),1,fp)!=1)
                  goto the_end;
                color=RGB(b,g,r);
                }
              }
            if(fseek(fp,m3+l3,SEEK_SET))
              goto the_end;
            }
          }
        else if(fseek(fp,m2+l2,SEEK_SET))
          goto the_end;
        }
      }
    else if(fseek(fp,m1+l1,SEEK_SET))
      goto the_end;
    }
  fclose(fp);
  return(model);
  the_end:
  fclose(fp);
  model.node=_free(model.node);
  model.elem=_free(model.elem);
  model.matl=_free(model.matl);
  return(model);
  }

void Write3DS(MODEL*model,char*fname)
  {
  MessageBox(NULL,"this feature not yet available","sorry...",MB_ICONSTOP|MB_OK);
  }
