/* read (or write) a TP2 3D View file
   written by Dudley J. Benton
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "..\\utilities\\colors.h"
#include "..\\utilities\\model.h"

MODEL Read3DV(char*fname)
  {
  char bufr[1024];
  int i,j,k,l,m,m1,m2,me,n,ne,nn;
  double x,y,z;
  DWORD color;
  FILE*fp;
  static MODEL model;
  memset(&model,0,sizeof(MODEL));
  if((fp=fopen(fname,"rt"))==NULL)
    {
    MessageBox(NULL,"can't open model",fname,MB_ICONSTOP|MB_OK);
    return(model);
    }
  l=0;
  while(1)
    {
    if(fgets(bufr,sizeof(bufr),fp))
      l++;
    else
      {
      fclose(fp);
      MessageBox(NULL,"unexpected end of file","EOF error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(bufr[0]!='*')
      break;
    }
  if(sscanf(bufr,"%i",&nn)!=1)
    {
    fclose(fp);
    MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
    return(model);
    }
  if(nn<3)
    {
    fclose(fp);
    MessageBox(NULL,"model contains less than 3 nodes","model error",MB_ICONSTOP|MB_OK);
    return(model);
    }
  for(n=0;n<nn;n++)
    {
    while(1)
      {
      if(fgets(bufr,sizeof(bufr),fp))
        l++;
      else
        {
        fclose(fp);
        MessageBox(NULL,"unexpected end of file","EOF error",MB_ICONSTOP|MB_OK);
        return(model);
        }
      if(bufr[0]!='*')
        break;
      }
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf",&x,&y,&z)!=3)
      {
      fclose(fp);
      MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    }
  while(1)
    {
    if(fgets(bufr,sizeof(bufr),fp))
      l++;
    else
      {
      fclose(fp);
      MessageBox(NULL,"unexpected end of file","EOF error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(bufr[0]!='*')
      break;
    }
  if(sscanf(bufr,"%i",&ne)!=1)
    {
    fclose(fp);
    MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
    return(model);
    }
  if(ne<1)
    {
    fclose(fp);
    MessageBox(NULL,"model contains less than 1 element","model error",MB_ICONSTOP|MB_OK);
    return(model);
    }
  for(me=n=0;n<ne;n++)
    {
    while(1)
      {
      if(fgets(bufr,sizeof(bufr),fp))
        l++;
      else
        {
        fclose(fp);
        MessageBox(NULL,"unexpected end of file","EOF error",MB_ICONSTOP|MB_OK);
        return(model);
        }
      if(bufr[0]!='*')
        break;
      }
    if(sscanf(bufr,"%i%*[ ,\t]%i%*[ ,\t]%i%n",&i,&j,&k,&m1)!=3)
      {
      fclose(fp);
      MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(i<1||i>nn)
      {
      fclose(fp);
      MessageBox(NULL,"no such node","model error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(j<1||j>nn)
      {
      fclose(fp);
      MessageBox(NULL,"no such node","model error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(k<1||k>nn)
      {
      fclose(fp);
      MessageBox(NULL,"no such node","model error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    me++;
    while(1)
      {
      m1++;
      while(strchr(", \t",bufr[m1]))
        m1++;
      if(bufr[m1]=='0'&&bufr[m1+1]=='x')
        goto next1;
      if(bufr[m1]<'1'||bufr[m1]>'9')
        goto next1;
      if(sscanf(bufr+m1,"%i%n",&k,&m2)!=1)
        {
        fclose(fp);
        MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
        return(model);
        }
      if(k<1||k>nn)
        {
        fclose(fp);
        MessageBox(NULL,"no such node","model error",MB_ICONSTOP|MB_OK);
        return(model);
        }
      me++;
      m1+=m2;
      }
next1:;
    }
  model.nodes=nn;
  if((model.node=calloc(nn,sizeof(NODE)))==NULL)
    {
    fclose(fp);
    MessageBox(NULL,"can't allocate memory","memory error",MB_ICONSTOP|MB_OK);
    memset(&model,0,sizeof(model));
    return(model);
    }
  model.elems=me;
  if((model.elem=calloc(me,sizeof(ELEM)))==NULL)
    {
    fclose(fp);
    MessageBox(NULL,"can't allocate memory","memory error",MB_ICONSTOP|MB_OK);
    free(model.node);
    memset(&model,0,sizeof(model));
    return(model);
    }
  rewind(fp);
  l=0;
  while(1)
    {
    if(fgets(bufr,sizeof(bufr),fp))
      l++;
    else
      {
      fclose(fp);
      MessageBox(NULL,"unexpected end of file","EOF error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(bufr[0]!='*')
      break;
    }
  if(sscanf(bufr,"%i",&nn)!=1)
    {
    fclose(fp);
    MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
    return(model);
    }
  for(n=0;n<nn;n++)
    {
    while(1)
      {
      if(fgets(bufr,sizeof(bufr),fp))
        l++;
      else
        {
        fclose(fp);
        MessageBox(NULL,"unexpected end of file","EOF error",MB_ICONSTOP|MB_OK);
        return(model);
        }
      if(bufr[0]!='*')
        break;
      }
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf",&x,&y,&z)!=3)
      {
      fclose(fp);
      MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    model.node[n].x=(float)x;
    model.node[n].y=(float)y;
    model.node[n].z=(float)z;
    }
  while(1)
    {
    if(fgets(bufr,sizeof(bufr),fp))
      l++;
    else
      {
      fclose(fp);
      MessageBox(NULL,"unexpected end of file","EOF error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(bufr[0]!='*')
      break;
    }
  if(sscanf(bufr,"%i",&ne)!=1)
    {
    fclose(fp);
    MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
    return(model);
    }
  for(me=n=0;n<ne;n++)
    {
    while(1)
      {
      if(fgets(bufr,sizeof(bufr),fp))
        l++;
      else
        {
        fclose(fp);
        MessageBox(NULL,"unexpected end of file","EOF error",MB_ICONSTOP|MB_OK);
        return(model);
        }
      if(bufr[0]!='*')
        break;
      }
    if(sscanf(bufr,"%i%*[ ,\t]%i%*[ ,\t]%i%n",&i,&j,&k,&m1)!=3)
      {
      fclose(fp);
      MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(i<1||i>nn)
      {
      fclose(fp);
      MessageBox(NULL,"no such node","model error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(j<1||j>nn)
      {
      fclose(fp);
      MessageBox(NULL,"no such node","model error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    if(k<1||k>nn)
      {
      fclose(fp);
      MessageBox(NULL,"no such node","model error",MB_ICONSTOP|MB_OK);
      return(model);
      }
    m=me;
    model.elem[me].i=i-1;
    model.elem[me].j=j-1;
    model.elem[me].k=k-1;
    model.elem[me++].m=-1;
    while(1)
      {
      m1++;
      while(strchr(", \t",bufr[m1]))
        m1++;
      if(bufr[m1]=='0'&&bufr[m1+1]=='x')
        goto next2;
      if(bufr[m1]<'1'||bufr[m1]>'9')
        goto next2;
      j=k;
      if(sscanf(bufr+m1,"%i%n",&k,&m2)!=1)
        {
        fclose(fp);
        MessageBox(NULL,"scan error on number of nodes","scan error",MB_ICONSTOP|MB_OK);
        return(model);
        }
      if(k<1||k>nn)
        {
        fclose(fp);
        MessageBox(NULL,"no such node","model error",MB_ICONSTOP|MB_OK);
        return(model);
        }
      model.elem[me].i=i-1;
      model.elem[me].j=j-1;
      model.elem[me].k=k-1;
      model.elem[me++].m=-1;
      m1+=m2;
      }
next2:
    color=Text2Color(bufr+m1);
    while(m<me)
      model.elem[m++].color=color;
    }
  fclose(fp);
  return(model);
  }

void Write3DV(MODEL*model,char*fname)
  {
  int i;
  DWORD color;
  FILE*fp;
  if((fp=fopen(fname,"wt"))==NULL)
    {
    MessageBox(NULL,"can't create output file",fname,MB_ICONSTOP|MB_OK);
    return;
    }
  fprintf(fp,"%i\n",model->nodes);
  for(i=0;i<model->nodes;i++)
    fprintf(fp,"%hG %hG %hG\n",model->node[i].x,model->node[i].y,model->node[i].z);
  fprintf(fp,"%i\n",model->elems);
  for(i=0;i<model->elems;i++)
    {
    if(model->colors)
      color=model->elem[i].color;
    else if(model->elem[i].m>=0)
      color=white;
    else
      color=model->elem[i].color;
    fprintf(fp,"%i %i %i RGB=0x%06X\n",model->elem[i].i+1,model->elem[i].j+1,model->elem[i].k+1,color);
    }
  fclose(fp);
  }
