/* read (or write) a TP2 3D Triangle (TRI) file
   written by Dudley J. Benton
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "..\\utilities\\colors.h"
#include "..\\utilities\\model.h"

MODEL ReadTRI(char*fname)
  {
  char bufr[1024],*ptr;
  int l;
  double X[3],Y[3],Z[3];
  FILE*fp;
  static MODEL model;
  memset(&model,0,sizeof(MODEL));
  if((fp=fopen(fname,"rt"))==NULL)
    {
    MessageBox(NULL,"can't open model",fname,MB_ICONSTOP|MB_OK);
    return(model);
    }
  while(fgets(bufr,sizeof(bufr),fp))
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%n",X,Y,Z,X+1,Y+1,Z+1,X+2,Y+2,Z+2,&l)==9)
      model.elems++;
  if(model.elems<1)
    {
    fclose(fp);
    MessageBox(NULL,"model contains no elements","model error",MB_ICONSTOP|MB_OK);
    memset(&model,0,sizeof(MODEL));
    return(model);
    }
  if((model.elem=calloc(model.elems,sizeof(ELEM)))==NULL)
    {
    fclose(fp);
    MessageBox(NULL,"can't allocate memory for elements","memory error",MB_ICONSTOP|MB_OK);
    memset(&model,0,sizeof(MODEL));
    return(model);
    }
  model.nodes=model.elems*3;
  if((model.node=calloc(model.nodes,sizeof(NODE)))==NULL)
    {
    fclose(fp);
    MessageBox(NULL,"can't allocate memory for nodes","memory error",MB_ICONSTOP|MB_OK);
    free(model.elem);
    memset(&model,0,sizeof(MODEL));
    return(model);
    }
  rewind(fp);
  model.elems=model.nodes=0;
  while(fgets(bufr,sizeof(bufr),fp))
    {
    if(sscanf(bufr,"%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%*[ ,\t]%lf%n",X,Y,Z,X+1,Y+1,Z+1,X+2,Y+2,Z+2,&l)==9)
      {
      model.node[model.nodes].x=(float)X[0];
      model.node[model.nodes].y=(float)Y[0];
      model.node[model.nodes].z=(float)Z[0];
      model.elem[model.elems].i=model.nodes++;
      model.node[model.nodes].x=(float)X[1];
      model.node[model.nodes].y=(float)Y[1];
      model.node[model.nodes].z=(float)Z[1];
      model.elem[model.elems].j=model.nodes++;
      model.node[model.nodes].x=(float)X[2];
      model.node[model.nodes].y=(float)Y[2];
      model.node[model.nodes].z=(float)Z[2];
      model.elem[model.elems].k=model.nodes++;
      ptr=bufr+l;
      while(*ptr==' ')
        ptr++;
      if(*ptr='\n')
        ptr="blue";
      model.elem[model.elems++].color=Text2Color(ptr);
      }
    }
  fclose(fp);
  return(model);
  }

void WriteTRI(MODEL*model,char*fname)
  {
  int e,i,j,k;
  FILE*fp;
  if((fp=fopen(fname,"wt"))==NULL)
    {
    MessageBox(NULL,"can't create output file",fname,MB_ICONSTOP|MB_OK);
    return;
    }
  for(e=0;e<model->elems;e++)
    {
    i=model->elem[e].i;
    j=model->elem[e].j;
    k=model->elem[e].k;
    fprintf(fp,"%hG,%hG,%hG ",model->node[i].x,model->node[i].y,model->node[i].z);
    fprintf(fp,"%hG,%hG,%hG ",model->node[j].x,model->node[j].y,model->node[j].z);
    fprintf(fp,"%hG,%hG,%hG ",model->node[k].x,model->node[k].y,model->node[k].z);
    fprintf(fp,"%s\n",Color2Text(model->elem[e].color));
    }
  fclose(fp);
  }
