/* read a TP2 3D Triangle (TRI) file and write out
   an Excel CSV (Comma Separated Values) file
   written by Dudley J. Benton
*/
#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <io.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "model.h"

MODEL ReadTRI(char*fname);

void WriteCSV(MODEL*model,char*fname)
  {
  int e,i,j,k;
  FILE*fp;
  if((fp=fopen(fname,"wt"))==NULL)
    {
    MessageBox(NULL,"can't create output file",fname,MB_ICONSTOP|MB_OK);
    return;
    }
  for(e=0;e<model->elems;e++)
    {
    i=model->elem[e].i;
    j=model->elem[e].j;
    k=model->elem[e].k;
    fprintf(fp,"\n");
    fprintf(fp,"%hG,%hG,%hG\n",model->node[i].x,model->node[i].y,model->node[i].z);
    fprintf(fp,"%hG,%hG,%hG\n",model->node[j].x,model->node[j].y,model->node[j].z);
    fprintf(fp,"%hG,%hG,%hG\n",model->node[k].x,model->node[k].y,model->node[k].z);
    }
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  char*try="... try\nTRItoCSV infile.tri\n";
  char fname[FILENAME_MAX],*ptr;
  MODEL model;
  printf("read TRI write CSV (Excel)\n");
  if(argc<2)
    {
    printf("no input file specified%s",try);
    return(__LINE__);
    }
  if(argc>2)
    {
    printf("too many arguments%s",try);
    return(__LINE__);
    }
  printf("input file: %s\n",argv[1]);
  strcpy(fname,argv[1]);
  if((ptr=strstr(fname,".tri"))==NULL)
    {
    printf("input file must have extension TRI%s",try);
    return(__LINE__);
    }
  strcpy(ptr,".csv");
  model=ReadTRI(argv[1]);
  printf("%i nodes\n",model.nodes);
  if(model.nodes<3)
    Abort(__LINE__,"can't read input file");
  printf("%i elements\n",model.elems);
  printf("output file: %s\n",fname);
  if(!_access(fname,0))
    {
    printf("output file must not already exist\n");
    return(__LINE__);
    }
  WriteCSV(&model,fname);
  return(0);
  }
