/*******************************************************************************
  OBJECTS AND PROCEDURES COMMON TO ANY WINDOWS APPLICATION USING OPENGL
*******************************************************************************/
#include "winapp.h"

HINSTANCE hInst; /* instance of main program */
HWND hMain;      /* handle of main program */
HWND hPlot;      /* handle of rendering window */
HWND hLogo;      /* handle of logo window */
HDC pDC;         /* plot window device context */

BOOL animate=FALSE;
BOOL back=TRUE;
BOOL fill=TRUE;

BITMAPINFOHEADER*bLogo;

void Abort(int code,char*format,...)
  {
  char bufr[1024],head[32];
  va_list arg_marker;
  va_start(arg_marker,format);
  vsprintf(bufr,format,arg_marker);
  sprintf(head,"Fatal Error Code %i",code);
  MessageBox(NULL,bufr,head,MB_ICONSTOP|MB_OK);
  _flushall();
  _fcloseall();
  ExitProcess(0);
  }

void CenterWindowOnScreen(HWND hWnd)
  {
  int high,left,top,wide;
  RECT wr;
  GetWindowRect(hWnd,&wr);
  wide=wr.right-wr.left;
  high=wr.bottom-wr.top;
  left=(GetSystemMetrics(SM_CXSCREEN)-wide)/2;
  top=(GetSystemMetrics(SM_CYSCREEN)-high)/2;
  MoveWindow(hWnd,left,top,wide,high,TRUE);
  }

void WindowText(HWND hWnd,char*format,...)
  {
  char bufr[128];
  va_list arg_marker;
  va_start(arg_marker,format);
  if(vsprintf(bufr,format,arg_marker)>=sizeof(bufr))
    FatalAppExit(0,"WindowText(): buffer overflow");
  SetWindowText(hWnd,bufr);
  }

void PaintBitmap(HDC hDC,RECT*rc,BITMAPINFOHEADER*bHead,int stretch)
  {
  int colors,h,H,l,t,w,W;
  BYTE*bits;
  l=rc->left;
  t=rc->top;
  w=rc->right-rc->left;
  h=rc->bottom-rc->top;
  W=bHead->biWidth;
  H=bHead->biHeight;
  if(bHead->biBitCount==24)
    colors=0;
  else
    {
    colors=bHead->biClrUsed;
    if(colors==0)
      colors=1<<bHead->biBitCount;
    }
  bits=((BYTE*)bHead)+sizeof(BITMAPINFOHEADER)+colors*sizeof(DWORD);
  if(stretch)
    StretchDIBits(hDC,l,t,w,h,0,0,W,H,bits,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  else if(w*abs(H)<abs(W)*h)
    StretchDIBits(hDC,l,t+(h-w*abs(H)/abs(W))/2,w,w*abs(H)/abs(W),0,0,W,H,bits,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  else
    StretchDIBits(hDC,l+(w-h*abs(W)/abs(H))/2,t,h*abs(W)/abs(H),h,0,0,W,H,bits,(BITMAPINFO*)bHead,DIB_RGB_COLORS,SRCCOPY);
  }

LRESULT WINAPI LogoProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_DESTROY)
    return(FALSE);
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    RECT rc;
    if((hDC=BeginPaint(hWnd,&ps))==NULL)
      Abort(__LINE__,"can't get device context to begin painting\nWin Err Code %i",GetLastError());
    GetClientRect(hWnd,&rc);
    if(bLogo)
      PaintBitmap(hDC,&rc,bLogo,TRUE);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

LRESULT WINAPI PlotProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_DESTROY)
    return(FALSE);
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    hDC=BeginPaint(hWnd,&ps);
    if(pDC)
      {
      if(hDC!=pDC)
        Abort(__LINE__,"OpenGL device context not preserved");
      glRepaint();
      }
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

LRESULT WINAPI AboutProc(HWND hWnd,DWORD wMsg,WPARAM wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    return(FALSE);
  if(wMsg==WM_COMMAND)
    {
    if(HIWORD(wParam)==BN_CLICKED||HIWORD(wParam)==BN_DOUBLECLICKED)
      return(EndDialog(hWnd,LOWORD(wParam)));
    else
      return(FALSE);
    }
  if(wMsg==WM_CTLCOLORDLG)
    return((DWORD)StockBrush(COLOR_BTNFACE));
  if(wMsg==WM_CTLCOLORSTATIC)
    {
    SetBkMode((HDC)wParam,TRANSPARENT);
    SetTextColor((HDC)wParam,BLACK);
    return((DWORD)StockBrush(COLOR_BTNFACE));
    }
  if(wMsg==WM_DESTROY)
    return(TRUE);
  if(wMsg==WM_INITDIALOG)
    {
    CenterWindowOnScreen(hWnd);
    BringWindowToTop(hWnd);
    return(TRUE);
    }
  return(FALSE);
  }

HWND CreateRadioButton(DWORD ID,char*text,BOOL check)
  {
  static HWND hWnd;
  if((hWnd=CreateWindow("BUTTON",text,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|BS_RADIOBUTTON,
      0,0,0,0,hMain,(void*)ID,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create radio button \"%s\"\nerror code %i",text,GetLastError());
  SendMessage(hWnd,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  SendMessage(hWnd,BM_SETCHECK,check,0);
  return(hWnd);
  }

HWND CreatePushButton(DWORD ID,char*text)
  {
  static HWND hWnd;
  if((hWnd=CreateWindow("BUTTON",text,
      WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE|BS_PUSHBUTTON,
      0,0,0,0,hMain,(void*)ID,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create pushbutton \"%s\"\nerror code %i",text,GetLastError());
  SendMessage(hWnd,WM_SETFONT,(DWORD)GetStockObject(ANSI_VAR_FONT),0);
  EnableWindow(hWnd,ID!=0&&text!=NULL);
  return(hWnd);
  }

void*GetResource(char*rName,char*type)
  {
  void*rLock;
  HGLOBAL rLoad;
  HRSRC rFind;
  if((rFind=FindResource(hInst,rName,type))==NULL)
    Abort(__LINE__,"can't find resource %s\nWindows error code %i",rName,GetLastError());
  if((rLoad=LoadResource(hInst,rFind))==NULL)
    Abort(__LINE__,"can't load resource %s\nWindows error code %i",rName,GetLastError());
  if((rLock=LockResource(rLoad))==NULL)
    Abort(__LINE__,"can't lock resource %s\nWindows error code %i",rName,GetLastError());
  return(rLock);
  }

BITMAPINFOHEADER*GetBitmap(char*name)
  {
  return((BITMAPINFOHEADER*)((BYTE*)GetResource(name,(char*)RT_BITMAP)));
  }

void Register(char*cName,void*wProc,HBRUSH hBrush,char*dCursor,int iExtra,DWORD style)
  {
  WNDCLASS wc;
  memset(&wc,0,sizeof(WNDCLASS));
  wc.hInstance    =hInst;
  wc.hIcon        =LoadIcon(hInst,"ICON");
  wc.style        =CS_HREDRAW|CS_VREDRAW|style;
  wc.lpszClassName=cName;
  wc.lpfnWndProc  =(WNDPROC)wProc;
  wc.hbrBackground=hBrush;
  wc.hCursor      =LoadCursor(NULL,dCursor);
  wc.cbWndExtra   =iExtra;
  if(!RegisterClass(&wc))
    Abort(__LINE__,"can't register '%s' class\nWindows error code %i",cName,GetLastError());
  }

void CaptureScreen(BOOL main,char*fname)
  {
  int h,high,i,j,w,wide,width;
  BITMAP bm;
  BITMAPINFO bi;
  BITMAPFILEHEADER bf;
  BYTE a,b,*bits,g,r;
  FILE*fp;
  HBITMAP hBM,hOld;
  HDC cDC,wDC;
  RECT rw;
  if(main)
    {
    wDC=GetWindowDC(hMain);
    GetWindowRect(hMain,&rw);
    }
  else
    {
    wDC=GetWindowDC(hPlot);
    GetWindowRect(hPlot,&rw);
    }
  high=rw.bottom-rw.top;
  wide=rw.right-rw.left;
  if(wDC==NULL)
    Abort(__LINE__,"can't get window device context\nWindows error code %i",GetLastError());
  if(!(hBM=CreateCompatibleBitmap(wDC,wide,high)))
    Abort(__LINE__,"can't create compatible bitmap\nWindows error code %i",GetLastError());
  if(GetObject(hBM,sizeof(BITMAP),&bm)!=sizeof(BITMAP))
    Abort(__LINE__,"can't get bitmap\nWindows error code %i",GetLastError());
  if(!(bm.bmBitsPixel==24||bm.bmBitsPixel==32))
    Abort(__LINE__,"compatible bitmap not 24 or 32 bits/pixel (%i)",bm.bmBitsPixel);
  if(!(cDC=CreateCompatibleDC(wDC)))
    Abort(__LINE__,"can't create compatible device context\nWindows error code %i",GetLastError());
  if((hOld=SelectObject(cDC,hBM))==NULL)
    Abort(__LINE__,"can't select bitmap into device context\nWindows error code %i",GetLastError());
  if(!BitBlt(cDC,0,0,wide,high,wDC,0,0,SRCCOPY))
    Abort(__LINE__,"BitBlt failed\nWindows error code %i",GetLastError());
  if((bits=calloc(bm.bmWidthBytes*bm.bmHeight,sizeof(BYTE)))==NULL)
    Abort(__LINE__,"can't allocate memory for bitmap");
  memset(&bi,0,sizeof(BITMAPINFO));
  bi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  bi.bmiHeader.biWidth=wide;
  bi.bmiHeader.biHeight=high;
  bi.bmiHeader.biPlanes=1;
  bi.bmiHeader.biSizeImage=bm.bmWidthBytes*bm.bmHeight;
  bi.bmiHeader.biBitCount=bm.bmBitsPixel;
  if(GetDIBits(cDC,hBM,0,high,bits,&bi,DIB_RGB_COLORS)!=high)
    Abort(__LINE__,"can't get bitiap bits\nWindows error code %i",GetLastError());
  memset(&bf,0,sizeof(BITMAPFILEHEADER));
  bf.bfType=0x4D42;
  bf.bfOffBits=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);
  bf.bfSize=bf.bfOffBits+bi.bmiHeader.biSizeImage;
  if((fp=fopen(fname,"wb"))==NULL)
    Abort(__LINE__,"can't create capture file %s",fname);
  if(fwrite(&bf,sizeof(BITMAPFILEHEADER),1,fp)!=1)
    Abort(__LINE__,"can't write bitmap file header");
  width=wide*3;
  while(width%4)
    width++;
  bi.bmiHeader.biSizeImage=width*high;
  bi.bmiHeader.biBitCount=24;
  if(fwrite(&bi.bmiHeader,sizeof(BITMAPINFOHEADER),1,fp)!=1)
    Abort(__LINE__,"can't write bitmap info header");
  if(bm.bmBitsPixel==24)
    {
    if(fwrite(bits,sizeof(BYTE),width*high,fp)!=width*high)
      Abort(__LINE__,"can't write bitmap bits");
    }
  else
    {
    BYTE*row;
    if((row=calloc(width,sizeof(BYTE)))==NULL)
      Abort(__LINE__,"can't allocate memory for row");
    for(h=0;h<high;h++)
      {
      i=bm.bmWidthBytes*h;
      for(w=j=0;w<wide;w++)
        {
        b=bits[i++];
        g=bits[i++];
        r=bits[i++];
        a=bits[i++];
        row[j++]=b;
        row[j++]=g;
        row[j++]=r;
        }
      if(fwrite(row,sizeof(BYTE),width,fp)!=width)
        Abort(__LINE__,"can't write bitmap row");
      }
    free(row);
    }
  fclose(fp);
  free(bits);
  if(SelectObject(cDC,hOld)!=hBM)
    Abort(__LINE__,"can't deselect bitmap from device context\nWindows error code %i",GetLastError());
  DeleteObject(hBM);
  DeleteDC(cDC);
  if(main)
    ReleaseDC(hMain,wDC);
  else
    ReleaseDC(hPlot,wDC);
  }
