#ifdef __SC__
typedef   signed short int __int16;
#endif
typedef unsigned long long QWORD;

enum{
  MZ_ZIP_MAX_IO_BUF_SIZE=64*1024,
  MZ_ZIP_MAX_ARCHIVE_FILENAME_SIZE=260,
  MZ_ZIP_MAX_ARCHIVE_FILE_COMMENT_SIZE=256
  };

typedef enum{
  MZ_ZIP_MODE_INVALID=0,
  MZ_ZIP_MODE_READING=1,
  MZ_ZIP_MODE_WRITING=2,
  MZ_ZIP_MODE_WRITING_HAS_BEEN_FINALIZED=3
  }mz_zip_mode;

enum{
  MZ_DEFAULT_STRATEGY=0,
  MZ_FILTERED=1,
  MZ_HUFFMAN_ONLY=2,
  MZ_RLE=3,
  MZ_FIXED=4
  };

enum{
  MZ_NO_COMPRESSION=0,
  MZ_BEST_SPEED=1,
  MZ_BEST_COMPRESSION=9,
  MZ_UBER_COMPRESSION=10,
  MZ_DEFAULT_LEVEL=6,
  MZ_DEFAULT_COMPRESSION=-1
  };

typedef size_t (*mz_file_read_func )(void*pOpaque, QWORD file_ofs, void*pBuf, size_t n);
typedef size_t (*mz_file_write_func)(void*pOpaque, QWORD file_ofs, void*pBuf, size_t n);
typedef void  *(*mz_alloc_func     )(void*opaque, size_t items, size_t size);
typedef void   (*mz_free_func      )(void*opaque, void*address);
typedef void  *(*mz_realloc_func   )(void*opaque, void*address, size_t items, size_t size);

typedef struct
  {
  DWORD m_file_index;
  DWORD m_central_dir_ofs;
  WORD m_version_made_by;
  WORD m_version_needed;
  WORD m_bit_flag;
  WORD m_method;
  time_t m_time;
  DWORD m_crc32;
  QWORD m_comp_size;
  QWORD m_uncomp_size;
  WORD m_internal_attr;
  DWORD m_external_attr;
  QWORD m_local_header_ofs;
  DWORD m_comment_size;
  char m_filename[MZ_ZIP_MAX_ARCHIVE_FILENAME_SIZE];
  char m_comment[MZ_ZIP_MAX_ARCHIVE_FILE_COMMENT_SIZE];
  }mz_zip_archive_file_stat;

typedef struct
  {
  void*m_p;
  size_t m_size, m_capacity;
  DWORD m_element_size;
  }mz_zip_array;

struct mz_internal_state;
struct mz_zip_internal_state_tag;
typedef struct mz_zip_internal_state_tag mz_zip_internal_state;

struct mz_zip_internal_state_tag
  {
  mz_zip_array m_central_dir;
  mz_zip_array m_central_dir_offsets;
  mz_zip_array m_sorted_central_dir_offsets;
  FILE *m_pFile;
  void*m_pMem;
  size_t m_mem_size;
  size_t m_mem_capacity;
  };

typedef struct mz_zip_archive_tag
  {
  QWORD m_archive_size;
  QWORD m_central_directory_file_ofs;
  DWORD m_total_files;
  mz_zip_mode m_zip_mode;
  DWORD m_file_offset_alignment;
  mz_alloc_func m_pAlloc;
  mz_free_func m_pFree;
  mz_realloc_func m_pRealloc;
  void*m_pAlloc_opaque;
  mz_file_read_func m_pRead;
  mz_file_write_func m_pWrite;
  void*m_pIO_opaque;
  mz_zip_internal_state *m_pState;
  }mz_zip_archive;

typedef struct mz_stream_s
  {
  unsigned char*next_in;
  unsigned int avail_in;
  DWORD total_in;
  unsigned char*next_out;
  unsigned int avail_out;
  DWORD total_out;
  char*msg;
  struct mz_internal_state *state;
  mz_alloc_func zalloc;
  mz_free_func zfree;
  void*opaque;
  int data_type;
  DWORD adler;
  DWORD reserved;
  }mz_stream;

int   mz_zip_add_mem_to_archive_file_in_place(char*pZip_filename, char*pArchive_name, void*pBuf, size_t buf_size, void*pComment, WORD comment_size, DWORD level_and_flags);
int   mz_zip_reader_end(mz_zip_archive *pZip);
int   mz_zip_reader_file_stat(mz_zip_archive *pZip, DWORD file_index, mz_zip_archive_file_stat *pStat);
int   mz_zip_reader_init_file(mz_zip_archive *pZip, char*pFilename, DWORD flags);
int   mz_zip_reader_init_mem(mz_zip_archive *pZip,void*pMem,size_t size,DWORD flags);
int   mz_zip_reader_is_file_a_directory(mz_zip_archive *pZip, DWORD file_index);
int   mz_zip_writer_add_mem(mz_zip_archive *pZip,char*pArchive_name,void*pBuf,size_t buf_size,DWORD level_and_flags);
int   mz_zip_writer_end(mz_zip_archive *pZip);
int   mz_zip_writer_finalize_heap_archive(mz_zip_archive *pZip,void**pBuf,size_t *pSize);
int   mz_zip_writer_init_heap(mz_zip_archive *pZip,size_t size_to_reserve_at_beginning,size_t initial_allocation_size);
void  mz_zip_time_to_dos_time(time_t time, WORD *pDOS_time, WORD *pDOS_date);
void *mz_zip_reader_extract_file_to_heap(mz_zip_archive *pZip, char*pFilename, size_t *pSize, DWORD flags);
void *mz_zip_reader_extract_to_heap(mz_zip_archive *pZip, DWORD file_index, size_t *pSize, DWORD flags);
DWORD mz_zip_reader_get_num_files(mz_zip_archive *pZip);

typedef struct{
  int deleting;
  time_t time;
  size_t uncomp_size;
  size_t comp_size;
  }THIS_FILE;
extern THIS_FILE this_file;

HANDLE FindFirstZip(mz_zip_archive*zip,char*lpFileName,WIN32_FIND_DATAA*lpFindData);
BOOL FindNextZip(HANDLE hFind,WIN32_FIND_DATAA*lpFindData);
BOOL FindCloseZip(HANDLE hFind);
BOOL IsFileMatch(char*fname,char*wild);
char*GetNameZip(HANDLE hFind,WIN32_FIND_DATAA*lpFindData);

typedef struct{
  BYTE*data;
  DWORD size;
  }BLOCK_DATA;

typedef struct{
  char name[FILENAME_MAX];
  BLOCK_DATA data;
  DWORD position;
  time_t time;
  }ZILE;

ZILE  *zopen(mz_zip_archive*zip,char*name,char*access);
time_t ztime(ZILE*zp);
DWORD  zsize(ZILE*zp);
void   zrewind(ZILE*zp);
int    zseek(ZILE*zp,long offset,int origin);
long   ztell(ZILE*zp);
char  *zgets(char*bufr,int n,ZILE*zp);
size_t zread(void*bufr,size_t size,size_t count,ZILE*zp);
size_t zwrite(void*bufr,size_t size,size_t count,ZILE*zp);
int    zclose(ZILE*zp);
