/*
 * jcapimin.c
 *
 * Copyright (C) 1994-1998, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains application interface code for the compression half
 * of the JPEG library.  These are the "minimum" API routines that may be
 * needed in either the normal full-compression case or the transcoding-only
 * case.
 *
 * Most of the routines intended to be called directly by an application
 * are in this file or in jcapistd.c.  But also see jcparam.c for
 * parameter-setup helper routines, jcomapi.c for routines shared by
 * compression and decompression, and jctrans.c for the transcoding case.
 */

/*
 * Small JPEG Decoder Library v0.93b
 * Last updated: Dec. 28, 2001 
 * Copyright (C) 1994-2000 Rich Geldreich
 * richgel@voicenet.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <setjmp.h>

#include "bitmap.h"

#define JPEG_INTERNALS
#define HAVE_PROTOTYPES
#define HAVE_UNSIGNED_CHAR
#define HAVE_UNSIGNED_SHORT
#undef CHAR_IS_UNSIGNED
#define HAVE_STDDEF_H
#define HAVE_STDLIB_H
#undef NEED_BSD_STRINGS
#undef NEED_SYS_TYPES_H
#undef NEED_FAR_POINTERS
#undef NEED_SHORT_EXTERNAL_NAMES
#undef INCOMPLETE_TYPES_BROKEN
#ifdef JPEG_INTERNALS
#undef RIGHT_SHIFT_IS_UNSIGNED
#endif
#ifdef JPEG_CJPEG_DJPEG
#endif
#define JCONFIG_INCLUDED
#ifdef HAVE_STDDEF_H
#endif
#ifdef HAVE_STDLIB_H
#endif
#ifdef NEED_SYS_TYPES_H
#endif
#ifdef NEED_BSD_STRINGS
#else
#define MEMZERO(target,size) memset((void *)(target), 0, (size_t)(size))
#define MEMCOPY(dest,src,size) memcpy((void *)(dest), (const void *)(src), (size_t)(size))
#endif
#define SIZEOF(object) ((size_t) sizeof(object))
#define JFREAD(file,buf,sizeofbuf)    ((size_t) fread((void *) (buf), (size_t) 1, (size_t) (sizeofbuf), (file)))
#define JFWRITE(file,buf,sizeofbuf)    ((size_t) fwrite((const void *) (buf), (size_t) 1, (size_t) (sizeofbuf), (file)))
#ifndef JPEGLIB_H
#define JPEGLIB_H
#ifndef JCONFIG_INCLUDED
#endif
#define BITS_IN_JSAMPLE  8
#define MAX_COMPONENTS  10
#if 8  == 8
#ifdef HAVE_UNSIGNED_CHAR
typedef unsigned char JSAMPLE;
#define GETJSAMPLE(value)  ((int) (value))
#else
#endif
#define MAXJSAMPLE 255
#define CENTERJSAMPLE 128
#endif
#if 8  == 12
#endif
typedef short JCOEF;
#ifdef HAVE_UNSIGNED_CHAR
typedef unsigned char JOCTET;
#define GETJOCTET(value)  (value)
#else
#endif
#ifdef HAVE_UNSIGNED_CHAR
typedef unsigned char UINT8;
#else
#endif
#ifdef HAVE_UNSIGNED_SHORT
typedef unsigned short UINT16;
#else
#endif
typedef unsigned int JDIMENSION;
#define JPEG_MAX_DIMENSION  65500L
#define METHODDEF(type)  static type
#define LOCAL(type)  static type
#define GLOBAL(type)  type
#define EXTERN(type)  extern type
#ifdef HAVE_PROTOTYPES
#define JMETHOD(type,methodname,arglist)  type (*methodname) arglist
#else
#endif
#ifdef NEED_FAR_POINTERS
#else
#endif
#ifndef HAVE_BOOLEAN
typedef int boolean;
#endif
#ifndef FALSE
#endif
#ifndef TRUE
#endif
#ifdef JPEG_INTERNALS
#define JPEG_INTERNAL_OPTIONS
#endif
#ifdef JPEG_INTERNAL_OPTIONS
#define DCT_ISLOW_SUPPORTED
#define DCT_IFAST_SUPPORTED
#define DCT_FLOAT_SUPPORTED
#undef  C_ARITH_CODING_SUPPORTED
#define C_MULTISCAN_FILES_SUPPORTED
#define C_PROGRESSIVE_SUPPORTED
#define ENTROPY_OPT_SUPPORTED
#define INPUT_SMOOTHING_SUPPORTED
#undef  D_ARITH_CODING_SUPPORTED
#define D_MULTISCAN_FILES_SUPPORTED
#define D_PROGRESSIVE_SUPPORTED
#define SAVE_MARKERS_SUPPORTED
#define BLOCK_SMOOTHING_SUPPORTED
#define IDCT_SCALING_SUPPORTED
#undef  UPSAMPLE_SCALING_SUPPORTED
#define UPSAMPLE_MERGING_SUPPORTED
#define QUANT_1PASS_SUPPORTED
#define QUANT_2PASS_SUPPORTED
#define RGB_RED  0
#define RGB_GREEN 1
#define RGB_BLUE 2
#define RGB_PIXELSIZE 3
#ifndef INLINE
#ifdef __GNUC__
#endif
#ifndef INLINE
#define INLINE
#endif
#endif
#ifndef MULTIPLIER
#define MULTIPLIER  int
#endif
#ifndef FAST_FLOAT
#ifdef HAVE_PROTOTYPES
#define FAST_FLOAT  float
#else
#endif
#endif
#endif
#define JPEG_LIB_VERSION  62
#define DCTSIZE      8
#define DCTSIZE2     64
#define NUM_QUANT_TBLS      4
#define NUM_HUFF_TBLS       4
#define NUM_ARITH_TBLS      16
#define MAX_COMPS_IN_SCAN   4
#define MAX_SAMP_FACTOR     4
#define C_MAX_BLOCKS_IN_MCU   10
#ifndef D_MAX_BLOCKS_IN_MCU
#define D_MAX_BLOCKS_IN_MCU   10
#endif
typedef JSAMPLE *JSAMPROW;
typedef JSAMPROW *JSAMPARRAY;
typedef JSAMPARRAY *JSAMPIMAGE;
typedef JCOEF JBLOCK[64];
typedef JBLOCK *JBLOCKROW;
typedef JBLOCKROW *JBLOCKARRAY;
typedef JBLOCKARRAY *JBLOCKIMAGE;
typedef JCOEF *JCOEFPTR;
typedef struct
 {
  UINT16 quantval[64];
  boolean sent_table;
  } JQUANT_TBL;
typedef struct
 {
  UINT8 bits[17];
  UINT8 huffval[256];
  boolean sent_table;
  } JHUFF_TBL;
typedef struct
 {
  int component_id;
  int component_index;
  int h_samp_factor;
  int v_samp_factor;
  int quant_tbl_no;
  int dc_tbl_no;
  int ac_tbl_no;
  JDIMENSION width_in_blocks;
  JDIMENSION height_in_blocks;
  int DCT_scaled_size;
  JDIMENSION downsampled_width;
  JDIMENSION downsampled_height;
  boolean component_needed;
  int MCU_width;
  int MCU_height;
  int MCU_blocks;
  int MCU_sample_width;
  int last_col_width;
  int last_row_height;
  JQUANT_TBL *quant_table;
  void *dct_table;
  } jpeg_component_info;
typedef struct
 {
  int comps_in_scan;
  int component_index[4];
  int Ss,
    Se;
  int Ah,
    Al;
  } jpeg_scan_info;
typedef struct jpeg_marker_struct *jpeg_saved_marker_ptr;
struct jpeg_marker_struct
 {
  jpeg_saved_marker_ptr next;
  UINT8 marker;
  unsigned int original_length;
  unsigned int data_length;
  JOCTET *data;
  };
typedef enum
 {
  JCS_UNKNOWN,
  JCS_GRAYSCALE,
  JCS_RGB,
  JCS_YCbCr,
  JCS_CMYK,
  JCS_YCCK
  } J_COLOR_SPACE;
typedef enum
 {
  JDCT_ISLOW,
  JDCT_IFAST,
  JDCT_FLOAT
  } J_DCT_METHOD;
#ifndef JDCT_DEFAULT
#define JDCT_DEFAULT  JDCT_ISLOW
#endif
#ifndef JDCT_FASTEST
#define JDCT_FASTEST  JDCT_IFAST
#endif
typedef enum
 {
  JDITHER_NONE,
  JDITHER_ORDERED,
  JDITHER_FS
  } J_DITHER_MODE;
#define jpeg_common_fields   struct jpeg_error_mgr * err;   struct jpeg_memory_mgr * mem;   struct jpeg_progress_mgr * progress;   void * client_data;    boolean is_decompressor;   int global_state
struct jpeg_common_struct
 {
  struct jpeg_error_mgr *err;
  struct jpeg_memory_mgr *mem;
  struct jpeg_progress_mgr *progress;
  void *client_data;
  boolean is_decompressor;
  int global_state;
  };
typedef struct jpeg_common_struct *j_common_ptr;
typedef struct jpeg_compress_struct *j_compress_ptr;
typedef struct jpeg_decompress_struct *j_decompress_ptr;
struct jpeg_compress_struct
 {
  struct jpeg_error_mgr *err;
  struct jpeg_memory_mgr *mem;
  struct jpeg_progress_mgr *progress;
  void *client_data;
  boolean is_decompressor;
  int global_state;
  struct jpeg_destination_mgr *dest;
  JDIMENSION image_width;
  JDIMENSION image_height;
  int input_components;
  J_COLOR_SPACE in_color_space;
  double input_gamma;
  int data_precision;
  int num_components;
  J_COLOR_SPACE jpeg_color_space;
  jpeg_component_info *comp_info;
  JQUANT_TBL *quant_tbl_ptrs[4];
  JHUFF_TBL *dc_huff_tbl_ptrs[4];
  JHUFF_TBL *ac_huff_tbl_ptrs[4];
  UINT8 arith_dc_L[16];
  UINT8 arith_dc_U[16];
  UINT8 arith_ac_K[16];
  int num_scans;
  const jpeg_scan_info *scan_info;
  boolean raw_data_in;
  boolean arith_code;
  boolean optimize_coding;
  boolean CCIR601_sampling;
  int smoothing_factor;
  J_DCT_METHOD dct_method;
  unsigned int restart_interval;
  int restart_in_rows;
  boolean write_JFIF_header;
  UINT8 JFIF_major_version;
  UINT8 JFIF_minor_version;
  UINT8 density_unit;
  UINT16 X_density;
  UINT16 Y_density;
  boolean write_Adobe_marker;
  JDIMENSION next_scanline;
  boolean progressive_mode;
  int max_h_samp_factor;
  int max_v_samp_factor;
  JDIMENSION total_iMCU_rows;
  int comps_in_scan;
  jpeg_component_info *cur_comp_info[4];
  JDIMENSION MCUs_per_row;
  JDIMENSION MCU_rows_in_scan;
  int blocks_in_MCU;
  int MCU_membership[10];
  int Ss,
    Se,
    Ah,
    Al;
  struct jpeg_comp_master *master;
  struct jpeg_c_main_controller *main;
  struct jpeg_c_prep_controller *prep;
  struct jpeg_c_coef_controller *coef;
  struct jpeg_marker_writer *marker;
  struct jpeg_color_converter *cconvert;
  struct jpeg_downsampler *downsample;
  struct jpeg_forward_dct *fdct;
  struct jpeg_entropy_encoder *entropy;
  jpeg_scan_info *script_space;
  int script_space_size;
  };
struct jpeg_decompress_struct
 {
  struct jpeg_error_mgr *err;
  struct jpeg_memory_mgr *mem;
  struct jpeg_progress_mgr *progress;
  void *client_data;
  boolean is_decompressor;
  int global_state;
  struct jpeg_source_mgr *src;
  JDIMENSION image_width;
  JDIMENSION image_height;
  int num_components;
  J_COLOR_SPACE jpeg_color_space;
  J_COLOR_SPACE out_color_space;
  unsigned int scale_num,
    scale_denom;
  double output_gamma;
  boolean buffered_image;
  boolean raw_data_out;
  J_DCT_METHOD dct_method;
  boolean do_fancy_upsampling;
  boolean do_block_smoothing;
  boolean quantize_colors;
  J_DITHER_MODE dither_mode;
  boolean two_pass_quantize;
  int desired_number_of_colors;
  boolean enable_1pass_quant;
  boolean enable_external_quant;
  boolean enable_2pass_quant;
  JDIMENSION output_width;
  JDIMENSION output_height;
  int out_color_components;
  int output_components;
  int rec_outbuf_height;
  int actual_number_of_colors;
  JSAMPARRAY colormap;
  JDIMENSION output_scanline;
  int input_scan_number;
  JDIMENSION input_iMCU_row;
  int output_scan_number;
  JDIMENSION output_iMCU_row;
  int (*coef_bits)[64];
  JQUANT_TBL *quant_tbl_ptrs[4];
  JHUFF_TBL *dc_huff_tbl_ptrs[4];
  JHUFF_TBL *ac_huff_tbl_ptrs[4];
  int data_precision;
  jpeg_component_info *comp_info;
  boolean progressive_mode;
  boolean arith_code;
  UINT8 arith_dc_L[16];
  UINT8 arith_dc_U[16];
  UINT8 arith_ac_K[16];
  unsigned int restart_interval;
  boolean saw_JFIF_marker;
  UINT8 JFIF_major_version;
  UINT8 JFIF_minor_version;
  UINT8 density_unit;
  UINT16 X_density;
  UINT16 Y_density;
  boolean saw_Adobe_marker;
  UINT8 Adobe_transform;
  boolean CCIR601_sampling;
  jpeg_saved_marker_ptr marker_list;
  int max_h_samp_factor;
  int max_v_samp_factor;
  int min_DCT_scaled_size;
  JDIMENSION total_iMCU_rows;
  JSAMPLE *sample_range_limit;
  int comps_in_scan;
  jpeg_component_info *cur_comp_info[4];
  JDIMENSION MCUs_per_row;
  JDIMENSION MCU_rows_in_scan;
  int blocks_in_MCU;
  int MCU_membership[10];
  int Ss,
    Se,
    Ah,
    Al;
  int unread_marker;
  struct jpeg_decomp_master *master;
  struct jpeg_d_main_controller *main;
  struct jpeg_d_coef_controller *coef;
  struct jpeg_d_post_controller *post;
  struct jpeg_input_controller *inputctl;
  struct jpeg_marker_reader *marker;
  struct jpeg_entropy_decoder *entropy;
  struct jpeg_inverse_dct *idct;
  struct jpeg_upsampler *upsample;
  struct jpeg_color_deconverter *cconvert;
  struct jpeg_color_quantizer *cquantize;
  };
struct jpeg_error_mgr
 {
  void (*error_exit) (j_common_ptr cinfo);
  void (*emit_message) (j_common_ptr cinfo, int msg_level);
  void (*output_message) (j_common_ptr cinfo);
  void (*format_message) (j_common_ptr cinfo, char *buffer);
#define JMSG_LENGTH_MAX  200
  void (*reset_error_mgr) (j_common_ptr cinfo);
  int msg_code;
#define JMSG_STR_PARM_MAX  80
  union
   {
    int i[8];
    char s[80];
    } msg_parm;
  int trace_level;
  long num_warnings;
  const char *const * jpeg_message_table;
  int last_jpeg_message;
  const char *const * addon_message_table;
  int first_addon_message;
  int last_addon_message;
  };
struct jpeg_progress_mgr
 {
  void (*progress_monitor) (j_common_ptr cinfo);
  long pass_counter;
  long pass_limit;
  int completed_passes;
  int total_passes;
  };
struct jpeg_destination_mgr
 {
  JOCTET *next_output_byte;
  size_t free_in_buffer;
  void (*init_destination) (j_compress_ptr cinfo);
    boolean(*empty_output_buffer) (j_compress_ptr cinfo);
  void (*term_destination) (j_compress_ptr cinfo);
  };
struct jpeg_source_mgr
 {
  const JOCTET *next_input_byte;
  size_t bytes_in_buffer;
  void (*init_source) (j_decompress_ptr cinfo);
    boolean(*fill_input_buffer) (j_decompress_ptr cinfo);
  void (*skip_input_data) (j_decompress_ptr cinfo, long num_bytes);
    boolean(*resync_to_restart) (j_decompress_ptr cinfo, int desired);
  void (*term_source) (j_decompress_ptr cinfo);
  };
#define JPOOL_PERMANENT 0
#define JPOOL_IMAGE 1
#define JPOOL_NUMPOOLS 2
typedef struct jvirt_sarray_control *jvirt_sarray_ptr;
typedef struct jvirt_barray_control *jvirt_barray_ptr;
struct jpeg_memory_mgr
 {
  void *(*alloc_small) (j_common_ptr cinfo, int pool_id, size_t sizeofobject);
  void *(*alloc_large) (j_common_ptr cinfo, int pool_id, size_t sizeofobject);
    JSAMPARRAY(*alloc_sarray) (j_common_ptr cinfo, int pool_id, JDIMENSION samplesperrow, JDIMENSION numrows);
    JBLOCKARRAY(*alloc_barray) (j_common_ptr cinfo, int pool_id, JDIMENSION blocksperrow, JDIMENSION numrows);
    jvirt_sarray_ptr(*request_virt_sarray) (j_common_ptr cinfo, int pool_id, boolean pre_zero, JDIMENSION samplesperrow, JDIMENSION numrows, JDIMENSION maxaccess);
    jvirt_barray_ptr(*request_virt_barray) (j_common_ptr cinfo, int pool_id, boolean pre_zero, JDIMENSION blocksperrow, JDIMENSION numrows, JDIMENSION maxaccess);
  void (*realize_virt_arrays) (j_common_ptr cinfo);
    JSAMPARRAY(*access_virt_sarray) (j_common_ptr cinfo, jvirt_sarray_ptr ptr, JDIMENSION start_row, JDIMENSION num_rows, boolean writable);
    JBLOCKARRAY(*access_virt_barray) (j_common_ptr cinfo, jvirt_barray_ptr ptr, JDIMENSION start_row, JDIMENSION num_rows, boolean writable);
  void (*free_pool) (j_common_ptr cinfo, int pool_id);
  void (*self_destruct) (j_common_ptr cinfo);
  long max_memory_to_use;
  long max_alloc_chunk;
  };
typedef boolean(*jpeg_marker_parser_method) (j_decompress_ptr cinfo);
#ifdef HAVE_PROTOTYPES
#define JPP(arglist) arglist
#else
#endif
#ifdef NEED_SHORT_EXTERNAL_NAMES
#endif
extern struct jpeg_error_mgr *jpeg_std_error
  (struct jpeg_error_mgr * err);
#define jpeg_create_compress(cinfo)     jpeg_CreateCompress((cinfo), JPEG_LIB_VERSION,    (size_t) sizeof(struct jpeg_compress_struct))
#define jpeg_create_decompress(cinfo)     jpeg_CreateDecompress((cinfo), JPEG_LIB_VERSION,      (size_t) sizeof(struct jpeg_decompress_struct))
extern void jpeg_CreateCompress(j_compress_ptr cinfo, int version, size_t structsize);
extern void jpeg_CreateDecompress(j_decompress_ptr cinfo, int version, size_t structsize);
extern void jpeg_destroy_compress(j_compress_ptr cinfo);
extern void jpeg_destroy_decompress(j_decompress_ptr cinfo);
extern void jpeg_stdio_dest(j_compress_ptr cinfo, FILE * outfile);
extern void jpeg_stdio_src(j_decompress_ptr cinfo, FILE * infile);
extern void jpeg_set_defaults(j_compress_ptr cinfo);
extern void jpeg_set_colorspace(j_compress_ptr cinfo, J_COLOR_SPACE colorspace);
extern void jpeg_default_colorspace(j_compress_ptr cinfo);
extern void jpeg_set_quality(j_compress_ptr cinfo, int quality, boolean force_baseline);
extern void jpeg_set_linear_quality(j_compress_ptr cinfo, int scale_factor, boolean force_baseline);
extern void jpeg_add_quant_table(j_compress_ptr cinfo, int which_tbl, const unsigned int *basic_table, int scale_factor, boolean force_baseline);
extern int jpeg_quality_scaling(int quality);
extern void jpeg_simple_progression(j_compress_ptr cinfo);
extern void jpeg_suppress_tables(j_compress_ptr cinfo, boolean suppress);
extern JQUANT_TBL *jpeg_alloc_quant_table(j_common_ptr cinfo);
extern JHUFF_TBL *jpeg_alloc_huff_table(j_common_ptr cinfo);
extern void jpeg_start_compress(j_compress_ptr cinfo, boolean write_all_tables);
extern JDIMENSION jpeg_write_scanlines(j_compress_ptr cinfo, JSAMPARRAY scanlines, JDIMENSION num_lines);
extern void jpeg_finish_compress(j_compress_ptr cinfo);
extern JDIMENSION jpeg_write_raw_data(j_compress_ptr cinfo, JSAMPIMAGE data, JDIMENSION num_lines);
extern void jpeg_write_marker
  (j_compress_ptr cinfo, int marker, const JOCTET * dataptr, unsigned int datalen);
extern void jpeg_write_m_header
  (j_compress_ptr cinfo, int marker, unsigned int datalen);
extern void jpeg_write_m_byte
  (j_compress_ptr cinfo, int val);
extern void jpeg_write_tables(j_compress_ptr cinfo);
extern int jpeg_read_header(j_decompress_ptr cinfo, boolean require_image);
#define JPEG_SUSPENDED  0
#define JPEG_HEADER_OK  1
#define JPEG_HEADER_TABLES_ONLY 2
extern boolean jpeg_start_decompress(j_decompress_ptr cinfo);
extern JDIMENSION jpeg_read_scanlines(j_decompress_ptr cinfo, JSAMPARRAY scanlines, JDIMENSION max_lines);
extern boolean jpeg_finish_decompress(j_decompress_ptr cinfo);
extern JDIMENSION jpeg_read_raw_data(j_decompress_ptr cinfo, JSAMPIMAGE data, JDIMENSION max_lines);
extern boolean jpeg_has_multiple_scans(j_decompress_ptr cinfo);
extern boolean jpeg_start_output(j_decompress_ptr cinfo, int scan_number);
extern boolean jpeg_finish_output(j_decompress_ptr cinfo);
extern boolean jpeg_input_complete(j_decompress_ptr cinfo);
extern void jpeg_new_colormap(j_decompress_ptr cinfo);
extern int jpeg_consume_input(j_decompress_ptr cinfo);
#define JPEG_REACHED_SOS 1
#define JPEG_REACHED_EOI 2
#define JPEG_ROW_COMPLETED 3
#define JPEG_SCAN_COMPLETED 4
extern void jpeg_calc_output_dimensions(j_decompress_ptr cinfo);
extern void jpeg_save_markers
  (j_decompress_ptr cinfo, int marker_code, unsigned int length_limit);
extern void jpeg_set_marker_processor
  (j_decompress_ptr cinfo, int marker_code, jpeg_marker_parser_method routine);
extern jvirt_barray_ptr *jpeg_read_coefficients(j_decompress_ptr cinfo);
extern void jpeg_write_coefficients(j_compress_ptr cinfo, jvirt_barray_ptr * coef_arrays);
extern void jpeg_copy_critical_parameters(j_decompress_ptr srcinfo, j_compress_ptr dstinfo);
extern void jpeg_abort_compress(j_compress_ptr cinfo);
extern void jpeg_abort_decompress(j_decompress_ptr cinfo);
extern void jpeg_abort(j_common_ptr cinfo);
extern void jpeg_destroy(j_common_ptr cinfo);
extern boolean jpeg_resync_to_restart(j_decompress_ptr cinfo, int desired);
#define JPEG_RST0 0xD0
#define JPEG_EOI 0xD9
#define JPEG_APP0 0xE0
#define JPEG_COM 0xFE
#ifdef INCOMPLETE_TYPES_BROKEN
#endif
#ifdef JPEG_INTERNALS
typedef enum
 {
  JBUF_PASS_THRU,
  JBUF_SAVE_SOURCE,
  JBUF_CRANK_DEST,
  JBUF_SAVE_AND_PASS
  } J_BUF_MODE;
#define CSTATE_START 100
#define CSTATE_SCANNING 101
#define CSTATE_RAW_OK 102
#define CSTATE_WRCOEFS 103
#define DSTATE_START 200
#define DSTATE_INHEADER 201
#define DSTATE_READY 202
#define DSTATE_PRELOAD 203
#define DSTATE_PRESCAN 204
#define DSTATE_SCANNING 205
#define DSTATE_RAW_OK 206
#define DSTATE_BUFIMAGE 207
#define DSTATE_BUFPOST 208
#define DSTATE_RDCOEFS 209
#define DSTATE_STOPPING 210
struct jpeg_comp_master
 {
  void (*prepare_for_pass) (j_compress_ptr cinfo);
  void (*pass_startup) (j_compress_ptr cinfo);
  void (*finish_pass) (j_compress_ptr cinfo);
  boolean call_pass_startup;
  boolean is_last_pass;
  };
struct jpeg_c_main_controller
 {
  void (*start_pass) (j_compress_ptr cinfo, J_BUF_MODE pass_mode);
  void (*process_data) (j_compress_ptr cinfo, JSAMPARRAY input_buf, JDIMENSION * in_row_ctr, JDIMENSION in_rows_avail);
  };
struct jpeg_c_prep_controller
 {
  void (*start_pass) (j_compress_ptr cinfo, J_BUF_MODE pass_mode);
  void (*pre_process_data) (j_compress_ptr cinfo, JSAMPARRAY input_buf, JDIMENSION * in_row_ctr, JDIMENSION in_rows_avail, JSAMPIMAGE output_buf, JDIMENSION * out_row_group_ctr, JDIMENSION out_row_groups_avail);
  };
struct jpeg_c_coef_controller
 {
  void (*start_pass) (j_compress_ptr cinfo, J_BUF_MODE pass_mode);
    boolean(*compress_data) (j_compress_ptr cinfo, JSAMPIMAGE input_buf);
  };
struct jpeg_color_converter
 {
  void (*start_pass) (j_compress_ptr cinfo);
  void (*color_convert) (j_compress_ptr cinfo, JSAMPARRAY input_buf, JSAMPIMAGE output_buf, JDIMENSION output_row, int num_rows);
  };
struct jpeg_downsampler
 {
  void (*start_pass) (j_compress_ptr cinfo);
  void (*downsample) (j_compress_ptr cinfo, JSAMPIMAGE input_buf, JDIMENSION in_row_index, JSAMPIMAGE output_buf, JDIMENSION out_row_group_index);
  boolean need_context_rows;
  };
struct jpeg_forward_dct
 {
  void (*start_pass) (j_compress_ptr cinfo);
  void (*forward_DCT) (j_compress_ptr cinfo, jpeg_component_info * compptr, JSAMPARRAY sample_data, JBLOCKROW coef_blocks, JDIMENSION start_row, JDIMENSION start_col, JDIMENSION num_blocks);
  };
struct jpeg_entropy_encoder
 {
  void (*start_pass) (j_compress_ptr cinfo, boolean gather_statistics);
    boolean(*encode_mcu) (j_compress_ptr cinfo, JBLOCKROW * MCU_data);
  void (*finish_pass) (j_compress_ptr cinfo);
  };
struct jpeg_marker_writer
 {
  void (*write_file_header) (j_compress_ptr cinfo);
  void (*write_frame_header) (j_compress_ptr cinfo);
  void (*write_scan_header) (j_compress_ptr cinfo);
  void (*write_file_trailer) (j_compress_ptr cinfo);
  void (*write_tables_only) (j_compress_ptr cinfo);
  void (*write_marker_header) (j_compress_ptr cinfo, int marker, unsigned int datalen);
  void (*write_marker_byte) (j_compress_ptr cinfo, int val);
  };
struct jpeg_decomp_master
 {
  void (*prepare_for_output_pass) (j_decompress_ptr cinfo);
  void (*finish_output_pass) (j_decompress_ptr cinfo);
  boolean is_dummy_pass;
  };
struct jpeg_input_controller
 {
  int (*consume_input) (j_decompress_ptr cinfo);
  void (*reset_input_controller) (j_decompress_ptr cinfo);
  void (*start_input_pass) (j_decompress_ptr cinfo);
  void (*finish_input_pass) (j_decompress_ptr cinfo);
  boolean has_multiple_scans;
  boolean eoi_reached;
  };
struct jpeg_d_main_controller
 {
  void (*start_pass) (j_decompress_ptr cinfo, J_BUF_MODE pass_mode);
  void (*process_data) (j_decompress_ptr cinfo, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
  };
struct jpeg_d_coef_controller
 {
  void (*start_input_pass) (j_decompress_ptr cinfo);
  int (*consume_data) (j_decompress_ptr cinfo);
  void (*start_output_pass) (j_decompress_ptr cinfo);
  int (*decompress_data) (j_decompress_ptr cinfo, JSAMPIMAGE output_buf);
  jvirt_barray_ptr *coef_arrays;
  };
struct jpeg_d_post_controller
 {
  void (*start_pass) (j_decompress_ptr cinfo, J_BUF_MODE pass_mode);
  void (*post_process_data) (j_decompress_ptr cinfo, JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr, JDIMENSION in_row_groups_avail, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
  };
struct jpeg_marker_reader
 {
  void (*reset_marker_reader) (j_decompress_ptr cinfo);
  int (*read_markers) (j_decompress_ptr cinfo);
  jpeg_marker_parser_method read_restart_marker;
  boolean saw_SOI;
  boolean saw_SOF;
  int next_restart_num;
  unsigned int discarded_bytes;
  };
struct jpeg_entropy_decoder
 {
  void (*start_pass) (j_decompress_ptr cinfo);
    boolean(*decode_mcu) (j_decompress_ptr cinfo, JBLOCKROW * MCU_data);
  boolean insufficient_data;
  };
typedef void (*inverse_DCT_method_ptr) (j_decompress_ptr cinfo, jpeg_component_info * compptr, JCOEFPTR coef_block, JSAMPARRAY output_buf, JDIMENSION output_col);
struct jpeg_inverse_dct
 {
  void (*start_pass) (j_decompress_ptr cinfo);
  inverse_DCT_method_ptr inverse_DCT[10];
  };
struct jpeg_upsampler
 {
  void (*start_pass) (j_decompress_ptr cinfo);
  void (*upsample) (j_decompress_ptr cinfo, JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr, JDIMENSION in_row_groups_avail, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
  boolean need_context_rows;
  };
struct jpeg_color_deconverter
 {
  void (*start_pass) (j_decompress_ptr cinfo);
  void (*color_convert) (j_decompress_ptr cinfo, JSAMPIMAGE input_buf, JDIMENSION input_row, JSAMPARRAY output_buf, int num_rows);
  };
struct jpeg_color_quantizer
 {
  void (*start_pass) (j_decompress_ptr cinfo, boolean is_pre_scan);
  void (*color_quantize) (j_decompress_ptr cinfo, JSAMPARRAY input_buf, JSAMPARRAY output_buf, int num_rows);
  void (*finish_pass) (j_decompress_ptr cinfo);
  void (*new_color_map) (j_decompress_ptr cinfo);
  };
#undef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#undef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#ifdef RIGHT_SHIFT_IS_UNSIGNED
#else
#define SHIFT_TEMPS
#define RIGHT_SHIFT(x,shft) ((x) >> (shft))
#endif
#ifdef NEED_SHORT_EXTERNAL_NAMES
#endif
extern void jinit_compress_master(j_compress_ptr cinfo);
extern void jinit_c_master_control(j_compress_ptr cinfo, boolean transcode_only);
extern void jinit_c_main_controller(j_compress_ptr cinfo, boolean need_full_buffer);
extern void jinit_c_prep_controller(j_compress_ptr cinfo, boolean need_full_buffer);
extern void jinit_c_coef_controller(j_compress_ptr cinfo, boolean need_full_buffer);
extern void jinit_color_converter(j_compress_ptr cinfo);
extern void jinit_downsampler(j_compress_ptr cinfo);
extern void jinit_forward_dct(j_compress_ptr cinfo);
extern void jinit_huff_encoder(j_compress_ptr cinfo);
extern void jinit_phuff_encoder(j_compress_ptr cinfo);
extern void jinit_marker_writer(j_compress_ptr cinfo);
extern void jinit_master_decompress(j_decompress_ptr cinfo);
extern void jinit_d_main_controller(j_decompress_ptr cinfo, boolean need_full_buffer);
extern void jinit_d_coef_controller(j_decompress_ptr cinfo, boolean need_full_buffer);
extern void jinit_d_post_controller(j_decompress_ptr cinfo, boolean need_full_buffer);
extern void jinit_input_controller(j_decompress_ptr cinfo);
extern void jinit_marker_reader(j_decompress_ptr cinfo);
extern void jinit_huff_decoder(j_decompress_ptr cinfo);
extern void jinit_phuff_decoder(j_decompress_ptr cinfo);
extern void jinit_inverse_dct(j_decompress_ptr cinfo);
extern void jinit_upsampler(j_decompress_ptr cinfo);
extern void jinit_color_deconverter(j_decompress_ptr cinfo);
extern void jinit_1pass_quantizer(j_decompress_ptr cinfo);
extern void jinit_2pass_quantizer(j_decompress_ptr cinfo);
extern void jinit_merged_upsampler(j_decompress_ptr cinfo);
extern void jinit_memory_mgr(j_common_ptr cinfo);
extern long jdiv_round_up(long a, long b);
extern long jround_up(long a, long b);
extern void jcopy_sample_rows(JSAMPARRAY input_array, int source_row, JSAMPARRAY output_array, int dest_row, int num_rows, JDIMENSION num_cols);
extern void jcopy_block_row(JBLOCKROW input_row, JBLOCKROW output_row, JDIMENSION num_blocks);
extern void jzero_far(void *target, size_t bytestozero);
#if 0
#endif
extern const int jpeg_natural_order[];
#ifdef INCOMPLETE_TYPES_BROKEN
#endif
#ifndef JMESSAGE
#ifndef JERROR_H
#define JMAKE_ENUM_LIST
#else
#endif
#endif
#ifdef JMAKE_ENUM_LIST
typedef enum
 {
#define JMESSAGE(code,string) code ,
#endif
  JMSG_NOMESSAGE,
  JERR_ARITH_NOTIMPL,
  JERR_BAD_ALIGN_TYPE,
  JERR_BAD_ALLOC_CHUNK,
  JERR_BAD_BUFFER_MODE,
  JERR_BAD_COMPONENT_ID,
  JERR_BAD_DCT_COEF,
  JERR_BAD_DCTSIZE,
  JERR_BAD_HUFF_TABLE,
  JERR_BAD_IN_COLORSPACE,
  JERR_BAD_J_COLORSPACE,
  JERR_BAD_LENGTH,
  JERR_BAD_LIB_VERSION,
  JERR_BAD_MCU_SIZE,
  JERR_BAD_POOL_ID,
  JERR_BAD_PRECISION,
  JERR_BAD_PROGRESSION,
  JERR_BAD_PROG_SCRIPT,
  JERR_BAD_SAMPLING,
  JERR_BAD_SCAN_SCRIPT,
  JERR_BAD_STATE,
  JERR_BAD_STRUCT_SIZE,
  JERR_BAD_VIRTUAL_ACCESS,
  JERR_BUFFER_SIZE,
  JERR_CANT_SUSPEND,
  JERR_CCIR601_NOTIMPL,
  JERR_COMPONENT_COUNT,
  JERR_CONVERSION_NOTIMPL,
  JERR_DAC_INDEX,
  JERR_DAC_VALUE,
  JERR_DHT_INDEX,
  JERR_DQT_INDEX,
  JERR_EMPTY_IMAGE,
  JERR_EMS_READ,
  JERR_EMS_WRITE,
  JERR_EOI_EXPECTED,
  JERR_FILE_READ,
  JERR_FILE_WRITE,
  JERR_FRACT_SAMPLE_NOTIMPL,
  JERR_HUFF_CLEN_OVERFLOW,
  JERR_HUFF_MISSING_CODE,
  JERR_IMAGE_TOO_BIG,
  JERR_INPUT_EMPTY,
  JERR_INPUT_EOF,
  JERR_MISMATCHED_QUANT_TABLE,
  JERR_MISSING_DATA,
  JERR_MODE_CHANGE,
  JERR_NOTIMPL,
  JERR_NOT_COMPILED,
  JERR_NO_BACKING_STORE,
  JERR_NO_HUFF_TABLE,
  JERR_NO_IMAGE,
  JERR_NO_QUANT_TABLE,
  JERR_NO_SOI,
  JERR_OUT_OF_MEMORY,
  JERR_QUANT_COMPONENTS,
  JERR_QUANT_FEW_COLORS,
  JERR_QUANT_MANY_COLORS,
  JERR_SOF_DUPLICATE,
  JERR_SOF_NO_SOS,
  JERR_SOF_UNSUPPORTED,
  JERR_SOI_DUPLICATE,
  JERR_SOS_NO_SOF,
  JERR_TFILE_CREATE,
  JERR_TFILE_READ,
  JERR_TFILE_SEEK,
  JERR_TFILE_WRITE,
  JERR_TOO_LITTLE_DATA,
  JERR_UNKNOWN_MARKER,
  JERR_VIRTUAL_BUG,
  JERR_WIDTH_OVERFLOW,
  JERR_XMS_READ,
  JERR_XMS_WRITE,
  JMSG_COPYRIGHT,
  JMSG_VERSION,
  JTRC_16BIT_TABLES,
  JTRC_ADOBE,
  JTRC_APP0,
  JTRC_APP14,
  JTRC_DAC,
  JTRC_DHT,
  JTRC_DQT,
  JTRC_DRI,
  JTRC_EMS_CLOSE,
  JTRC_EMS_OPEN,
  JTRC_EOI,
  JTRC_HUFFBITS,
  JTRC_JFIF,
  JTRC_JFIF_BADTHUMBNAILSIZE,
  JTRC_JFIF_EXTENSION,
  JTRC_JFIF_THUMBNAIL,
  JTRC_MISC_MARKER,
  JTRC_PARMLESS_MARKER,
  JTRC_QUANTVALS,
  JTRC_QUANT_3_NCOLORS,
  JTRC_QUANT_NCOLORS,
  JTRC_QUANT_SELECTED,
  JTRC_RECOVERY_ACTION,
  JTRC_RST,
  JTRC_SMOOTH_NOTIMPL,
  JTRC_SOF,
  JTRC_SOF_COMPONENT,
  JTRC_SOI,
  JTRC_SOS,
  JTRC_SOS_COMPONENT,
  JTRC_SOS_PARAMS,
  JTRC_TFILE_CLOSE,
  JTRC_TFILE_OPEN,
  JTRC_THUMB_JPEG,
  JTRC_THUMB_PALETTE,
  JTRC_THUMB_RGB,
  JTRC_UNKNOWN_IDS,
  JTRC_XMS_CLOSE,
  JTRC_XMS_OPEN,
  JWRN_ADOBE_XFORM,
  JWRN_BOGUS_PROGRESSION,
  JWRN_EXTRANEOUS_DATA,
  JWRN_HIT_MARKER,
  JWRN_HUFF_BAD_CODE,
  JWRN_JFIF_MAJOR,
  JWRN_JPEG_EOF,
  JWRN_MUST_RESYNC,
  JWRN_NOT_SEQUENTIAL,
  JWRN_TOO_MUCH_DATA,
#ifdef JMAKE_ENUM_LIST
  JMSG_LASTMSGCODE
  } J_MESSAGE_CODE;
#undef JMAKE_ENUM_LIST
#endif
#undef JMESSAGE
#ifndef JERROR_H
#define JERROR_H
#define ERREXIT(cinfo,code)    ((cinfo)->err->msg_code = (code),    (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXIT1(cinfo,code,p1)    ((cinfo)->err->msg_code = (code),    (cinfo)->err->msg_parm.i[0] = (p1),    (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXIT2(cinfo,code,p1,p2)    ((cinfo)->err->msg_code = (code),    (cinfo)->err->msg_parm.i[0] = (p1),    (cinfo)->err->msg_parm.i[1] = (p2),    (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXIT3(cinfo,code,p1,p2,p3)    ((cinfo)->err->msg_code = (code),    (cinfo)->err->msg_parm.i[0] = (p1),    (cinfo)->err->msg_parm.i[1] = (p2),    (cinfo)->err->msg_parm.i[2] = (p3),    (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXIT4(cinfo,code,p1,p2,p3,p4)    ((cinfo)->err->msg_code = (code),    (cinfo)->err->msg_parm.i[0] = (p1),    (cinfo)->err->msg_parm.i[1] = (p2),    (cinfo)->err->msg_parm.i[2] = (p3),    (cinfo)->err->msg_parm.i[3] = (p4),    (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define ERREXITS(cinfo,code,str)    ((cinfo)->err->msg_code = (code),    strncpy((cinfo)->err->msg_parm.s, (str), JMSG_STR_PARM_MAX),    (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)))
#define MAKESTMT(stuff)  do { stuff } while (0)
#define WARNMS(cinfo,code)    ((cinfo)->err->msg_code = (code),    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1))
#define WARNMS1(cinfo,code,p1)    ((cinfo)->err->msg_code = (code),    (cinfo)->err->msg_parm.i[0] = (p1),    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1))
#define WARNMS2(cinfo,code,p1,p2)    ((cinfo)->err->msg_code = (code),    (cinfo)->err->msg_parm.i[0] = (p1),    (cinfo)->err->msg_parm.i[1] = (p2),    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1))
#define TRACEMS(cinfo,lvl,code)    ((cinfo)->err->msg_code = (code),    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)))
#define TRACEMS1(cinfo,lvl,code,p1)    ((cinfo)->err->msg_code = (code),    (cinfo)->err->msg_parm.i[0] = (p1),    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)))
#define TRACEMS2(cinfo,lvl,code,p1,p2)    ((cinfo)->err->msg_code = (code),    (cinfo)->err->msg_parm.i[0] = (p1),    (cinfo)->err->msg_parm.i[1] = (p2),    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)))
#define TRACEMS3(cinfo,lvl,code,p1,p2,p3)    MAKESTMT(int * _mp = (cinfo)->err->msg_parm.i;     _mp[0] = (p1); _mp[1] = (p2); _mp[2] = (p3);     (cinfo)->err->msg_code = (code);     (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)); )
#define TRACEMS4(cinfo,lvl,code,p1,p2,p3,p4)    MAKESTMT(int * _mp = (cinfo)->err->msg_parm.i;     _mp[0] = (p1); _mp[1] = (p2); _mp[2] = (p3); _mp[3] = (p4);     (cinfo)->err->msg_code = (code);     (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)); )
#define TRACEMS5(cinfo,lvl,code,p1,p2,p3,p4,p5)    MAKESTMT(int * _mp = (cinfo)->err->msg_parm.i;     _mp[0] = (p1); _mp[1] = (p2); _mp[2] = (p3); _mp[3] = (p4);     _mp[4] = (p5);     (cinfo)->err->msg_code = (code);     (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)); )
#define TRACEMS8(cinfo,lvl,code,p1,p2,p3,p4,p5,p6,p7,p8)    MAKESTMT(int * _mp = (cinfo)->err->msg_parm.i;     _mp[0] = (p1); _mp[1] = (p2); _mp[2] = (p3); _mp[3] = (p4);     _mp[4] = (p5); _mp[5] = (p6); _mp[6] = (p7); _mp[7] = (p8);     (cinfo)->err->msg_code = (code);     (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)); )
#define TRACEMSS(cinfo,lvl,code,str)    ((cinfo)->err->msg_code = (code),    strncpy((cinfo)->err->msg_parm.s, (str), JMSG_STR_PARM_MAX),    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (lvl)))
#endif
#endif
#endif
void
  jpeg_CreateCompress(j_compress_ptr cinfo, int version, size_t structsize)
 {
  int i;
  cinfo->mem = ((void *) 0);
  if (version != 62)
    ((cinfo)->err->msg_code = (JERR_BAD_LIB_VERSION), (cinfo)->err->msg_parm.i[0] = (62), (cinfo)->err->msg_parm.i[1] = (version), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (structsize != ((size_t) sizeof(struct jpeg_compress_struct)))
    ((cinfo)->err->msg_code = (JERR_BAD_STRUCT_SIZE), (cinfo)->err->msg_parm.i[0] = ((int) ((size_t) sizeof(struct jpeg_compress_struct))), (cinfo)->err->msg_parm.i[1] = ((int) structsize), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
   {
    struct jpeg_error_mgr *err = cinfo->err;
    void *client_data = cinfo->client_data;
    memset((void *) (cinfo), 0, (size_t) (((size_t) sizeof(struct jpeg_compress_struct))));
    cinfo->err = err;
    cinfo->client_data = client_data;
    }
  cinfo->is_decompressor = 0;
  jinit_memory_mgr((j_common_ptr) cinfo);
  cinfo->progress = ((void *) 0);
  cinfo->dest = ((void *) 0);
  cinfo->comp_info = ((void *) 0);
  for (i = 0; i < 4; i++)
    cinfo->quant_tbl_ptrs[i] = ((void *) 0);
  for (i = 0; i < 4; i++)
   {
    cinfo->dc_huff_tbl_ptrs[i] = ((void *) 0);
    cinfo->ac_huff_tbl_ptrs[i] = ((void *) 0);
    }
  cinfo->script_space = ((void *) 0);
  cinfo->input_gamma = 1.0;
  cinfo->global_state = 100;
  }
void
  jpeg_destroy_compress(j_compress_ptr cinfo)
 {
  jpeg_destroy((j_common_ptr) cinfo);
  }
void
  jpeg_abort_compress(j_compress_ptr cinfo)
 {
  jpeg_abort((j_common_ptr) cinfo);
  }
void
  jpeg_suppress_tables(j_compress_ptr cinfo, boolean suppress)
 {
  int i;
  JQUANT_TBL *qtbl;
  JHUFF_TBL *htbl;
  for (i = 0; i < 4; i++)
   {
    if ((qtbl = cinfo->quant_tbl_ptrs[i]) != ((void *) 0))
      qtbl->sent_table = suppress;
    }
  for (i = 0; i < 4; i++)
   {
    if ((htbl = cinfo->dc_huff_tbl_ptrs[i]) != ((void *) 0))
      htbl->sent_table = suppress;
    if ((htbl = cinfo->ac_huff_tbl_ptrs[i]) != ((void *) 0))
      htbl->sent_table = suppress;
    }
  }
void
  jpeg_finish_compress(j_compress_ptr cinfo)
 {
  JDIMENSION iMCU_row;
  if (cinfo->global_state == 101 ||
   cinfo->global_state == 102)
   {
    if (cinfo->next_scanline < cinfo->image_height)
      ((cinfo)->err->msg_code = (JERR_TOO_LITTLE_DATA), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    (*cinfo->master->finish_pass) (cinfo);
    }
  else if (cinfo->global_state != 103)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  while (!cinfo->master->is_last_pass)
   {
    (*cinfo->master->prepare_for_pass) (cinfo);
    for (iMCU_row = 0; iMCU_row < cinfo->total_iMCU_rows; iMCU_row++)
     {
      if (cinfo->progress != ((void *) 0))
       {
        cinfo->progress->pass_counter = (long) iMCU_row;
        cinfo->progress->pass_limit = (long) cinfo->total_iMCU_rows;
        (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
        }
      if (!(*cinfo->coef->compress_data) (cinfo, (JSAMPIMAGE) ((void *) 0)))
        ((cinfo)->err->msg_code = (JERR_CANT_SUSPEND), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      }
    (*cinfo->master->finish_pass) (cinfo);
    }
  (*cinfo->marker->write_file_trailer) (cinfo);
  (*cinfo->dest->term_destination) (cinfo);
  jpeg_abort((j_common_ptr) cinfo);
  }
void
  jpeg_write_marker(j_compress_ptr cinfo, int marker,
   const JOCTET * dataptr, unsigned int datalen)
 {
  void (*write_marker_byte) (j_compress_ptr info, int val);
  if (cinfo->next_scanline != 0 ||
   (cinfo->global_state != 101 &&
    cinfo->global_state != 102 &&
    cinfo->global_state != 103))
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  (*cinfo->marker->write_marker_header) (cinfo, marker, datalen);
  write_marker_byte = cinfo->marker->write_marker_byte;
  while (datalen--)
   {
    (*write_marker_byte) (cinfo, *dataptr);
    dataptr++;
    }
  }
void
  jpeg_write_m_header(j_compress_ptr cinfo, int marker, unsigned int datalen)
 {
  if (cinfo->next_scanline != 0 ||
   (cinfo->global_state != 101 &&
    cinfo->global_state != 102 &&
    cinfo->global_state != 103))
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  (*cinfo->marker->write_marker_header) (cinfo, marker, datalen);
  }
void
  jpeg_write_m_byte(j_compress_ptr cinfo, int val)
 {
  (*cinfo->marker->write_marker_byte) (cinfo, val);
  }
void
  jpeg_write_tables(j_compress_ptr cinfo)
 {
  if (cinfo->global_state != 100)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  (*cinfo->err->reset_error_mgr) ((j_common_ptr) cinfo);
  (*cinfo->dest->init_destination) (cinfo);
  jinit_marker_writer(cinfo);
  (*cinfo->marker->write_tables_only) (cinfo);
  (*cinfo->dest->term_destination) (cinfo);
  }
#define JPEG_INTERNALS
void
  jpeg_abort(j_common_ptr cinfo)
 {
  int pool;
  if (cinfo->mem == ((void *) 0))
    return;
  for (pool = 2 - 1; pool > 0; pool--)
   {
    (*cinfo->mem->free_pool) (cinfo, pool);
    }
  if (cinfo->is_decompressor)
   {
    cinfo->global_state = 200;
    ((j_decompress_ptr) cinfo)->marker_list = ((void *) 0);
    }
  else
   {
    cinfo->global_state = 100;
    }
  }
void
  jpeg_destroy(j_common_ptr cinfo)
 {
  if (cinfo->mem != ((void *) 0))
    (*cinfo->mem->self_destruct) (cinfo);
  cinfo->mem = ((void *) 0);
  cinfo->global_state = 0;
  }
JQUANT_TBL *
  jpeg_alloc_quant_table(j_common_ptr cinfo)
 {
  JQUANT_TBL *tbl;
  tbl = (JQUANT_TBL *)
   (*cinfo->mem->alloc_small) (cinfo, 0, ((size_t) sizeof(JQUANT_TBL)));
  tbl->sent_table = 0;
  return tbl;
  }
JHUFF_TBL *
  jpeg_alloc_huff_table(j_common_ptr cinfo)
 {
  JHUFF_TBL *tbl;
  tbl = (JHUFF_TBL *)
   (*cinfo->mem->alloc_small) (cinfo, 0, ((size_t) sizeof(JHUFF_TBL)));
  tbl->sent_table = 0;
  return tbl;
  }
#define JPEG_INTERNALS
void
  jpeg_CreateDecompress(j_decompress_ptr cinfo, int version, size_t structsize)
 {
  int i;
  cinfo->mem = ((void *) 0);
  if (version != 62)
    ((cinfo)->err->msg_code = (JERR_BAD_LIB_VERSION), (cinfo)->err->msg_parm.i[0] = (62), (cinfo)->err->msg_parm.i[1] = (version), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (structsize != ((size_t) sizeof(struct jpeg_decompress_struct)))
    ((cinfo)->err->msg_code = (JERR_BAD_STRUCT_SIZE), (cinfo)->err->msg_parm.i[0] = ((int) ((size_t) sizeof(struct jpeg_decompress_struct))), (cinfo)->err->msg_parm.i[1] = ((int) structsize), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
   {
    struct jpeg_error_mgr *err = cinfo->err;
    void *client_data = cinfo->client_data;
    memset((void *) (cinfo), 0, (size_t) (((size_t) sizeof(struct jpeg_decompress_struct))));
    cinfo->err = err;
    cinfo->client_data = client_data;
    }
  cinfo->is_decompressor = 1;
  jinit_memory_mgr((j_common_ptr) cinfo);
  cinfo->progress = ((void *) 0);
  cinfo->src = ((void *) 0);
  for (i = 0; i < 4; i++)
    cinfo->quant_tbl_ptrs[i] = ((void *) 0);
  for (i = 0; i < 4; i++)
   {
    cinfo->dc_huff_tbl_ptrs[i] = ((void *) 0);
    cinfo->ac_huff_tbl_ptrs[i] = ((void *) 0);
    }
  cinfo->marker_list = ((void *) 0);
  jinit_marker_reader(cinfo);
  jinit_input_controller(cinfo);
  cinfo->global_state = 200;
  }
void
  jpeg_destroy_decompress(j_decompress_ptr cinfo)
 {
  jpeg_destroy((j_common_ptr) cinfo);
  }
void
  jpeg_abort_decompress(j_decompress_ptr cinfo)
 {
  jpeg_abort((j_common_ptr) cinfo);
  }
static void
  default_decompress_parms(j_decompress_ptr cinfo)
 {
  switch (cinfo->num_components)
   {
 case 1:
    cinfo->jpeg_color_space = JCS_GRAYSCALE;
    cinfo->out_color_space = JCS_GRAYSCALE;
    break;
   case 3:
    if (cinfo->saw_JFIF_marker)
     {
      cinfo->jpeg_color_space = JCS_YCbCr;
      }
    else if (cinfo->saw_Adobe_marker)
     {
      switch (cinfo->Adobe_transform)
       {
       case 0:
        cinfo->jpeg_color_space = JCS_RGB;
        break;
       case 1:
        cinfo->jpeg_color_space = JCS_YCbCr;
        break;
       default:
        ((cinfo)->err->msg_code = (JWRN_ADOBE_XFORM), (cinfo)->err->msg_parm.i[0] = (cinfo->Adobe_transform), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
        cinfo->jpeg_color_space = JCS_YCbCr;
        break;
        }
      }
    else
     {
      int cid0 = cinfo->comp_info[0].component_id;
      int cid1 = cinfo->comp_info[1].component_id;
      int cid2 = cinfo->comp_info[2].component_id;
      if (cid0 == 1 && cid1 == 2 && cid2 == 3)
        cinfo->jpeg_color_space = JCS_YCbCr;
      else if (cid0 == 82 && cid1 == 71 && cid2 == 66)
        cinfo->jpeg_color_space = JCS_RGB;
      else
       {
        do
         {
          int *_mp = (cinfo)->err->msg_parm.i;
          _mp[0] = (cid0);
          _mp[1] = (cid1);
          _mp[2] = (cid2);
          (cinfo)->err->msg_code = (JTRC_UNKNOWN_IDS);
          (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1));
          }
        while (0);
        cinfo->jpeg_color_space = JCS_YCbCr;
        }
      }
    cinfo->out_color_space = JCS_RGB;
    break;
   case 4:
    if (cinfo->saw_Adobe_marker)
     {
      switch (cinfo->Adobe_transform)
       {
       case 0:
        cinfo->jpeg_color_space = JCS_CMYK;
        break;
       case 2:
        cinfo->jpeg_color_space = JCS_YCCK;
        break;
       default:
        ((cinfo)->err->msg_code = (JWRN_ADOBE_XFORM), (cinfo)->err->msg_parm.i[0] = (cinfo->Adobe_transform), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
        cinfo->jpeg_color_space = JCS_YCCK;
        break;
        }
      }
    else
     {
      cinfo->jpeg_color_space = JCS_CMYK;
      }
    cinfo->out_color_space = JCS_CMYK;
    break;
   default:
    cinfo->jpeg_color_space = JCS_UNKNOWN;
    cinfo->out_color_space = JCS_UNKNOWN;
    break;
    }
  cinfo->scale_num = 1;
  cinfo->scale_denom = 1;
  cinfo->output_gamma = 1.0;
  cinfo->buffered_image = 0;
  cinfo->raw_data_out = 0;
  cinfo->dct_method = JDCT_ISLOW;
  cinfo->do_fancy_upsampling = 1;
  cinfo->do_block_smoothing = 1;
  cinfo->quantize_colors = 0;
  cinfo->dither_mode = JDITHER_FS;
#ifdef QUANT_2PASS_SUPPORTED
  cinfo->two_pass_quantize = 1;
#else
#endif
  cinfo->desired_number_of_colors = 256;
  cinfo->colormap = ((void *) 0);
  cinfo->enable_1pass_quant = 0;
  cinfo->enable_external_quant = 0;
  cinfo->enable_2pass_quant = 0;
  }
int
  jpeg_read_header(j_decompress_ptr cinfo, boolean require_image)
 {
  int retcode;
  if (cinfo->global_state != 200 &&
   cinfo->global_state != 201)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  retcode = jpeg_consume_input(cinfo);
  switch (retcode)
   {
   case 1:
    retcode = 1;
    break;
   case 2:
    if (require_image)
      ((cinfo)->err->msg_code = (JERR_NO_IMAGE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    jpeg_abort((j_common_ptr) cinfo);
    retcode = 2;
    break;
   case 0:
    break;
    }
  return retcode;
  }
int
  jpeg_consume_input(j_decompress_ptr cinfo)
 {
  int retcode = 0;
  switch (cinfo->global_state)
   {
   case 200:
    (*cinfo->inputctl->reset_input_controller) (cinfo);
    (*cinfo->src->init_source) (cinfo);
    cinfo->global_state = 201;
   case 201:
    retcode = (*cinfo->inputctl->consume_input) (cinfo);
    if (retcode == 1)
     {
      default_decompress_parms(cinfo);
      cinfo->global_state = 202;
      }
    break;
   case 202:
    retcode = 1;
    break;
   case 203:
   case 204:
   case 205:
   case 206:
   case 207:
   case 208:
   case 210:
    retcode = (*cinfo->inputctl->consume_input) (cinfo);
    break;
   default:
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  return retcode;
  }
boolean
jpeg_input_complete(j_decompress_ptr cinfo)
 {
  if (cinfo->global_state < 200 ||
   cinfo->global_state > 210)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  return cinfo->inputctl->eoi_reached;
  }
boolean
jpeg_has_multiple_scans(j_decompress_ptr cinfo)
 {
  if (cinfo->global_state < 202 ||
   cinfo->global_state > 210)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  return cinfo->inputctl->has_multiple_scans;
  }
boolean
jpeg_finish_decompress(j_decompress_ptr cinfo)
 {
  if ((cinfo->global_state == 205 ||
    cinfo->global_state == 206) && !cinfo->buffered_image)
   {
    if (cinfo->output_scanline < cinfo->output_height)
      ((cinfo)->err->msg_code = (JERR_TOO_LITTLE_DATA), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    (*cinfo->master->finish_output_pass) (cinfo);
    cinfo->global_state = 210;
    }
  else if (cinfo->global_state == 207)
   {
    cinfo->global_state = 210;
    }
  else if (cinfo->global_state != 210)
   {
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  while (!cinfo->inputctl->eoi_reached)
   {
    if ((*cinfo->inputctl->consume_input) (cinfo) == 0)
      return 0;
    }
  (*cinfo->src->term_source) (cinfo);
  jpeg_abort((j_common_ptr) cinfo);
  return 1;
  }
#define JPEG_INTERNALS
#define AM_MEMORY_MANAGER
#ifdef NEED_SHORT_EXTERNAL_NAMES
#endif
extern void *jpeg_get_small(j_common_ptr cinfo, size_t sizeofobject);
extern void jpeg_free_small(j_common_ptr cinfo, void *object, size_t sizeofobject);
extern void *jpeg_get_large(j_common_ptr cinfo, size_t sizeofobject);
extern void jpeg_free_large(j_common_ptr cinfo, void *object, size_t sizeofobject);
#ifndef MAX_ALLOC_CHUNK
#define MAX_ALLOC_CHUNK  1000000000L
#endif
extern long jpeg_mem_available(j_common_ptr cinfo, long min_bytes_needed, long max_bytes_needed, long already_allocated);
#define TEMP_NAME_LENGTH   64
#ifdef USE_MSDOS_MEMMGR
#endif
#ifdef USE_MAC_MEMMGR
#endif
typedef struct backing_store_struct *backing_store_ptr;
typedef struct backing_store_struct
 {
  void (*read_backing_store) (j_common_ptr cinfo, backing_store_ptr info, void *buffer_address, long file_offset, long byte_count);
  void (*write_backing_store) (j_common_ptr cinfo, backing_store_ptr info, void *buffer_address, long file_offset, long byte_count);
  void (*close_backing_store) (j_common_ptr cinfo, backing_store_ptr info);
#ifdef USE_MSDOS_MEMMGR
#else
#ifdef USE_MAC_MEMMGR
#else
  FILE *temp_file;
  char temp_name[64];
#endif
#endif
  } backing_store_info;
extern void jpeg_open_backing_store(j_common_ptr cinfo, backing_store_ptr info, long total_bytes_needed);
extern long jpeg_mem_init(j_common_ptr cinfo);
extern void jpeg_mem_term(j_common_ptr cinfo);
#ifndef NO_GETENV
#ifndef HAVE_STDLIB_H
#endif
#endif
#ifndef ALIGN_TYPE
#define ALIGN_TYPE  double
#endif
typedef union small_pool_struct *small_pool_ptr;
typedef union small_pool_struct
 {
  struct
   {
    small_pool_ptr next;
    size_t bytes_used;
    size_t bytes_left;
    } hdr;
  double dummy;
  } small_pool_hdr;
typedef union large_pool_struct *large_pool_ptr;
typedef union large_pool_struct
 {
  struct
   {
    large_pool_ptr next;
    size_t bytes_used;
    size_t bytes_left;
    } hdr;
  double dummy;
  } large_pool_hdr;
typedef struct
 {
  struct jpeg_memory_mgr pub;
  small_pool_ptr small_list[2];
  large_pool_ptr large_list[2];
  jvirt_sarray_ptr virt_sarray_list;
  jvirt_barray_ptr virt_barray_list;
  long total_space_allocated;
  JDIMENSION last_rowsperchunk;
  } my_memory_mgr;
typedef my_memory_mgr *my_mem_ptr;
struct jvirt_sarray_control
 {
  JSAMPARRAY mem_buffer;
  JDIMENSION rows_in_array;
  JDIMENSION samplesperrow;
  JDIMENSION maxaccess;
  JDIMENSION rows_in_mem;
  JDIMENSION rowsperchunk;
  JDIMENSION cur_start_row;
  JDIMENSION first_undef_row;
  boolean pre_zero;
  boolean dirty;
  boolean b_s_open;
  jvirt_sarray_ptr next;
  backing_store_info b_s_info;
  };
struct jvirt_barray_control
 {
  JBLOCKARRAY mem_buffer;
  JDIMENSION rows_in_array;
  JDIMENSION blocksperrow;
  JDIMENSION maxaccess;
  JDIMENSION rows_in_mem;
  JDIMENSION rowsperchunk;
  JDIMENSION cur_start_row;
  JDIMENSION first_undef_row;
  boolean pre_zero;
  boolean dirty;
  boolean b_s_open;
  jvirt_barray_ptr next;
  backing_store_info b_s_info;
  };
#ifdef MEM_STATS
#endif
static void
  out_of_memory(j_common_ptr cinfo, int which)
 {
#ifdef MEM_STATS
#endif
  ((cinfo)->err->msg_code = (JERR_OUT_OF_MEMORY), (cinfo)->err->msg_parm.i[0] = (which), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  }
static const size_t first_pool_slop[2] =
 {
  1600,
  16000
  };
static const size_t extra_pool_slop[2] =
 {
  0,
  5000
  };
#define MIN_SLOP  50
static void *
  alloc_small(j_common_ptr cinfo, int pool_id, size_t sizeofobject)
 {
  my_mem_ptr mem = (my_mem_ptr) cinfo->mem;
  small_pool_ptr hdr_ptr,
    prev_hdr_ptr;
  char *data_ptr;
  size_t odd_bytes,
    min_request,
    slop;
  if (sizeofobject > (size_t) (1000000000L - ((size_t) sizeof(small_pool_hdr))))
    out_of_memory(cinfo, 1);
  odd_bytes = sizeofobject % ((size_t) sizeof(double));
  if (odd_bytes > 0)
    sizeofobject += ((size_t) sizeof(double)) - odd_bytes;
  if (pool_id < 0 || pool_id >= 2)
    ((cinfo)->err->msg_code = (JERR_BAD_POOL_ID), (cinfo)->err->msg_parm.i[0] = (pool_id), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  prev_hdr_ptr = ((void *) 0);
  hdr_ptr = mem->small_list[pool_id];
  while (hdr_ptr != ((void *) 0))
   {
    if (hdr_ptr->hdr.bytes_left >= sizeofobject)
      break;
    prev_hdr_ptr = hdr_ptr;
    hdr_ptr = hdr_ptr->hdr.next;
    }
  if (hdr_ptr == ((void *) 0))
   {
    min_request = sizeofobject + ((size_t) sizeof(small_pool_hdr));
    if (prev_hdr_ptr == ((void *) 0))
      slop = first_pool_slop[pool_id];
    else
      slop = extra_pool_slop[pool_id];
    if (slop > (size_t) (1000000000L - min_request))
      slop = (size_t) (1000000000L - min_request);
    for (;;)
     {
      hdr_ptr = (small_pool_ptr) jpeg_get_small(cinfo, min_request + slop);
      if (hdr_ptr != ((void *) 0))
        break;
      slop /= 2;
      if (slop < 50)
        out_of_memory(cinfo, 2);
      }
    mem->total_space_allocated += min_request + slop;
    hdr_ptr->hdr.next = ((void *) 0);
    hdr_ptr->hdr.bytes_used = 0;
    hdr_ptr->hdr.bytes_left = sizeofobject + slop;
    if (prev_hdr_ptr == ((void *) 0))
      mem->small_list[pool_id] = hdr_ptr;
    else
      prev_hdr_ptr->hdr.next = hdr_ptr;
    }
  data_ptr = (char *) (hdr_ptr + 1);
  data_ptr += hdr_ptr->hdr.bytes_used;
  hdr_ptr->hdr.bytes_used += sizeofobject;
  hdr_ptr->hdr.bytes_left -= sizeofobject;
  return (void *) data_ptr;
  }
static void *
  alloc_large(j_common_ptr cinfo, int pool_id, size_t sizeofobject)
 {
  my_mem_ptr mem = (my_mem_ptr) cinfo->mem;
  large_pool_ptr hdr_ptr;
  size_t odd_bytes;
  if (sizeofobject > (size_t) (1000000000L - ((size_t) sizeof(large_pool_hdr))))
    out_of_memory(cinfo, 3);
  odd_bytes = sizeofobject % ((size_t) sizeof(double));
  if (odd_bytes > 0)
    sizeofobject += ((size_t) sizeof(double)) - odd_bytes;
  if (pool_id < 0 || pool_id >= 2)
    ((cinfo)->err->msg_code = (JERR_BAD_POOL_ID), (cinfo)->err->msg_parm.i[0] = (pool_id), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  hdr_ptr = (large_pool_ptr) jpeg_get_large(cinfo, sizeofobject +
   ((size_t) sizeof(large_pool_hdr)));
  if (hdr_ptr == ((void *) 0))
    out_of_memory(cinfo, 4);
  mem->total_space_allocated += sizeofobject + ((size_t) sizeof(large_pool_hdr));
  hdr_ptr->hdr.next = mem->large_list[pool_id];
  hdr_ptr->hdr.bytes_used = sizeofobject;
  hdr_ptr->hdr.bytes_left = 0;
  mem->large_list[pool_id] = hdr_ptr;
  return (void *) (hdr_ptr + 1);
  }
static JSAMPARRAY
  alloc_sarray(j_common_ptr cinfo, int pool_id,
   JDIMENSION samplesperrow, JDIMENSION numrows)
 {
  my_mem_ptr mem = (my_mem_ptr) cinfo->mem;
  JSAMPARRAY result;
  JSAMPROW workspace;
  JDIMENSION rowsperchunk,
    currow,
    i;
  long ltemp;
  ltemp = (1000000000L - ((size_t) sizeof(large_pool_hdr))) /
   ((long) samplesperrow * ((size_t) sizeof(JSAMPLE)));
  if (ltemp <= 0)
    ((cinfo)->err->msg_code = (JERR_WIDTH_OVERFLOW), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (ltemp < (long) numrows)
    rowsperchunk = (JDIMENSION) ltemp;
  else
    rowsperchunk = numrows;
  mem->last_rowsperchunk = rowsperchunk;
  result = (JSAMPARRAY) alloc_small(cinfo, pool_id,
   (size_t) (numrows * ((size_t) sizeof(JSAMPROW))));
  currow = 0;
  while (currow < numrows)
   {
    rowsperchunk = ((rowsperchunk) < (numrows - currow) ? (rowsperchunk) : (numrows - currow));
    workspace = (JSAMPROW) alloc_large(cinfo, pool_id,
     (size_t) ((size_t) rowsperchunk * (size_t) samplesperrow
      * ((size_t) sizeof(JSAMPLE))));
    for (i = rowsperchunk; i > 0; i--)
     {
      result[currow++] = workspace;
      workspace += samplesperrow;
      }
    }
  return result;
  }
static JBLOCKARRAY
  alloc_barray(j_common_ptr cinfo, int pool_id,
   JDIMENSION blocksperrow, JDIMENSION numrows)
 {
  my_mem_ptr mem = (my_mem_ptr) cinfo->mem;
  JBLOCKARRAY result;
  JBLOCKROW workspace;
  JDIMENSION rowsperchunk,
    currow,
    i;
  long ltemp;
  ltemp = (1000000000L - ((size_t) sizeof(large_pool_hdr))) /
   ((long) blocksperrow * ((size_t) sizeof(JBLOCK)));
  if (ltemp <= 0)
    ((cinfo)->err->msg_code = (JERR_WIDTH_OVERFLOW), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (ltemp < (long) numrows)
    rowsperchunk = (JDIMENSION) ltemp;
  else
    rowsperchunk = numrows;
  mem->last_rowsperchunk = rowsperchunk;
  result = (JBLOCKARRAY) alloc_small(cinfo, pool_id,
   (size_t) (numrows * ((size_t) sizeof(JBLOCKROW))));
  currow = 0;
  while (currow < numrows)
   {
    rowsperchunk = ((rowsperchunk) < (numrows - currow) ? (rowsperchunk) : (numrows - currow));
    workspace = (JBLOCKROW) alloc_large(cinfo, pool_id,
     (size_t) ((size_t) rowsperchunk * (size_t) blocksperrow
      * ((size_t) sizeof(JBLOCK))));
    for (i = rowsperchunk; i > 0; i--)
     {
      result[currow++] = workspace;
      workspace += blocksperrow;
      }
    }
  return result;
  }
static jvirt_sarray_ptr
  request_virt_sarray(j_common_ptr cinfo, int pool_id, boolean pre_zero,
   JDIMENSION samplesperrow, JDIMENSION numrows,
   JDIMENSION maxaccess)
 {
  my_mem_ptr mem = (my_mem_ptr) cinfo->mem;
  jvirt_sarray_ptr result;
  if (pool_id != 1)
    ((cinfo)->err->msg_code = (JERR_BAD_POOL_ID), (cinfo)->err->msg_parm.i[0] = (pool_id), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  result = (jvirt_sarray_ptr) alloc_small(cinfo, pool_id,
   ((size_t) sizeof(struct jvirt_sarray_control)));
  result->mem_buffer = ((void *) 0);
  result->rows_in_array = numrows;
  result->samplesperrow = samplesperrow;
  result->maxaccess = maxaccess;
  result->pre_zero = pre_zero;
  result->b_s_open = 0;
  result->next = mem->virt_sarray_list;
  mem->virt_sarray_list = result;
  return result;
  }
static jvirt_barray_ptr
  request_virt_barray(j_common_ptr cinfo, int pool_id, boolean pre_zero,
   JDIMENSION blocksperrow, JDIMENSION numrows,
   JDIMENSION maxaccess)
 {
  my_mem_ptr mem = (my_mem_ptr) cinfo->mem;
  jvirt_barray_ptr result;
  if (pool_id != 1)
    ((cinfo)->err->msg_code = (JERR_BAD_POOL_ID), (cinfo)->err->msg_parm.i[0] = (pool_id), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  result = (jvirt_barray_ptr) alloc_small(cinfo, pool_id,
   ((size_t) sizeof(struct jvirt_barray_control)));
  result->mem_buffer = ((void *) 0);
  result->rows_in_array = numrows;
  result->blocksperrow = blocksperrow;
  result->maxaccess = maxaccess;
  result->pre_zero = pre_zero;
  result->b_s_open = 0;
  result->next = mem->virt_barray_list;
  mem->virt_barray_list = result;
  return result;
  }
static void
  realize_virt_arrays(j_common_ptr cinfo)
 {
  my_mem_ptr mem = (my_mem_ptr) cinfo->mem;
  long space_per_minheight,
    maximum_space,
    avail_mem;
  long minheights,
    max_minheights;
  jvirt_sarray_ptr sptr;
  jvirt_barray_ptr bptr;
  space_per_minheight = 0;
  maximum_space = 0;
  for (sptr = mem->virt_sarray_list; sptr != ((void *) 0); sptr = sptr->next)
   {
    if (sptr->mem_buffer == ((void *) 0))
     {
      space_per_minheight += (long) sptr->maxaccess *
       (long) sptr->samplesperrow * ((size_t) sizeof(JSAMPLE));
      maximum_space += (long) sptr->rows_in_array *
       (long) sptr->samplesperrow * ((size_t) sizeof(JSAMPLE));
      }
    }
  for (bptr = mem->virt_barray_list; bptr != ((void *) 0); bptr = bptr->next)
   {
    if (bptr->mem_buffer == ((void *) 0))
     {
      space_per_minheight += (long) bptr->maxaccess *
       (long) bptr->blocksperrow * ((size_t) sizeof(JBLOCK));
      maximum_space += (long) bptr->rows_in_array *
       (long) bptr->blocksperrow * ((size_t) sizeof(JBLOCK));
      }
    }
  if (space_per_minheight <= 0)
    return;
  avail_mem = jpeg_mem_available(cinfo, space_per_minheight, maximum_space,
   mem->total_space_allocated);
  if (avail_mem >= maximum_space)
    max_minheights = 1000000000L;
  else
   {
    max_minheights = avail_mem / space_per_minheight;
    if (max_minheights <= 0)
      max_minheights = 1;
    }
  for (sptr = mem->virt_sarray_list; sptr != ((void *) 0); sptr = sptr->next)
   {
    if (sptr->mem_buffer == ((void *) 0))
     {
      minheights = ((long) sptr->rows_in_array - 1L) / sptr->maxaccess + 1L;
      if (minheights <= max_minheights)
       {
        sptr->rows_in_mem = sptr->rows_in_array;
        }
      else
       {
        sptr->rows_in_mem = (JDIMENSION) (max_minheights * sptr->maxaccess);
        jpeg_open_backing_store(cinfo, &sptr->b_s_info,
         (long) sptr->rows_in_array *
         (long) sptr->samplesperrow *
         (long) ((size_t) sizeof(JSAMPLE)));
        sptr->b_s_open = 1;
        }
      sptr->mem_buffer = alloc_sarray(cinfo, 1,
       sptr->samplesperrow, sptr->rows_in_mem);
      sptr->rowsperchunk = mem->last_rowsperchunk;
      sptr->cur_start_row = 0;
      sptr->first_undef_row = 0;
      sptr->dirty = 0;
      }
    }
  for (bptr = mem->virt_barray_list; bptr != ((void *) 0); bptr = bptr->next)
   {
    if (bptr->mem_buffer == ((void *) 0))
     {
      minheights = ((long) bptr->rows_in_array - 1L) / bptr->maxaccess + 1L;
      if (minheights <= max_minheights)
       {
        bptr->rows_in_mem = bptr->rows_in_array;
        }
      else
       {
        bptr->rows_in_mem = (JDIMENSION) (max_minheights * bptr->maxaccess);
        jpeg_open_backing_store(cinfo, &bptr->b_s_info,
         (long) bptr->rows_in_array *
         (long) bptr->blocksperrow *
         (long) ((size_t) sizeof(JBLOCK)));
        bptr->b_s_open = 1;
        }
      bptr->mem_buffer = alloc_barray(cinfo, 1,
       bptr->blocksperrow, bptr->rows_in_mem);
      bptr->rowsperchunk = mem->last_rowsperchunk;
      bptr->cur_start_row = 0;
      bptr->first_undef_row = 0;
      bptr->dirty = 0;
      }
    }
  }
static void
  do_sarray_io(j_common_ptr cinfo, jvirt_sarray_ptr ptr, boolean writing)
 {
  long bytesperrow,
    file_offset,
    byte_count,
    rows,
    thisrow,
    i;
  bytesperrow = (long) ptr->samplesperrow * ((size_t) sizeof(JSAMPLE));
  file_offset = ptr->cur_start_row * bytesperrow;
  for (i = 0; i < (long) ptr->rows_in_mem; i += ptr->rowsperchunk)
   {
    rows = (((long) ptr->rowsperchunk) < ((long) ptr->rows_in_mem - i) ? ((long) ptr->rowsperchunk) : ((long) ptr->rows_in_mem - i));
    thisrow = (long) ptr->cur_start_row + i;
    rows = ((rows) < ((long) ptr->first_undef_row - thisrow) ? (rows) : ((long) ptr->first_undef_row - thisrow));
    rows = ((rows) < ((long) ptr->rows_in_array - thisrow) ? (rows) : ((long) ptr->rows_in_array - thisrow));
    if (rows <= 0)
      break;
    byte_count = rows * bytesperrow;
    if (writing)
      (*ptr->b_s_info.write_backing_store) (cinfo, &ptr->b_s_info,
       (void *) ptr->mem_buffer[i],
       file_offset, byte_count);
    else
      (*ptr->b_s_info.read_backing_store) (cinfo, &ptr->b_s_info,
       (void *) ptr->mem_buffer[i],
       file_offset, byte_count);
    file_offset += byte_count;
    }
  }
static void
  do_barray_io(j_common_ptr cinfo, jvirt_barray_ptr ptr, boolean writing)
 {
  long bytesperrow,
    file_offset,
    byte_count,
    rows,
    thisrow,
    i;
  bytesperrow = (long) ptr->blocksperrow * ((size_t) sizeof(JBLOCK));
  file_offset = ptr->cur_start_row * bytesperrow;
  for (i = 0; i < (long) ptr->rows_in_mem; i += ptr->rowsperchunk)
   {
    rows = (((long) ptr->rowsperchunk) < ((long) ptr->rows_in_mem - i) ? ((long) ptr->rowsperchunk) : ((long) ptr->rows_in_mem - i));
    thisrow = (long) ptr->cur_start_row + i;
    rows = ((rows) < ((long) ptr->first_undef_row - thisrow) ? (rows) : ((long) ptr->first_undef_row - thisrow));
    rows = ((rows) < ((long) ptr->rows_in_array - thisrow) ? (rows) : ((long) ptr->rows_in_array - thisrow));
    if (rows <= 0)
      break;
    byte_count = rows * bytesperrow;
    if (writing)
      (*ptr->b_s_info.write_backing_store) (cinfo, &ptr->b_s_info,
       (void *) ptr->mem_buffer[i],
       file_offset, byte_count);
    else
      (*ptr->b_s_info.read_backing_store) (cinfo, &ptr->b_s_info,
       (void *) ptr->mem_buffer[i],
       file_offset, byte_count);
    file_offset += byte_count;
    }
  }
static JSAMPARRAY
  access_virt_sarray(j_common_ptr cinfo, jvirt_sarray_ptr ptr,
   JDIMENSION start_row, JDIMENSION num_rows,
   boolean writable)
 {
  JDIMENSION end_row = start_row + num_rows;
  JDIMENSION undef_row;
  if (end_row > ptr->rows_in_array || num_rows > ptr->maxaccess ||
   ptr->mem_buffer == ((void *) 0))
    ((cinfo)->err->msg_code = (JERR_BAD_VIRTUAL_ACCESS), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (start_row < ptr->cur_start_row ||
   end_row > ptr->cur_start_row + ptr->rows_in_mem)
   {
    if (!ptr->b_s_open)
      ((cinfo)->err->msg_code = (JERR_VIRTUAL_BUG), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (ptr->dirty)
     {
      do_sarray_io(cinfo, ptr, 1);
      ptr->dirty = 0;
      }
    if (start_row > ptr->cur_start_row)
     {
      ptr->cur_start_row = start_row;
      }
    else
     {
      long ltemp;
      ltemp = (long) end_row - (long) ptr->rows_in_mem;
      if (ltemp < 0)
        ltemp = 0;
      ptr->cur_start_row = (JDIMENSION) ltemp;
      }
    do_sarray_io(cinfo, ptr, 0);
    }
  if (ptr->first_undef_row < end_row)
   {
    if (ptr->first_undef_row < start_row)
     {
      if (writable)
        ((cinfo)->err->msg_code = (JERR_BAD_VIRTUAL_ACCESS), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      undef_row = start_row;
      }
    else
     {
      undef_row = ptr->first_undef_row;
      }
    if (writable)
      ptr->first_undef_row = end_row;
    if (ptr->pre_zero)
     {
      size_t bytesperrow = (size_t) ptr->samplesperrow * ((size_t) sizeof(JSAMPLE));
      undef_row -= ptr->cur_start_row;
      end_row -= ptr->cur_start_row;
      while (undef_row < end_row)
       {
        jzero_far((void *) ptr->mem_buffer[undef_row], bytesperrow);
        undef_row++;
        }
      }
    else
     {
      if (!writable)
        ((cinfo)->err->msg_code = (JERR_BAD_VIRTUAL_ACCESS), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      }
    }
  if (writable)
    ptr->dirty = 1;
  return ptr->mem_buffer + (start_row - ptr->cur_start_row);
  }
static JBLOCKARRAY
  access_virt_barray(j_common_ptr cinfo, jvirt_barray_ptr ptr,
   JDIMENSION start_row, JDIMENSION num_rows,
   boolean writable)
 {
  JDIMENSION end_row = start_row + num_rows;
  JDIMENSION undef_row;
  if (end_row > ptr->rows_in_array || num_rows > ptr->maxaccess ||
   ptr->mem_buffer == ((void *) 0))
    ((cinfo)->err->msg_code = (JERR_BAD_VIRTUAL_ACCESS), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (start_row < ptr->cur_start_row ||
   end_row > ptr->cur_start_row + ptr->rows_in_mem)
   {
    if (!ptr->b_s_open)
      ((cinfo)->err->msg_code = (JERR_VIRTUAL_BUG), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (ptr->dirty)
     {
      do_barray_io(cinfo, ptr, 1);
      ptr->dirty = 0;
      }
    if (start_row > ptr->cur_start_row)
     {
      ptr->cur_start_row = start_row;
      }
    else
     {
      long ltemp;
      ltemp = (long) end_row - (long) ptr->rows_in_mem;
      if (ltemp < 0)
        ltemp = 0;
      ptr->cur_start_row = (JDIMENSION) ltemp;
      }
    do_barray_io(cinfo, ptr, 0);
    }
  if (ptr->first_undef_row < end_row)
   {
    if (ptr->first_undef_row < start_row)
     {
      if (writable)
        ((cinfo)->err->msg_code = (JERR_BAD_VIRTUAL_ACCESS), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      undef_row = start_row;
      }
    else
     {
      undef_row = ptr->first_undef_row;
      }
    if (writable)
      ptr->first_undef_row = end_row;
    if (ptr->pre_zero)
     {
      size_t bytesperrow = (size_t) ptr->blocksperrow * ((size_t) sizeof(JBLOCK));
      undef_row -= ptr->cur_start_row;
      end_row -= ptr->cur_start_row;
      while (undef_row < end_row)
       {
        jzero_far((void *) ptr->mem_buffer[undef_row], bytesperrow);
        undef_row++;
        }
      }
    else
     {
      if (!writable)
        ((cinfo)->err->msg_code = (JERR_BAD_VIRTUAL_ACCESS), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      }
    }
  if (writable)
    ptr->dirty = 1;
  return ptr->mem_buffer + (start_row - ptr->cur_start_row);
  }
static void
  free_pool(j_common_ptr cinfo, int pool_id)
 {
  my_mem_ptr mem = (my_mem_ptr) cinfo->mem;
  small_pool_ptr shdr_ptr;
  large_pool_ptr lhdr_ptr;
  size_t space_freed;
  if (pool_id < 0 || pool_id >= 2)
    ((cinfo)->err->msg_code = (JERR_BAD_POOL_ID), (cinfo)->err->msg_parm.i[0] = (pool_id), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
#ifdef MEM_STATS
#endif
  if (pool_id == 1)
   {
    jvirt_sarray_ptr sptr;
    jvirt_barray_ptr bptr;
    for (sptr = mem->virt_sarray_list; sptr != ((void *) 0); sptr = sptr->next)
     {
      if (sptr->b_s_open)
       {
        sptr->b_s_open = 0;
        (*sptr->b_s_info.close_backing_store) (cinfo, &sptr->b_s_info);
        }
      }
    mem->virt_sarray_list = ((void *) 0);
    for (bptr = mem->virt_barray_list; bptr != ((void *) 0); bptr = bptr->next)
     {
      if (bptr->b_s_open)
       {
        bptr->b_s_open = 0;
        (*bptr->b_s_info.close_backing_store) (cinfo, &bptr->b_s_info);
        }
      }
    mem->virt_barray_list = ((void *) 0);
    }
  lhdr_ptr = mem->large_list[pool_id];
  mem->large_list[pool_id] = ((void *) 0);
  while (lhdr_ptr != ((void *) 0))
   {
    large_pool_ptr next_lhdr_ptr = lhdr_ptr->hdr.next;
    space_freed = lhdr_ptr->hdr.bytes_used +
     lhdr_ptr->hdr.bytes_left +
     ((size_t) sizeof(large_pool_hdr));
    jpeg_free_large(cinfo, (void *) lhdr_ptr, space_freed);
    mem->total_space_allocated -= space_freed;
    lhdr_ptr = next_lhdr_ptr;
    }
  shdr_ptr = mem->small_list[pool_id];
  mem->small_list[pool_id] = ((void *) 0);
  while (shdr_ptr != ((void *) 0))
   {
    small_pool_ptr next_shdr_ptr = shdr_ptr->hdr.next;
    space_freed = shdr_ptr->hdr.bytes_used +
     shdr_ptr->hdr.bytes_left +
     ((size_t) sizeof(small_pool_hdr));
    jpeg_free_small(cinfo, (void *) shdr_ptr, space_freed);
    mem->total_space_allocated -= space_freed;
    shdr_ptr = next_shdr_ptr;
    }
  }
static void
  self_destruct(j_common_ptr cinfo)
 {
  int pool;
  for (pool = 2 - 1; pool >= 0; pool--)
   {
    free_pool(cinfo, pool);
    }
  jpeg_free_small(cinfo, (void *) cinfo->mem, ((size_t) sizeof(my_memory_mgr)));
  cinfo->mem = ((void *) 0);
  jpeg_mem_term(cinfo);
  }
void
  jinit_memory_mgr(j_common_ptr cinfo)
 {
  my_mem_ptr mem;
  long max_to_use;
  int pool;
  size_t test_mac;
  cinfo->mem = ((void *) 0);
  if ((((size_t) sizeof(double)) & (((size_t) sizeof(double)) -1)) != 0)
    ((cinfo)->err->msg_code = (JERR_BAD_ALIGN_TYPE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  test_mac = (size_t) 1000000000L;
  if ((long) test_mac != 1000000000L ||
   (1000000000L % ((size_t) sizeof(double))) != 0)
    ((cinfo)->err->msg_code = (JERR_BAD_ALLOC_CHUNK), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  max_to_use = jpeg_mem_init(cinfo);
  mem = (my_mem_ptr) jpeg_get_small(cinfo, ((size_t) sizeof(my_memory_mgr)));
  if (mem == ((void *) 0))
   {
    jpeg_mem_term(cinfo);
    ((cinfo)->err->msg_code = (JERR_OUT_OF_MEMORY), (cinfo)->err->msg_parm.i[0] = (0), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  mem->pub.alloc_small = alloc_small;
  mem->pub.alloc_large = alloc_large;
  mem->pub.alloc_sarray = alloc_sarray;
  mem->pub.alloc_barray = alloc_barray;
  mem->pub.request_virt_sarray = request_virt_sarray;
  mem->pub.request_virt_barray = request_virt_barray;
  mem->pub.realize_virt_arrays = realize_virt_arrays;
  mem->pub.access_virt_sarray = access_virt_sarray;
  mem->pub.access_virt_barray = access_virt_barray;
  mem->pub.free_pool = free_pool;
  mem->pub.self_destruct = self_destruct;
  mem->pub.max_alloc_chunk = 1000000000L;
  mem->pub.max_memory_to_use = max_to_use;
  for (pool = 2 - 1; pool >= 0; pool--)
   {
    mem->small_list[pool] = ((void *) 0);
    mem->large_list[pool] = ((void *) 0);
    }
  mem->virt_sarray_list = ((void *) 0);
  mem->virt_barray_list = ((void *) 0);
  mem->total_space_allocated = ((size_t) sizeof(my_memory_mgr));
  cinfo->mem = &mem->pub;
#ifndef NO_GETENV
   {
    char *memenv;
    if ((memenv = getenv("JPEGMEM")) != ((void *) 0))
     {
      char ch = 'x';
      if (sscanf(memenv, "%ld%c", &max_to_use, &ch) > 0)
       {
        if (ch == 'm' || ch == 'M')
          max_to_use *= 1000L;
        mem->pub.max_memory_to_use = max_to_use * 1000L;
        }
      }
    }
#endif
  }
#define JPEG_INTERNALS
#ifndef HAVE_STDLIB_H
#endif
#ifndef SEEK_SET
#endif
void *
  jpeg_get_small(j_common_ptr cinfo, size_t sizeofobject)
 {
  return (void *) malloc(sizeofobject);
  }
void
  jpeg_free_small(j_common_ptr cinfo, void *object, size_t sizeofobject)
 {
  free(object);
  }
void *
  jpeg_get_large(j_common_ptr cinfo, size_t sizeofobject)
 {
  return (void *) malloc(sizeofobject);
  }
void
  jpeg_free_large(j_common_ptr cinfo, void *object, size_t sizeofobject)
 {
  free(object);
  }
#ifndef DEFAULT_MAX_MEM
#define DEFAULT_MAX_MEM  1000000L
#endif
long
  jpeg_mem_available(j_common_ptr cinfo, long min_bytes_needed,
   long max_bytes_needed, long already_allocated)
 {
  return cinfo->mem->max_memory_to_use - already_allocated;
  }
static void
  read_backing_store(j_common_ptr cinfo, backing_store_ptr info,
   void *buffer_address,
   long file_offset, long byte_count)
 {
  if (fseek(info->temp_file, file_offset, 0))
    ((cinfo)->err->msg_code = (JERR_TFILE_SEEK), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (((size_t) fread((void *) (buffer_address), (size_t) 1, (size_t) (byte_count), (info->temp_file)))
   != (size_t) byte_count)
    ((cinfo)->err->msg_code = (JERR_TFILE_READ), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  }
static void
  write_backing_store(j_common_ptr cinfo, backing_store_ptr info,
   void *buffer_address,
   long file_offset, long byte_count)
 {
  if (fseek(info->temp_file, file_offset, 0))
    ((cinfo)->err->msg_code = (JERR_TFILE_SEEK), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (((size_t) fwrite((const void *) (buffer_address), (size_t) 1, (size_t) (byte_count), (info->temp_file)))
   != (size_t) byte_count)
    ((cinfo)->err->msg_code = (JERR_TFILE_WRITE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  }
static void
  close_backing_store(j_common_ptr cinfo, backing_store_ptr info)
 {
  fclose(info->temp_file);
  }
void
  jpeg_open_backing_store(j_common_ptr cinfo, backing_store_ptr info,
   long total_bytes_needed)
 {
  if ((info->temp_file = tmpfile()) == ((void *) 0))
    ((cinfo)->err->msg_code = (JERR_TFILE_CREATE), strncpy((cinfo)->err->msg_parm.s, (""), 80), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  info->read_backing_store = read_backing_store;
  info->write_backing_store = write_backing_store;
  info->close_backing_store = close_backing_store;
  }
long
  jpeg_mem_init(j_common_ptr cinfo)
 {
  return 1000000L;
  }
void
  jpeg_mem_term(j_common_ptr cinfo)
 {
  }
#define JPEG_INTERNALS
#if 0
#endif
const int jpeg_natural_order[64 + 16] = {
  0, 1, 8, 16, 9, 2, 3, 10,
  17, 24, 32, 25, 18, 11, 4, 5,
  12, 19, 26, 33, 40, 48, 41, 34,
  27, 20, 13, 6, 7, 14, 21, 28,
  35, 42, 49, 56, 57, 50, 43, 36,
  29, 22, 15, 23, 30, 37, 44, 51,
  58, 59, 52, 45, 38, 31, 39, 46,
  53, 60, 61, 54, 47, 55, 62, 63,
  63, 63, 63, 63, 63, 63, 63, 63,
  63, 63, 63, 63, 63, 63, 63, 63
  };
long
  jdiv_round_up(long a, long b)
 {
  return (a + b - 1L) / b;
  }
long
  jround_up(long a, long b)
 {
  a += b - 1L;
  return a - (a % b);
  }
#ifndef NEED_FAR_POINTERS
#define FMEMCOPY(dest,src,size) MEMCOPY(dest,src,size)
#define FMEMZERO(target,size) MEMZERO(target,size)
#else
#endif
void
  jcopy_sample_rows(JSAMPARRAY input_array, int source_row,
   JSAMPARRAY output_array, int dest_row,
   int num_rows, JDIMENSION num_cols)
 {
  register JSAMPROW inptr,
    outptr;
#ifdef FMEMCOPY
  register size_t count = (size_t) (num_cols * ((size_t) sizeof(JSAMPLE)));
#else
#endif
  register int row;
  input_array += source_row;
  output_array += dest_row;
  for (row = num_rows; row > 0; row--)
   {
    inptr = *input_array++;
    outptr = *output_array++;
#ifdef FMEMCOPY
    memcpy((void *) (outptr), (const void *) (inptr), (size_t) (count));
#else
#endif
    }
  }
void
  jcopy_block_row(JBLOCKROW input_row, JBLOCKROW output_row,
   JDIMENSION num_blocks)
 {
#ifdef FMEMCOPY
  memcpy((void *) (output_row), (const void *) (input_row), (size_t) (num_blocks * (64 * ((size_t) sizeof(JCOEF)))));
#else
#endif
  }
void
  jzero_far(void *target, size_t bytestozero)
 {
#ifdef FMEMZERO
  memset((void *) (target), 0, (size_t) (bytestozero));
#else
#endif
  }
#define JPEG_INTERNALS
void
  jpeg_start_compress(j_compress_ptr cinfo, boolean write_all_tables)
 {
  if (cinfo->global_state != 100)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (write_all_tables)
    jpeg_suppress_tables(cinfo, 0);
  (*cinfo->err->reset_error_mgr) ((j_common_ptr) cinfo);
  (*cinfo->dest->init_destination) (cinfo);
  jinit_compress_master(cinfo);
  (*cinfo->master->prepare_for_pass) (cinfo);
  cinfo->next_scanline = 0;
  cinfo->global_state = (cinfo->raw_data_in ? 102 : 101);
  }
JDIMENSION
jpeg_write_scanlines(j_compress_ptr cinfo, JSAMPARRAY scanlines,
 JDIMENSION num_lines)
 {
  JDIMENSION row_ctr,
    rows_left;
  if (cinfo->global_state != 101)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->next_scanline >= cinfo->image_height)
    ((cinfo)->err->msg_code = (JWRN_TOO_MUCH_DATA), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
  if (cinfo->progress != ((void *) 0))
   {
    cinfo->progress->pass_counter = (long) cinfo->next_scanline;
    cinfo->progress->pass_limit = (long) cinfo->image_height;
    (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
    }
  if (cinfo->master->call_pass_startup)
    (*cinfo->master->pass_startup) (cinfo);
  rows_left = cinfo->image_height - cinfo->next_scanline;
  if (num_lines > rows_left)
    num_lines = rows_left;
  row_ctr = 0;
  (*cinfo->main->process_data) (cinfo, scanlines, &row_ctr, num_lines);
  cinfo->next_scanline += row_ctr;
  return row_ctr;
  }
JDIMENSION
jpeg_write_raw_data(j_compress_ptr cinfo, JSAMPIMAGE data,
 JDIMENSION num_lines)
 {
  JDIMENSION lines_per_iMCU_row;
  if (cinfo->global_state != 102)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->next_scanline >= cinfo->image_height)
   {
    ((cinfo)->err->msg_code = (JWRN_TOO_MUCH_DATA), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
    return 0;
    }
  if (cinfo->progress != ((void *) 0))
   {
    cinfo->progress->pass_counter = (long) cinfo->next_scanline;
    cinfo->progress->pass_limit = (long) cinfo->image_height;
    (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
    }
  if (cinfo->master->call_pass_startup)
    (*cinfo->master->pass_startup) (cinfo);
  lines_per_iMCU_row = cinfo->max_v_samp_factor * 8;
  if (num_lines < lines_per_iMCU_row)
    ((cinfo)->err->msg_code = (JERR_BUFFER_SIZE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (!(*cinfo->coef->compress_data) (cinfo, data))
   {
    return 0;
    }
  cinfo->next_scanline += lines_per_iMCU_row;
  return lines_per_iMCU_row;
  }
#define JPEG_INTERNALS
static boolean output_pass_setup(j_decompress_ptr cinfo);
boolean
jpeg_start_decompress(j_decompress_ptr cinfo)
 {
  if (cinfo->global_state == 202)
   {
    jinit_master_decompress(cinfo);
    if (cinfo->buffered_image)
     {
      cinfo->global_state = 207;
      return 1;
      }
    cinfo->global_state = 203;
    }
  if (cinfo->global_state == 203)
   {
    if (cinfo->inputctl->has_multiple_scans)
     {
#ifdef D_MULTISCAN_FILES_SUPPORTED
      for (;;)
       {
        int retcode;
        if (cinfo->progress != ((void *) 0))
          (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
        retcode = (*cinfo->inputctl->consume_input) (cinfo);
        if (retcode == 0)
          return 0;
        if (retcode == 2)
          break;
        if (cinfo->progress != ((void *) 0) &&
         (retcode == 3 || retcode == 1))
         {
          if (++cinfo->progress->pass_counter >= cinfo->progress->pass_limit)
           {
            cinfo->progress->pass_limit += (long) cinfo->total_iMCU_rows;
            }
          }
        }
#else
#endif
      }
    cinfo->output_scan_number = cinfo->input_scan_number;
    }
  else if (cinfo->global_state != 204)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  return output_pass_setup(cinfo);
  }
static boolean
  output_pass_setup(j_decompress_ptr cinfo)
 {
  if (cinfo->global_state != 204)
   {
    (*cinfo->master->prepare_for_output_pass) (cinfo);
    cinfo->output_scanline = 0;
    cinfo->global_state = 204;
    }
  while (cinfo->master->is_dummy_pass)
   {
#ifdef QUANT_2PASS_SUPPORTED
    while (cinfo->output_scanline < cinfo->output_height)
     {
      JDIMENSION last_scanline;
      if (cinfo->progress != ((void *) 0))
       {
        cinfo->progress->pass_counter = (long) cinfo->output_scanline;
        cinfo->progress->pass_limit = (long) cinfo->output_height;
        (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
        }
      last_scanline = cinfo->output_scanline;
      (*cinfo->main->process_data) (cinfo, (JSAMPARRAY) ((void *) 0),
       &cinfo->output_scanline, (JDIMENSION) 0);
      if (cinfo->output_scanline == last_scanline)
        return 0;
      }
    (*cinfo->master->finish_output_pass) (cinfo);
    (*cinfo->master->prepare_for_output_pass) (cinfo);
    cinfo->output_scanline = 0;
#else
#endif
    }
  cinfo->global_state = cinfo->raw_data_out ? 206 : 205;
  return 1;
  }
JDIMENSION
jpeg_read_scanlines(j_decompress_ptr cinfo, JSAMPARRAY scanlines,
 JDIMENSION max_lines)
 {
  JDIMENSION row_ctr;
  if (cinfo->global_state != 205)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->output_scanline >= cinfo->output_height)
   {
    ((cinfo)->err->msg_code = (JWRN_TOO_MUCH_DATA), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
    return 0;
    }
  if (cinfo->progress != ((void *) 0))
   {
    cinfo->progress->pass_counter = (long) cinfo->output_scanline;
    cinfo->progress->pass_limit = (long) cinfo->output_height;
    (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
    }
  row_ctr = 0;
  (*cinfo->main->process_data) (cinfo, scanlines, &row_ctr, max_lines);
  cinfo->output_scanline += row_ctr;
  return row_ctr;
  }
JDIMENSION
jpeg_read_raw_data(j_decompress_ptr cinfo, JSAMPIMAGE data,
 JDIMENSION max_lines)
 {
  JDIMENSION lines_per_iMCU_row;
  if (cinfo->global_state != 206)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->output_scanline >= cinfo->output_height)
   {
    ((cinfo)->err->msg_code = (JWRN_TOO_MUCH_DATA), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
    return 0;
    }
  if (cinfo->progress != ((void *) 0))
   {
    cinfo->progress->pass_counter = (long) cinfo->output_scanline;
    cinfo->progress->pass_limit = (long) cinfo->output_height;
    (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
    }
  lines_per_iMCU_row = cinfo->max_v_samp_factor * cinfo->min_DCT_scaled_size;
  if (max_lines < lines_per_iMCU_row)
    ((cinfo)->err->msg_code = (JERR_BUFFER_SIZE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (!(*cinfo->coef->decompress_data) (cinfo, data))
    return 0;
  cinfo->output_scanline += lines_per_iMCU_row;
  return lines_per_iMCU_row;
  }
#ifdef D_MULTISCAN_FILES_SUPPORTED
boolean
jpeg_start_output(j_decompress_ptr cinfo, int scan_number)
 {
  if (cinfo->global_state != 207 &&
   cinfo->global_state != 204)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (scan_number <= 0)
    scan_number = 1;
  if (cinfo->inputctl->eoi_reached &&
   scan_number > cinfo->input_scan_number)
    scan_number = cinfo->input_scan_number;
  cinfo->output_scan_number = scan_number;
  return output_pass_setup(cinfo);
  }
boolean
jpeg_finish_output(j_decompress_ptr cinfo)
 {
  if ((cinfo->global_state == 205 ||
    cinfo->global_state == 206) && cinfo->buffered_image)
   {
    (*cinfo->master->finish_output_pass) (cinfo);
    cinfo->global_state = 208;
    }
  else if (cinfo->global_state != 208)
   {
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  while (cinfo->input_scan_number <= cinfo->output_scan_number &&
   !cinfo->inputctl->eoi_reached)
   {
    if ((*cinfo->inputctl->consume_input) (cinfo) == 0)
      return 0;
    }
  cinfo->global_state = 207;
  return 1;
  }
#endif
#define JPEG_INTERNALS
void
  jpeg_add_quant_table(j_compress_ptr cinfo, int which_tbl,
   const unsigned int *basic_table,
   int scale_factor, boolean force_baseline)
 {
  JQUANT_TBL **qtblptr;
  int i;
  long temp;
  if (cinfo->global_state != 100)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (which_tbl < 0 || which_tbl >= 4)
    ((cinfo)->err->msg_code = (JERR_DQT_INDEX), (cinfo)->err->msg_parm.i[0] = (which_tbl), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  qtblptr = &cinfo->quant_tbl_ptrs[which_tbl];
  if (*qtblptr == ((void *) 0))
    *qtblptr = jpeg_alloc_quant_table((j_common_ptr) cinfo);
  for (i = 0; i < 64; i++)
   {
    temp = ((long) basic_table[i] * scale_factor + 50L) / 100L;
    if (temp <= 0L)
      temp = 1L;
    if (temp > 32767L)
      temp = 32767L;
    if (force_baseline && temp > 255L)
      temp = 255L;
    (*qtblptr)->quantval[i] = (UINT16) temp;
    }
  (*qtblptr)->sent_table = 0;
  }
void
  jpeg_set_linear_quality(j_compress_ptr cinfo, int scale_factor,
   boolean force_baseline)
 {
   static const unsigned int std_luminance_quant_tbl[64] = {
    16, 11, 10, 16, 24, 40, 51, 61,
    12, 12, 14, 19, 26, 58, 60, 55,
    14, 13, 16, 24, 40, 57, 69, 56,
    14, 17, 22, 29, 51, 87, 80, 62,
    18, 22, 37, 56, 68, 109, 103, 77,
    24, 35, 55, 64, 81, 104, 113, 92,
    49, 64, 78, 87, 103, 121, 120, 101,
    72, 92, 95, 98, 112, 100, 103, 99
    };
   static const unsigned int std_chrominance_quant_tbl[64] = {
    17, 18, 24, 47, 99, 99, 99, 99,
    18, 21, 26, 66, 99, 99, 99, 99,
    24, 26, 56, 99, 99, 99, 99, 99,
    47, 66, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99
    };
  jpeg_add_quant_table(cinfo, 0, std_luminance_quant_tbl,
   scale_factor, force_baseline);
  jpeg_add_quant_table(cinfo, 1, std_chrominance_quant_tbl,
   scale_factor, force_baseline);
  }
int
  jpeg_quality_scaling(int quality)
 {
  if (quality <= 0)
    quality = 1;
  if (quality > 100)
    quality = 100;
  if (quality < 50)
    quality = 5000 / quality;
  else
    quality = 200 - quality * 2;
  return quality;
  }
void
  jpeg_set_quality(j_compress_ptr cinfo, int quality, boolean force_baseline)
 {
  quality = jpeg_quality_scaling(quality);
  jpeg_set_linear_quality(cinfo, quality, force_baseline);
  }
static void
  add_huff_table(j_compress_ptr cinfo,
   JHUFF_TBL ** htblptr, const UINT8 * bits, const UINT8 * val)
 {
  int nsymbols,
    len;
  if (*htblptr == ((void *) 0))
    *htblptr = jpeg_alloc_huff_table((j_common_ptr) cinfo);
  memcpy((void *) ((*htblptr)->bits), (const void *) (bits), (size_t) (((size_t) sizeof((*htblptr)->bits))));
  nsymbols = 0;
  for (len = 1; len <= 16; len++)
    nsymbols += bits[len];
  if (nsymbols < 1 || nsymbols > 256)
    ((cinfo)->err->msg_code = (JERR_BAD_HUFF_TABLE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  memcpy((void *) ((*htblptr)->huffval), (const void *) (val), (size_t) (nsymbols * ((size_t) sizeof(UINT8))));
  (*htblptr)->sent_table = 0;
  }
static void
  std_huff_tables(j_compress_ptr cinfo)
 {
  static const UINT8 bits_dc_luminance[17] =
  {0, 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0};
  static const UINT8 val_dc_luminance[] =
  {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11};
  static const UINT8 bits_dc_chrominance[17] =
  {0, 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0};
  static const UINT8 val_dc_chrominance[] =
  {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11};
  static const UINT8 bits_ac_luminance[17] =
  {0, 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d};
  static const UINT8 val_ac_luminance[] =
   {0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
    0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
    0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
    0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
    0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
    0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
    0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
    0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
    0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
    0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
    0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
    0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
    0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
    0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
    0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
    0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
    0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
    0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
    0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
    0xf9, 0xfa};
  static const UINT8 bits_ac_chrominance[17] =
  {0, 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77};
  static const UINT8 val_ac_chrominance[] =
   {0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
    0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
    0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
    0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
    0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
    0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
    0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
    0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
    0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
    0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
    0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
    0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
    0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
    0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
    0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
    0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
    0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
    0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
    0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
    0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
    0xf9, 0xfa};
  add_huff_table(cinfo, &cinfo->dc_huff_tbl_ptrs[0],
   bits_dc_luminance, val_dc_luminance);
  add_huff_table(cinfo, &cinfo->ac_huff_tbl_ptrs[0],
   bits_ac_luminance, val_ac_luminance);
  add_huff_table(cinfo, &cinfo->dc_huff_tbl_ptrs[1],
   bits_dc_chrominance, val_dc_chrominance);
  add_huff_table(cinfo, &cinfo->ac_huff_tbl_ptrs[1],
   bits_ac_chrominance, val_ac_chrominance);
  }
void
  jpeg_set_defaults(j_compress_ptr cinfo)
 {
  int i;
  if (cinfo->global_state != 100)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->comp_info == ((void *) 0))
    cinfo->comp_info = (jpeg_component_info *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 0,
     10 * ((size_t) sizeof(jpeg_component_info)));
  cinfo->data_precision = 8;
  jpeg_set_quality(cinfo, 75, 1);
  std_huff_tables(cinfo);
  for (i = 0; i < 16; i++)
   {
    cinfo->arith_dc_L[i] = 0;
    cinfo->arith_dc_U[i] = 1;
    cinfo->arith_ac_K[i] = 5;
    }
  cinfo->scan_info = ((void *) 0);
  cinfo->num_scans = 0;
  cinfo->raw_data_in = 0;
  cinfo->arith_code = 0;
  cinfo->optimize_coding = 0;
  if (cinfo->data_precision > 8)
    cinfo->optimize_coding = 1;
  cinfo->CCIR601_sampling = 0;
  cinfo->smoothing_factor = 0;
  cinfo->dct_method = JDCT_ISLOW;
  cinfo->restart_interval = 0;
  cinfo->restart_in_rows = 0;
  cinfo->JFIF_major_version = 1;
  cinfo->JFIF_minor_version = 1;
  cinfo->density_unit = 0;
  cinfo->X_density = 1;
  cinfo->Y_density = 1;
  jpeg_default_colorspace(cinfo);
  }
void
  jpeg_default_colorspace(j_compress_ptr cinfo)
 {
  switch (cinfo->in_color_space)
   {
 case JCS_GRAYSCALE:
    jpeg_set_colorspace(cinfo, JCS_GRAYSCALE);
    break;
   case JCS_RGB:
    jpeg_set_colorspace(cinfo, JCS_YCbCr);
    break;
   case JCS_YCbCr:
    jpeg_set_colorspace(cinfo, JCS_YCbCr);
    break;
   case JCS_CMYK:
    jpeg_set_colorspace(cinfo, JCS_CMYK);
    break;
   case JCS_YCCK:
    jpeg_set_colorspace(cinfo, JCS_YCCK);
    break;
   case JCS_UNKNOWN:
    jpeg_set_colorspace(cinfo, JCS_UNKNOWN);
    break;
   default:
    ((cinfo)->err->msg_code = (JERR_BAD_IN_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  }
void
  jpeg_set_colorspace(j_compress_ptr cinfo, J_COLOR_SPACE colorspace)
 {
  jpeg_component_info *compptr;
  int ci;
#define SET_COMP(index,id,hsamp,vsamp,quant,dctbl,actbl)    (compptr = &cinfo->comp_info[index],    compptr->component_id = (id),    compptr->h_samp_factor = (hsamp),    compptr->v_samp_factor = (vsamp),    compptr->quant_tbl_no = (quant),    compptr->dc_tbl_no = (dctbl),    compptr->ac_tbl_no = (actbl) )
  if (cinfo->global_state != 100)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  cinfo->jpeg_color_space = colorspace;
  cinfo->write_JFIF_header = 0;
  cinfo->write_Adobe_marker = 0;
  switch (colorspace)
   {
   case JCS_GRAYSCALE:
    cinfo->write_JFIF_header = 1;
    cinfo->num_components = 1;
    (compptr = &cinfo->comp_info[0], compptr->component_id = (1), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    break;
   case JCS_RGB:
    cinfo->write_Adobe_marker = 1;
    cinfo->num_components = 3;
    (compptr = &cinfo->comp_info[0], compptr->component_id = (0x52), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    (compptr = &cinfo->comp_info[1], compptr->component_id = (0x47), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    (compptr = &cinfo->comp_info[2], compptr->component_id = (0x42), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    break;
   case JCS_YCbCr:
    cinfo->write_JFIF_header = 1;
    cinfo->num_components = 3;
    (compptr = &cinfo->comp_info[0], compptr->component_id = (1), compptr->h_samp_factor = (2), compptr->v_samp_factor = (2), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    (compptr = &cinfo->comp_info[1], compptr->component_id = (2), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (1), compptr->dc_tbl_no = (1), compptr->ac_tbl_no = (1));
    (compptr = &cinfo->comp_info[2], compptr->component_id = (3), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (1), compptr->dc_tbl_no = (1), compptr->ac_tbl_no = (1));
    break;
   case JCS_CMYK:
    cinfo->write_Adobe_marker = 1;
    cinfo->num_components = 4;
    (compptr = &cinfo->comp_info[0], compptr->component_id = (0x43), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    (compptr = &cinfo->comp_info[1], compptr->component_id = (0x4D), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    (compptr = &cinfo->comp_info[2], compptr->component_id = (0x59), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    (compptr = &cinfo->comp_info[3], compptr->component_id = (0x4B), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    break;
   case JCS_YCCK:
    cinfo->write_Adobe_marker = 1;
    cinfo->num_components = 4;
    (compptr = &cinfo->comp_info[0], compptr->component_id = (1), compptr->h_samp_factor = (2), compptr->v_samp_factor = (2), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    (compptr = &cinfo->comp_info[1], compptr->component_id = (2), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (1), compptr->dc_tbl_no = (1), compptr->ac_tbl_no = (1));
    (compptr = &cinfo->comp_info[2], compptr->component_id = (3), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (1), compptr->dc_tbl_no = (1), compptr->ac_tbl_no = (1));
    (compptr = &cinfo->comp_info[3], compptr->component_id = (4), compptr->h_samp_factor = (2), compptr->v_samp_factor = (2), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
    break;
   case JCS_UNKNOWN:
    cinfo->num_components = cinfo->input_components;
    if (cinfo->num_components < 1 || cinfo->num_components > 10)
      ((cinfo)->err->msg_code = (JERR_COMPONENT_COUNT), (cinfo)->err->msg_parm.i[0] = (cinfo->num_components), (cinfo)->err->msg_parm.i[1] = (10), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    for (ci = 0; ci < cinfo->num_components; ci++)
     {
      (compptr = &cinfo->comp_info[ci], compptr->component_id = (ci), compptr->h_samp_factor = (1), compptr->v_samp_factor = (1), compptr->quant_tbl_no = (0), compptr->dc_tbl_no = (0), compptr->ac_tbl_no = (0));
      }
    break;
   default:
    ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  }
#ifdef C_PROGRESSIVE_SUPPORTED
static jpeg_scan_info *
  fill_a_scan(jpeg_scan_info * scanptr, int ci,
   int Ss, int Se, int Ah, int Al)
 {
  scanptr->comps_in_scan = 1;
  scanptr->component_index[0] = ci;
  scanptr->Ss = Ss;
  scanptr->Se = Se;
  scanptr->Ah = Ah;
  scanptr->Al = Al;
  scanptr++;
  return scanptr;
  }
static jpeg_scan_info *
  fill_scans(jpeg_scan_info * scanptr, int ncomps,
   int Ss, int Se, int Ah, int Al)
 {
  int ci;
  for (ci = 0; ci < ncomps; ci++)
   {
    scanptr->comps_in_scan = 1;
    scanptr->component_index[0] = ci;
    scanptr->Ss = Ss;
    scanptr->Se = Se;
    scanptr->Ah = Ah;
    scanptr->Al = Al;
    scanptr++;
    }
  return scanptr;
  }
static jpeg_scan_info *
  fill_dc_scans(jpeg_scan_info * scanptr, int ncomps, int Ah, int Al)
 {
  int ci;
  if (ncomps <= 4)
   {
    scanptr->comps_in_scan = ncomps;
    for (ci = 0; ci < ncomps; ci++)
      scanptr->component_index[ci] = ci;
    scanptr->Ss = scanptr->Se = 0;
    scanptr->Ah = Ah;
    scanptr->Al = Al;
    scanptr++;
    }
  else
   {
    scanptr = fill_scans(scanptr, ncomps, 0, 0, Ah, Al);
    }
  return scanptr;
  }
void
  jpeg_simple_progression(j_compress_ptr cinfo)
 {
  int ncomps = cinfo->num_components;
  int nscans;
  jpeg_scan_info *scanptr;
  if (cinfo->global_state != 100)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (ncomps == 3 && cinfo->jpeg_color_space == JCS_YCbCr)
   {
    nscans = 10;
    }
  else
   {
    if (ncomps > 4)
      nscans = 6 * ncomps;
    else
      nscans = 2 + 4 * ncomps;
    }
  if (cinfo->script_space == ((void *) 0) || cinfo->script_space_size < nscans)
   {
    cinfo->script_space_size = ((nscans) > (10) ? (nscans) : (10));
    cinfo->script_space = (jpeg_scan_info *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 0,
     cinfo->script_space_size * ((size_t) sizeof(jpeg_scan_info)));
    }
  scanptr = cinfo->script_space;
  cinfo->scan_info = scanptr;
  cinfo->num_scans = nscans;
  if (ncomps == 3 && cinfo->jpeg_color_space == JCS_YCbCr)
   {
    scanptr = fill_dc_scans(scanptr, ncomps, 0, 1);
    scanptr = fill_a_scan(scanptr, 0, 1, 5, 0, 2);
    scanptr = fill_a_scan(scanptr, 2, 1, 63, 0, 1);
    scanptr = fill_a_scan(scanptr, 1, 1, 63, 0, 1);
    scanptr = fill_a_scan(scanptr, 0, 6, 63, 0, 2);
    scanptr = fill_a_scan(scanptr, 0, 1, 63, 2, 1);
    scanptr = fill_dc_scans(scanptr, ncomps, 1, 0);
    scanptr = fill_a_scan(scanptr, 2, 1, 63, 1, 0);
    scanptr = fill_a_scan(scanptr, 1, 1, 63, 1, 0);
    scanptr = fill_a_scan(scanptr, 0, 1, 63, 1, 0);
    }
  else
   {
    scanptr = fill_dc_scans(scanptr, ncomps, 0, 1);
    scanptr = fill_scans(scanptr, ncomps, 1, 5, 0, 2);
    scanptr = fill_scans(scanptr, ncomps, 6, 63, 0, 2);
    scanptr = fill_scans(scanptr, ncomps, 1, 63, 2, 1);
    scanptr = fill_dc_scans(scanptr, ncomps, 1, 0);
    scanptr = fill_scans(scanptr, ncomps, 1, 63, 1, 0);
    }
  }
#endif
typedef struct
 {
  struct jpeg_source_mgr pub;
  FILE *infile;
  JOCTET *buffer;
  boolean start_of_file;
  } my_source_mgr;
typedef my_source_mgr *my_src_ptr;
#define INPUT_BUF_SIZE  4096
static void
  init_source(j_decompress_ptr cinfo)
 {
  my_src_ptr src = (my_src_ptr) cinfo->src;
  src->start_of_file = 1;
  }
static boolean
  fill_input_buffer(j_decompress_ptr cinfo)
 {
  my_src_ptr src = (my_src_ptr) cinfo->src;
  size_t nbytes;
  nbytes = ((size_t) fread((void *) (src->buffer), (size_t) 1, (size_t) (4096), (src->infile)));
  if (nbytes <= 0)
   {
    if (src->start_of_file)
      ((cinfo)->err->msg_code = (JERR_INPUT_EMPTY), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    ((cinfo)->err->msg_code = (JWRN_JPEG_EOF), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
    src->buffer[0] = (JOCTET) 0xFF;
    src->buffer[1] = (JOCTET) 0xD9;
    nbytes = 2;
    }
  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = nbytes;
  src->start_of_file = 0;
  return 1;
  }
static void
  skip_input_data(j_decompress_ptr cinfo, long num_bytes)
 {
  my_src_ptr src = (my_src_ptr) cinfo->src;
  if (num_bytes > 0)
   {
    while (num_bytes > (long) src->pub.bytes_in_buffer)
     {
      num_bytes -= (long) src->pub.bytes_in_buffer;
      (void) fill_input_buffer(cinfo);
      }
    src->pub.next_input_byte += (size_t) num_bytes;
    src->pub.bytes_in_buffer -= (size_t) num_bytes;
    }
  }
static void
  term_source(j_decompress_ptr cinfo)
 {
  }
void
  jpeg_stdio_src(j_decompress_ptr cinfo, FILE * infile)
 {
  my_src_ptr src;
  if (cinfo->src == ((void *) 0))
   {
    cinfo->src = (struct jpeg_source_mgr *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 0,
     ((size_t) sizeof(my_source_mgr)));
    src = (my_src_ptr) cinfo->src;
    src->buffer = (JOCTET *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 0,
     4096 * ((size_t) sizeof(JOCTET)));
    }
  src = (my_src_ptr) cinfo->src;
  src->pub.init_source = init_source;
  src->pub.fill_input_buffer = fill_input_buffer;
  src->pub.skip_input_data = skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart;
  src->pub.term_source = term_source;
  src->infile = infile;
  src->pub.bytes_in_buffer = 0;
  src->pub.next_input_byte = ((void *) 0);
  }
typedef struct
 {
  struct jpeg_destination_mgr pub;
  FILE *outfile;
  JOCTET *buffer;
  } my_destination_mgr;
typedef my_destination_mgr *my_dest_ptr;
#define OUTPUT_BUF_SIZE  4096
static void
  init_destination(j_compress_ptr cinfo)
 {
  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
  dest->buffer = (JOCTET *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   4096 * ((size_t) sizeof(JOCTET)));
  dest->pub.next_output_byte = dest->buffer;
  dest->pub.free_in_buffer = 4096;
  }
static boolean
  empty_output_buffer(j_compress_ptr cinfo)
 {
  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
  if (((size_t) fwrite((const void *) (dest->buffer), (size_t) 1, (size_t) (4096), (dest->outfile))) !=
   (size_t) 4096)
    ((cinfo)->err->msg_code = (JERR_FILE_WRITE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  dest->pub.next_output_byte = dest->buffer;
  dest->pub.free_in_buffer = 4096;
  return 1;
  }
static void
  term_destination(j_compress_ptr cinfo)
 {
  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
  size_t datacount = 4096 - dest->pub.free_in_buffer;
  if (datacount > 0)
   {
    if (((size_t) fwrite((const void *) (dest->buffer), (size_t) 1, (size_t) (datacount), (dest->outfile))) != datacount)
      ((cinfo)->err->msg_code = (JERR_FILE_WRITE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  fflush(dest->outfile);
  if (((dest->outfile)->_flag & 0x20))
    ((cinfo)->err->msg_code = (JERR_FILE_WRITE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  }
void
  jpeg_stdio_dest(j_compress_ptr cinfo, FILE * outfile)
 {
  my_dest_ptr dest;
  if (cinfo->dest == ((void *) 0))
   {
    cinfo->dest = (struct jpeg_destination_mgr *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 0,
     ((size_t) sizeof(my_destination_mgr)));
    }
  dest = (my_dest_ptr) cinfo->dest;
  dest->pub.init_destination = init_destination;
  dest->pub.empty_output_buffer = empty_output_buffer;
  dest->pub.term_destination = term_destination;
  dest->outfile = outfile;
  }
#define JVERSION "6b  27-Mar-1998"
#define JCOPYRIGHT "Copyright (C) 1998, Thomas G. Lane"
#ifdef USE_WINDOWS_MESSAGEBOX
#endif
#ifndef EXIT_FAILURE
#endif
#ifdef NEED_SHORT_EXTERNAL_NAMES
#endif
#define JMESSAGE(code,string) string ,
const char *const jpeg_std_message_table[] = {
  ((void *) 0)
  };
static void
  error_exit(j_common_ptr cinfo)
 {
  (*cinfo->err->output_message) (cinfo);
  jpeg_destroy(cinfo);
  exit(1);
  }
static void
  output_message(j_common_ptr cinfo)
 {
  char buffer[200];
  (*cinfo->err->format_message) (cinfo, buffer);
#ifdef USE_WINDOWS_MESSAGEBOX
#else
  printf("%s\n", buffer);
#endif
  }
static void
  emit_message(j_common_ptr cinfo, int msg_level)
 {
  struct jpeg_error_mgr *err = cinfo->err;
  if (msg_level < 0)
   {
    if (err->num_warnings == 0 || err->trace_level >= 3)
      (*err->output_message) (cinfo);
    err->num_warnings++;
    }
  else
   {
    if (err->trace_level >= msg_level)
      (*err->output_message) (cinfo);
    }
  }
static void
  format_message(j_common_ptr cinfo, char *buffer)
 {
  struct jpeg_error_mgr *err = cinfo->err;
  int msg_code = err->msg_code;
  const char *msgtext = ((void *) 0);
  const char *msgptr;
  char ch;
  boolean isstring;
  if (msg_code > 0 && msg_code <= err->last_jpeg_message)
   {
    msgtext = err->jpeg_message_table[msg_code];
    }
  else if (err->addon_message_table != ((void *) 0) &&
     msg_code >= err->first_addon_message &&
   msg_code <= err->last_addon_message)
   {
    msgtext = err->addon_message_table[msg_code - err->first_addon_message];
    }
  if (msgtext == ((void *) 0))
   {
    err->msg_parm.i[0] = msg_code;
    msgtext = err->jpeg_message_table[0];
    }
  isstring = 0;
  msgptr = msgtext;
  while ((ch = *msgptr++) != '\0')
   {
    if (ch == '%')
     {
      if (*msgptr == 's')
        isstring = 1;
      break;
      }
    }
  if (isstring)
    sprintf(buffer, msgtext, err->msg_parm.s);
  else
    sprintf(buffer, msgtext,
     err->msg_parm.i[0], err->msg_parm.i[1],
     err->msg_parm.i[2], err->msg_parm.i[3],
     err->msg_parm.i[4], err->msg_parm.i[5],
     err->msg_parm.i[6], err->msg_parm.i[7]);
  }
static void
  reset_error_mgr(j_common_ptr cinfo)
 {
  cinfo->err->num_warnings = 0;
  cinfo->err->msg_code = 0;
  }
struct jpeg_error_mgr *
  jpeg_std_error(struct jpeg_error_mgr * err)
 {
  err->error_exit = error_exit;
  err->emit_message = emit_message;
  err->output_message = output_message;
  err->format_message = format_message;
  err->reset_error_mgr = reset_error_mgr;
  err->trace_level = 0;
  err->num_warnings = 0;
  err->msg_code = 0;
  err->jpeg_message_table = jpeg_std_message_table;
  err->last_jpeg_message = (int) JMSG_LASTMSGCODE - 1;
  err->addon_message_table = ((void *) 0);
  err->first_addon_message = 0;
  err->last_addon_message = 0;
  return err;
  }
#define JPEG_INTERNALS
typedef enum
 {
  M_SOF0 = 0xc0,
  M_SOF1 = 0xc1,
  M_SOF2 = 0xc2,
  M_SOF3 = 0xc3,
  M_SOF5 = 0xc5,
  M_SOF6 = 0xc6,
  M_SOF7 = 0xc7,
  M_JPG = 0xc8,
  M_SOF9 = 0xc9,
  M_SOF10 = 0xca,
  M_SOF11 = 0xcb,
  M_SOF13 = 0xcd,
  M_SOF14 = 0xce,
  M_SOF15 = 0xcf,
  M_DHT = 0xc4,
  M_DAC = 0xcc,
  M_RST0 = 0xd0,
  M_RST1 = 0xd1,
  M_RST2 = 0xd2,
  M_RST3 = 0xd3,
  M_RST4 = 0xd4,
  M_RST5 = 0xd5,
  M_RST6 = 0xd6,
  M_RST7 = 0xd7,
  M_SOI = 0xd8,
  M_EOI = 0xd9,
  M_SOS = 0xda,
  M_DQT = 0xdb,
  M_DNL = 0xdc,
  M_DRI = 0xdd,
  M_DHP = 0xde,
  M_EXP = 0xdf,
  M_APP0 = 0xe0,
  M_APP1 = 0xe1,
  M_APP2 = 0xe2,
  M_APP3 = 0xe3,
  M_APP4 = 0xe4,
  M_APP5 = 0xe5,
  M_APP6 = 0xe6,
  M_APP7 = 0xe7,
  M_APP8 = 0xe8,
  M_APP9 = 0xe9,
  M_APP10 = 0xea,
  M_APP11 = 0xeb,
  M_APP12 = 0xec,
  M_APP13 = 0xed,
  M_APP14 = 0xee,
  M_APP15 = 0xef,
  M_JPG0 = 0xf0,
  M_JPG13 = 0xfd,
  M_COM = 0xfe,
  M_TEM = 0x01,
  M_ERROR = 0x100
  } JPEG_MARKER;
typedef struct
 {
  struct jpeg_marker_writer pub;
  unsigned int last_restart_interval;
  } my_marker_writer;
typedef my_marker_writer *my_marker_ptr;
static void
  emit_byte(j_compress_ptr cinfo, int val)
 {
  struct jpeg_destination_mgr *dest = cinfo->dest;
  *(dest->next_output_byte)++ = (JOCTET) val;
  if (--dest->free_in_buffer == 0)
   {
    if (!(*dest->empty_output_buffer) (cinfo))
      ((cinfo)->err->msg_code = (JERR_CANT_SUSPEND), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  }
static void
  emit_marker(j_compress_ptr cinfo, JPEG_MARKER mark)
 {
  emit_byte(cinfo, 0xFF);
  emit_byte(cinfo, (int) mark);
  }
static void
  emit_2bytes(j_compress_ptr cinfo, int value)
 {
  emit_byte(cinfo, (value >> 8) & 0xFF);
  emit_byte(cinfo, value & 0xFF);
  }
static int
  emit_dqt(j_compress_ptr cinfo, int index)
 {
  JQUANT_TBL *qtbl = cinfo->quant_tbl_ptrs[index];
  int prec;
  int i;
  if (qtbl == ((void *) 0))
    ((cinfo)->err->msg_code = (JERR_NO_QUANT_TABLE), (cinfo)->err->msg_parm.i[0] = (index), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  prec = 0;
  for (i = 0; i < 64; i++)
   {
    if (qtbl->quantval[i] > 255)
      prec = 1;
    }
  if (!qtbl->sent_table)
   {
    emit_marker(cinfo, M_DQT);
    emit_2bytes(cinfo, prec ? 64 * 2 + 1 + 2 : 64 + 1 + 2);
    emit_byte(cinfo, index + (prec << 4));
    for (i = 0; i < 64; i++)
     {
      unsigned int qval = qtbl->quantval[jpeg_natural_order[i]];
      if (prec)
        emit_byte(cinfo, (int) (qval >> 8));
      emit_byte(cinfo, (int) (qval & 0xFF));
      }
    qtbl->sent_table = 1;
    }
  return prec;
  }
static void
  emit_dht(j_compress_ptr cinfo, int index, boolean is_ac)
 {
  JHUFF_TBL *htbl;
  int length,
    i;
  if (is_ac)
   {
    htbl = cinfo->ac_huff_tbl_ptrs[index];
    index += 0x10;
    }
  else
   {
    htbl = cinfo->dc_huff_tbl_ptrs[index];
    }
  if (htbl == ((void *) 0))
    ((cinfo)->err->msg_code = (JERR_NO_HUFF_TABLE), (cinfo)->err->msg_parm.i[0] = (index), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (!htbl->sent_table)
   {
    emit_marker(cinfo, M_DHT);
    length = 0;
    for (i = 1; i <= 16; i++)
      length += htbl->bits[i];
    emit_2bytes(cinfo, length + 2 + 1 + 16);
    emit_byte(cinfo, index);
    for (i = 1; i <= 16; i++)
      emit_byte(cinfo, htbl->bits[i]);
    for (i = 0; i < length; i++)
      emit_byte(cinfo, htbl->huffval[i]);
    htbl->sent_table = 1;
    }
  }
static void
  emit_dac(j_compress_ptr cinfo)
 {
#ifdef C_ARITH_CODING_SUPPORTED
#endif
  }
static void
  emit_dri(j_compress_ptr cinfo)
 {
  emit_marker(cinfo, M_DRI);
  emit_2bytes(cinfo, 4);
  emit_2bytes(cinfo, (int) cinfo->restart_interval);
  }
static void
  emit_sof(j_compress_ptr cinfo, JPEG_MARKER code)
 {
  int ci;
  jpeg_component_info *compptr;
  emit_marker(cinfo, code);
  emit_2bytes(cinfo, 3 * cinfo->num_components + 2 + 5 + 1);
  if ((long) cinfo->image_height > 65535L ||
   (long) cinfo->image_width > 65535L)
    ((cinfo)->err->msg_code = (JERR_IMAGE_TOO_BIG), (cinfo)->err->msg_parm.i[0] = ((unsigned int) 65535), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  emit_byte(cinfo, cinfo->data_precision);
  emit_2bytes(cinfo, (int) cinfo->image_height);
  emit_2bytes(cinfo, (int) cinfo->image_width);
  emit_byte(cinfo, cinfo->num_components);
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    emit_byte(cinfo, compptr->component_id);
    emit_byte(cinfo, (compptr->h_samp_factor << 4) + compptr->v_samp_factor);
    emit_byte(cinfo, compptr->quant_tbl_no);
    }
  }
static void
  emit_sos(j_compress_ptr cinfo)
 {
  int i,
    td,
    ta;
  jpeg_component_info *compptr;
  emit_marker(cinfo, M_SOS);
  emit_2bytes(cinfo, 2 * cinfo->comps_in_scan + 2 + 1 + 3);
  emit_byte(cinfo, cinfo->comps_in_scan);
  for (i = 0; i < cinfo->comps_in_scan; i++)
   {
    compptr = cinfo->cur_comp_info[i];
    emit_byte(cinfo, compptr->component_id);
    td = compptr->dc_tbl_no;
    ta = compptr->ac_tbl_no;
    if (cinfo->progressive_mode)
     {
      if (cinfo->Ss == 0)
       {
        ta = 0;
        if (cinfo->Ah != 0 && !cinfo->arith_code)
          td = 0;
        }
      else
       {
        td = 0;
        }
      }
    emit_byte(cinfo, (td << 4) + ta);
    }
  emit_byte(cinfo, cinfo->Ss);
  emit_byte(cinfo, cinfo->Se);
  emit_byte(cinfo, (cinfo->Ah << 4) + cinfo->Al);
  }
static void
  emit_jfif_app0(j_compress_ptr cinfo)
 {
  emit_marker(cinfo, M_APP0);
  emit_2bytes(cinfo, 2 + 4 + 1 + 2 + 1 + 2 + 2 + 1 + 1);
  emit_byte(cinfo, 0x4A);
  emit_byte(cinfo, 0x46);
  emit_byte(cinfo, 0x49);
  emit_byte(cinfo, 0x46);
  emit_byte(cinfo, 0);
  emit_byte(cinfo, cinfo->JFIF_major_version);
  emit_byte(cinfo, cinfo->JFIF_minor_version);
  emit_byte(cinfo, cinfo->density_unit);
  emit_2bytes(cinfo, (int) cinfo->X_density);
  emit_2bytes(cinfo, (int) cinfo->Y_density);
  emit_byte(cinfo, 0);
  emit_byte(cinfo, 0);
  }
static void
  emit_adobe_app14(j_compress_ptr cinfo)
 {
  emit_marker(cinfo, M_APP14);
  emit_2bytes(cinfo, 2 + 5 + 2 + 2 + 2 + 1);
  emit_byte(cinfo, 0x41);
  emit_byte(cinfo, 0x64);
  emit_byte(cinfo, 0x6F);
  emit_byte(cinfo, 0x62);
  emit_byte(cinfo, 0x65);
  emit_2bytes(cinfo, 100);
  emit_2bytes(cinfo, 0);
  emit_2bytes(cinfo, 0);
  switch (cinfo->jpeg_color_space)
   {
   case JCS_YCbCr:
    emit_byte(cinfo, 1);
    break;
   case JCS_YCCK:
    emit_byte(cinfo, 2);
    break;
   default:
    emit_byte(cinfo, 0);
    break;
    }
  }
static void
  write_marker_header(j_compress_ptr cinfo, int marker, unsigned int datalen)
 {
  if (datalen > (unsigned int) 65533)
    ((cinfo)->err->msg_code = (JERR_BAD_LENGTH), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  emit_marker(cinfo, (JPEG_MARKER) marker);
  emit_2bytes(cinfo, (int) (datalen + 2));
  }
static void
  write_marker_byte(j_compress_ptr cinfo, int val)
 {
  emit_byte(cinfo, val);
  }
static void
  write_file_header(j_compress_ptr cinfo)
 {
  my_marker_ptr marker = (my_marker_ptr) cinfo->marker;
  emit_marker(cinfo, M_SOI);
  marker->last_restart_interval = 0;
  if (cinfo->write_JFIF_header)
    emit_jfif_app0(cinfo);
  if (cinfo->write_Adobe_marker)
    emit_adobe_app14(cinfo);
  }
static void
  write_frame_header(j_compress_ptr cinfo)
 {
  int ci,
    prec;
  boolean is_baseline;
  jpeg_component_info *compptr;
  prec = 0;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    prec += emit_dqt(cinfo, compptr->quant_tbl_no);
    }
  if (cinfo->arith_code || cinfo->progressive_mode ||
   cinfo->data_precision != 8)
   {
    is_baseline = 0;
    }
  else
   {
    is_baseline = 1;
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
     ci++, compptr++)
     {
      if (compptr->dc_tbl_no > 1 || compptr->ac_tbl_no > 1)
        is_baseline = 0;
      }
    if (prec && is_baseline)
     {
      is_baseline = 0;
      ((cinfo)->err->msg_code = (JTRC_16BIT_TABLES), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (0)));
      }
    }
  if (cinfo->arith_code)
   {
    emit_sof(cinfo, M_SOF9);
    }
  else
   {
    if (cinfo->progressive_mode)
      emit_sof(cinfo, M_SOF2);
    else if (is_baseline)
      emit_sof(cinfo, M_SOF0);
    else
      emit_sof(cinfo, M_SOF1);
    }
  }
static void
  write_scan_header(j_compress_ptr cinfo)
 {
  my_marker_ptr marker = (my_marker_ptr) cinfo->marker;
  int i;
  jpeg_component_info *compptr;
  if (cinfo->arith_code)
   {
    emit_dac(cinfo);
    }
  else
   {
    for (i = 0; i < cinfo->comps_in_scan; i++)
     {
      compptr = cinfo->cur_comp_info[i];
      if (cinfo->progressive_mode)
       {
        if (cinfo->Ss == 0)
         {
          if (cinfo->Ah == 0)
            emit_dht(cinfo, compptr->dc_tbl_no, 0);
          }
        else
         {
          emit_dht(cinfo, compptr->ac_tbl_no, 1);
          }
        }
      else
       {
        emit_dht(cinfo, compptr->dc_tbl_no, 0);
        emit_dht(cinfo, compptr->ac_tbl_no, 1);
        }
      }
    }
  if (cinfo->restart_interval != marker->last_restart_interval)
   {
    emit_dri(cinfo);
    marker->last_restart_interval = cinfo->restart_interval;
    }
  emit_sos(cinfo);
  }
static void
  write_file_trailer(j_compress_ptr cinfo)
 {
  emit_marker(cinfo, M_EOI);
  }
static void
  write_tables_only(j_compress_ptr cinfo)
 {
  int i;
  emit_marker(cinfo, M_SOI);
  for (i = 0; i < 4; i++)
   {
    if (cinfo->quant_tbl_ptrs[i] != ((void *) 0))
      (void) emit_dqt(cinfo, i);
    }
  if (!cinfo->arith_code)
   {
    for (i = 0; i < 4; i++)
     {
      if (cinfo->dc_huff_tbl_ptrs[i] != ((void *) 0))
        emit_dht(cinfo, i, 0);
      if (cinfo->ac_huff_tbl_ptrs[i] != ((void *) 0))
        emit_dht(cinfo, i, 1);
      }
    }
  emit_marker(cinfo, M_EOI);
  }
void
  jinit_marker_writer(j_compress_ptr cinfo)
 {
  my_marker_ptr marker;
  marker = (my_marker_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_marker_writer)));
  cinfo->marker = (struct jpeg_marker_writer *) marker;
  marker->pub.write_file_header = write_file_header;
  marker->pub.write_frame_header = write_frame_header;
  marker->pub.write_scan_header = write_scan_header;
  marker->pub.write_file_trailer = write_file_trailer;
  marker->pub.write_tables_only = write_tables_only;
  marker->pub.write_marker_header = write_marker_header;
  marker->pub.write_marker_byte = write_marker_byte;
  marker->last_restart_interval = 0;
  }
#define JPEG_INTERNALS
typedef struct
 {
  struct jpeg_input_controller pub;
  boolean inheaders;
  } my_input_controller;
typedef my_input_controller *my_inputctl_ptr;
static int consume_markers(j_decompress_ptr cinfo);
static void
  initial_setup(j_decompress_ptr cinfo)
 {
  int ci;
  jpeg_component_info *compptr;
  if ((long) cinfo->image_height > (long) 65500L ||
   (long) cinfo->image_width > (long) 65500L)
    ((cinfo)->err->msg_code = (JERR_IMAGE_TOO_BIG), (cinfo)->err->msg_parm.i[0] = ((unsigned int) 65500L), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->data_precision != 8)
    ((cinfo)->err->msg_code = (JERR_BAD_PRECISION), (cinfo)->err->msg_parm.i[0] = (cinfo->data_precision), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->num_components > 10)
    ((cinfo)->err->msg_code = (JERR_COMPONENT_COUNT), (cinfo)->err->msg_parm.i[0] = (cinfo->num_components), (cinfo)->err->msg_parm.i[1] = (10), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  cinfo->max_h_samp_factor = 1;
  cinfo->max_v_samp_factor = 1;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    if (compptr->h_samp_factor <= 0 || compptr->h_samp_factor > 4 ||
     compptr->v_samp_factor <= 0 || compptr->v_samp_factor > 4)
      ((cinfo)->err->msg_code = (JERR_BAD_SAMPLING), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    cinfo->max_h_samp_factor = ((cinfo->max_h_samp_factor) > (compptr->h_samp_factor) ? (cinfo->max_h_samp_factor) : (compptr->h_samp_factor));
    cinfo->max_v_samp_factor = ((cinfo->max_v_samp_factor) > (compptr->v_samp_factor) ? (cinfo->max_v_samp_factor) : (compptr->v_samp_factor));
    }
  cinfo->min_DCT_scaled_size = 8;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    compptr->DCT_scaled_size = 8;
    compptr->width_in_blocks = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width * (long) compptr->h_samp_factor,
     (long) (cinfo->max_h_samp_factor * 8));
    compptr->height_in_blocks = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height * (long) compptr->v_samp_factor,
     (long) (cinfo->max_v_samp_factor * 8));
    compptr->downsampled_width = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width * (long) compptr->h_samp_factor,
     (long) cinfo->max_h_samp_factor);
    compptr->downsampled_height = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height * (long) compptr->v_samp_factor,
     (long) cinfo->max_v_samp_factor);
    compptr->component_needed = 1;
    compptr->quant_table = ((void *) 0);
    }
  cinfo->total_iMCU_rows = (JDIMENSION)
   jdiv_round_up((long) cinfo->image_height,
   (long) (cinfo->max_v_samp_factor * 8));
  if (cinfo->comps_in_scan < cinfo->num_components || cinfo->progressive_mode)
    cinfo->inputctl->has_multiple_scans = 1;
  else
    cinfo->inputctl->has_multiple_scans = 0;
  }
static void
  per_scan_setup(j_decompress_ptr cinfo)
 {
  int ci,
    mcublks,
    tmp;
  jpeg_component_info *compptr;
  if (cinfo->comps_in_scan == 1)
   {
    compptr = cinfo->cur_comp_info[0];
    cinfo->MCUs_per_row = compptr->width_in_blocks;
    cinfo->MCU_rows_in_scan = compptr->height_in_blocks;
    compptr->MCU_width = 1;
    compptr->MCU_height = 1;
    compptr->MCU_blocks = 1;
    compptr->MCU_sample_width = compptr->DCT_scaled_size;
    compptr->last_col_width = 1;
    tmp = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
    if (tmp == 0)
      tmp = compptr->v_samp_factor;
    compptr->last_row_height = tmp;
    cinfo->blocks_in_MCU = 1;
    cinfo->MCU_membership[0] = 0;
    }
  else
   {
    if (cinfo->comps_in_scan <= 0 || cinfo->comps_in_scan > 4)
      ((cinfo)->err->msg_code = (JERR_COMPONENT_COUNT), (cinfo)->err->msg_parm.i[0] = (cinfo->comps_in_scan), (cinfo)->err->msg_parm.i[1] = (4), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    cinfo->MCUs_per_row = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width,
     (long) (cinfo->max_h_samp_factor * 8));
    cinfo->MCU_rows_in_scan = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height,
     (long) (cinfo->max_v_samp_factor * 8));
    cinfo->blocks_in_MCU = 0;
    for (ci = 0; ci < cinfo->comps_in_scan; ci++)
     {
      compptr = cinfo->cur_comp_info[ci];
      compptr->MCU_width = compptr->h_samp_factor;
      compptr->MCU_height = compptr->v_samp_factor;
      compptr->MCU_blocks = compptr->MCU_width * compptr->MCU_height;
      compptr->MCU_sample_width = compptr->MCU_width * compptr->DCT_scaled_size;
      tmp = (int) (compptr->width_in_blocks % compptr->MCU_width);
      if (tmp == 0)
        tmp = compptr->MCU_width;
      compptr->last_col_width = tmp;
      tmp = (int) (compptr->height_in_blocks % compptr->MCU_height);
      if (tmp == 0)
        tmp = compptr->MCU_height;
      compptr->last_row_height = tmp;
      mcublks = compptr->MCU_blocks;
      if (cinfo->blocks_in_MCU + mcublks > 10)
        ((cinfo)->err->msg_code = (JERR_BAD_MCU_SIZE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      while (mcublks-- > 0)
       {
        cinfo->MCU_membership[cinfo->blocks_in_MCU++] = ci;
        }
      }
    }
  }
static void
  latch_quant_tables(j_decompress_ptr cinfo)
 {
  int ci,
    qtblno;
  jpeg_component_info *compptr;
  JQUANT_TBL *qtbl;
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    if (compptr->quant_table != ((void *) 0))
      continue;
    qtblno = compptr->quant_tbl_no;
    if (qtblno < 0 || qtblno >= 4 ||
     cinfo->quant_tbl_ptrs[qtblno] == ((void *) 0))
      ((cinfo)->err->msg_code = (JERR_NO_QUANT_TABLE), (cinfo)->err->msg_parm.i[0] = (qtblno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    qtbl = (JQUANT_TBL *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
     ((size_t) sizeof(JQUANT_TBL)));
    memcpy((void *) (qtbl), (const void *) (cinfo->quant_tbl_ptrs[qtblno]), (size_t) (((size_t) sizeof(JQUANT_TBL))));
    compptr->quant_table = qtbl;
    }
  }
static void
  start_input_pass(j_decompress_ptr cinfo)
 {
  per_scan_setup(cinfo);
  latch_quant_tables(cinfo);
  (*cinfo->entropy->start_pass) (cinfo);
  (*cinfo->coef->start_input_pass) (cinfo);
  cinfo->inputctl->consume_input = cinfo->coef->consume_data;
  }
static void
  finish_input_pass(j_decompress_ptr cinfo)
 {
  cinfo->inputctl->consume_input = consume_markers;
  }
static int
  consume_markers(j_decompress_ptr cinfo)
 {
  my_inputctl_ptr inputctl = (my_inputctl_ptr) cinfo->inputctl;
  int val;
  if (inputctl->pub.eoi_reached)
    return 2;
  val = (*cinfo->marker->read_markers) (cinfo);
  switch (val)
   {
   case 1:
    if (inputctl->inheaders)
     {
      initial_setup(cinfo);
      inputctl->inheaders = 0;
      }
    else
     {
      if (!inputctl->pub.has_multiple_scans)
        ((cinfo)->err->msg_code = (JERR_EOI_EXPECTED), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      start_input_pass(cinfo);
      }
    break;
   case 2:
    inputctl->pub.eoi_reached = 1;
    if (inputctl->inheaders)
     {
      if (cinfo->marker->saw_SOF)
        ((cinfo)->err->msg_code = (JERR_SOF_NO_SOS), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      }
    else
     {
      if (cinfo->output_scan_number > cinfo->input_scan_number)
        cinfo->output_scan_number = cinfo->input_scan_number;
      }
    break;
   case 0:
    break;
    }
  return val;
  }
static void
  reset_input_controller(j_decompress_ptr cinfo)
 {
  my_inputctl_ptr inputctl = (my_inputctl_ptr) cinfo->inputctl;
  inputctl->pub.consume_input = consume_markers;
  inputctl->pub.has_multiple_scans = 0;
  inputctl->pub.eoi_reached = 0;
  inputctl->inheaders = 1;
  (*cinfo->err->reset_error_mgr) ((j_common_ptr) cinfo);
  (*cinfo->marker->reset_marker_reader) (cinfo);
  cinfo->coef_bits = ((void *) 0);
  }
void
  jinit_input_controller(j_decompress_ptr cinfo)
 {
  my_inputctl_ptr inputctl;
  inputctl = (my_inputctl_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 0,
   ((size_t) sizeof(my_input_controller)));
  cinfo->inputctl = (struct jpeg_input_controller *) inputctl;
  inputctl->pub.consume_input = consume_markers;
  inputctl->pub.reset_input_controller = reset_input_controller;
  inputctl->pub.start_input_pass = start_input_pass;
  inputctl->pub.finish_input_pass = finish_input_pass;
  inputctl->pub.has_multiple_scans = 0;
  inputctl->pub.eoi_reached = 0;
  inputctl->inheaders = 1;
  }
#define JPEG_INTERNALS
typedef struct
 {
  struct jpeg_marker_reader pub;
  jpeg_marker_parser_method process_COM;
  jpeg_marker_parser_method process_APPn[16];
  unsigned int length_limit_COM;
  unsigned int length_limit_APPn[16];
  jpeg_saved_marker_ptr cur_marker;
  unsigned int bytes_read;
  } my_marker_reader;
typedef my_marker_reader *my_marker_ptr1;
#define INPUT_VARS(cinfo)   struct jpeg_source_mgr * datasrc = (cinfo)->src;   const JOCTET * next_input_byte = datasrc->next_input_byte;   size_t bytes_in_buffer = datasrc->bytes_in_buffer
#define INPUT_SYNC(cinfo)   ( datasrc->next_input_byte = next_input_byte,     datasrc->bytes_in_buffer = bytes_in_buffer )
#define INPUT_RELOAD(cinfo)   ( next_input_byte = datasrc->next_input_byte,     bytes_in_buffer = datasrc->bytes_in_buffer )
#define MAKE_BYTE_AVAIL(cinfo,action)   if (bytes_in_buffer == 0) {     if (! (*datasrc->fill_input_buffer) (cinfo))       { action; }     INPUT_RELOAD(cinfo);   }
#define INPUT_BYTE(cinfo,V,action)   MAKESTMT( MAKE_BYTE_AVAIL(cinfo,action);     bytes_in_buffer--;     V = GETJOCTET(*next_input_byte++); )
#define INPUT_2BYTES(cinfo,V,action)   MAKESTMT( MAKE_BYTE_AVAIL(cinfo,action);     bytes_in_buffer--;     V = ((unsigned int) GETJOCTET(*next_input_byte++)) << 8;     MAKE_BYTE_AVAIL(cinfo,action);     bytes_in_buffer--;     V += GETJOCTET(*next_input_byte++); )
static boolean
  get_soi(j_decompress_ptr cinfo)
 {
  int i;
  ((cinfo)->err->msg_code = (JTRC_SOI), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
  if (cinfo->marker->saw_SOI)
    ((cinfo)->err->msg_code = (JERR_SOI_DUPLICATE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  for (i = 0; i < 16; i++)
   {
    cinfo->arith_dc_L[i] = 0;
    cinfo->arith_dc_U[i] = 1;
    cinfo->arith_ac_K[i] = 5;
    }
  cinfo->restart_interval = 0;
  cinfo->jpeg_color_space = JCS_UNKNOWN;
  cinfo->CCIR601_sampling = 0;
  cinfo->saw_JFIF_marker = 0;
  cinfo->JFIF_major_version = 1;
  cinfo->JFIF_minor_version = 1;
  cinfo->density_unit = 0;
  cinfo->X_density = 1;
  cinfo->Y_density = 1;
  cinfo->saw_Adobe_marker = 0;
  cinfo->Adobe_transform = 0;
  cinfo->marker->saw_SOI = 1;
  return 1;
  }
static boolean
  get_sof(j_decompress_ptr cinfo, boolean is_prog, boolean is_arith)
 {
  int length;
  int c,
    ci;
  jpeg_component_info *compptr;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  cinfo->progressive_mode = is_prog;
  cinfo->arith_code = is_arith;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length += (*next_input_byte++);
    }
  while (0);
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    cinfo->data_precision = (*next_input_byte++);
    }
  while (0);
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    cinfo->image_height = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    cinfo->image_height += (*next_input_byte++);
    }
  while (0);
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    cinfo->image_width = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    cinfo->image_width += (*next_input_byte++);
    }
  while (0);
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    cinfo->num_components = (*next_input_byte++);
    }
  while (0);
  length -= 8;
  do
   {
    int *_mp = (cinfo)->err->msg_parm.i;
    _mp[0] = (cinfo->unread_marker);
    _mp[1] = ((int) cinfo->image_width);
    _mp[2] = ((int) cinfo->image_height);
    _mp[3] = (cinfo->num_components);
    (cinfo)->err->msg_code = (JTRC_SOF);
    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1));
    }
  while (0);
  if (cinfo->marker->saw_SOF)
    ((cinfo)->err->msg_code = (JERR_SOF_DUPLICATE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->image_height <= 0 || cinfo->image_width <= 0
   || cinfo->num_components <= 0)
    ((cinfo)->err->msg_code = (JERR_EMPTY_IMAGE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (length != (cinfo->num_components * 3))
    ((cinfo)->err->msg_code = (JERR_BAD_LENGTH), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->comp_info == ((void *) 0))
    cinfo->comp_info = (jpeg_component_info *) (*cinfo->mem->alloc_small)
     ((j_common_ptr) cinfo, 1,
     cinfo->num_components * ((size_t) sizeof(jpeg_component_info)));
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    compptr->component_index = ci;
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      compptr->component_id = (*next_input_byte++);
      }
    while (0);
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      c = (*next_input_byte++);
      }
    while (0);
    compptr->h_samp_factor = (c >> 4) & 15;
    compptr->v_samp_factor = (c) & 15;
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      compptr->quant_tbl_no = (*next_input_byte++);
      }
    while (0);
    do
     {
      int *_mp = (cinfo)->err->msg_parm.i;
      _mp[0] = (compptr->component_id);
      _mp[1] = (compptr->h_samp_factor);
      _mp[2] = (compptr->v_samp_factor);
      _mp[3] = (compptr->quant_tbl_no);
      (cinfo)->err->msg_code = (JTRC_SOF_COMPONENT);
      (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1));
      }
    while (0);
    }
  cinfo->marker->saw_SOF = 1;
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  return 1;
  }
static boolean
  get_sos(j_decompress_ptr cinfo)
 {
  int length;
  int i,
    ci,
    n,
    c,
    cc;
  jpeg_component_info *compptr;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  if (!cinfo->marker->saw_SOF)
    ((cinfo)->err->msg_code = (JERR_SOS_NO_SOF), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length += (*next_input_byte++);
    }
  while (0);
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    n = (*next_input_byte++);
    }
  while (0);
  ((cinfo)->err->msg_code = (JTRC_SOS), (cinfo)->err->msg_parm.i[0] = (n), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
  if (length != (n * 2 + 6) || n < 1 || n > 4)
    ((cinfo)->err->msg_code = (JERR_BAD_LENGTH), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  cinfo->comps_in_scan = n;
  for (i = 0; i < n; i++)
   {
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      cc = (*next_input_byte++);
      }
    while (0);
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      c = (*next_input_byte++);
      }
    while (0);
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
     ci++, compptr++)
     {
      if (cc == compptr->component_id)
        goto id_found;
      }
    ((cinfo)->err->msg_code = (JERR_BAD_COMPONENT_ID), (cinfo)->err->msg_parm.i[0] = (cc), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
id_found:
    cinfo->cur_comp_info[i] = compptr;
    compptr->dc_tbl_no = (c >> 4) & 15;
    compptr->ac_tbl_no = (c) & 15;
    do
     {
      int *_mp = (cinfo)->err->msg_parm.i;
      _mp[0] = (cc);
      _mp[1] = (compptr->dc_tbl_no);
      _mp[2] = (compptr->ac_tbl_no);
      (cinfo)->err->msg_code = (JTRC_SOS_COMPONENT);
      (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1));
      }
    while (0);
    }
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    c = (*next_input_byte++);
    }
  while (0);
  cinfo->Ss = c;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    c = (*next_input_byte++);
    }
  while (0);
  cinfo->Se = c;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    c = (*next_input_byte++);
    }
  while (0);
  cinfo->Ah = (c >> 4) & 15;
  cinfo->Al = (c) & 15;
  do
   {
    int *_mp = (cinfo)->err->msg_parm.i;
    _mp[0] = (cinfo->Ss);
    _mp[1] = (cinfo->Se);
    _mp[2] = (cinfo->Ah);
    _mp[3] = (cinfo->Al);
    (cinfo)->err->msg_code = (JTRC_SOS_PARAMS);
    (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1));
    }
  while (0);
  cinfo->marker->next_restart_num = 0;
  cinfo->input_scan_number++;
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  return 1;
  }
#ifdef D_ARITH_CODING_SUPPORTED
#else
#define get_dac(cinfo)  skip_variable(cinfo)
#endif
static boolean
  get_dht(j_decompress_ptr cinfo)
 {
  int length;
  UINT8 bits[17];
  UINT8 huffval[256];
  int i,
    index,
    count;
  JHUFF_TBL **htblptr;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length += (*next_input_byte++);
    }
  while (0);
  length -= 2;
  while (length > 16)
   {
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      index = (*next_input_byte++);
      }
    while (0);
    ((cinfo)->err->msg_code = (JTRC_DHT), (cinfo)->err->msg_parm.i[0] = (index), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
    bits[0] = 0;
    count = 0;
    for (i = 1; i <= 16; i++)
     {
      do
       {
        if (bytes_in_buffer == 0)
         {
          if (!(*datasrc->fill_input_buffer) (cinfo))
           {
            return 0;
            } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
          };
        bytes_in_buffer--;
        bits[i] = (*next_input_byte++);
        }
      while (0);
      count += bits[i];
      }
    length -= 1 + 16;
    do
     {
      int *_mp = (cinfo)->err->msg_parm.i;
      _mp[0] = (bits[1]);
      _mp[1] = (bits[2]);
      _mp[2] = (bits[3]);
      _mp[3] = (bits[4]);
      _mp[4] = (bits[5]);
      _mp[5] = (bits[6]);
      _mp[6] = (bits[7]);
      _mp[7] = (bits[8]);
      (cinfo)->err->msg_code = (JTRC_HUFFBITS);
      (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (2));
      }
    while (0);
    do
     {
      int *_mp = (cinfo)->err->msg_parm.i;
      _mp[0] = (bits[9]);
      _mp[1] = (bits[10]);
      _mp[2] = (bits[11]);
      _mp[3] = (bits[12]);
      _mp[4] = (bits[13]);
      _mp[5] = (bits[14]);
      _mp[6] = (bits[15]);
      _mp[7] = (bits[16]);
      (cinfo)->err->msg_code = (JTRC_HUFFBITS);
      (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (2));
      }
    while (0);
    if (count > 256 || ((int) count) > length)
      ((cinfo)->err->msg_code = (JERR_BAD_HUFF_TABLE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    for (i = 0; i < count; i++)
      do
       {
        if (bytes_in_buffer == 0)
         {
          if (!(*datasrc->fill_input_buffer) (cinfo))
           {
            return 0;
            } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
          };
        bytes_in_buffer--;
        huffval[i] = (*next_input_byte++);
        }
      while (0);
    length -= count;
    if (index & 0x10)
     {
      index -= 0x10;
      htblptr = &cinfo->ac_huff_tbl_ptrs[index];
      }
    else
     {
      htblptr = &cinfo->dc_huff_tbl_ptrs[index];
      }
    if (index < 0 || index >= 4)
      ((cinfo)->err->msg_code = (JERR_DHT_INDEX), (cinfo)->err->msg_parm.i[0] = (index), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (*htblptr == ((void *) 0))
      *htblptr = jpeg_alloc_huff_table((j_common_ptr) cinfo);
    memcpy((void *) ((*htblptr)->bits), (const void *) (bits), (size_t) (((size_t) sizeof((*htblptr)->bits))));
    memcpy((void *) ((*htblptr)->huffval), (const void *) (huffval), (size_t) (((size_t) sizeof((*htblptr)->huffval))));
    }
  if (length != 0)
    ((cinfo)->err->msg_code = (JERR_BAD_LENGTH), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  return 1;
  }
static boolean
  get_dqt(j_decompress_ptr cinfo)
 {
  int length;
  int n,
    i,
    prec;
  unsigned int tmp;
  JQUANT_TBL *quant_ptr;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length += (*next_input_byte++);
    }
  while (0);
  length -= 2;
  while (length > 0)
   {
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      n = (*next_input_byte++);
      }
    while (0);
    prec = n >> 4;
    n &= 0x0F;
    ((cinfo)->err->msg_code = (JTRC_DQT), (cinfo)->err->msg_parm.i[0] = (n), (cinfo)->err->msg_parm.i[1] = (prec), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
    if (n >= 4)
      ((cinfo)->err->msg_code = (JERR_DQT_INDEX), (cinfo)->err->msg_parm.i[0] = (n), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (cinfo->quant_tbl_ptrs[n] == ((void *) 0))
      cinfo->quant_tbl_ptrs[n] = jpeg_alloc_quant_table((j_common_ptr) cinfo);
    quant_ptr = cinfo->quant_tbl_ptrs[n];
    for (i = 0; i < 64; i++)
     {
      if (prec)
        do
         {
          if (bytes_in_buffer == 0)
           {
            if (!(*datasrc->fill_input_buffer) (cinfo))
             {
              return 0;
              } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
            };
          bytes_in_buffer--;
          tmp = ((unsigned int) (*next_input_byte++)) << 8;
          if (bytes_in_buffer == 0)
           {
            if (!(*datasrc->fill_input_buffer) (cinfo))
             {
              return 0;
              } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
            };
          bytes_in_buffer--;
          tmp += (*next_input_byte++);
          }
        while (0);
      else
      do
       {
        if (bytes_in_buffer == 0)
         {
          if (!(*datasrc->fill_input_buffer) (cinfo))
           {
            return 0;
            } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
          };
        bytes_in_buffer--;
        tmp = (*next_input_byte++);
        }
      while (0);
      quant_ptr->quantval[jpeg_natural_order[i]] = (UINT16) tmp;
      }
    if (cinfo->err->trace_level >= 2)
     {
      for (i = 0; i < 64; i += 8)
       {
        do
         {
          int *_mp = (cinfo)->err->msg_parm.i;
          _mp[0] = (quant_ptr->quantval[i]);
          _mp[1] = (quant_ptr->quantval[i + 1]);
          _mp[2] = (quant_ptr->quantval[i + 2]);
          _mp[3] = (quant_ptr->quantval[i + 3]);
          _mp[4] = (quant_ptr->quantval[i + 4]);
          _mp[5] = (quant_ptr->quantval[i + 5]);
          _mp[6] = (quant_ptr->quantval[i + 6]);
          _mp[7] = (quant_ptr->quantval[i + 7]);
          (cinfo)->err->msg_code = (JTRC_QUANTVALS);
          (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (2));
          }
        while (0);
        }
      }
    length -= 64 + 1;
    if (prec)
      length -= 64;
    }
  if (length != 0)
    ((cinfo)->err->msg_code = (JERR_BAD_LENGTH), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  return 1;
  }
static boolean
  get_dri(j_decompress_ptr cinfo)
 {
  int length;
  unsigned int tmp;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length += (*next_input_byte++);
    }
  while (0);
  if (length != 4)
    ((cinfo)->err->msg_code = (JERR_BAD_LENGTH), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    tmp = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    tmp += (*next_input_byte++);
    }
  while (0);
  ((cinfo)->err->msg_code = (JTRC_DRI), (cinfo)->err->msg_parm.i[0] = (tmp), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
  cinfo->restart_interval = tmp;
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  return 1;
  }
#define APP0_DATA_LEN 14
#define APP14_DATA_LEN 12
#define APPN_DATA_LEN 14
static void
  examine_app0(j_decompress_ptr cinfo, JOCTET * data,
   unsigned int datalen, int remaining)
 {
  int totallen = (int) datalen + remaining;
  if (datalen >= 14 &&
   (data[0]) == 0x4A &&
   (data[1]) == 0x46 &&
   (data[2]) == 0x49 &&
   (data[3]) == 0x46 &&
   (data[4]) == 0)
   {
    cinfo->saw_JFIF_marker = 1;
    cinfo->JFIF_major_version = (data[5]);
    cinfo->JFIF_minor_version = (data[6]);
    cinfo->density_unit = (data[7]);
    cinfo->X_density = ((data[8]) << 8) + (data[9]);
    cinfo->Y_density = ((data[10]) << 8) + (data[11]);
    if (cinfo->JFIF_major_version != 1)
      ((cinfo)->err->msg_code = (JWRN_JFIF_MAJOR), (cinfo)->err->msg_parm.i[0] = (cinfo->JFIF_major_version), (cinfo)->err->msg_parm.i[1] = (cinfo->JFIF_minor_version), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
    do
     {
      int *_mp = (cinfo)->err->msg_parm.i;
      _mp[0] = (cinfo->JFIF_major_version);
      _mp[1] = (cinfo->JFIF_minor_version);
      _mp[2] = (cinfo->X_density);
      _mp[3] = (cinfo->Y_density);
      _mp[4] = (cinfo->density_unit);
      (cinfo)->err->msg_code = (JTRC_JFIF);
      (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1));
      }
    while (0);
    if ((data[12]) | (data[13]))
      ((cinfo)->err->msg_code = (JTRC_JFIF_THUMBNAIL), (cinfo)->err->msg_parm.i[0] = ((data[12])), (cinfo)->err->msg_parm.i[1] = ((data[13])), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
    totallen -= 14;
    if (totallen !=
     ((int) (data[12]) * (int) (data[13]) * (int) 3))
      ((cinfo)->err->msg_code = (JTRC_JFIF_BADTHUMBNAILSIZE), (cinfo)->err->msg_parm.i[0] = ((int) totallen), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
    }
  else if (datalen >= 6 &&
     (data[0]) == 0x4A &&
     (data[1]) == 0x46 &&
     (data[2]) == 0x58 &&
     (data[3]) == 0x58 &&
   (data[4]) == 0)
   {
    switch ((data[5]))
     {
     case 0x10:
      ((cinfo)->err->msg_code = (JTRC_THUMB_JPEG), (cinfo)->err->msg_parm.i[0] = ((int) totallen), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
      break;
     case 0x11:
      ((cinfo)->err->msg_code = (JTRC_THUMB_PALETTE), (cinfo)->err->msg_parm.i[0] = ((int) totallen), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
      break;
     case 0x13:
      ((cinfo)->err->msg_code = (JTRC_THUMB_RGB), (cinfo)->err->msg_parm.i[0] = ((int) totallen), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
      break;
     default:
      ((cinfo)->err->msg_code = (JTRC_JFIF_EXTENSION), (cinfo)->err->msg_parm.i[0] = ((data[5])), (cinfo)->err->msg_parm.i[1] = ((int) totallen), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
      break;
      }
    }
  else
   {
    ((cinfo)->err->msg_code = (JTRC_APP0), (cinfo)->err->msg_parm.i[0] = ((int) totallen), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
    }
  }
static void
  examine_app14(j_decompress_ptr cinfo, JOCTET * data,
   unsigned int datalen, int remaining)
 {
  unsigned int version,
    flags0,
    flags1,
    transform;
  if (datalen >= 12 &&
   (data[0]) == 0x41 &&
   (data[1]) == 0x64 &&
   (data[2]) == 0x6F &&
   (data[3]) == 0x62 &&
   (data[4]) == 0x65)
   {
    version = ((data[5]) << 8) + (data[6]);
    flags0 = ((data[7]) << 8) + (data[8]);
    flags1 = ((data[9]) << 8) + (data[10]);
    transform = (data[11]);
    do
     {
      int *_mp = (cinfo)->err->msg_parm.i;
      _mp[0] = (version);
      _mp[1] = (flags0);
      _mp[2] = (flags1);
      _mp[3] = (transform);
      (cinfo)->err->msg_code = (JTRC_ADOBE);
      (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1));
      }
    while (0);
    cinfo->saw_Adobe_marker = 1;
    cinfo->Adobe_transform = (UINT8) transform;
    }
  else
   {
    ((cinfo)->err->msg_code = (JTRC_APP14), (cinfo)->err->msg_parm.i[0] = ((int) (datalen + remaining)), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
    }
  }
static boolean
  get_interesting_appn(j_decompress_ptr cinfo)
 {
  int length;
  JOCTET b[14];
  unsigned int i,
    numtoread;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length += (*next_input_byte++);
    }
  while (0);
  length -= 2;
  if (length >= 14)
    numtoread = 14;
  else if (length > 0)
    numtoread = (unsigned int) length;
  else
    numtoread = 0;
  for (i = 0; i < numtoread; i++)
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      b[i] = (*next_input_byte++);
      }
    while (0);
  length -= numtoread;
  switch (cinfo->unread_marker)
   {
   case M_APP0:
    examine_app0(cinfo, (JOCTET *) b, numtoread, length);
    break;
   case M_APP14:
    examine_app14(cinfo, (JOCTET *) b, numtoread, length);
    break;
   default:
    ((cinfo)->err->msg_code = (JERR_UNKNOWN_MARKER), (cinfo)->err->msg_parm.i[0] = (cinfo->unread_marker), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  if (length > 0)
    (*cinfo->src->skip_input_data) (cinfo, (long) length);
  return 1;
  }
#ifdef SAVE_MARKERS_SUPPORTED
static boolean
  save_marker(j_decompress_ptr cinfo)
 {
  my_marker_ptr1 marker = (my_marker_ptr1) cinfo->marker;
  jpeg_saved_marker_ptr cur_marker = marker->cur_marker;
  unsigned int bytes_read,
    data_length;
  JOCTET *data;
  int length = 0;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  if (cur_marker == ((void *) 0))
   {
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      length = ((unsigned int) (*next_input_byte++)) << 8;
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      length += (*next_input_byte++);
      }
    while (0);
    length -= 2;
    if (length >= 0)
     {
      unsigned int limit;
      if (cinfo->unread_marker == (int) M_COM)
        limit = marker->length_limit_COM;
      else
        limit = marker->length_limit_APPn[cinfo->unread_marker - (int) M_APP0];
      if ((unsigned int) length < limit)
        limit = (unsigned int) length;
      cur_marker = (jpeg_saved_marker_ptr)
       (*cinfo->mem->alloc_large) ((j_common_ptr) cinfo, 1,
       ((size_t) sizeof(struct jpeg_marker_struct)) + limit);
      cur_marker->next = ((void *) 0);
      cur_marker->marker = (UINT8) cinfo->unread_marker;
      cur_marker->original_length = (unsigned int) length;
      cur_marker->data_length = limit;
      data = cur_marker->data = (JOCTET *) (cur_marker + 1);
      marker->cur_marker = cur_marker;
      marker->bytes_read = 0;
      bytes_read = 0;
      data_length = limit;
      }
    else
     {
      bytes_read = data_length = 0;
      data = ((void *) 0);
      }
    }
  else
   {
    bytes_read = marker->bytes_read;
    data_length = cur_marker->data_length;
    data = cur_marker->data + bytes_read;
    }
  while (bytes_read < data_length)
   {
    (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
    marker->bytes_read = bytes_read;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    while (bytes_read < data_length && bytes_in_buffer > 0)
     {
      *data++ = *next_input_byte++;
      bytes_in_buffer--;
      bytes_read++;
      }
    }
  if (cur_marker != ((void *) 0))
   {
    if (cinfo->marker_list == ((void *) 0))
     {
      cinfo->marker_list = cur_marker;
      }
    else
     {
      jpeg_saved_marker_ptr prev = cinfo->marker_list;
      while (prev->next != ((void *) 0))
        prev = prev->next;
      prev->next = cur_marker;
      }
    data = cur_marker->data;
    length = cur_marker->original_length - data_length;
    }
  marker->cur_marker = ((void *) 0);
  switch (cinfo->unread_marker)
   {
   case M_APP0:
    examine_app0(cinfo, data, data_length, length);
    break;
   case M_APP14:
    examine_app14(cinfo, data, data_length, length);
    break;
   default:
    ((cinfo)->err->msg_code = (JTRC_MISC_MARKER), (cinfo)->err->msg_parm.i[0] = (cinfo->unread_marker), (cinfo)->err->msg_parm.i[1] = ((int) (data_length + length)), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
    break;
    }
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  if (length > 0)
    (*cinfo->src->skip_input_data) (cinfo, (long) length);
  return 1;
  }
#endif
static boolean
  skip_variable(j_decompress_ptr cinfo)
 {
  int length;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length = ((unsigned int) (*next_input_byte++)) << 8;
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    length += (*next_input_byte++);
    }
  while (0);
  length -= 2;
  ((cinfo)->err->msg_code = (JTRC_MISC_MARKER), (cinfo)->err->msg_parm.i[0] = (cinfo->unread_marker), (cinfo)->err->msg_parm.i[1] = ((int) length), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  if (length > 0)
    (*cinfo->src->skip_input_data) (cinfo, (long) length);
  return 1;
  }
static boolean
  next_marker(j_decompress_ptr cinfo)
 {
  int c;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  for (;;)
   {
    do
     {
      if (bytes_in_buffer == 0)
       {
        if (!(*datasrc->fill_input_buffer) (cinfo))
         {
          return 0;
          } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
        };
      bytes_in_buffer--;
      c = (*next_input_byte++);
      }
    while (0);
    while (c != 0xFF)
     {
      cinfo->marker->discarded_bytes++;
      (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
      do
       {
        if (bytes_in_buffer == 0)
         {
          if (!(*datasrc->fill_input_buffer) (cinfo))
           {
            return 0;
            } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
          };
        bytes_in_buffer--;
        c = (*next_input_byte++);
        }
      while (0);
      }
    do
     {
      do
       {
        if (bytes_in_buffer == 0)
         {
          if (!(*datasrc->fill_input_buffer) (cinfo))
           {
            return 0;
            } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
          };
        bytes_in_buffer--;
        c = (*next_input_byte++);
        }
      while (0);
      }
    while (c == 0xFF);
    if (c != 0)
      break;
    cinfo->marker->discarded_bytes += 2;
    (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
    }
  if (cinfo->marker->discarded_bytes != 0)
   {
    ((cinfo)->err->msg_code = (JWRN_EXTRANEOUS_DATA), (cinfo)->err->msg_parm.i[0] = (cinfo->marker->discarded_bytes), (cinfo)->err->msg_parm.i[1] = (c), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
    cinfo->marker->discarded_bytes = 0;
    }
  cinfo->unread_marker = c;
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  return 1;
  }
static boolean
  first_marker(j_decompress_ptr cinfo)
 {
  int c,
    c2;
  struct jpeg_source_mgr *datasrc = (cinfo)->src;
  const JOCTET *next_input_byte = datasrc->next_input_byte;
  size_t bytes_in_buffer = datasrc->bytes_in_buffer;
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    c = (*next_input_byte++);
    }
  while (0);
  do
   {
    if (bytes_in_buffer == 0)
     {
      if (!(*datasrc->fill_input_buffer) (cinfo))
       {
        return 0;
        } (next_input_byte = datasrc->next_input_byte, bytes_in_buffer = datasrc->bytes_in_buffer);
      };
    bytes_in_buffer--;
    c2 = (*next_input_byte++);
    }
  while (0);
  if (c != 0xFF || c2 != (int) M_SOI)
    ((cinfo)->err->msg_code = (JERR_NO_SOI), (cinfo)->err->msg_parm.i[0] = (c), (cinfo)->err->msg_parm.i[1] = (c2), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  cinfo->unread_marker = c2;
  (datasrc->next_input_byte = next_input_byte, datasrc->bytes_in_buffer = bytes_in_buffer);
  return 1;
  }
static int
  read_markers(j_decompress_ptr cinfo)
 {
  for (;;)
   {
    if (cinfo->unread_marker == 0)
     {
      if (!cinfo->marker->saw_SOI)
       {
        if (!first_marker(cinfo))
          return 0;
        }
      else
       {
        if (!next_marker(cinfo))
          return 0;
        }
      }
    switch (cinfo->unread_marker)
     {
     case M_SOI:
      if (!get_soi(cinfo))
        return 0;
      break;
     case M_SOF0:
     case M_SOF1:
      if (!get_sof(cinfo, 0, 0))
        return 0;
      break;
     case M_SOF2:
      if (!get_sof(cinfo, 1, 0))
        return 0;
      break;
     case M_SOF9:
      if (!get_sof(cinfo, 0, 1))
        return 0;
      break;
     case M_SOF10:
      if (!get_sof(cinfo, 1, 1))
        return 0;
      break;
     case M_SOF3:
     case M_SOF5:
     case M_SOF6:
     case M_SOF7:
     case M_JPG:
     case M_SOF11:
     case M_SOF13:
     case M_SOF14:
     case M_SOF15:
      ((cinfo)->err->msg_code = (JERR_SOF_UNSUPPORTED), (cinfo)->err->msg_parm.i[0] = (cinfo->unread_marker), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      break;
     case M_SOS:
      if (!get_sos(cinfo))
        return 0;
      cinfo->unread_marker = 0;
      return 1;
     case M_EOI:
      ((cinfo)->err->msg_code = (JTRC_EOI), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
      cinfo->unread_marker = 0;
      return 2;
     case M_DAC:
      if (!skip_variable(cinfo))
        return 0;
      break;
     case M_DHT:
      if (!get_dht(cinfo))
        return 0;
      break;
     case M_DQT:
      if (!get_dqt(cinfo))
        return 0;
      break;
     case M_DRI:
      if (!get_dri(cinfo))
        return 0;
      break;
     case M_APP0:
     case M_APP1:
     case M_APP2:
     case M_APP3:
     case M_APP4:
     case M_APP5:
     case M_APP6:
     case M_APP7:
     case M_APP8:
     case M_APP9:
     case M_APP10:
     case M_APP11:
     case M_APP12:
     case M_APP13:
     case M_APP14:
     case M_APP15:
      if (!(*((my_marker_ptr1) cinfo->marker)->process_APPn[
         cinfo->unread_marker - (int) M_APP0]) (cinfo))
        return 0;
      break;
     case M_COM:
      if (!(*((my_marker_ptr1) cinfo->marker)->process_COM) (cinfo))
        return 0;
      break;
     case M_RST0:
     case M_RST1:
     case M_RST2:
     case M_RST3:
     case M_RST4:
     case M_RST5:
     case M_RST6:
     case M_RST7:
     case M_TEM:
      ((cinfo)->err->msg_code = (JTRC_PARMLESS_MARKER), (cinfo)->err->msg_parm.i[0] = (cinfo->unread_marker), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
      break;
     case M_DNL:
      if (!skip_variable(cinfo))
        return 0;
      break;
     default:
      ((cinfo)->err->msg_code = (JERR_UNKNOWN_MARKER), (cinfo)->err->msg_parm.i[0] = (cinfo->unread_marker), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      break;
      }
    cinfo->unread_marker = 0;
    }
  }
static boolean
  read_restart_marker(j_decompress_ptr cinfo)
 {
  if (cinfo->unread_marker == 0)
   {
    if (!next_marker(cinfo))
      return 0;
    }
  if (cinfo->unread_marker ==
   ((int) M_RST0 + cinfo->marker->next_restart_num))
   {
    ((cinfo)->err->msg_code = (JTRC_RST), (cinfo)->err->msg_parm.i[0] = (cinfo->marker->next_restart_num), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (3)));
    cinfo->unread_marker = 0;
    }
  else
   {
    if (!(*cinfo->src->resync_to_restart) (cinfo,
      cinfo->marker->next_restart_num))
      return 0;
    }
  cinfo->marker->next_restart_num = (cinfo->marker->next_restart_num + 1) & 7;
  return 1;
  }
boolean
jpeg_resync_to_restart(j_decompress_ptr cinfo, int desired)
 {
  int marker = cinfo->unread_marker;
  int action = 1;
  ((cinfo)->err->msg_code = (JWRN_MUST_RESYNC), (cinfo)->err->msg_parm.i[0] = (marker), (cinfo)->err->msg_parm.i[1] = (desired), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
  for (;;)
   {
    if (marker < (int) M_SOF0)
      action = 2;
    else if (marker < (int) M_RST0 || marker > (int) M_RST7)
      action = 3;
    else
     {
      if (marker == ((int) M_RST0 + ((desired + 1) & 7)) ||
       marker == ((int) M_RST0 + ((desired + 2) & 7)))
        action = 3;
      else if (marker == ((int) M_RST0 + ((desired - 1) & 7)) ||
       marker == ((int) M_RST0 + ((desired - 2) & 7)))
        action = 2;
      else
        action = 1;
      }
    ((cinfo)->err->msg_code = (JTRC_RECOVERY_ACTION), (cinfo)->err->msg_parm.i[0] = (marker), (cinfo)->err->msg_parm.i[1] = (action), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (4)));
    switch (action)
     {
     case 1:
      cinfo->unread_marker = 0;
      return 1;
     case 2:
      if (!next_marker(cinfo))
        return 0;
      marker = cinfo->unread_marker;
      break;
     case 3:
      return 1;
      }
    }
  }
static void
  reset_marker_reader(j_decompress_ptr cinfo)
 {
  my_marker_ptr1 marker = (my_marker_ptr1) cinfo->marker;
  cinfo->comp_info = ((void *) 0);
  cinfo->input_scan_number = 0;
  cinfo->unread_marker = 0;
  marker->pub.saw_SOI = 0;
  marker->pub.saw_SOF = 0;
  marker->pub.discarded_bytes = 0;
  marker->cur_marker = ((void *) 0);
  }
void
  jinit_marker_reader(j_decompress_ptr cinfo)
 {
  my_marker_ptr1 marker;
  int i;
  marker = (my_marker_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 0,
   ((size_t) sizeof(my_marker_reader)));
  cinfo->marker = (struct jpeg_marker_reader *) marker;
  marker->pub.reset_marker_reader = reset_marker_reader;
  marker->pub.read_markers = read_markers;
  marker->pub.read_restart_marker = read_restart_marker;
  marker->process_COM = skip_variable;
  marker->length_limit_COM = 0;
  for (i = 0; i < 16; i++)
   {
    marker->process_APPn[i] = skip_variable;
    marker->length_limit_APPn[i] = 0;
    }
  marker->process_APPn[0] = get_interesting_appn;
  marker->process_APPn[14] = get_interesting_appn;
  reset_marker_reader(cinfo);
  }
#ifdef SAVE_MARKERS_SUPPORTED
void
  jpeg_save_markers(j_decompress_ptr cinfo, int marker_code,
   unsigned int length_limit)
 {
  my_marker_ptr1 marker = (my_marker_ptr1) cinfo->marker;
  long maxlength;
  jpeg_marker_parser_method processor;
  maxlength = cinfo->mem->max_alloc_chunk - ((size_t) sizeof(struct jpeg_marker_struct));
  if (((long) length_limit) > maxlength)
    length_limit = (unsigned int) maxlength;
  if (length_limit)
   {
    processor = save_marker;
    if (marker_code == (int) M_APP0 && length_limit < 14)
      length_limit = 14;
    else if (marker_code == (int) M_APP14 && length_limit < 12)
      length_limit = 12;
    }
  else
   {
    processor = skip_variable;
    if (marker_code == (int) M_APP0 || marker_code == (int) M_APP14)
      processor = get_interesting_appn;
    }
  if (marker_code == (int) M_COM)
   {
    marker->process_COM = processor;
    marker->length_limit_COM = length_limit;
    }
  else if (marker_code >= (int) M_APP0 && marker_code <= (int) M_APP15)
   {
    marker->process_APPn[marker_code - (int) M_APP0] = processor;
    marker->length_limit_APPn[marker_code - (int) M_APP0] = length_limit;
    }
  else
    ((cinfo)->err->msg_code = (JERR_UNKNOWN_MARKER), (cinfo)->err->msg_parm.i[0] = (marker_code), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  }
#endif
void
  jpeg_set_marker_processor(j_decompress_ptr cinfo, int marker_code,
   jpeg_marker_parser_method routine)
 {
  my_marker_ptr1 marker = (my_marker_ptr1) cinfo->marker;
  if (marker_code == (int) M_COM)
    marker->process_COM = routine;
  else if (marker_code >= (int) M_APP0 && marker_code <= (int) M_APP15)
    marker->process_APPn[marker_code - (int) M_APP0] = routine;
  else
    ((cinfo)->err->msg_code = (JERR_UNKNOWN_MARKER), (cinfo)->err->msg_parm.i[0] = (marker_code), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  }
#define JPEG_INTERNALS
void
  jinit_compress_master(j_compress_ptr cinfo)
 {
  jinit_c_master_control(cinfo, 0);
  if (!cinfo->raw_data_in)
   {
    jinit_color_converter(cinfo);
    jinit_downsampler(cinfo);
    jinit_c_prep_controller(cinfo, 0);
    }
  jinit_forward_dct(cinfo);
  if (cinfo->arith_code)
   {
    ((cinfo)->err->msg_code = (JERR_ARITH_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  else
   {
    if (cinfo->progressive_mode)
     {
#ifdef C_PROGRESSIVE_SUPPORTED
      jinit_phuff_encoder(cinfo);
#else
#endif
      }
    else
      jinit_huff_encoder(cinfo);
    }
  jinit_c_coef_controller(cinfo,
   (boolean) (cinfo->num_scans > 1 || cinfo->optimize_coding));
  jinit_c_main_controller(cinfo, 0);
  jinit_marker_writer(cinfo);
  (*cinfo->mem->realize_virt_arrays) ((j_common_ptr) cinfo);
  (*cinfo->marker->write_file_header) (cinfo);
  }
#define JPEG_INTERNALS
typedef struct
 {
  struct jpeg_decomp_master pub;
  int pass_number;
  boolean using_merged_upsample;
  struct jpeg_color_quantizer *quantizer_1pass;
  struct jpeg_color_quantizer *quantizer_2pass;
  } my_decomp_master;
typedef my_decomp_master *my_master_ptr1;
static boolean
  use_merged_upsample(j_decompress_ptr cinfo)
 {
#ifdef UPSAMPLE_MERGING_SUPPORTED
  if (cinfo->do_fancy_upsampling || cinfo->CCIR601_sampling)
    return 0;
  if (cinfo->jpeg_color_space != JCS_YCbCr || cinfo->num_components != 3 ||
   cinfo->out_color_space != JCS_RGB ||
   cinfo->out_color_components != 3)
    return 0;
  if (cinfo->comp_info[0].h_samp_factor != 2 ||
   cinfo->comp_info[1].h_samp_factor != 1 ||
   cinfo->comp_info[2].h_samp_factor != 1 ||
   cinfo->comp_info[0].v_samp_factor > 2 ||
   cinfo->comp_info[1].v_samp_factor != 1 ||
   cinfo->comp_info[2].v_samp_factor != 1)
    return 0;
  if (cinfo->comp_info[0].DCT_scaled_size != cinfo->min_DCT_scaled_size ||
   cinfo->comp_info[1].DCT_scaled_size != cinfo->min_DCT_scaled_size ||
   cinfo->comp_info[2].DCT_scaled_size != cinfo->min_DCT_scaled_size)
    return 0;
  return 1;
#else
#endif
  }
void
  jpeg_calc_output_dimensions(j_decompress_ptr cinfo)
 {
#ifdef IDCT_SCALING_SUPPORTED
  int ci;
  jpeg_component_info *compptr;
#endif
  if (cinfo->global_state != 202)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
#ifdef IDCT_SCALING_SUPPORTED
  if (cinfo->scale_num * 8 <= cinfo->scale_denom)
   {
    cinfo->output_width = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width, 8L);
    cinfo->output_height = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height, 8L);
    cinfo->min_DCT_scaled_size = 1;
    }
  else if (cinfo->scale_num * 4 <= cinfo->scale_denom)
   {
    cinfo->output_width = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width, 4L);
    cinfo->output_height = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height, 4L);
    cinfo->min_DCT_scaled_size = 2;
    }
  else if (cinfo->scale_num * 2 <= cinfo->scale_denom)
   {
    cinfo->output_width = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width, 2L);
    cinfo->output_height = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height, 2L);
    cinfo->min_DCT_scaled_size = 4;
    }
  else
   {
    cinfo->output_width = cinfo->image_width;
    cinfo->output_height = cinfo->image_height;
    cinfo->min_DCT_scaled_size = 8;
    }
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    int ssize = cinfo->min_DCT_scaled_size;
    while (ssize < 8 &&
     (compptr->h_samp_factor * ssize * 2 <=
      cinfo->max_h_samp_factor * cinfo->min_DCT_scaled_size) &&
     (compptr->v_samp_factor * ssize * 2 <=
      cinfo->max_v_samp_factor * cinfo->min_DCT_scaled_size))
     {
      ssize = ssize * 2;
      }
    compptr->DCT_scaled_size = ssize;
    }
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    compptr->downsampled_width = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width *
     (long) (compptr->h_samp_factor * compptr->DCT_scaled_size),
     (long) (cinfo->max_h_samp_factor * 8));
    compptr->downsampled_height = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height *
     (long) (compptr->v_samp_factor * compptr->DCT_scaled_size),
     (long) (cinfo->max_v_samp_factor * 8));
    }
#else
#endif
  switch (cinfo->out_color_space)
   {
   case JCS_GRAYSCALE:
    cinfo->out_color_components = 1;
    break;
   case JCS_RGB:
#if 3  != 3
#endif
   case JCS_YCbCr:
    cinfo->out_color_components = 3;
    break;
   case JCS_CMYK:
   case JCS_YCCK:
    cinfo->out_color_components = 4;
    break;
   default:
    cinfo->out_color_components = cinfo->num_components;
    break;
    }
  cinfo->output_components = (cinfo->quantize_colors ? 1 :
   cinfo->out_color_components);
  if (use_merged_upsample(cinfo))
    cinfo->rec_outbuf_height = cinfo->max_v_samp_factor;
  else
    cinfo->rec_outbuf_height = 1;
  }
static void
  prepare_range_limit_table(j_decompress_ptr cinfo)
 {
  JSAMPLE *table;
  int i;
  table = (JSAMPLE *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (5 * (255 + 1) + 128) * ((size_t) sizeof(JSAMPLE)));
  table += (255 + 1);
  cinfo->sample_range_limit = table;
  memset((void *) (table - (255 + 1)), 0, (size_t) ((255 + 1) * ((size_t) sizeof(JSAMPLE))));
  for (i = 0; i <= 255; i++)
    table[i] = (JSAMPLE) i;
  table += 128;
  for (i = 128; i < 2 * (255 + 1); i++)
    table[i] = 255;
  memset((void *) (table + (2 * (255 + 1))), 0, (size_t) ((2 * (255 + 1) - 128) * ((size_t) sizeof(JSAMPLE))));
  memcpy((void *) (table + (4 * (255 + 1) - 128)), (const void *) (cinfo->sample_range_limit), (size_t) (128 * ((size_t) sizeof(JSAMPLE))));
  }
static void
  master_selection(j_decompress_ptr cinfo)
 {
  my_master_ptr1 master = (my_master_ptr1) cinfo->master;
  boolean use_c_buffer;
  long samplesperrow;
  JDIMENSION jd_samplesperrow;
  jpeg_calc_output_dimensions(cinfo);
  prepare_range_limit_table(cinfo);
  samplesperrow = (long) cinfo->output_width * (long) cinfo->out_color_components;
  jd_samplesperrow = (JDIMENSION) samplesperrow;
  if ((long) jd_samplesperrow != samplesperrow)
    ((cinfo)->err->msg_code = (JERR_WIDTH_OVERFLOW), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  master->pass_number = 0;
  master->using_merged_upsample = use_merged_upsample(cinfo);
  master->quantizer_1pass = ((void *) 0);
  master->quantizer_2pass = ((void *) 0);
  if (!cinfo->quantize_colors || !cinfo->buffered_image)
   {
    cinfo->enable_1pass_quant = 0;
    cinfo->enable_external_quant = 0;
    cinfo->enable_2pass_quant = 0;
    }
  if (cinfo->quantize_colors)
   {
    if (cinfo->raw_data_out)
      ((cinfo)->err->msg_code = (JERR_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (cinfo->out_color_components != 3)
     {
      cinfo->enable_1pass_quant = 1;
      cinfo->enable_external_quant = 0;
      cinfo->enable_2pass_quant = 0;
      cinfo->colormap = ((void *) 0);
      }
    else if (cinfo->colormap != ((void *) 0))
     {
      cinfo->enable_external_quant = 1;
      }
    else if (cinfo->two_pass_quantize)
     {
      cinfo->enable_2pass_quant = 1;
      }
    else
     {
      cinfo->enable_1pass_quant = 1;
      }
    if (cinfo->enable_1pass_quant)
     {
#ifdef QUANT_1PASS_SUPPORTED
      jinit_1pass_quantizer(cinfo);
      master->quantizer_1pass = cinfo->cquantize;
#else
#endif
      }
    if (cinfo->enable_2pass_quant || cinfo->enable_external_quant)
     {
#ifdef QUANT_2PASS_SUPPORTED
      jinit_2pass_quantizer(cinfo);
      master->quantizer_2pass = cinfo->cquantize;
#else
#endif
      }
    }
  if (!cinfo->raw_data_out)
   {
    if (master->using_merged_upsample)
     {
#ifdef UPSAMPLE_MERGING_SUPPORTED
      jinit_merged_upsampler(cinfo);
#else
#endif
      }
    else
     {
      jinit_color_deconverter(cinfo);
      jinit_upsampler(cinfo);
      }
    jinit_d_post_controller(cinfo, cinfo->enable_2pass_quant);
    }
  jinit_inverse_dct(cinfo);
  if (cinfo->arith_code)
   {
    ((cinfo)->err->msg_code = (JERR_ARITH_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  else
   {
    if (cinfo->progressive_mode)
     {
#ifdef D_PROGRESSIVE_SUPPORTED
      jinit_phuff_decoder(cinfo);
#else
#endif
      }
    else
      jinit_huff_decoder(cinfo);
    }
  use_c_buffer = cinfo->inputctl->has_multiple_scans || cinfo->buffered_image;
  jinit_d_coef_controller(cinfo, use_c_buffer);
  if (!cinfo->raw_data_out)
    jinit_d_main_controller(cinfo, 0);
  (*cinfo->mem->realize_virt_arrays) ((j_common_ptr) cinfo);
  (*cinfo->inputctl->start_input_pass) (cinfo);
#ifdef D_MULTISCAN_FILES_SUPPORTED
  if (cinfo->progress != ((void *) 0) && !cinfo->buffered_image &&
   cinfo->inputctl->has_multiple_scans)
   {
    int nscans;
    if (cinfo->progressive_mode)
     {
      nscans = 2 + 3 * cinfo->num_components;
      }
    else
     {
      nscans = cinfo->num_components;
      }
    cinfo->progress->pass_counter = 0L;
    cinfo->progress->pass_limit = (long) cinfo->total_iMCU_rows * nscans;
    cinfo->progress->completed_passes = 0;
    cinfo->progress->total_passes = (cinfo->enable_2pass_quant ? 3 : 2);
    master->pass_number++;
    }
#endif
  }
static void
  prepare_for_output_pass(j_decompress_ptr cinfo)
 {
  my_master_ptr1 master = (my_master_ptr1) cinfo->master;
  if (master->pub.is_dummy_pass)
   {
#ifdef QUANT_2PASS_SUPPORTED
    master->pub.is_dummy_pass = 0;
    (*cinfo->cquantize->start_pass) (cinfo, 0);
    (*cinfo->post->start_pass) (cinfo, JBUF_CRANK_DEST);
    (*cinfo->main->start_pass) (cinfo, JBUF_CRANK_DEST);
#else
#endif
    }
  else
   {
    if (cinfo->quantize_colors && cinfo->colormap == ((void *) 0))
     {
      if (cinfo->two_pass_quantize && cinfo->enable_2pass_quant)
       {
        cinfo->cquantize = master->quantizer_2pass;
        master->pub.is_dummy_pass = 1;
        }
      else if (cinfo->enable_1pass_quant)
       {
        cinfo->cquantize = master->quantizer_1pass;
        }
      else
       {
        ((cinfo)->err->msg_code = (JERR_MODE_CHANGE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
        }
      }
    (*cinfo->idct->start_pass) (cinfo);
    (*cinfo->coef->start_output_pass) (cinfo);
    if (!cinfo->raw_data_out)
     {
      if (!master->using_merged_upsample)
        (*cinfo->cconvert->start_pass) (cinfo);
      (*cinfo->upsample->start_pass) (cinfo);
      if (cinfo->quantize_colors)
        (*cinfo->cquantize->start_pass) (cinfo, master->pub.is_dummy_pass);
      (*cinfo->post->start_pass) (cinfo,
       (master->pub.is_dummy_pass ? JBUF_SAVE_AND_PASS : JBUF_PASS_THRU));
      (*cinfo->main->start_pass) (cinfo, JBUF_PASS_THRU);
      }
    }
  if (cinfo->progress != ((void *) 0))
   {
    cinfo->progress->completed_passes = master->pass_number;
    cinfo->progress->total_passes = master->pass_number +
     (master->pub.is_dummy_pass ? 2 : 1);
    if (cinfo->buffered_image && !cinfo->inputctl->eoi_reached)
     {
      cinfo->progress->total_passes += (cinfo->enable_2pass_quant ? 2 : 1);
      }
    }
  }
static void
  finish_output_pass(j_decompress_ptr cinfo)
 {
  my_master_ptr1 master = (my_master_ptr1) cinfo->master;
  if (cinfo->quantize_colors)
    (*cinfo->cquantize->finish_pass) (cinfo);
  master->pass_number++;
  }
#ifdef D_MULTISCAN_FILES_SUPPORTED
void
  jpeg_new_colormap(j_decompress_ptr cinfo)
 {
  my_master_ptr1 master = (my_master_ptr1) cinfo->master;
  if (cinfo->global_state != 207)
    ((cinfo)->err->msg_code = (JERR_BAD_STATE), (cinfo)->err->msg_parm.i[0] = (cinfo->global_state), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->quantize_colors && cinfo->enable_external_quant &&
   cinfo->colormap != ((void *) 0))
   {
    cinfo->cquantize = master->quantizer_2pass;
    (*cinfo->cquantize->new_color_map) (cinfo);
    master->pub.is_dummy_pass = 0;
    }
  else
    ((cinfo)->err->msg_code = (JERR_MODE_CHANGE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  }
#endif
void
  jinit_master_decompress(j_decompress_ptr cinfo)
 {
  my_master_ptr1 master;
  master = (my_master_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_decomp_master)));
  cinfo->master = (struct jpeg_decomp_master *) master;
  master->pub.prepare_for_output_pass = prepare_for_output_pass;
  master->pub.finish_output_pass = finish_output_pass;
  master->pub.is_dummy_pass = 0;
  master_selection(cinfo);
  }
#define JPEG_INTERNALS
#undef FULL_MAIN_BUFFER_SUPPORTED
typedef struct
 {
  struct jpeg_c_main_controller pub;
  JDIMENSION cur_iMCU_row;
  JDIMENSION rowgroup_ctr;
  boolean suspended;
  J_BUF_MODE pass_mode;
  JSAMPARRAY buffer[10];
#ifdef FULL_MAIN_BUFFER_SUPPORTED
#endif
  } my_main_controller;
typedef my_main_controller *my_main_ptr;
static void process_data_simple_main
  (j_compress_ptr cinfo, JSAMPARRAY input_buf, JDIMENSION * in_row_ctr, JDIMENSION in_rows_avail);
#ifdef FULL_MAIN_BUFFER_SUPPORTED
#endif
static void
  start_pass_main(j_compress_ptr cinfo, J_BUF_MODE pass_mode)
 {
  my_main_ptr main = (my_main_ptr) cinfo->main;
  if (cinfo->raw_data_in)
    return;
  main->cur_iMCU_row = 0;
  main->rowgroup_ctr = 0;
  main->suspended = 0;
  main->pass_mode = pass_mode;
  switch (pass_mode)
   {
   case JBUF_PASS_THRU:
#ifdef FULL_MAIN_BUFFER_SUPPORTED
#endif
    main->pub.process_data = process_data_simple_main;
    break;
#ifdef FULL_MAIN_BUFFER_SUPPORTED
#endif
   default:
    ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  }
static void
  process_data_simple_main(j_compress_ptr cinfo,
   JSAMPARRAY input_buf, JDIMENSION * in_row_ctr,
   JDIMENSION in_rows_avail)
 {
  my_main_ptr main = (my_main_ptr) cinfo->main;
  while (main->cur_iMCU_row < cinfo->total_iMCU_rows)
   {
    if (main->rowgroup_ctr < 8)
      (*cinfo->prep->pre_process_data) (cinfo,
       input_buf, in_row_ctr, in_rows_avail,
       main->buffer, &main->rowgroup_ctr,
       (JDIMENSION) 8);
    if (main->rowgroup_ctr != 8)
      return;
    if (!(*cinfo->coef->compress_data) (cinfo, main->buffer))
     {
      if (!main->suspended)
       {
        (*in_row_ctr)--;
        main->suspended = 1;
        }
      return;
      }
    if (main->suspended)
     {
      (*in_row_ctr)++;
      main->suspended = 0;
      }
    main->rowgroup_ctr = 0;
    main->cur_iMCU_row++;
    }
  }
#ifdef FULL_MAIN_BUFFER_SUPPORTED
#endif
void
  jinit_c_main_controller(j_compress_ptr cinfo, boolean need_full_buffer)
 {
  my_main_ptr main;
  int ci;
  jpeg_component_info *compptr;
  main = (my_main_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_main_controller)));
  cinfo->main = (struct jpeg_c_main_controller *) main;
  main->pub.start_pass = start_pass_main;
  if (cinfo->raw_data_in)
    return;
  if (need_full_buffer)
   {
#ifdef FULL_MAIN_BUFFER_SUPPORTED
#else
    ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
#endif
    }
  else
   {
#ifdef FULL_MAIN_BUFFER_SUPPORTED
#endif
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
     ci++, compptr++)
     {
      main->buffer[ci] = (*cinfo->mem->alloc_sarray)
       ((j_common_ptr) cinfo, 1,
       compptr->width_in_blocks * 8,
       (JDIMENSION) (compptr->v_samp_factor * 8));
      }
    }
  }
#define JPEG_INTERNALS
#if 8  == 8
#define MAX_COEF_BITS 10
#else
#endif
typedef struct
 {
  unsigned int ehufco[256];
  char ehufsi[256];
  } c_derived_tbl;
#ifdef NEED_SHORT_EXTERNAL_NAMES
#endif
extern void jpeg_make_c_derived_tbl
  (j_compress_ptr cinfo, boolean isDC, int tblno, c_derived_tbl ** pdtbl);
extern void jpeg_gen_optimal_table
  (j_compress_ptr cinfo, JHUFF_TBL * htbl, long freq[]);
typedef struct
 {
  int put_buffer;
  int put_bits;
  int last_dc_val[4];
  } savable_state;
#ifndef NO_STRUCT_ASSIGN
#define ASSIGN_STATE(dest,src)  ((dest) = (src))
#else
#endif
typedef struct
 {
  struct jpeg_entropy_encoder pub;
  savable_state saved;
  unsigned int restarts_to_go;
  int next_restart_num;
  c_derived_tbl *dc_derived_tbls[4];
  c_derived_tbl *ac_derived_tbls[4];
#ifdef ENTROPY_OPT_SUPPORTED
  long *dc_count_ptrs[4];
  long *ac_count_ptrs[4];
#endif
  } huff_entropy_encoder;
typedef huff_entropy_encoder *huff_entropy_ptr;
typedef struct
 {
  JOCTET *next_output_byte;
  size_t free_in_buffer;
  savable_state cur;
  j_compress_ptr cinfo;
  } working_state;
static boolean encode_mcu_huff(j_compress_ptr cinfo, JBLOCKROW * MCU_data);
static void finish_pass_huff(j_compress_ptr cinfo);
#ifdef ENTROPY_OPT_SUPPORTED
static boolean encode_mcu_gather(j_compress_ptr cinfo, JBLOCKROW * MCU_data);
static void finish_pass_gather(j_compress_ptr cinfo);
#endif
static void
  start_pass_huff(j_compress_ptr cinfo, boolean gather_statistics)
 {
  huff_entropy_ptr entropy = (huff_entropy_ptr) cinfo->entropy;
  int ci,
    dctbl,
    actbl;
  jpeg_component_info *compptr;
  if (gather_statistics)
   {
#ifdef ENTROPY_OPT_SUPPORTED
    entropy->pub.encode_mcu = encode_mcu_gather;
    entropy->pub.finish_pass = finish_pass_gather;
#else
#endif
    }
  else
   {
    entropy->pub.encode_mcu = encode_mcu_huff;
    entropy->pub.finish_pass = finish_pass_huff;
    }
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    dctbl = compptr->dc_tbl_no;
    actbl = compptr->ac_tbl_no;
    if (gather_statistics)
     {
#ifdef ENTROPY_OPT_SUPPORTED
      if (dctbl < 0 || dctbl >= 4)
        ((cinfo)->err->msg_code = (JERR_NO_HUFF_TABLE), (cinfo)->err->msg_parm.i[0] = (dctbl), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      if (actbl < 0 || actbl >= 4)
        ((cinfo)->err->msg_code = (JERR_NO_HUFF_TABLE), (cinfo)->err->msg_parm.i[0] = (actbl), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      if (entropy->dc_count_ptrs[dctbl] == ((void *) 0))
        entropy->dc_count_ptrs[dctbl] = (long *)
         (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
         257 * ((size_t) sizeof(long)));
      memset((void *) (entropy->dc_count_ptrs[dctbl]), 0, (size_t) (257 * ((size_t) sizeof(long))));
      if (entropy->ac_count_ptrs[actbl] == ((void *) 0))
        entropy->ac_count_ptrs[actbl] = (long *)
         (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
         257 * ((size_t) sizeof(long)));
      memset((void *) (entropy->ac_count_ptrs[actbl]), 0, (size_t) (257 * ((size_t) sizeof(long))));
#endif
      }
    else
     {
      jpeg_make_c_derived_tbl(cinfo, 1, dctbl,
       &entropy->dc_derived_tbls[dctbl]);
      jpeg_make_c_derived_tbl(cinfo, 0, actbl,
       &entropy->ac_derived_tbls[actbl]);
      }
    entropy->saved.last_dc_val[ci] = 0;
    }
  entropy->saved.put_buffer = 0;
  entropy->saved.put_bits = 0;
  entropy->restarts_to_go = cinfo->restart_interval;
  entropy->next_restart_num = 0;
  }
void
  jpeg_make_c_derived_tbl(j_compress_ptr cinfo, boolean isDC, int tblno,
   c_derived_tbl ** pdtbl)
 {
  JHUFF_TBL *htbl;
  c_derived_tbl *dtbl;
  int p,
    i,
    l,
    lastp,
    si,
    maxsymbol;
  char huffsize[257];
  unsigned int huffcode[257];
  unsigned int code;
  if (tblno < 0 || tblno >= 4)
    ((cinfo)->err->msg_code = (JERR_NO_HUFF_TABLE), (cinfo)->err->msg_parm.i[0] = (tblno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  htbl =
   isDC ? cinfo->dc_huff_tbl_ptrs[tblno] : cinfo->ac_huff_tbl_ptrs[tblno];
  if (htbl == ((void *) 0))
    ((cinfo)->err->msg_code = (JERR_NO_HUFF_TABLE), (cinfo)->err->msg_parm.i[0] = (tblno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (*pdtbl == ((void *) 0))
    *pdtbl = (c_derived_tbl *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
     ((size_t) sizeof(c_derived_tbl)));
  dtbl = *pdtbl;
  p = 0;
  for (l = 1; l <= 16; l++)
   {
    i = (int) htbl->bits[l];
    if (i < 0 || p + i > 256)
      ((cinfo)->err->msg_code = (JERR_BAD_HUFF_TABLE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    while (i--)
      huffsize[p++] = (char) l;
    }
  huffsize[p] = 0;
  lastp = p;
  code = 0;
  si = huffsize[0];
  p = 0;
  while (huffsize[p])
   {
    while (((int) huffsize[p]) == si)
     {
      huffcode[p++] = code;
      code++;
      }
    if (((int) code) >= (((int) 1) << si))
      ((cinfo)->err->msg_code = (JERR_BAD_HUFF_TABLE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    code <<= 1;
    si++;
    }
  memset((void *) (dtbl->ehufsi), 0, (size_t) (((size_t) sizeof(dtbl->ehufsi))));
  maxsymbol = isDC ? 15 : 255;
  for (p = 0; p < lastp; p++)
   {
    i = htbl->huffval[p];
    if (i < 0 || i > maxsymbol || dtbl->ehufsi[i])
      ((cinfo)->err->msg_code = (JERR_BAD_HUFF_TABLE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    dtbl->ehufco[i] = huffcode[p];
    dtbl->ehufsi[i] = huffsize[p];
    }
  }
#define emit_byte(state,val,action)   { *(state)->next_output_byte++ = (JOCTET) (val);     if (--(state)->free_in_buffer == 0)       if (! dump_buffer(state))         { action; } }
static boolean
  dump_buffer(working_state * state)
 {
  struct jpeg_destination_mgr *dest = state->cinfo->dest;
  if (!(*dest->empty_output_buffer) (state->cinfo))
    return 0;
  state->next_output_byte = dest->next_output_byte;
  state->free_in_buffer = dest->free_in_buffer;
  return 1;
  }
static boolean
  emit_bits(working_state * state, unsigned int code, int size)
 {
  register int put_buffer = (int) code;
  register int put_bits = state->cur.put_bits;
  if (size == 0)
    ((state->cinfo)->err->msg_code = (JERR_HUFF_MISSING_CODE), (*(state->cinfo)->err->error_exit) ((j_common_ptr) (state->cinfo)));
  put_buffer &= (((int) 1) << size) - 1;
  put_bits += size;
  put_buffer <<= 24 - put_bits;
  put_buffer |= state->cur.put_buffer;
  while (put_bits >= 8)
   {
    int c = (int) ((put_buffer >> 16) & 0xFF);
     {
      *(state)->next_output_byte++ = (JOCTET) (c);
      if (--(state)->free_in_buffer == 0)
        if (!dump_buffer(state))
         {
          return 0;
          }
      };
    if (c == 0xFF)
     {
       {
        *(state)->next_output_byte++ = (JOCTET) (0);
        if (--(state)->free_in_buffer == 0)
          if (!dump_buffer(state))
           {
            return 0;
            }
        };
      }
    put_buffer <<= 8;
    put_bits -= 8;
    }
  state->cur.put_buffer = put_buffer;
  state->cur.put_bits = put_bits;
  return 1;
  }
static boolean
  flush_bits(working_state * state)
 {
  if (!emit_bits(state, 0x7F, 7))
    return 0;
  state->cur.put_buffer = 0;
  state->cur.put_bits = 0;
  return 1;
  }
static boolean
  encode_one_block(working_state * state, JCOEFPTR block, int last_dc_val,
   c_derived_tbl * dctbl, c_derived_tbl * actbl)
 {
  register int temp,
    temp2;
  register int nbits;
  register int k,
    r,
    i;
  temp = temp2 = block[0] - last_dc_val;
  if (temp < 0)
   {
    temp = -temp;
    temp2--;
    }
  nbits = 0;
  while (temp)
   {
    nbits++;
    temp >>= 1;
    }
  if (nbits > 10 + 1)
    ((state->cinfo)->err->msg_code = (JERR_BAD_DCT_COEF), (*(state->cinfo)->err->error_exit) ((j_common_ptr) (state->cinfo)));
  if (!emit_bits(state, dctbl->ehufco[nbits], dctbl->ehufsi[nbits]))
    return 0;
  if (nbits)
    if (!emit_bits(state, (unsigned int) temp2, nbits))
      return 0;
  r = 0;
  for (k = 1; k < 64; k++)
   {
    if ((temp = block[jpeg_natural_order[k]]) == 0)
     {
      r++;
      }
    else
     {
      while (r > 15)
       {
        if (!emit_bits(state, actbl->ehufco[0xF0], actbl->ehufsi[0xF0]))
          return 0;
        r -= 16;
        }
      temp2 = temp;
      if (temp < 0)
       {
        temp = -temp;
        temp2--;
        }
      nbits = 1;
      while ((temp >>= 1))
        nbits++;
      if (nbits > 10)
        ((state->cinfo)->err->msg_code = (JERR_BAD_DCT_COEF), (*(state->cinfo)->err->error_exit) ((j_common_ptr) (state->cinfo)));
      i = (r << 4) + nbits;
      if (!emit_bits(state, actbl->ehufco[i], actbl->ehufsi[i]))
        return 0;
      if (!emit_bits(state, (unsigned int) temp2, nbits))
        return 0;
      r = 0;
      }
    }
  if (r > 0)
    if (!emit_bits(state, actbl->ehufco[0], actbl->ehufsi[0]))
      return 0;
  return 1;
  }
static boolean
  emit_restart(working_state * state, int restart_num)
 {
  int ci;
  if (!flush_bits(state))
    return 0;
   {
    *(state)->next_output_byte++ = (JOCTET) (0xFF);
    if (--(state)->free_in_buffer == 0)
      if (!dump_buffer(state))
       {
        return 0;
        }
    };
   {
    *(state)->next_output_byte++ = (JOCTET) (0xD0 + restart_num);
    if (--(state)->free_in_buffer == 0)
      if (!dump_buffer(state))
       {
        return 0;
        }
    };
  for (ci = 0; ci < state->cinfo->comps_in_scan; ci++)
    state->cur.last_dc_val[ci] = 0;
  return 1;
  }
static boolean
  encode_mcu_huff(j_compress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  huff_entropy_ptr entropy = (huff_entropy_ptr) cinfo->entropy;
  working_state state;
  int blkn,
    ci;
  jpeg_component_info *compptr;
  state.next_output_byte = cinfo->dest->next_output_byte;
  state.free_in_buffer = cinfo->dest->free_in_buffer;
  ((state.cur) = (entropy->saved));
  state.cinfo = cinfo;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
      if (!emit_restart(&state, entropy->next_restart_num))
        return 0;
    }
  for (blkn = 0; blkn < cinfo->blocks_in_MCU; blkn++)
   {
    ci = cinfo->MCU_membership[blkn];
    compptr = cinfo->cur_comp_info[ci];
    if (!encode_one_block(&state,
      MCU_data[blkn][0], state.cur.last_dc_val[ci],
      entropy->dc_derived_tbls[compptr->dc_tbl_no],
      entropy->ac_derived_tbls[compptr->ac_tbl_no]))
      return 0;
    state.cur.last_dc_val[ci] = MCU_data[blkn][0][0];
    }
  cinfo->dest->next_output_byte = state.next_output_byte;
  cinfo->dest->free_in_buffer = state.free_in_buffer;
  ((entropy->saved) = (state.cur));
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
     {
      entropy->restarts_to_go = cinfo->restart_interval;
      entropy->next_restart_num++;
      entropy->next_restart_num &= 7;
      }
    entropy->restarts_to_go--;
    }
  return 1;
  }
static void
  finish_pass_huff(j_compress_ptr cinfo)
 {
  huff_entropy_ptr entropy = (huff_entropy_ptr) cinfo->entropy;
  working_state state;
  state.next_output_byte = cinfo->dest->next_output_byte;
  state.free_in_buffer = cinfo->dest->free_in_buffer;
  ((state.cur) = (entropy->saved));
  state.cinfo = cinfo;
  if (!flush_bits(&state))
    ((cinfo)->err->msg_code = (JERR_CANT_SUSPEND), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  cinfo->dest->next_output_byte = state.next_output_byte;
  cinfo->dest->free_in_buffer = state.free_in_buffer;
  ((entropy->saved) = (state.cur));
  }
#ifdef ENTROPY_OPT_SUPPORTED
static void
  htest_one_block(j_compress_ptr cinfo, JCOEFPTR block, int last_dc_val,
   long dc_counts[], long ac_counts[])
 {
  register int temp;
  register int nbits;
  register int k,
    r;
  temp = block[0] - last_dc_val;
  if (temp < 0)
    temp = -temp;
  nbits = 0;
  while (temp)
   {
    nbits++;
    temp >>= 1;
    }
  if (nbits > 10 + 1)
    ((cinfo)->err->msg_code = (JERR_BAD_DCT_COEF), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  dc_counts[nbits]++;
  r = 0;
  for (k = 1; k < 64; k++)
   {
    if ((temp = block[jpeg_natural_order[k]]) == 0)
     {
      r++;
      }
    else
     {
      while (r > 15)
       {
        ac_counts[0xF0]++;
        r -= 16;
        }
      if (temp < 0)
        temp = -temp;
      nbits = 1;
      while ((temp >>= 1))
        nbits++;
      if (nbits > 10)
        ((cinfo)->err->msg_code = (JERR_BAD_DCT_COEF), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      ac_counts[(r << 4) + nbits]++;
      r = 0;
      }
    }
  if (r > 0)
    ac_counts[0]++;
  }
static boolean
  encode_mcu_gather(j_compress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  huff_entropy_ptr entropy = (huff_entropy_ptr) cinfo->entropy;
  int blkn,
    ci;
  jpeg_component_info *compptr;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
     {
      for (ci = 0; ci < cinfo->comps_in_scan; ci++)
        entropy->saved.last_dc_val[ci] = 0;
      entropy->restarts_to_go = cinfo->restart_interval;
      }
    entropy->restarts_to_go--;
    }
  for (blkn = 0; blkn < cinfo->blocks_in_MCU; blkn++)
   {
    ci = cinfo->MCU_membership[blkn];
    compptr = cinfo->cur_comp_info[ci];
    htest_one_block(cinfo, MCU_data[blkn][0], entropy->saved.last_dc_val[ci],
     entropy->dc_count_ptrs[compptr->dc_tbl_no],
     entropy->ac_count_ptrs[compptr->ac_tbl_no]);
    entropy->saved.last_dc_val[ci] = MCU_data[blkn][0][0];
    }
  return 1;
  }
void
  jpeg_gen_optimal_table(j_compress_ptr cinfo, JHUFF_TBL * htbl, long freq[])
 {
#define MAX_CLEN 32
  UINT8 bits[32 + 1];
  int codesize[257];
  int others[257];
  int c1,
    c2;
  int p,
    i,
    j;
  long v;
  memset((void *) (bits), 0, (size_t) (((size_t) sizeof(bits))));
  memset((void *) (codesize), 0, (size_t) (((size_t) sizeof(codesize))));
  for (i = 0; i < 257; i++)
    others[i] = -1;
  freq[256] = 1;
  for (;;)
   {
    c1 = -1;
    v = 1000000000L;
    for (i = 0; i <= 256; i++)
     {
      if (freq[i] && freq[i] <= v)
       {
        v = freq[i];
        c1 = i;
        }
      }
    c2 = -1;
    v = 1000000000L;
    for (i = 0; i <= 256; i++)
     {
      if (freq[i] && freq[i] <= v && i != c1)
       {
        v = freq[i];
        c2 = i;
        }
      }
    if (c2 < 0)
      break;
    freq[c1] += freq[c2];
    freq[c2] = 0;
    codesize[c1]++;
    while (others[c1] >= 0)
     {
      c1 = others[c1];
      codesize[c1]++;
      }
    others[c1] = c2;
    codesize[c2]++;
    while (others[c2] >= 0)
     {
      c2 = others[c2];
      codesize[c2]++;
      }
    }
  for (i = 0; i <= 256; i++)
   {
    if (codesize[i])
     {
      if (codesize[i] > 32)
        ((cinfo)->err->msg_code = (JERR_HUFF_CLEN_OVERFLOW), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      bits[codesize[i]]++;
      }
    }
  for (i = 32; i > 16; i--)
   {
    while (bits[i] > 0)
     {
      j = i - 2;
      while (bits[j] == 0)
        j--;
      bits[i] -= 2;
      bits[i - 1]++;
      bits[j + 1] += 2;
      bits[j]--;
      }
    }
  while (bits[i] == 0)
    i--;
  bits[i]--;
  memcpy((void *) (htbl->bits), (const void *) (bits), (size_t) (((size_t) sizeof(htbl->bits))));
  p = 0;
  for (i = 1; i <= 32; i++)
   {
    for (j = 0; j <= 255; j++)
     {
      if (codesize[j] == i)
       {
        htbl->huffval[p] = (UINT8) j;
        p++;
        }
      }
    }
  htbl->sent_table = 0;
  }
static void
  finish_pass_gather(j_compress_ptr cinfo)
 {
  huff_entropy_ptr entropy = (huff_entropy_ptr) cinfo->entropy;
  int ci,
    dctbl,
    actbl;
  jpeg_component_info *compptr;
  JHUFF_TBL **htblptr;
  boolean did_dc[4];
  boolean did_ac[4];
  memset((void *) (did_dc), 0, (size_t) (((size_t) sizeof(did_dc))));
  memset((void *) (did_ac), 0, (size_t) (((size_t) sizeof(did_ac))));
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    dctbl = compptr->dc_tbl_no;
    actbl = compptr->ac_tbl_no;
    if (!did_dc[dctbl])
     {
      htblptr = &cinfo->dc_huff_tbl_ptrs[dctbl];
      if (*htblptr == ((void *) 0))
        *htblptr = jpeg_alloc_huff_table((j_common_ptr) cinfo);
      jpeg_gen_optimal_table(cinfo, *htblptr, entropy->dc_count_ptrs[dctbl]);
      did_dc[dctbl] = 1;
      }
    if (!did_ac[actbl])
     {
      htblptr = &cinfo->ac_huff_tbl_ptrs[actbl];
      if (*htblptr == ((void *) 0))
        *htblptr = jpeg_alloc_huff_table((j_common_ptr) cinfo);
      jpeg_gen_optimal_table(cinfo, *htblptr, entropy->ac_count_ptrs[actbl]);
      did_ac[actbl] = 1;
      }
    }
  }
#endif
void
  jinit_huff_encoder(j_compress_ptr cinfo)
 {
  huff_entropy_ptr entropy;
  int i;
  entropy = (huff_entropy_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(huff_entropy_encoder)));
  cinfo->entropy = (struct jpeg_entropy_encoder *) entropy;
  entropy->pub.start_pass = start_pass_huff;
  for (i = 0; i < 4; i++)
   {
    entropy->dc_derived_tbls[i] = entropy->ac_derived_tbls[i] = ((void *) 0);
#ifdef ENTROPY_OPT_SUPPORTED
    entropy->dc_count_ptrs[i] = entropy->ac_count_ptrs[i] = ((void *) 0);
#endif
    }
  }
#define JPEG_INTERNALS
#ifdef NEED_SHORT_EXTERNAL_NAMES
#endif
#define HUFF_LOOKAHEAD 8
typedef struct
 {
  int maxcode[18];
  int valoffset[17];
  JHUFF_TBL *pub;
  int look_nbits[1 << 8];
  UINT8 look_sym[1 << 8];
  } d_derived_tbl;
extern void jpeg_make_d_derived_tbl
  (j_decompress_ptr cinfo, boolean isDC, int tblno, d_derived_tbl ** pdtbl);
typedef int bit_buf_type;
#define BIT_BUF_SIZE  32
typedef struct
 {
  bit_buf_type get_buffer;
  int bits_left;
  } bitread_perm_state;
typedef struct
 {
  const JOCTET *next_input_byte;
  size_t bytes_in_buffer;
  bit_buf_type get_buffer;
  int bits_left;
  j_decompress_ptr cinfo;
  } bitread_working_state;
#define BITREAD_STATE_VARS   register bit_buf_type get_buffer;   register int bits_left;   bitread_working_state br_state
#define BITREAD_LOAD_STATE(cinfop,permstate)   br_state.cinfo = cinfop;  br_state.next_input_byte = cinfop->src->next_input_byte;  br_state.bytes_in_buffer = cinfop->src->bytes_in_buffer;  get_buffer = permstate.get_buffer;  bits_left = permstate.bits_left;
#define BITREAD_SAVE_STATE(cinfop,permstate)   cinfop->src->next_input_byte = br_state.next_input_byte;  cinfop->src->bytes_in_buffer = br_state.bytes_in_buffer;  permstate.get_buffer = get_buffer;  permstate.bits_left = bits_left
#define CHECK_BIT_BUFFER(state,nbits,action)  { if (bits_left < (nbits)) {       if (! jpeg_fill_bit_buffer(&(state),get_buffer,bits_left,nbits))         { action; }       get_buffer = (state).get_buffer; bits_left = (state).bits_left; } }
#define GET_BITS(nbits)  (((int) (get_buffer >> (bits_left -= (nbits)))) & ((1<<(nbits))-1))
#define PEEK_BITS(nbits)  (((int) (get_buffer >> (bits_left -  (nbits)))) & ((1<<(nbits))-1))
#define DROP_BITS(nbits)  (bits_left -= (nbits))
extern boolean jpeg_fill_bit_buffer
  (bitread_working_state * state, register bit_buf_type get_buffer, register int bits_left, int nbits);
#define HUFF_DECODE(result,state,htbl,failaction,slowlabel) { register int nb, look;   if (bits_left < HUFF_LOOKAHEAD) {     if (! jpeg_fill_bit_buffer(&state,get_buffer,bits_left, 0)) {failaction;}     get_buffer = state.get_buffer; bits_left = state.bits_left;     if (bits_left < HUFF_LOOKAHEAD) {       nb = 1; goto slowlabel;     }   }   look = PEEK_BITS(HUFF_LOOKAHEAD);   if ((nb = htbl->look_nbits[look]) != 0) {     DROP_BITS(nb);     result = htbl->look_sym[look];   } else {     nb = HUFF_LOOKAHEAD+1; slowlabel:     if ((result=jpeg_huff_decode(&state,get_buffer,bits_left,htbl,nb)) < 0)  { failaction; }     get_buffer = state.get_buffer; bits_left = state.bits_left;   } }
extern int jpeg_huff_decode
  (bitread_working_state * state, register bit_buf_type get_buffer, register int bits_left, d_derived_tbl * htbl, int min_bits);
typedef struct
 {
  int last_dc_val[4];
  } savable_state1;
#ifndef NO_STRUCT_ASSIGN
#define ASSIGN_STATE(dest,src)  ((dest) = (src))
#else
#endif
typedef struct
 {
  struct jpeg_entropy_decoder pub;
  bitread_perm_state bitstate;
  savable_state1 saved;
  unsigned int restarts_to_go;
  d_derived_tbl *dc_derived_tbls[4];
  d_derived_tbl *ac_derived_tbls[4];
  d_derived_tbl *dc_cur_tbls[10];
  d_derived_tbl *ac_cur_tbls[10];
  boolean dc_needed[10];
  boolean ac_needed[10];
  } huff_entropy_decoder;
typedef huff_entropy_decoder *huff_entropy_ptr1;
static void
  start_pass_huff_decoder(j_decompress_ptr cinfo)
 {
  huff_entropy_ptr1 entropy = (huff_entropy_ptr1) cinfo->entropy;
  int ci,
    blkn,
    dctbl,
    actbl;
  jpeg_component_info *compptr;
  if (cinfo->Ss != 0 || cinfo->Se != 64 - 1 ||
   cinfo->Ah != 0 || cinfo->Al != 0)
    ((cinfo)->err->msg_code = (JWRN_NOT_SEQUENTIAL), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    dctbl = compptr->dc_tbl_no;
    actbl = compptr->ac_tbl_no;
    jpeg_make_d_derived_tbl(cinfo, 1, dctbl,
     &entropy->dc_derived_tbls[dctbl]);
    jpeg_make_d_derived_tbl(cinfo, 0, actbl,
     &entropy->ac_derived_tbls[actbl]);
    entropy->saved.last_dc_val[ci] = 0;
    }
  for (blkn = 0; blkn < cinfo->blocks_in_MCU; blkn++)
   {
    ci = cinfo->MCU_membership[blkn];
    compptr = cinfo->cur_comp_info[ci];
    entropy->dc_cur_tbls[blkn] = entropy->dc_derived_tbls[compptr->dc_tbl_no];
    entropy->ac_cur_tbls[blkn] = entropy->ac_derived_tbls[compptr->ac_tbl_no];
    if (compptr->component_needed)
     {
      entropy->dc_needed[blkn] = 1;
      entropy->ac_needed[blkn] = (compptr->DCT_scaled_size > 1);
      }
    else
     {
      entropy->dc_needed[blkn] = entropy->ac_needed[blkn] = 0;
      }
    }
  entropy->bitstate.bits_left = 0;
  entropy->bitstate.get_buffer = 0;
  entropy->pub.insufficient_data = 0;
  entropy->restarts_to_go = cinfo->restart_interval;
  }
void
  jpeg_make_d_derived_tbl(j_decompress_ptr cinfo, boolean isDC, int tblno,
   d_derived_tbl ** pdtbl)
 {
  JHUFF_TBL *htbl;
  d_derived_tbl *dtbl;
  int p,
    i,
    l,
    si,
    numsymbols;
  int lookbits,
    ctr;
  char huffsize[257];
  unsigned int huffcode[257];
  unsigned int code;
  if (tblno < 0 || tblno >= 4)
    ((cinfo)->err->msg_code = (JERR_NO_HUFF_TABLE), (cinfo)->err->msg_parm.i[0] = (tblno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  htbl =
   isDC ? cinfo->dc_huff_tbl_ptrs[tblno] : cinfo->ac_huff_tbl_ptrs[tblno];
  if (htbl == ((void *) 0))
    ((cinfo)->err->msg_code = (JERR_NO_HUFF_TABLE), (cinfo)->err->msg_parm.i[0] = (tblno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (*pdtbl == ((void *) 0))
    *pdtbl = (d_derived_tbl *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
     ((size_t) sizeof(d_derived_tbl)));
  dtbl = *pdtbl;
  dtbl->pub = htbl;
  p = 0;
  for (l = 1; l <= 16; l++)
   {
    i = (int) htbl->bits[l];
    if (i < 0 || p + i > 256)
      ((cinfo)->err->msg_code = (JERR_BAD_HUFF_TABLE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    while (i--)
      huffsize[p++] = (char) l;
    }
  huffsize[p] = 0;
  numsymbols = p;
  code = 0;
  si = huffsize[0];
  p = 0;
  while (huffsize[p])
   {
    while (((int) huffsize[p]) == si)
     {
      huffcode[p++] = code;
      code++;
      }
    if (((int) code) >= (((int) 1) << si))
      ((cinfo)->err->msg_code = (JERR_BAD_HUFF_TABLE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    code <<= 1;
    si++;
    }
  p = 0;
  for (l = 1; l <= 16; l++)
   {
    if (htbl->bits[l])
     {
      dtbl->valoffset[l] = (int) p - (int) huffcode[p];
      p += htbl->bits[l];
      dtbl->maxcode[l] = huffcode[p - 1];
      }
    else
     {
      dtbl->maxcode[l] = -1;
      }
    }
  dtbl->maxcode[17] = 0xFFFFFL;
  memset((void *) (dtbl->look_nbits), 0, (size_t) (((size_t) sizeof(dtbl->look_nbits))));
  p = 0;
  for (l = 1; l <= 8; l++)
   {
    for (i = 1; i <= (int) htbl->bits[l]; i++, p++)
     {
      lookbits = huffcode[p] << (8 - l);
      for (ctr = 1 << (8 - l); ctr > 0; ctr--)
       {
        dtbl->look_nbits[lookbits] = l;
        dtbl->look_sym[lookbits] = htbl->huffval[p];
        lookbits++;
        }
      }
    }
  if (isDC)
   {
    for (i = 0; i < numsymbols; i++)
     {
      int sym = htbl->huffval[i];
      if (sym < 0 || sym > 15)
        ((cinfo)->err->msg_code = (JERR_BAD_HUFF_TABLE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      }
    }
  }
#ifdef SLOW_SHIFT_32
#else
#define MIN_GET_BITS  (BIT_BUF_SIZE-7)
#endif
boolean
jpeg_fill_bit_buffer(bitread_working_state * state,
 register bit_buf_type get_buffer, register int bits_left,
 int nbits)
 {
  register const JOCTET *next_input_byte = state->next_input_byte;
  register size_t bytes_in_buffer = state->bytes_in_buffer;
  j_decompress_ptr cinfo = state->cinfo;
  if (cinfo->unread_marker == 0)
   {
    while (bits_left < (32 - 7))
     {
      register int c;
      if (bytes_in_buffer == 0)
       {
        if (!(*cinfo->src->fill_input_buffer) (cinfo))
          return 0;
        next_input_byte = cinfo->src->next_input_byte;
        bytes_in_buffer = cinfo->src->bytes_in_buffer;
        }
      bytes_in_buffer--;
      c = (*next_input_byte++);
      if (c == 0xFF)
       {
        do
         {
          if (bytes_in_buffer == 0)
           {
            if (!(*cinfo->src->fill_input_buffer) (cinfo))
              return 0;
            next_input_byte = cinfo->src->next_input_byte;
            bytes_in_buffer = cinfo->src->bytes_in_buffer;
            }
          bytes_in_buffer--;
          c = (*next_input_byte++);
          }
        while (c == 0xFF);
        if (c == 0)
         {
          c = 0xFF;
          }
        else
         {
          cinfo->unread_marker = c;
          goto no_more_bytes;
          }
        }
      get_buffer = (get_buffer << 8) | c;
      bits_left += 8;
      }
    }
  else
   {
no_more_bytes:
    if (nbits > bits_left)
     {
      if (!cinfo->entropy->insufficient_data)
       {
        ((cinfo)->err->msg_code = (JWRN_HIT_MARKER), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
        cinfo->entropy->insufficient_data = 1;
        }
      get_buffer <<= (32 - 7) - bits_left;
      bits_left = (32 - 7);
      }
    }
  state->next_input_byte = next_input_byte;
  state->bytes_in_buffer = bytes_in_buffer;
  state->get_buffer = get_buffer;
  state->bits_left = bits_left;
  return 1;
  }
int
  jpeg_huff_decode(bitread_working_state * state,
   register bit_buf_type get_buffer, register int bits_left,
   d_derived_tbl * htbl, int min_bits)
 {
  register int l = min_bits;
  register int code;
   {
    if (bits_left < (l))
     {
      if (!jpeg_fill_bit_buffer(&(*state), get_buffer, bits_left, l))
       {
        return -1;
        } get_buffer = (*state).get_buffer;
      bits_left = (*state).bits_left;
      }
    };
  code = (((int) (get_buffer >> (bits_left -= (l)))) & ((1 << (l)) - 1));
  while (code > htbl->maxcode[l])
   {
    code <<= 1;
     {
      if (bits_left < (1))
       {
        if (!jpeg_fill_bit_buffer(&(*state), get_buffer, bits_left, 1))
         {
          return -1;
          } get_buffer = (*state).get_buffer;
        bits_left = (*state).bits_left;
        }
      };
    code |= (((int) (get_buffer >> (bits_left -= (1)))) & ((1 << (1)) - 1));
    l++;
    }
  state->get_buffer = get_buffer;
  state->bits_left = bits_left;
  if (l > 16)
   {
    ((state->cinfo)->err->msg_code = (JWRN_HUFF_BAD_CODE), (*(state->cinfo)->err->emit_message) ((j_common_ptr) (state->cinfo), -1));
    return 0;
    }
  return htbl->pub->huffval[(int) (code + htbl->valoffset[l])];
  }
#ifdef AVOID_TABLES
#else
#define HUFF_EXTEND(x,s)  ((x) < extend_test[s] ? (x) + extend_offset[s] : (x))
static const int extend_test[16] =
 {0, 0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080,
  0x0100, 0x0200, 0x0400, 0x0800, 0x1000, 0x2000, 0x4000};
static const int extend_offset[16] =
 {0, ((-1) << 1) + 1, ((-1) << 2) + 1, ((-1) << 3) + 1, ((-1) << 4) + 1,
  ((-1) << 5) + 1, ((-1) << 6) + 1, ((-1) << 7) + 1, ((-1) << 8) + 1,
  ((-1) << 9) + 1, ((-1) << 10) + 1, ((-1) << 11) + 1, ((-1) << 12) + 1,
  ((-1) << 13) + 1, ((-1) << 14) + 1, ((-1) << 15) + 1};
#endif
static boolean
  process_restart(j_decompress_ptr cinfo)
 {
  huff_entropy_ptr1 entropy = (huff_entropy_ptr1) cinfo->entropy;
  int ci;
  cinfo->marker->discarded_bytes += entropy->bitstate.bits_left / 8;
  entropy->bitstate.bits_left = 0;
  if (!(*cinfo->marker->read_restart_marker) (cinfo))
    return 0;
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
    entropy->saved.last_dc_val[ci] = 0;
  entropy->restarts_to_go = cinfo->restart_interval;
  if (cinfo->unread_marker == 0)
    entropy->pub.insufficient_data = 0;
  return 1;
  }
static boolean
  decode_mcu(j_decompress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  huff_entropy_ptr1 entropy = (huff_entropy_ptr1) cinfo->entropy;
  int blkn;
  register bit_buf_type get_buffer;
  register int bits_left;
  bitread_working_state br_state;
  savable_state1 state;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
      if (!process_restart(cinfo))
        return 0;
    }
  if (!entropy->pub.insufficient_data)
   {
    br_state.cinfo = cinfo;
    br_state.next_input_byte = cinfo->src->next_input_byte;
    br_state.bytes_in_buffer = cinfo->src->bytes_in_buffer;
    get_buffer = entropy->bitstate.get_buffer;
    bits_left = entropy->bitstate.bits_left;;
    ((state) = (entropy->saved));
    for (blkn = 0; blkn < cinfo->blocks_in_MCU; blkn++)
     {
      JBLOCKROW block = MCU_data[blkn];
      d_derived_tbl *dctbl = entropy->dc_cur_tbls[blkn];
      d_derived_tbl *actbl = entropy->ac_cur_tbls[blkn];
      register int s,
        k,
        r;
       {
        register int nb,
          look;
        if (bits_left < 8)
         {
          if (!jpeg_fill_bit_buffer(&br_state, get_buffer, bits_left, 0))
           {
            return 0;
            } get_buffer = br_state.get_buffer;
          bits_left = br_state.bits_left;
          if (bits_left < 8)
           {
            nb = 1;
            goto label1;
            }
          } look = (((int) (get_buffer >> (bits_left - (8)))) & ((1 << (8)) - 1));
        if ((nb = dctbl->look_nbits[look]) != 0)
         {
          (bits_left -= (nb));
          s = dctbl->look_sym[look];
          }
        else
         {
          nb = 8 + 1;
      label1:if ((s = jpeg_huff_decode(&br_state, get_buffer, bits_left, dctbl, nb)) < 0)
           {
            return 0;
            } get_buffer = br_state.get_buffer;
          bits_left = br_state.bits_left;
          }
        };
      if (s)
       {
         {
          if (bits_left < (s))
           {
            if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, s))
             {
              return 0;
              } get_buffer = (br_state).get_buffer;
            bits_left = (br_state).bits_left;
            }
          };
        r = (((int) (get_buffer >> (bits_left -= (s)))) & ((1 << (s)) - 1));
        s = ((r) < extend_test[s] ? (r) + extend_offset[s] : (r));
        }
      if (entropy->dc_needed[blkn])
       {
        int ci = cinfo->MCU_membership[blkn];
        s += state.last_dc_val[ci];
        state.last_dc_val[ci] = s;
        (*block)[0] = (JCOEF) s;
        }
      if (entropy->ac_needed[blkn])
       {
        for (k = 1; k < 64; k++)
         {
           {
            register int nb,
              look;
            if (bits_left < 8)
             {
              if (!jpeg_fill_bit_buffer(&br_state, get_buffer, bits_left, 0))
               {
                return 0;
                } get_buffer = br_state.get_buffer;
              bits_left = br_state.bits_left;
              if (bits_left < 8)
               {
                nb = 1;
                goto label2;
                }
              } look = (((int) (get_buffer >> (bits_left - (8)))) & ((1 << (8)) - 1));
            if ((nb = actbl->look_nbits[look]) != 0)
             {
              (bits_left -= (nb));
              s = actbl->look_sym[look];
              }
            else
             {
              nb = 8 + 1;
          label2:if ((s = jpeg_huff_decode(&br_state, get_buffer, bits_left, actbl, nb)) < 0)
               {
                return 0;
                } get_buffer = br_state.get_buffer;
              bits_left = br_state.bits_left;
              }
            };
          r = s >> 4;
          s &= 15;
          if (s)
           {
            k += r;
             {
              if (bits_left < (s))
               {
                if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, s))
                 {
                  return 0;
                  } get_buffer = (br_state).get_buffer;
                bits_left = (br_state).bits_left;
                }
              };
            r = (((int) (get_buffer >> (bits_left -= (s)))) & ((1 << (s)) - 1));
            s = ((r) < extend_test[s] ? (r) + extend_offset[s] : (r));
            (*block)[jpeg_natural_order[k]] = (JCOEF) s;
            }
          else
           {
            if (r != 15)
              break;
            k += 15;
            }
          }
        }
      else
       {
        for (k = 1; k < 64; k++)
         {
           {
            register int nb,
              look;
            if (bits_left < 8)
             {
              if (!jpeg_fill_bit_buffer(&br_state, get_buffer, bits_left, 0))
               {
                return 0;
                } get_buffer = br_state.get_buffer;
              bits_left = br_state.bits_left;
              if (bits_left < 8)
               {
                nb = 1;
                goto label3;
                }
              } look = (((int) (get_buffer >> (bits_left - (8)))) & ((1 << (8)) - 1));
            if ((nb = actbl->look_nbits[look]) != 0)
             {
              (bits_left -= (nb));
              s = actbl->look_sym[look];
              }
            else
             {
              nb = 8 + 1;
          label3:if ((s = jpeg_huff_decode(&br_state, get_buffer, bits_left, actbl, nb)) < 0)
               {
                return 0;
                } get_buffer = br_state.get_buffer;
              bits_left = br_state.bits_left;
              }
            };
          r = s >> 4;
          s &= 15;
          if (s)
           {
            k += r;
             {
              if (bits_left < (s))
               {
                if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, s))
                 {
                  return 0;
                  } get_buffer = (br_state).get_buffer;
                bits_left = (br_state).bits_left;
                }
              };
            (bits_left -= (s));
            }
          else
           {
            if (r != 15)
              break;
            k += 15;
            }
          }
        }
      }
    cinfo->src->next_input_byte = br_state.next_input_byte;
    cinfo->src->bytes_in_buffer = br_state.bytes_in_buffer;
    entropy->bitstate.get_buffer = get_buffer;
    entropy->bitstate.bits_left = bits_left;
    ((entropy->saved) = (state));
    }
  entropy->restarts_to_go--;
  return 1;
  }
void
  jinit_huff_decoder(j_decompress_ptr cinfo)
 {
  huff_entropy_ptr1 entropy;
  int i;
  entropy = (huff_entropy_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(huff_entropy_decoder)));
  cinfo->entropy = (struct jpeg_entropy_decoder *) entropy;
  entropy->pub.start_pass = start_pass_huff_decoder;
  entropy->pub.decode_mcu = decode_mcu;
  for (i = 0; i < 4; i++)
   {
    entropy->dc_derived_tbls[i] = entropy->ac_derived_tbls[i] = ((void *) 0);
    }
  }
#define JPEG_INTERNALS
#if 8  == 8
typedef int DCTELEM;
#else
#endif
typedef void (*forward_DCT_method_ptr) (DCTELEM * data);
typedef void (*float_DCT_method_ptr) (float *data);
typedef int ISLOW_MULT_TYPE;
#if 8  == 8
typedef int IFAST_MULT_TYPE;
#define IFAST_SCALE_BITS  2
#else
#endif
typedef float FLOAT_MULT_TYPE;
#define IDCT_range_limit(cinfo)  ((cinfo)->sample_range_limit + CENTERJSAMPLE)
#define RANGE_MASK  (MAXJSAMPLE * 4 + 3)
#ifdef NEED_SHORT_EXTERNAL_NAMES
#endif
extern void jpeg_fdct_islow(DCTELEM * data);
extern void jpeg_fdct_ifast(DCTELEM * data);
extern void jpeg_fdct_float(float *data);
extern void jpeg_idct_islow
  (j_decompress_ptr cinfo, jpeg_component_info * compptr, JCOEFPTR coef_block, JSAMPARRAY output_buf, JDIMENSION output_col);
extern void jpeg_idct_ifast
  (j_decompress_ptr cinfo, jpeg_component_info * compptr, JCOEFPTR coef_block, JSAMPARRAY output_buf, JDIMENSION output_col);
extern void jpeg_idct_float
  (j_decompress_ptr cinfo, jpeg_component_info * compptr, JCOEFPTR coef_block, JSAMPARRAY output_buf, JDIMENSION output_col);
extern void jpeg_idct_4x4
  (j_decompress_ptr cinfo, jpeg_component_info * compptr, JCOEFPTR coef_block, JSAMPARRAY output_buf, JDIMENSION output_col);
extern void jpeg_idct_2x2
  (j_decompress_ptr cinfo, jpeg_component_info * compptr, JCOEFPTR coef_block, JSAMPARRAY output_buf, JDIMENSION output_col);
extern void jpeg_idct_1x1
  (j_decompress_ptr cinfo, jpeg_component_info * compptr, JCOEFPTR coef_block, JSAMPARRAY output_buf, JDIMENSION output_col);
#define ONE 1
#define CONST_SCALE (ONE << CONST_BITS)
#define FIX(x) ((int) ((x) * CONST_SCALE + 0.5))
#define DESCALE(x,n)  RIGHT_SHIFT((x) + (ONE << ((n)-1)), n)
#ifdef SHORTxSHORT_32
#endif
#ifdef SHORTxLCONST_32
#endif
#ifndef MULTIPLY16C16
#define MULTIPLY16C16(var,const)  ((var) * (const))
#endif
#ifdef SHORTxSHORT_32
#endif
#ifndef MULTIPLY16V16
#define MULTIPLY16V16(var1,var2)  ((var1) * (var2))
#endif
typedef struct
 {
  struct jpeg_inverse_dct pub;
  int cur_method[10];
  } my_idct_controller;
typedef my_idct_controller *my_idct_ptr;
typedef union
 {
  ISLOW_MULT_TYPE islow_array[64];
#ifdef DCT_IFAST_SUPPORTED
  IFAST_MULT_TYPE ifast_array[64];
#endif
#ifdef DCT_FLOAT_SUPPORTED
  FLOAT_MULT_TYPE float_array[64];
#endif
  } multiplier_table;
#ifdef DCT_ISLOW_SUPPORTED
#define PROVIDE_ISLOW_TABLES
#else
#endif
static void
  start_pass(j_decompress_ptr cinfo)
 {
  my_idct_ptr idct = (my_idct_ptr) cinfo->idct;
  int ci,
    i;
  jpeg_component_info *compptr;
  int method = 0;
  inverse_DCT_method_ptr method_ptr = ((void *) 0);
  JQUANT_TBL *qtbl;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    switch (compptr->DCT_scaled_size)
     {
#ifdef IDCT_SCALING_SUPPORTED
     case 1:
      method_ptr = jpeg_idct_1x1;
      method = JDCT_ISLOW;
      break;
     case 2:
      method_ptr = jpeg_idct_2x2;
      method = JDCT_ISLOW;
      break;
     case 4:
      method_ptr = jpeg_idct_4x4;
      method = JDCT_ISLOW;
      break;
#endif
     case 8:
      switch (cinfo->dct_method)
       {
#ifdef DCT_ISLOW_SUPPORTED
       case JDCT_ISLOW:
        method_ptr = jpeg_idct_islow;
        method = JDCT_ISLOW;
        break;
#endif
#ifdef DCT_IFAST_SUPPORTED
       case JDCT_IFAST:
        method_ptr = jpeg_idct_ifast;
        method = JDCT_IFAST;
        break;
#endif
#ifdef DCT_FLOAT_SUPPORTED
       case JDCT_FLOAT:
        method_ptr = jpeg_idct_float;
        method = JDCT_FLOAT;
        break;
#endif
       default:
        ((cinfo)->err->msg_code = (JERR_NOT_COMPILED), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
        break;
        }
      break;
     default:
      ((cinfo)->err->msg_code = (JERR_BAD_DCTSIZE), (cinfo)->err->msg_parm.i[0] = (compptr->DCT_scaled_size), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      break;
      }
    idct->pub.inverse_DCT[ci] = method_ptr;
    if (!compptr->component_needed || idct->cur_method[ci] == method)
      continue;
    qtbl = compptr->quant_table;
    if (qtbl == ((void *) 0))
      continue;
    idct->cur_method[ci] = method;
    switch (method)
     {
#ifdef PROVIDE_ISLOW_TABLES
     case JDCT_ISLOW:
       {
        ISLOW_MULT_TYPE *ismtbl = (ISLOW_MULT_TYPE *) compptr->dct_table;
        for (i = 0; i < 64; i++)
         {
          ismtbl[i] = (ISLOW_MULT_TYPE) qtbl->quantval[i];
          }
        }
      break;
#endif
#ifdef DCT_IFAST_SUPPORTED
     case JDCT_IFAST:
       {
        IFAST_MULT_TYPE *ifmtbl = (IFAST_MULT_TYPE *) compptr->dct_table;
#define CONST_BITS 14
         static const short aanscales[64] = {
          16384, 22725, 21407, 19266, 16384, 12873, 8867, 4520,
          22725, 31521, 29692, 26722, 22725, 17855, 12299, 6270,
          21407, 29692, 27969, 25172, 21407, 16819, 11585, 5906,
          19266, 26722, 25172, 22654, 19266, 15137, 10426, 5315,
          16384, 22725, 21407, 19266, 16384, 12873, 8867, 4520,
          12873, 17855, 16819, 15137, 12873, 10114, 6967, 3552,
          8867, 12299, 11585, 10426, 8867, 6967, 4799, 2446,
          4520, 6270, 5906, 5315, 4520, 3552, 2446, 1247
          };
        for (i = 0; i < 64; i++)
         {
          ifmtbl[i] = (IFAST_MULT_TYPE)
           ((((((int) qtbl->quantval[i]) * ((int) aanscales[i]))) + (((int) 1) << ((14 - 2) - 1))) >> (14 - 2));
          }
        }
      break;
#endif
#ifdef DCT_FLOAT_SUPPORTED
     case JDCT_FLOAT:
       {
        FLOAT_MULT_TYPE *fmtbl = (FLOAT_MULT_TYPE *) compptr->dct_table;
        int row,
          col;
         static const double aanscalefactor[8] = {
          1.0, 1.387039845, 1.306562965, 1.175875602,
          1.0, 0.785694958, 0.541196100, 0.275899379
          };
        i = 0;
        for (row = 0; row < 8; row++)
         {
          for (col = 0; col < 8; col++)
           {
            fmtbl[i] = (FLOAT_MULT_TYPE)
             ((double) qtbl->quantval[i] *
             aanscalefactor[row] * aanscalefactor[col]);
            i++;
            }
          }
        }
      break;
#endif
     default:
      ((cinfo)->err->msg_code = (JERR_NOT_COMPILED), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      break;
      }
    }
  }
void
  jinit_inverse_dct(j_decompress_ptr cinfo)
 {
  my_idct_ptr idct;
  int ci;
  jpeg_component_info *compptr;
  idct = (my_idct_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_idct_controller)));
  cinfo->idct = (struct jpeg_inverse_dct *) idct;
  idct->pub.start_pass = start_pass;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    compptr->dct_table =
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
     ((size_t) sizeof(multiplier_table)));
    memset((void *) (compptr->dct_table), 0, (size_t) (((size_t) sizeof(multiplier_table))));
    idct->cur_method[ci] = -1;
    }
  }
#define JPEG_INTERNALS
#ifdef QUANT_1PASS_SUPPORTED
#define ODITHER_SIZE  16
#define ODITHER_CELLS (ODITHER_SIZE*ODITHER_SIZE)
#define ODITHER_MASK  (ODITHER_SIZE-1)
typedef int ODITHER_MATRIX[16][16];
typedef int (*ODITHER_MATRIX_PTR)[16];
static const UINT8 base_dither_matrix[16][16] = {
  {0, 192, 48, 240, 12, 204, 60, 252, 3, 195, 51, 243, 15, 207, 63, 255},
  {128, 64, 176, 112, 140, 76, 188, 124, 131, 67, 179, 115, 143, 79, 191, 127},
  {32, 224, 16, 208, 44, 236, 28, 220, 35, 227, 19, 211, 47, 239, 31, 223},
  {160, 96, 144, 80, 172, 108, 156, 92, 163, 99, 147, 83, 175, 111, 159, 95},
  {8, 200, 56, 248, 4, 196, 52, 244, 11, 203, 59, 251, 7, 199, 55, 247},
  {136, 72, 184, 120, 132, 68, 180, 116, 139, 75, 187, 123, 135, 71, 183, 119},
  {40, 232, 24, 216, 36, 228, 20, 212, 43, 235, 27, 219, 39, 231, 23, 215},
  {168, 104, 152, 88, 164, 100, 148, 84, 171, 107, 155, 91, 167, 103, 151, 87},
  {2, 194, 50, 242, 14, 206, 62, 254, 1, 193, 49, 241, 13, 205, 61, 253},
  {130, 66, 178, 114, 142, 78, 190, 126, 129, 65, 177, 113, 141, 77, 189, 125},
  {34, 226, 18, 210, 46, 238, 30, 222, 33, 225, 17, 209, 45, 237, 29, 221},
  {162, 98, 146, 82, 174, 110, 158, 94, 161, 97, 145, 81, 173, 109, 157, 93},
  {10, 202, 58, 250, 6, 198, 54, 246, 9, 201, 57, 249, 5, 197, 53, 245},
  {138, 74, 186, 122, 134, 70, 182, 118, 137, 73, 185, 121, 133, 69, 181, 117},
  {42, 234, 26, 218, 38, 230, 22, 214, 41, 233, 25, 217, 37, 229, 21, 213},
  {170, 106, 154, 90, 166, 102, 150, 86, 169, 105, 153, 89, 165, 101, 149, 85}
  };
#if 8  == 8
typedef short FSERROR;
typedef int LOCFSERROR;
#else
#endif
typedef FSERROR *FSERRPTR;
#define MAX_Q_COMPS 4
typedef struct
 {
  struct jpeg_color_quantizer pub;
  JSAMPARRAY sv_colormap;
  int sv_actual;
  JSAMPARRAY colorindex;
  boolean is_padded;
  int Ncolors[4];
  int row_index;
  ODITHER_MATRIX_PTR odither[4];
  FSERRPTR fserrors[4];
  boolean on_odd_row;
  } my_cquantizer;
typedef my_cquantizer *my_cquantize_ptr;
static int
  select_ncolors(j_decompress_ptr cinfo, int Ncolors[])
 {
  int nc = cinfo->out_color_components;
  int max_colors = cinfo->desired_number_of_colors;
  int total_colors,
    iroot,
    i,
    j;
  boolean changed;
  long temp;
  static const int RGB_order[3] = {1, 0, 2};
  iroot = 1;
  do
   {
    iroot++;
    temp = iroot;
    for (i = 1; i < nc; i++)
      temp *= iroot;
    }
  while (temp <= (long) max_colors);
  iroot--;
  if (iroot < 2)
    ((cinfo)->err->msg_code = (JERR_QUANT_FEW_COLORS), (cinfo)->err->msg_parm.i[0] = ((int) temp), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  total_colors = 1;
  for (i = 0; i < nc; i++)
   {
    Ncolors[i] = iroot;
    total_colors *= iroot;
    }
  do
   {
    changed = 0;
    for (i = 0; i < nc; i++)
     {
      j = (cinfo->out_color_space == JCS_RGB ? RGB_order[i] : i);
      temp = total_colors / Ncolors[j];
      temp *= Ncolors[j] + 1;
      if (temp > (long) max_colors)
        break;
      Ncolors[j]++;
      total_colors = (int) temp;
      changed = 1;
      }
    }
  while (changed);
  return total_colors;
  }
static int
  output_value(j_decompress_ptr cinfo, int ci, int j, int maxj)
 {
  return (int) (((int) j * 255 + maxj / 2) / maxj);
  }
static int
  largest_input_value(j_decompress_ptr cinfo, int ci, int j, int maxj)
 {
  return (int) (((int) (2 * j + 1) * 255 + maxj) / (2 * maxj));
  }
static void
  create_colormap(j_decompress_ptr cinfo)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  JSAMPARRAY colormap;
  int total_colors;
  int i,
    j,
    k,
    nci,
    blksize,
    blkdist,
    ptr,
    val;
  total_colors = select_ncolors(cinfo, cquantize->Ncolors);
  if (cinfo->out_color_components == 3)
    do
     {
      int *_mp = (cinfo)->err->msg_parm.i;
      _mp[0] = (total_colors);
      _mp[1] = (cquantize->Ncolors[0]);
      _mp[2] = (cquantize->Ncolors[1]);
      _mp[3] = (cquantize->Ncolors[2]);
      (cinfo)->err->msg_code = (JTRC_QUANT_3_NCOLORS);
      (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1));
      }
    while (0);
  else
  ((cinfo)->err->msg_code = (JTRC_QUANT_NCOLORS), (cinfo)->err->msg_parm.i[0] = (total_colors), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
  colormap = (*cinfo->mem->alloc_sarray)
   ((j_common_ptr) cinfo, 1,
   (JDIMENSION) total_colors, (JDIMENSION) cinfo->out_color_components);
  blkdist = total_colors;
  for (i = 0; i < cinfo->out_color_components; i++)
   {
    nci = cquantize->Ncolors[i];
    blksize = blkdist / nci;
    for (j = 0; j < nci; j++)
     {
      val = output_value(cinfo, i, j, nci - 1);
      for (ptr = j * blksize; ptr < total_colors; ptr += blkdist)
       {
        for (k = 0; k < blksize; k++)
          colormap[i][ptr + k] = (JSAMPLE) val;
        }
      }
    blkdist = blksize;
    }
  cquantize->sv_colormap = colormap;
  cquantize->sv_actual = total_colors;
  }
static void
  create_colorindex(j_decompress_ptr cinfo)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  JSAMPROW indexptr;
  int i,
    j,
    k,
    nci,
    blksize,
    val,
    pad;
  if (cinfo->dither_mode == JDITHER_ORDERED)
   {
    pad = 255 * 2;
    cquantize->is_padded = 1;
    }
  else
   {
    pad = 0;
    cquantize->is_padded = 0;
    }
  cquantize->colorindex = (*cinfo->mem->alloc_sarray)
   ((j_common_ptr) cinfo, 1,
   (JDIMENSION) (255 + 1 + pad),
   (JDIMENSION) cinfo->out_color_components);
  blksize = cquantize->sv_actual;
  for (i = 0; i < cinfo->out_color_components; i++)
   {
    nci = cquantize->Ncolors[i];
    blksize = blksize / nci;
    if (pad)
      cquantize->colorindex[i] += 255;
    indexptr = cquantize->colorindex[i];
    val = 0;
    k = largest_input_value(cinfo, i, 0, nci - 1);
    for (j = 0; j <= 255; j++)
     {
      while (j > k)
        k = largest_input_value(cinfo, i, ++val, nci - 1);
      indexptr[j] = (JSAMPLE) (val * blksize);
      }
    if (pad)
      for (j = 1; j <= 255; j++)
       {
        indexptr[-j] = indexptr[0];
        indexptr[255 + j] = indexptr[255];
        }
    }
  }
static ODITHER_MATRIX_PTR
  make_odither_array(j_decompress_ptr cinfo, int ncolors)
 {
  ODITHER_MATRIX_PTR odither;
  int j,
    k;
  int num,
    den;
  odither = (ODITHER_MATRIX_PTR)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(ODITHER_MATRIX)));
  den = 2 * (16 * 16) * ((int) (ncolors - 1));
  for (j = 0; j < 16; j++)
   {
    for (k = 0; k < 16; k++)
     {
      num = ((int) ((16 * 16) - 1 - 2 * ((int) base_dither_matrix[j][k])))
       * 255;
      odither[j][k] = (int) (num < 0 ? -((-num) / den) : num / den);
      }
    }
  return odither;
  }
static void
  create_odither_tables(j_decompress_ptr cinfo)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  ODITHER_MATRIX_PTR odither;
  int i,
    j,
    nci;
  for (i = 0; i < cinfo->out_color_components; i++)
   {
    nci = cquantize->Ncolors[i];
    odither = ((void *) 0);
    for (j = 0; j < i; j++)
     {
      if (nci == cquantize->Ncolors[j])
       {
        odither = cquantize->odither[j];
        break;
        }
      }
    if (odither == ((void *) 0))
      odither = make_odither_array(cinfo, nci);
    cquantize->odither[i] = odither;
    }
  }
static void
  color_quantize(j_decompress_ptr cinfo, JSAMPARRAY input_buf,
   JSAMPARRAY output_buf, int num_rows)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  JSAMPARRAY colorindex = cquantize->colorindex;
  register int pixcode,
    ci;
  register JSAMPROW ptrin,
    ptrout;
  int row;
  JDIMENSION col;
  JDIMENSION width = cinfo->output_width;
  register int nc = cinfo->out_color_components;
  for (row = 0; row < num_rows; row++)
   {
    ptrin = input_buf[row];
    ptrout = output_buf[row];
    for (col = width; col > 0; col--)
     {
      pixcode = 0;
      for (ci = 0; ci < nc; ci++)
       {
        pixcode += ((int) (colorindex[ci][((int) (*ptrin++))]));
        }
      *ptrout++ = (JSAMPLE) pixcode;
      }
    }
  }
static void
  color_quantize3(j_decompress_ptr cinfo, JSAMPARRAY input_buf,
   JSAMPARRAY output_buf, int num_rows)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  register int pixcode;
  register JSAMPROW ptrin,
    ptrout;
  JSAMPROW colorindex0 = cquantize->colorindex[0];
  JSAMPROW colorindex1 = cquantize->colorindex[1];
  JSAMPROW colorindex2 = cquantize->colorindex[2];
  int row;
  JDIMENSION col;
  JDIMENSION width = cinfo->output_width;
  for (row = 0; row < num_rows; row++)
   {
    ptrin = input_buf[row];
    ptrout = output_buf[row];
    for (col = width; col > 0; col--)
     {
      pixcode = ((int) (colorindex0[((int) (*ptrin++))]));
      pixcode += ((int) (colorindex1[((int) (*ptrin++))]));
      pixcode += ((int) (colorindex2[((int) (*ptrin++))]));
      *ptrout++ = (JSAMPLE) pixcode;
      }
    }
  }
static void
  quantize_ord_dither(j_decompress_ptr cinfo, JSAMPARRAY input_buf,
   JSAMPARRAY output_buf, int num_rows)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  register JSAMPROW input_ptr;
  register JSAMPROW output_ptr;
  JSAMPROW colorindex_ci;
  int *dither;
  int row_index,
    col_index;
  int nc = cinfo->out_color_components;
  int ci;
  int row;
  JDIMENSION col;
  JDIMENSION width = cinfo->output_width;
  for (row = 0; row < num_rows; row++)
   {
    jzero_far((void *) output_buf[row],
     (size_t) (width * ((size_t) sizeof(JSAMPLE))));
    row_index = cquantize->row_index;
    for (ci = 0; ci < nc; ci++)
     {
      input_ptr = input_buf[row] + ci;
      output_ptr = output_buf[row];
      colorindex_ci = cquantize->colorindex[ci];
      dither = cquantize->odither[ci][row_index];
      col_index = 0;
      for (col = width; col > 0; col--)
       {
        *output_ptr += colorindex_ci[((int) (*input_ptr)) + dither[col_index]];
        input_ptr += nc;
        output_ptr++;
        col_index = (col_index + 1) & (16 - 1);
        }
      }
    row_index = (row_index + 1) & (16 - 1);
    cquantize->row_index = row_index;
    }
  }
static void
  quantize3_ord_dither(j_decompress_ptr cinfo, JSAMPARRAY input_buf,
   JSAMPARRAY output_buf, int num_rows)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  register int pixcode;
  register JSAMPROW input_ptr;
  register JSAMPROW output_ptr;
  JSAMPROW colorindex0 = cquantize->colorindex[0];
  JSAMPROW colorindex1 = cquantize->colorindex[1];
  JSAMPROW colorindex2 = cquantize->colorindex[2];
  int *dither0;
  int *dither1;
  int *dither2;
  int row_index,
    col_index;
  int row;
  JDIMENSION col;
  JDIMENSION width = cinfo->output_width;
  for (row = 0; row < num_rows; row++)
   {
    row_index = cquantize->row_index;
    input_ptr = input_buf[row];
    output_ptr = output_buf[row];
    dither0 = cquantize->odither[0][row_index];
    dither1 = cquantize->odither[1][row_index];
    dither2 = cquantize->odither[2][row_index];
    col_index = 0;
    for (col = width; col > 0; col--)
     {
      pixcode = ((int) (colorindex0[((int) (*input_ptr++)) + dither0[col_index]]));
      pixcode += ((int) (colorindex1[((int) (*input_ptr++)) + dither1[col_index]]));
      pixcode += ((int) (colorindex2[((int) (*input_ptr++)) + dither2[col_index]]));
      *output_ptr++ = (JSAMPLE) pixcode;
      col_index = (col_index + 1) & (16 - 1);
      }
    row_index = (row_index + 1) & (16 - 1);
    cquantize->row_index = row_index;
    }
  }
static void
  quantize_fs_dither(j_decompress_ptr cinfo, JSAMPARRAY input_buf,
   JSAMPARRAY output_buf, int num_rows)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  register LOCFSERROR cur;
  LOCFSERROR belowerr;
  LOCFSERROR bpreverr;
  LOCFSERROR bnexterr;
  LOCFSERROR delta;
  register FSERRPTR errorptr;
  register JSAMPROW input_ptr;
  register JSAMPROW output_ptr;
  JSAMPROW colorindex_ci;
  JSAMPROW colormap_ci;
  int pixcode;
  int nc = cinfo->out_color_components;
  int dir;
  int dirnc;
  int ci;
  int row;
  JDIMENSION col;
  JDIMENSION width = cinfo->output_width;
  JSAMPLE *range_limit = cinfo->sample_range_limit;
  for (row = 0; row < num_rows; row++)
   {
    jzero_far((void *) output_buf[row],
     (size_t) (width * ((size_t) sizeof(JSAMPLE))));
    for (ci = 0; ci < nc; ci++)
     {
      input_ptr = input_buf[row] + ci;
      output_ptr = output_buf[row];
      if (cquantize->on_odd_row)
       {
        input_ptr += (width - 1) * nc;
        output_ptr += width - 1;
        dir = -1;
        dirnc = -nc;
        errorptr = cquantize->fserrors[ci] + (width + 1);
        }
      else
       {
        dir = 1;
        dirnc = nc;
        errorptr = cquantize->fserrors[ci];
        }
      colorindex_ci = cquantize->colorindex[ci];
      colormap_ci = cquantize->sv_colormap[ci];
      cur = 0;
      belowerr = bpreverr = 0;
      for (col = width; col > 0; col--)
       {
        cur = ((cur + errorptr[dir] + 8) >> (4));
        cur += ((int) (*input_ptr));
        cur = ((int) (range_limit[cur]));
        pixcode = ((int) (colorindex_ci[cur]));
        *output_ptr += (JSAMPLE) pixcode;
        cur -= ((int) (colormap_ci[pixcode]));
        bnexterr = cur;
        delta = cur * 2;
        cur += delta;
        errorptr[0] = (FSERROR) (bpreverr + cur);
        cur += delta;
        bpreverr = belowerr + cur;
        belowerr = bnexterr;
        cur += delta;
        input_ptr += dirnc;
        output_ptr += dir;
        errorptr += dir;
        }
      errorptr[0] = (FSERROR) bpreverr;
      }
    cquantize->on_odd_row = (cquantize->on_odd_row ? 0 : 1);
    }
  }
static void
  alloc_fs_workspace(j_decompress_ptr cinfo)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  size_t arraysize;
  int i;
  arraysize = (size_t) ((cinfo->output_width + 2) * ((size_t) sizeof(FSERROR)));
  for (i = 0; i < cinfo->out_color_components; i++)
   {
    cquantize->fserrors[i] = (FSERRPTR)
     (*cinfo->mem->alloc_large) ((j_common_ptr) cinfo, 1, arraysize);
    }
  }
static void
  start_pass_1_quant(j_decompress_ptr cinfo, boolean is_pre_scan)
 {
  my_cquantize_ptr cquantize = (my_cquantize_ptr) cinfo->cquantize;
  size_t arraysize;
  int i;
  cinfo->colormap = cquantize->sv_colormap;
  cinfo->actual_number_of_colors = cquantize->sv_actual;
  switch (cinfo->dither_mode)
   {
   case JDITHER_NONE:
    if (cinfo->out_color_components == 3)
      cquantize->pub.color_quantize = color_quantize3;
    else
      cquantize->pub.color_quantize = color_quantize;
    break;
   case JDITHER_ORDERED:
    if (cinfo->out_color_components == 3)
      cquantize->pub.color_quantize = quantize3_ord_dither;
    else
      cquantize->pub.color_quantize = quantize_ord_dither;
    cquantize->row_index = 0;
    if (!cquantize->is_padded)
      create_colorindex(cinfo);
    if (cquantize->odither[0] == ((void *) 0))
      create_odither_tables(cinfo);
    break;
   case JDITHER_FS:
    cquantize->pub.color_quantize = quantize_fs_dither;
    cquantize->on_odd_row = 0;
    if (cquantize->fserrors[0] == ((void *) 0))
      alloc_fs_workspace(cinfo);
    arraysize = (size_t) ((cinfo->output_width + 2) * ((size_t) sizeof(FSERROR)));
    for (i = 0; i < cinfo->out_color_components; i++)
      jzero_far((void *) cquantize->fserrors[i], arraysize);
    break;
   default:
    ((cinfo)->err->msg_code = (JERR_NOT_COMPILED), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  }
static void
  finish_pass_1_quant(j_decompress_ptr cinfo)
 {
  }
static void
  new_color_map_1_quant(j_decompress_ptr cinfo)
 {
  ((cinfo)->err->msg_code = (JERR_MODE_CHANGE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  }
void
  jinit_1pass_quantizer(j_decompress_ptr cinfo)
 {
  my_cquantize_ptr cquantize;
  cquantize = (my_cquantize_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_cquantizer)));
  cinfo->cquantize = (struct jpeg_color_quantizer *) cquantize;
  cquantize->pub.start_pass = start_pass_1_quant;
  cquantize->pub.finish_pass = finish_pass_1_quant;
  cquantize->pub.new_color_map = new_color_map_1_quant;
  cquantize->fserrors[0] = ((void *) 0);
  cquantize->odither[0] = ((void *) 0);
  if (cinfo->out_color_components > 4)
    ((cinfo)->err->msg_code = (JERR_QUANT_COMPONENTS), (cinfo)->err->msg_parm.i[0] = (4), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->desired_number_of_colors > (255 + 1))
    ((cinfo)->err->msg_code = (JERR_QUANT_MANY_COLORS), (cinfo)->err->msg_parm.i[0] = (255 + 1), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  create_colormap(cinfo);
  create_colorindex(cinfo);
  if (cinfo->dither_mode == JDITHER_FS)
    alloc_fs_workspace(cinfo);
  }
#endif
#define JPEG_INTERNALS
#ifdef QUANT_2PASS_SUPPORTED
#define R_SCALE 2
#define G_SCALE 3
#define B_SCALE 1
#if 0  == 0
#define C0_SCALE R_SCALE
#endif
#if 2  == 0
#endif
#if 1  == 1
#define C1_SCALE G_SCALE
#endif
#if 0  == 2
#endif
#if 2  == 2
#define C2_SCALE B_SCALE
#endif
#define MAXNUMCOLORS  (MAXJSAMPLE+1)
#define HIST_C0_BITS  5
#define HIST_C1_BITS  6
#define HIST_C2_BITS  5
#define HIST_C0_ELEMS  (1<<HIST_C0_BITS)
#define HIST_C1_ELEMS  (1<<HIST_C1_BITS)
#define HIST_C2_ELEMS  (1<<HIST_C2_BITS)
#define C0_SHIFT  (BITS_IN_JSAMPLE-HIST_C0_BITS)
#define C1_SHIFT  (BITS_IN_JSAMPLE-HIST_C1_BITS)
#define C2_SHIFT  (BITS_IN_JSAMPLE-HIST_C2_BITS)
typedef UINT16 histcell;
typedef histcell *histptr;
typedef histcell hist1d[(1 << 5)];
typedef hist1d *hist2d;
typedef hist2d *hist3d;
#if 8  == 8
typedef short FSERROR;
typedef int LOCFSERROR;
#else
#endif
typedef FSERROR *FSERRPTR;
typedef struct
 {
  struct jpeg_color_quantizer pub;
  JSAMPARRAY sv_colormap;
  int desired;
  hist3d histogram;
  boolean needs_zeroed;
  FSERRPTR fserrors;
  boolean on_odd_row;
  int *error_limiter;
  } my_cquantizer1;
typedef my_cquantizer1 *my_cquantizer_ptr1;
static void
  prescan_quantize(j_decompress_ptr cinfo, JSAMPARRAY input_buf,
   JSAMPARRAY output_buf, int num_rows)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  register JSAMPROW ptr;
  register histptr histp;
  register hist3d histogram = cquantize->histogram;
  int row;
  JDIMENSION col;
  JDIMENSION width = cinfo->output_width;
  for (row = 0; row < num_rows; row++)
   {
    ptr = input_buf[row];
    for (col = width; col > 0; col--)
     {
      histp = &histogram[((int) (ptr[0])) >> (8 - 5)]
       [((int) (ptr[1])) >> (8 - 6)]
       [((int) (ptr[2])) >> (8 - 5)];
      if (++(*histp) <= 0)
        (*histp)--;
      ptr += 3;
      }
    }
  }
typedef struct
 {
  int c0min,
    c0max;
  int c1min,
    c1max;
  int c2min,
    c2max;
  int volume;
  long colorcount;
  } box;
typedef box *boxptr;
static boxptr
  find_biggest_color_pop(boxptr boxlist, int numboxes)
 {
  register boxptr boxp;
  register int i;
  register long maxc = 0;
  boxptr which = ((void *) 0);
  for (i = 0, boxp = boxlist; i < numboxes; i++, boxp++)
   {
    if (boxp->colorcount > maxc && boxp->volume > 0)
     {
      which = boxp;
      maxc = boxp->colorcount;
      }
    }
  return which;
  }
static boxptr
  find_biggest_volume(boxptr boxlist, int numboxes)
 {
  register boxptr boxp;
  register int i;
  register int maxv = 0;
  boxptr which = ((void *) 0);
  for (i = 0, boxp = boxlist; i < numboxes; i++, boxp++)
   {
    if (boxp->volume > maxv)
     {
      which = boxp;
      maxv = boxp->volume;
      }
    }
  return which;
  }
static void
  update_box(j_decompress_ptr cinfo, boxptr boxp)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  hist3d histogram = cquantize->histogram;
  histptr histp;
  int c0,
    c1,
    c2;
  int c0min,
    c0max,
    c1min,
    c1max,
    c2min,
    c2max;
  int dist0,
    dist1,
    dist2;
  long ccount;
  c0min = boxp->c0min;
  c0max = boxp->c0max;
  c1min = boxp->c1min;
  c1max = boxp->c1max;
  c2min = boxp->c2min;
  c2max = boxp->c2max;
  if (c0max > c0min)
    for (c0 = c0min; c0 <= c0max; c0++)
      for (c1 = c1min; c1 <= c1max; c1++)
       {
        histp = &histogram[c0][c1][c2min];
        for (c2 = c2min; c2 <= c2max; c2++)
          if (*histp++ != 0)
           {
            boxp->c0min = c0min = c0;
            goto have_c0min;
            }
        }
have_c0min:
  if (c0max > c0min)
    for (c0 = c0max; c0 >= c0min; c0--)
      for (c1 = c1min; c1 <= c1max; c1++)
       {
        histp = &histogram[c0][c1][c2min];
        for (c2 = c2min; c2 <= c2max; c2++)
          if (*histp++ != 0)
           {
            boxp->c0max = c0max = c0;
            goto have_c0max;
            }
        }
have_c0max:
  if (c1max > c1min)
    for (c1 = c1min; c1 <= c1max; c1++)
      for (c0 = c0min; c0 <= c0max; c0++)
       {
        histp = &histogram[c0][c1][c2min];
        for (c2 = c2min; c2 <= c2max; c2++)
          if (*histp++ != 0)
           {
            boxp->c1min = c1min = c1;
            goto have_c1min;
            }
        }
have_c1min:
  if (c1max > c1min)
    for (c1 = c1max; c1 >= c1min; c1--)
      for (c0 = c0min; c0 <= c0max; c0++)
       {
        histp = &histogram[c0][c1][c2min];
        for (c2 = c2min; c2 <= c2max; c2++)
          if (*histp++ != 0)
           {
            boxp->c1max = c1max = c1;
            goto have_c1max;
            }
        }
have_c1max:
  if (c2max > c2min)
    for (c2 = c2min; c2 <= c2max; c2++)
      for (c0 = c0min; c0 <= c0max; c0++)
       {
        histp = &histogram[c0][c1min][c2];
        for (c1 = c1min; c1 <= c1max; c1++, histp += (1 << 5))
          if (*histp != 0)
           {
            boxp->c2min = c2min = c2;
            goto have_c2min;
            }
        }
have_c2min:
  if (c2max > c2min)
    for (c2 = c2max; c2 >= c2min; c2--)
      for (c0 = c0min; c0 <= c0max; c0++)
       {
        histp = &histogram[c0][c1min][c2];
        for (c1 = c1min; c1 <= c1max; c1++, histp += (1 << 5))
          if (*histp != 0)
           {
            boxp->c2max = c2max = c2;
            goto have_c2max;
            }
        }
have_c2max:
  dist0 = ((c0max - c0min) << (8 - 5)) * 2;
  dist1 = ((c1max - c1min) << (8 - 6)) * 3;
  dist2 = ((c2max - c2min) << (8 - 5)) * 1;
  boxp->volume = dist0 * dist0 + dist1 * dist1 + dist2 * dist2;
  ccount = 0;
  for (c0 = c0min; c0 <= c0max; c0++)
    for (c1 = c1min; c1 <= c1max; c1++)
     {
      histp = &histogram[c0][c1][c2min];
      for (c2 = c2min; c2 <= c2max; c2++, histp++)
        if (*histp != 0)
         {
          ccount++;
          }
      }
  boxp->colorcount = ccount;
  }
static int
  median_cut(j_decompress_ptr cinfo, boxptr boxlist, int numboxes,
   int desired_colors)
 {
  int n,
    lb;
  int c0,
    c1,
    c2,
    cmax;
  register boxptr b1,
    b2;
  while (numboxes < desired_colors)
   {
    if (numboxes * 2 <= desired_colors)
     {
      b1 = find_biggest_color_pop(boxlist, numboxes);
      }
    else
     {
      b1 = find_biggest_volume(boxlist, numboxes);
      }
    if (b1 == ((void *) 0))
      break;
    b2 = &boxlist[numboxes];
    b2->c0max = b1->c0max;
    b2->c1max = b1->c1max;
    b2->c2max = b1->c2max;
    b2->c0min = b1->c0min;
    b2->c1min = b1->c1min;
    b2->c2min = b1->c2min;
    c0 = ((b1->c0max - b1->c0min) << (8 - 5)) * 2;
    c1 = ((b1->c1max - b1->c1min) << (8 - 6)) * 3;
    c2 = ((b1->c2max - b1->c2min) << (8 - 5)) * 1;
#if 0  == 0
    cmax = c1;
    n = 1;
    if (c0 > cmax)
     {
      cmax = c0;
      n = 0;
      }
    if (c2 > cmax)
     {
      n = 2;
      }
#else
#endif
    switch (n)
     {
     case 0:
      lb = (b1->c0max + b1->c0min) / 2;
      b1->c0max = lb;
      b2->c0min = lb + 1;
      break;
     case 1:
      lb = (b1->c1max + b1->c1min) / 2;
      b1->c1max = lb;
      b2->c1min = lb + 1;
      break;
     case 2:
      lb = (b1->c2max + b1->c2min) / 2;
      b1->c2max = lb;
      b2->c2min = lb + 1;
      break;
      }
    update_box(cinfo, b1);
    update_box(cinfo, b2);
    numboxes++;
    }
  return numboxes;
  }
static void
  compute_color(j_decompress_ptr cinfo, boxptr boxp, int icolor)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  hist3d histogram = cquantize->histogram;
  histptr histp;
  int c0,
    c1,
    c2;
  int c0min,
    c0max,
    c1min,
    c1max,
    c2min,
    c2max;
  long count;
  long total = 0;
  long c0total = 0;
  long c1total = 0;
  long c2total = 0;
  c0min = boxp->c0min;
  c0max = boxp->c0max;
  c1min = boxp->c1min;
  c1max = boxp->c1max;
  c2min = boxp->c2min;
  c2max = boxp->c2max;
  for (c0 = c0min; c0 <= c0max; c0++)
    for (c1 = c1min; c1 <= c1max; c1++)
     {
      histp = &histogram[c0][c1][c2min];
      for (c2 = c2min; c2 <= c2max; c2++)
       {
        if ((count = *histp++) != 0)
         {
          total += count;
          c0total += ((c0 << (8 - 5)) + ((1 << (8 - 5)) >> 1)) * count;
          c1total += ((c1 << (8 - 6)) + ((1 << (8 - 6)) >> 1)) * count;
          c2total += ((c2 << (8 - 5)) + ((1 << (8 - 5)) >> 1)) * count;
          }
        }
      }
  cinfo->colormap[0][icolor] = (JSAMPLE) ((c0total + (total >> 1)) / total);
  cinfo->colormap[1][icolor] = (JSAMPLE) ((c1total + (total >> 1)) / total);
  cinfo->colormap[2][icolor] = (JSAMPLE) ((c2total + (total >> 1)) / total);
  }
static void
  select_colors(j_decompress_ptr cinfo, int desired_colors)
 {
  boxptr boxlist;
  int numboxes;
  int i;
  boxlist = (boxptr) (*cinfo->mem->alloc_small)
   ((j_common_ptr) cinfo, 1, desired_colors * ((size_t) sizeof(box)));
  numboxes = 1;
  boxlist[0].c0min = 0;
  boxlist[0].c0max = 255 >> (8 - 5);
  boxlist[0].c1min = 0;
  boxlist[0].c1max = 255 >> (8 - 6);
  boxlist[0].c2min = 0;
  boxlist[0].c2max = 255 >> (8 - 5);
  update_box(cinfo, &boxlist[0]);
  numboxes = median_cut(cinfo, boxlist, numboxes, desired_colors);
  for (i = 0; i < numboxes; i++)
    compute_color(cinfo, &boxlist[i], i);
  cinfo->actual_number_of_colors = numboxes;
  ((cinfo)->err->msg_code = (JTRC_QUANT_SELECTED), (cinfo)->err->msg_parm.i[0] = (numboxes), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (1)));
  }
#define BOX_C0_LOG  (HIST_C0_BITS-3)
#define BOX_C1_LOG  (HIST_C1_BITS-3)
#define BOX_C2_LOG  (HIST_C2_BITS-3)
#define BOX_C0_ELEMS  (1<<BOX_C0_LOG)
#define BOX_C1_ELEMS  (1<<BOX_C1_LOG)
#define BOX_C2_ELEMS  (1<<BOX_C2_LOG)
#define BOX_C0_SHIFT  (C0_SHIFT + BOX_C0_LOG)
#define BOX_C1_SHIFT  (C1_SHIFT + BOX_C1_LOG)
#define BOX_C2_SHIFT  (C2_SHIFT + BOX_C2_LOG)
static int
  find_nearby_colors(j_decompress_ptr cinfo, int minc0, int minc1, int minc2,
   JSAMPLE colorlist[])
 {
  int numcolors = cinfo->actual_number_of_colors;
  int maxc0,
    maxc1,
    maxc2;
  int centerc0,
    centerc1,
    centerc2;
  int i,
    x,
    ncolors;
  int minmaxdist,
    min_dist,
    max_dist,
    tdist;
  int mindist[(255 + 1)];
  maxc0 = minc0 + ((1 << ((8 - 5) + (5 - 3))) - (1 << (8 - 5)));
  centerc0 = (minc0 + maxc0) >> 1;
  maxc1 = minc1 + ((1 << ((8 - 6) + (6 - 3))) - (1 << (8 - 6)));
  centerc1 = (minc1 + maxc1) >> 1;
  maxc2 = minc2 + ((1 << ((8 - 5) + (5 - 3))) - (1 << (8 - 5)));
  centerc2 = (minc2 + maxc2) >> 1;
  minmaxdist = 0x7FFFFFFFL;
  for (i = 0; i < numcolors; i++)
   {
    x = ((int) (cinfo->colormap[0][i]));
    if (x < minc0)
     {
      tdist = (x - minc0) * 2;
      min_dist = tdist * tdist;
      tdist = (x - maxc0) * 2;
      max_dist = tdist * tdist;
      }
    else if (x > maxc0)
     {
      tdist = (x - maxc0) * 2;
      min_dist = tdist * tdist;
      tdist = (x - minc0) * 2;
      max_dist = tdist * tdist;
      }
    else
     {
      min_dist = 0;
      if (x <= centerc0)
       {
        tdist = (x - maxc0) * 2;
        max_dist = tdist * tdist;
        }
      else
       {
        tdist = (x - minc0) * 2;
        max_dist = tdist * tdist;
        }
      }
    x = ((int) (cinfo->colormap[1][i]));
    if (x < minc1)
     {
      tdist = (x - minc1) * 3;
      min_dist += tdist * tdist;
      tdist = (x - maxc1) * 3;
      max_dist += tdist * tdist;
      }
    else if (x > maxc1)
     {
      tdist = (x - maxc1) * 3;
      min_dist += tdist * tdist;
      tdist = (x - minc1) * 3;
      max_dist += tdist * tdist;
      }
    else
     {
      if (x <= centerc1)
       {
        tdist = (x - maxc1) * 3;
        max_dist += tdist * tdist;
        }
      else
       {
        tdist = (x - minc1) * 3;
        max_dist += tdist * tdist;
        }
      }
    x = ((int) (cinfo->colormap[2][i]));
    if (x < minc2)
     {
      tdist = (x - minc2) * 1;
      min_dist += tdist * tdist;
      tdist = (x - maxc2) * 1;
      max_dist += tdist * tdist;
      }
    else if (x > maxc2)
     {
      tdist = (x - maxc2) * 1;
      min_dist += tdist * tdist;
      tdist = (x - minc2) * 1;
      max_dist += tdist * tdist;
      }
    else
     {
      if (x <= centerc2)
       {
        tdist = (x - maxc2) * 1;
        max_dist += tdist * tdist;
        }
      else
       {
        tdist = (x - minc2) * 1;
        max_dist += tdist * tdist;
        }
      }
    mindist[i] = min_dist;
    if (max_dist < minmaxdist)
      minmaxdist = max_dist;
    }
  ncolors = 0;
  for (i = 0; i < numcolors; i++)
   {
    if (mindist[i] <= minmaxdist)
      colorlist[ncolors++] = (JSAMPLE) i;
    }
  return ncolors;
  }
static void
  find_best_colors(j_decompress_ptr cinfo, int minc0, int minc1, int minc2,
   int numcolors, JSAMPLE colorlist[], JSAMPLE bestcolor[])
 {
  int ic0,
    ic1,
    ic2;
  int i,
    icolor;
  register int *bptr;
  JSAMPLE *cptr;
  int dist0,
    dist1;
  register int dist2;
  int xx0,
    xx1;
  register int xx2;
  int inc0,
    inc1,
    inc2;
  int bestdist[(1 << (5 - 3)) * (1 << (6 - 3)) * (1 << (5 - 3))];
  bptr = bestdist;
  for (i = (1 << (5 - 3)) * (1 << (6 - 3)) * (1 << (5 - 3)) - 1; i >= 0; i--)
    *bptr++ = 0x7FFFFFFFL;
#define STEP_C0  ((1 << C0_SHIFT) * C0_SCALE)
#define STEP_C1  ((1 << C1_SHIFT) * C1_SCALE)
#define STEP_C2  ((1 << C2_SHIFT) * C2_SCALE)
  for (i = 0; i < numcolors; i++)
   {
    icolor = ((int) (colorlist[i]));
    inc0 = (minc0 - ((int) (cinfo->colormap[0][icolor]))) * 2;
    dist0 = inc0 * inc0;
    inc1 = (minc1 - ((int) (cinfo->colormap[1][icolor]))) * 3;
    dist0 += inc1 * inc1;
    inc2 = (minc2 - ((int) (cinfo->colormap[2][icolor]))) * 1;
    dist0 += inc2 * inc2;
    inc0 = inc0 * (2 * ((1 << (8 - 5)) * 2)) + ((1 << (8 - 5)) * 2) * ((1 << (8 - 5)) * 2);
    inc1 = inc1 * (2 * ((1 << (8 - 6)) * 3)) + ((1 << (8 - 6)) * 3) * ((1 << (8 - 6)) * 3);
    inc2 = inc2 * (2 * ((1 << (8 - 5)) * 1)) + ((1 << (8 - 5)) * 1) * ((1 << (8 - 5)) * 1);
    bptr = bestdist;
    cptr = bestcolor;
    xx0 = inc0;
    for (ic0 = (1 << (5 - 3)) - 1; ic0 >= 0; ic0--)
     {
      dist1 = dist0;
      xx1 = inc1;
      for (ic1 = (1 << (6 - 3)) - 1; ic1 >= 0; ic1--)
       {
        dist2 = dist1;
        xx2 = inc2;
        for (ic2 = (1 << (5 - 3)) - 1; ic2 >= 0; ic2--)
         {
          if (dist2 < *bptr)
           {
            *bptr = dist2;
            *cptr = (JSAMPLE) icolor;
            }
          dist2 += xx2;
          xx2 += 2 * ((1 << (8 - 5)) * 1) * ((1 << (8 - 5)) * 1);
          bptr++;
          cptr++;
          }
        dist1 += xx1;
        xx1 += 2 * ((1 << (8 - 6)) * 3) * ((1 << (8 - 6)) * 3);
        }
      dist0 += xx0;
      xx0 += 2 * ((1 << (8 - 5)) * 2) * ((1 << (8 - 5)) * 2);
      }
    }
  }
static void
  fill_inverse_cmap(j_decompress_ptr cinfo, int c0, int c1, int c2)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  hist3d histogram = cquantize->histogram;
  int minc0,
    minc1,
    minc2;
  int ic0,
    ic1,
    ic2;
  register JSAMPLE *cptr;
  register histptr cachep;
  JSAMPLE colorlist[(255 + 1)];
  int numcolors;
  JSAMPLE bestcolor[(1 << (5 - 3)) * (1 << (6 - 3)) * (1 << (5 - 3))];
  c0 >>= (5 - 3);
  c1 >>= (6 - 3);
  c2 >>= (5 - 3);
  minc0 = (c0 << ((8 - 5) + (5 - 3))) + ((1 << (8 - 5)) >> 1);
  minc1 = (c1 << ((8 - 6) + (6 - 3))) + ((1 << (8 - 6)) >> 1);
  minc2 = (c2 << ((8 - 5) + (5 - 3))) + ((1 << (8 - 5)) >> 1);
  numcolors = find_nearby_colors(cinfo, minc0, minc1, minc2, colorlist);
  find_best_colors(cinfo, minc0, minc1, minc2, numcolors, colorlist,
   bestcolor);
  c0 <<= (5 - 3);
  c1 <<= (6 - 3);
  c2 <<= (5 - 3);
  cptr = bestcolor;
  for (ic0 = 0; ic0 < (1 << (5 - 3)); ic0++)
   {
    for (ic1 = 0; ic1 < (1 << (6 - 3)); ic1++)
     {
      cachep = &histogram[c0 + ic0][c1 + ic1][c2];
      for (ic2 = 0; ic2 < (1 << (5 - 3)); ic2++)
       {
        *cachep++ = (histcell) (((int) (*cptr++)) + 1);
        }
      }
    }
  }
static void
  pass2_no_dither(j_decompress_ptr cinfo,
   JSAMPARRAY input_buf, JSAMPARRAY output_buf, int num_rows)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  hist3d histogram = cquantize->histogram;
  register JSAMPROW inptr,
    outptr;
  register histptr cachep;
  register int c0,
    c1,
    c2;
  int row;
  JDIMENSION col;
  JDIMENSION width = cinfo->output_width;
  for (row = 0; row < num_rows; row++)
   {
    inptr = input_buf[row];
    outptr = output_buf[row];
    for (col = width; col > 0; col--)
     {
      c0 = ((int) (*inptr++)) >> (8 - 5);
      c1 = ((int) (*inptr++)) >> (8 - 6);
      c2 = ((int) (*inptr++)) >> (8 - 5);
      cachep = &histogram[c0][c1][c2];
      if (*cachep == 0)
        fill_inverse_cmap(cinfo, c0, c1, c2);
      *outptr++ = (JSAMPLE) (*cachep - 1);
      }
    }
  }
static void
  pass2_fs_dither(j_decompress_ptr cinfo,
   JSAMPARRAY input_buf, JSAMPARRAY output_buf, int num_rows)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  hist3d histogram = cquantize->histogram;
  register LOCFSERROR cur0,
    cur1,
    cur2;
  LOCFSERROR belowerr0,
    belowerr1,
    belowerr2;
  LOCFSERROR bpreverr0,
    bpreverr1,
    bpreverr2;
  register FSERRPTR errorptr;
  JSAMPROW inptr;
  JSAMPROW outptr;
  histptr cachep;
  int dir;
  int dir3;
  int row;
  JDIMENSION col;
  JDIMENSION width = cinfo->output_width;
  JSAMPLE *range_limit = cinfo->sample_range_limit;
  int *error_limit = cquantize->error_limiter;
  JSAMPROW colormap0 = cinfo->colormap[0];
  JSAMPROW colormap1 = cinfo->colormap[1];
  JSAMPROW colormap2 = cinfo->colormap[2];
  for (row = 0; row < num_rows; row++)
   {
    inptr = input_buf[row];
    outptr = output_buf[row];
    if (cquantize->on_odd_row)
     {
      inptr += (width - 1) * 3;
      outptr += width - 1;
      dir = -1;
      dir3 = -3;
      errorptr = cquantize->fserrors + (width + 1) * 3;
      cquantize->on_odd_row = 0;
      }
    else
     {
      dir = 1;
      dir3 = 3;
      errorptr = cquantize->fserrors;
      cquantize->on_odd_row = 1;
      }
    cur0 = cur1 = cur2 = 0;
    belowerr0 = belowerr1 = belowerr2 = 0;
    bpreverr0 = bpreverr1 = bpreverr2 = 0;
    for (col = width; col > 0; col--)
     {
      cur0 = ((cur0 + errorptr[dir3 + 0] + 8) >> (4));
      cur1 = ((cur1 + errorptr[dir3 + 1] + 8) >> (4));
      cur2 = ((cur2 + errorptr[dir3 + 2] + 8) >> (4));
      cur0 = error_limit[cur0];
      cur1 = error_limit[cur1];
      cur2 = error_limit[cur2];
      cur0 += ((int) (inptr[0]));
      cur1 += ((int) (inptr[1]));
      cur2 += ((int) (inptr[2]));
      cur0 = ((int) (range_limit[cur0]));
      cur1 = ((int) (range_limit[cur1]));
      cur2 = ((int) (range_limit[cur2]));
      cachep = &histogram[cur0 >> (8 - 5)][cur1 >> (8 - 6)][cur2 >> (8 - 5)];
      if (*cachep == 0)
        fill_inverse_cmap(cinfo, cur0 >> (8 - 5), cur1 >> (8 - 6), cur2 >> (8 - 5));
       {
        register int pixcode = *cachep - 1;
        *outptr = (JSAMPLE) pixcode;
        cur0 -= ((int) (colormap0[pixcode]));
        cur1 -= ((int) (colormap1[pixcode]));
        cur2 -= ((int) (colormap2[pixcode]));
        }
       {
        register LOCFSERROR bnexterr,
          delta;
        bnexterr = cur0;
        delta = cur0 * 2;
        cur0 += delta;
        errorptr[0] = (FSERROR) (bpreverr0 + cur0);
        cur0 += delta;
        bpreverr0 = belowerr0 + cur0;
        belowerr0 = bnexterr;
        cur0 += delta;
        bnexterr = cur1;
        delta = cur1 * 2;
        cur1 += delta;
        errorptr[1] = (FSERROR) (bpreverr1 + cur1);
        cur1 += delta;
        bpreverr1 = belowerr1 + cur1;
        belowerr1 = bnexterr;
        cur1 += delta;
        bnexterr = cur2;
        delta = cur2 * 2;
        cur2 += delta;
        errorptr[2] = (FSERROR) (bpreverr2 + cur2);
        cur2 += delta;
        bpreverr2 = belowerr2 + cur2;
        belowerr2 = bnexterr;
        cur2 += delta;
        }
      inptr += dir3;
      outptr += dir;
      errorptr += dir3;
      }
    errorptr[0] = (FSERROR) bpreverr0;
    errorptr[1] = (FSERROR) bpreverr1;
    errorptr[2] = (FSERROR) bpreverr2;
    }
  }
static void
  init_error_limit(j_decompress_ptr cinfo)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  int *table;
  int in,
    out;
  table = (int *) (*cinfo->mem->alloc_small)
   ((j_common_ptr) cinfo, 1, (255 * 2 + 1) * ((size_t) sizeof(int)));
  table += 255;
  cquantize->error_limiter = table;
#define STEPSIZE ((MAXJSAMPLE+1)/16)
  out = 0;
  for (in = 0; in < ((255 + 1) / 16); in++, out++)
   {
    table[in] = out;
    table[-in] = -out;
    }
  for (; in < ((255 + 1) / 16) * 3; in++, out += (in & 1) ? 0 : 1)
   {
    table[in] = out;
    table[-in] = -out;
    }
  for (; in <= 255; in++)
   {
    table[in] = out;
    table[-in] = -out;
    }
#undef STEPSIZE
  }
static void
  finish_pass1(j_decompress_ptr cinfo)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  cinfo->colormap = cquantize->sv_colormap;
  select_colors(cinfo, cquantize->desired);
  cquantize->needs_zeroed = 1;
  }
static void
  finish_pass2(j_decompress_ptr cinfo)
 {
  }
static void
  start_pass_2_quant(j_decompress_ptr cinfo, boolean is_pre_scan)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  hist3d histogram = cquantize->histogram;
  int i;
  if (cinfo->dither_mode != JDITHER_NONE)
    cinfo->dither_mode = JDITHER_FS;
  if (is_pre_scan)
   {
    cquantize->pub.color_quantize = prescan_quantize;
    cquantize->pub.finish_pass = finish_pass1;
    cquantize->needs_zeroed = 1;
    }
  else
   {
    if (cinfo->dither_mode == JDITHER_FS)
      cquantize->pub.color_quantize = pass2_fs_dither;
    else
      cquantize->pub.color_quantize = pass2_no_dither;
    cquantize->pub.finish_pass = finish_pass2;
    i = cinfo->actual_number_of_colors;
    if (i < 1)
      ((cinfo)->err->msg_code = (JERR_QUANT_FEW_COLORS), (cinfo)->err->msg_parm.i[0] = (1), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (i > (255 + 1))
      ((cinfo)->err->msg_code = (JERR_QUANT_MANY_COLORS), (cinfo)->err->msg_parm.i[0] = ((255 + 1)), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (cinfo->dither_mode == JDITHER_FS)
     {
      size_t arraysize = (size_t) ((cinfo->output_width + 2) *
       (3 * ((size_t) sizeof(FSERROR))));
      if (cquantize->fserrors == ((void *) 0))
        cquantize->fserrors = (FSERRPTR) (*cinfo->mem->alloc_large)
         ((j_common_ptr) cinfo, 1, arraysize);
      jzero_far((void *) cquantize->fserrors, arraysize);
      if (cquantize->error_limiter == ((void *) 0))
        init_error_limit(cinfo);
      cquantize->on_odd_row = 0;
      }
    }
  if (cquantize->needs_zeroed)
   {
    for (i = 0; i < (1 << 5); i++)
     {
      jzero_far((void *) histogram[i],
       (1 << 6) * (1 << 5) * ((size_t) sizeof(histcell)));
      }
    cquantize->needs_zeroed = 0;
    }
  }
static void
  new_color_map_2_quant(j_decompress_ptr cinfo)
 {
  my_cquantizer_ptr1 cquantize = (my_cquantizer_ptr1) cinfo->cquantize;
  cquantize->needs_zeroed = 1;
  }
void
  jinit_2pass_quantizer(j_decompress_ptr cinfo)
 {
  my_cquantizer_ptr1 cquantize;
  int i;
  cquantize = (my_cquantizer_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_cquantizer1)));
  cinfo->cquantize = (struct jpeg_color_quantizer *) cquantize;
  cquantize->pub.start_pass = start_pass_2_quant;
  cquantize->pub.new_color_map = new_color_map_2_quant;
  cquantize->fserrors = ((void *) 0);
  cquantize->error_limiter = ((void *) 0);
  if (cinfo->out_color_components != 3)
    ((cinfo)->err->msg_code = (JERR_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  cquantize->histogram = (hist3d) (*cinfo->mem->alloc_small)
   ((j_common_ptr) cinfo, 1, (1 << 5) * ((size_t) sizeof(hist2d)));
  for (i = 0; i < (1 << 5); i++)
   {
    cquantize->histogram[i] = (hist2d) (*cinfo->mem->alloc_large)
     ((j_common_ptr) cinfo, 1,
     (1 << 6) * (1 << 5) * ((size_t) sizeof(histcell)));
    }
  cquantize->needs_zeroed = 1;
  if (cinfo->enable_2pass_quant)
   {
    int desired = cinfo->desired_number_of_colors;
    if (desired < 8)
      ((cinfo)->err->msg_code = (JERR_QUANT_FEW_COLORS), (cinfo)->err->msg_parm.i[0] = (8), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (desired > (255 + 1))
      ((cinfo)->err->msg_code = (JERR_QUANT_MANY_COLORS), (cinfo)->err->msg_parm.i[0] = ((255 + 1)), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    cquantize->sv_colormap = (*cinfo->mem->alloc_sarray)
     ((j_common_ptr) cinfo, 1, (JDIMENSION) desired, (JDIMENSION) 3);
    cquantize->desired = desired;
    }
  else
    cquantize->sv_colormap = ((void *) 0);
  if (cinfo->dither_mode != JDITHER_NONE)
    cinfo->dither_mode = JDITHER_FS;
  if (cinfo->dither_mode == JDITHER_FS)
   {
    cquantize->fserrors = (FSERRPTR) (*cinfo->mem->alloc_large)
     ((j_common_ptr) cinfo, 1,
     (size_t) ((cinfo->output_width + 2) * (3 * ((size_t) sizeof(FSERROR)))));
    init_error_limit(cinfo);
    }
  }
#endif
#define JPEG_INTERNALS
#ifdef IDCT_SCALING_SUPPORTED
#if 8  != 8
#endif
#if 8  == 8
#undef CONST_BITS
#define CONST_BITS  13
#define PASS1_BITS  2
#else
#endif
#if 13  == 13
#define FIX_0_211164243  1730
#define FIX_0_509795579  4176
#define FIX_0_601344887  4926
#define FIX_0_720959822  5906
#define FIX_0_765366865  6270
#define FIX_0_850430095  6967
#define FIX_0_899976223  7373
#define FIX_1_061594337  8697
#define FIX_1_272758580  10426
#define FIX_1_451774981  11893
#define FIX_1_847759065  15137
#define FIX_2_172734803  17799
#define FIX_2_562915447  20995
#define FIX_3_624509785  29692
#else
#endif
#if 8  == 8
#define MULTIPLY(var,const)  MULTIPLY16C16(var,const)
#else
#endif
#define DEQUANTIZE(coef,quantval)  (((ISLOW_MULT_TYPE) (coef)) * (quantval))
void
  jpeg_idct_4x4(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JCOEFPTR coef_block,
   JSAMPARRAY output_buf, JDIMENSION output_col)
 {
  int tmp0,
    tmp2,
    tmp10,
    tmp12;
  int z1,
    z2,
    z3,
    z4;
  JCOEFPTR inptr;
  ISLOW_MULT_TYPE *quantptr;
  int *wsptr;
  JSAMPROW outptr;
  JSAMPLE *range_limit = ((cinfo)->sample_range_limit + 128);
  int ctr;
  int workspace[8 * 4];
  inptr = coef_block;
  quantptr = (ISLOW_MULT_TYPE *) compptr->dct_table;
  wsptr = workspace;
  for (ctr = 8; ctr > 0; inptr++, quantptr++, wsptr++, ctr--)
   {
    if (ctr == 8 - 4)
      continue;
    if (inptr[8 * 1] == 0 && inptr[8 * 2] == 0 &&
     inptr[8 * 3] == 0 && inptr[8 * 5] == 0 &&
     inptr[8 * 6] == 0 && inptr[8 * 7] == 0)
     {
      int dcval = (((ISLOW_MULT_TYPE) (inptr[8 * 0])) * (quantptr[8 * 0])) << 2;
      wsptr[8 * 0] = dcval;
      wsptr[8 * 1] = dcval;
      wsptr[8 * 2] = dcval;
      wsptr[8 * 3] = dcval;
      continue;
      }
    tmp0 = (((ISLOW_MULT_TYPE) (inptr[8 * 0])) * (quantptr[8 * 0]));
    tmp0 <<= (13 + 1);
    z2 = (((ISLOW_MULT_TYPE) (inptr[8 * 2])) * (quantptr[8 * 2]));
    z3 = (((ISLOW_MULT_TYPE) (inptr[8 * 6])) * (quantptr[8 * 6]));
    tmp2 = ((z2) * (((int) 15137))) + ((z3) * (-((int) 6270)));
    tmp10 = tmp0 + tmp2;
    tmp12 = tmp0 - tmp2;
    z1 = (((ISLOW_MULT_TYPE) (inptr[8 * 7])) * (quantptr[8 * 7]));
    z2 = (((ISLOW_MULT_TYPE) (inptr[8 * 5])) * (quantptr[8 * 5]));
    z3 = (((ISLOW_MULT_TYPE) (inptr[8 * 3])) * (quantptr[8 * 3]));
    z4 = (((ISLOW_MULT_TYPE) (inptr[8 * 1])) * (quantptr[8 * 1]));
    tmp0 = ((z1) * (-((int) 1730)))
     + ((z2) * (((int) 11893)))
     + ((z3) * (-((int) 17799)))
     + ((z4) * (((int) 8697)));
    tmp2 = ((z1) * (-((int) 4176)))
     + ((z2) * (-((int) 4926)))
     + ((z3) * (((int) 7373)))
     + ((z4) * (((int) 20995)));
    wsptr[8 * 0] = (int) (((tmp10 + tmp2) + (((int) 1) << ((13 - 2 + 1) - 1))) >> (13 - 2 + 1));
    wsptr[8 * 3] = (int) (((tmp10 - tmp2) + (((int) 1) << ((13 - 2 + 1) - 1))) >> (13 - 2 + 1));
    wsptr[8 * 1] = (int) (((tmp12 + tmp0) + (((int) 1) << ((13 - 2 + 1) - 1))) >> (13 - 2 + 1));
    wsptr[8 * 2] = (int) (((tmp12 - tmp0) + (((int) 1) << ((13 - 2 + 1) - 1))) >> (13 - 2 + 1));
    }
  wsptr = workspace;
  for (ctr = 0; ctr < 4; ctr++)
   {
    outptr = output_buf[ctr] + output_col;
#ifndef NO_ZERO_ROW_TEST
    if (wsptr[1] == 0 && wsptr[2] == 0 && wsptr[3] == 0 &&
     wsptr[5] == 0 && wsptr[6] == 0 && wsptr[7] == 0)
     {
      JSAMPLE dcval = range_limit[(int) ((((int) wsptr[0]) + (((int) 1) << ((2 + 3) - 1))) >> (2 + 3))
       & (255 * 4 + 3)];
      outptr[0] = dcval;
      outptr[1] = dcval;
      outptr[2] = dcval;
      outptr[3] = dcval;
      wsptr += 8;
      continue;
      }
#endif
    tmp0 = ((int) wsptr[0]) << (13 + 1);
    tmp2 = (((int) wsptr[2]) * (((int) 15137)))
     + (((int) wsptr[6]) * (-((int) 6270)));
    tmp10 = tmp0 + tmp2;
    tmp12 = tmp0 - tmp2;
    z1 = (int) wsptr[7];
    z2 = (int) wsptr[5];
    z3 = (int) wsptr[3];
    z4 = (int) wsptr[1];
    tmp0 = ((z1) * (-((int) 1730)))
     + ((z2) * (((int) 11893)))
     + ((z3) * (-((int) 17799)))
     + ((z4) * (((int) 8697)));
    tmp2 = ((z1) * (-((int) 4176)))
     + ((z2) * (-((int) 4926)))
     + ((z3) * (((int) 7373)))
     + ((z4) * (((int) 20995)));
    outptr[0] = range_limit[(int) (((tmp10 + tmp2) + (((int) 1) << ((13 + 2 + 3 + 1) - 1))) >> (13 + 2 + 3 + 1))
     & (255 * 4 + 3)];
    outptr[3] = range_limit[(int) (((tmp10 - tmp2) + (((int) 1) << ((13 + 2 + 3 + 1) - 1))) >> (13 + 2 + 3 + 1))
     & (255 * 4 + 3)];
    outptr[1] = range_limit[(int) (((tmp12 + tmp0) + (((int) 1) << ((13 + 2 + 3 + 1) - 1))) >> (13 + 2 + 3 + 1))
     & (255 * 4 + 3)];
    outptr[2] = range_limit[(int) (((tmp12 - tmp0) + (((int) 1) << ((13 + 2 + 3 + 1) - 1))) >> (13 + 2 + 3 + 1))
     & (255 * 4 + 3)];
    wsptr += 8;
    }
  }
void
  jpeg_idct_2x2(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JCOEFPTR coef_block,
   JSAMPARRAY output_buf, JDIMENSION output_col)
 {
  int tmp0,
    tmp10,
    z1;
  JCOEFPTR inptr;
  ISLOW_MULT_TYPE *quantptr;
  int *wsptr;
  JSAMPROW outptr;
  JSAMPLE *range_limit = ((cinfo)->sample_range_limit + 128);
  int ctr;
  int workspace[8 * 2];
  inptr = coef_block;
  quantptr = (ISLOW_MULT_TYPE *) compptr->dct_table;
  wsptr = workspace;
  for (ctr = 8; ctr > 0; inptr++, quantptr++, wsptr++, ctr--)
   {
    if (ctr == 8 - 2 || ctr == 8 - 4 || ctr == 8 - 6)
      continue;
    if (inptr[8 * 1] == 0 && inptr[8 * 3] == 0 &&
     inptr[8 * 5] == 0 && inptr[8 * 7] == 0)
     {
      int dcval = (((ISLOW_MULT_TYPE) (inptr[8 * 0])) * (quantptr[8 * 0])) << 2;
      wsptr[8 * 0] = dcval;
      wsptr[8 * 1] = dcval;
      continue;
      }
    z1 = (((ISLOW_MULT_TYPE) (inptr[8 * 0])) * (quantptr[8 * 0]));
    tmp10 = z1 << (13 + 2);
    z1 = (((ISLOW_MULT_TYPE) (inptr[8 * 7])) * (quantptr[8 * 7]));
    tmp0 = ((z1) * (-((int) 5906)));
    z1 = (((ISLOW_MULT_TYPE) (inptr[8 * 5])) * (quantptr[8 * 5]));
    tmp0 += ((z1) * (((int) 6967)));
    z1 = (((ISLOW_MULT_TYPE) (inptr[8 * 3])) * (quantptr[8 * 3]));
    tmp0 += ((z1) * (-((int) 10426)));
    z1 = (((ISLOW_MULT_TYPE) (inptr[8 * 1])) * (quantptr[8 * 1]));
    tmp0 += ((z1) * (((int) 29692)));
    wsptr[8 * 0] = (int) (((tmp10 + tmp0) + (((int) 1) << ((13 - 2 + 2) - 1))) >> (13 - 2 + 2));
    wsptr[8 * 1] = (int) (((tmp10 - tmp0) + (((int) 1) << ((13 - 2 + 2) - 1))) >> (13 - 2 + 2));
    }
  wsptr = workspace;
  for (ctr = 0; ctr < 2; ctr++)
   {
    outptr = output_buf[ctr] + output_col;
#ifndef NO_ZERO_ROW_TEST
    if (wsptr[1] == 0 && wsptr[3] == 0 && wsptr[5] == 0 && wsptr[7] == 0)
     {
      JSAMPLE dcval = range_limit[(int) ((((int) wsptr[0]) + (((int) 1) << ((2 + 3) - 1))) >> (2 + 3))
       & (255 * 4 + 3)];
      outptr[0] = dcval;
      outptr[1] = dcval;
      wsptr += 8;
      continue;
      }
#endif
    tmp10 = ((int) wsptr[0]) << (13 + 2);
    tmp0 = (((int) wsptr[7]) * (-((int) 5906)))
     + (((int) wsptr[5]) * (((int) 6967)))
     + (((int) wsptr[3]) * (-((int) 10426)))
     + (((int) wsptr[1]) * (((int) 29692)));
    outptr[0] = range_limit[(int) (((tmp10 + tmp0) + (((int) 1) << ((13 + 2 + 3 + 2) - 1))) >> (13 + 2 + 3 + 2))
     & (255 * 4 + 3)];
    outptr[1] = range_limit[(int) (((tmp10 - tmp0) + (((int) 1) << ((13 + 2 + 3 + 2) - 1))) >> (13 + 2 + 3 + 2))
     & (255 * 4 + 3)];
    wsptr += 8;
    }
  }
void
  jpeg_idct_1x1(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JCOEFPTR coef_block,
   JSAMPARRAY output_buf, JDIMENSION output_col)
 {
  int dcval;
  ISLOW_MULT_TYPE *quantptr;
  JSAMPLE *range_limit = ((cinfo)->sample_range_limit + 128);
  quantptr = (ISLOW_MULT_TYPE *) compptr->dct_table;
  dcval = (((ISLOW_MULT_TYPE) (coef_block[0])) * (quantptr[0]));
  dcval = (int) ((((int) dcval) + (((int) 1) << ((3) - 1))) >> (3));
  output_buf[0][output_col] = range_limit[dcval & (255 * 4 + 3)];
  }
#endif
#define JPEG_INTERNALS
#ifdef DCT_FLOAT_SUPPORTED
#if 8  != 8
#endif
#undef DEQUANTIZE
#define DEQUANTIZE(coef,quantval)  (((FAST_FLOAT) (coef)) * (quantval))
void
  jpeg_idct_float(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JCOEFPTR coef_block,
   JSAMPARRAY output_buf, JDIMENSION output_col)
 {
  float tmp0,
    tmp1,
    tmp2,
    tmp3,
    tmp4,
    tmp5,
    tmp6,
    tmp7;
  float tmp10,
    tmp11,
    tmp12,
    tmp13;
  float z5,
    z10,
    z11,
    z12,
    z13;
  JCOEFPTR inptr;
  FLOAT_MULT_TYPE *quantptr;
  float *wsptr;
  JSAMPROW outptr;
  JSAMPLE *range_limit = ((cinfo)->sample_range_limit + 128);
  int ctr;
  float workspace[64];
  inptr = coef_block;
  quantptr = (FLOAT_MULT_TYPE *) compptr->dct_table;
  wsptr = workspace;
  for (ctr = 8; ctr > 0; ctr--)
   {
    if (inptr[8 * 1] == 0 && inptr[8 * 2] == 0 &&
     inptr[8 * 3] == 0 && inptr[8 * 4] == 0 &&
     inptr[8 * 5] == 0 && inptr[8 * 6] == 0 &&
     inptr[8 * 7] == 0)
     {
      float dcval = (((float) (inptr[8 * 0])) * (quantptr[8 * 0]));
      wsptr[8 * 0] = dcval;
      wsptr[8 * 1] = dcval;
      wsptr[8 * 2] = dcval;
      wsptr[8 * 3] = dcval;
      wsptr[8 * 4] = dcval;
      wsptr[8 * 5] = dcval;
      wsptr[8 * 6] = dcval;
      wsptr[8 * 7] = dcval;
      inptr++;
      quantptr++;
      wsptr++;
      continue;
      }
    tmp0 = (((float) (inptr[8 * 0])) * (quantptr[8 * 0]));
    tmp1 = (((float) (inptr[8 * 2])) * (quantptr[8 * 2]));
    tmp2 = (((float) (inptr[8 * 4])) * (quantptr[8 * 4]));
    tmp3 = (((float) (inptr[8 * 6])) * (quantptr[8 * 6]));
    tmp10 = tmp0 + tmp2;
    tmp11 = tmp0 - tmp2;
    tmp13 = tmp1 + tmp3;
    tmp12 = (tmp1 - tmp3) * ((float) 1.414213562) - tmp13;
    tmp0 = tmp10 + tmp13;
    tmp3 = tmp10 - tmp13;
    tmp1 = tmp11 + tmp12;
    tmp2 = tmp11 - tmp12;
    tmp4 = (((float) (inptr[8 * 1])) * (quantptr[8 * 1]));
    tmp5 = (((float) (inptr[8 * 3])) * (quantptr[8 * 3]));
    tmp6 = (((float) (inptr[8 * 5])) * (quantptr[8 * 5]));
    tmp7 = (((float) (inptr[8 * 7])) * (quantptr[8 * 7]));
    z13 = tmp6 + tmp5;
    z10 = tmp6 - tmp5;
    z11 = tmp4 + tmp7;
    z12 = tmp4 - tmp7;
    tmp7 = z11 + z13;
    tmp11 = (z11 - z13) * ((float) 1.414213562);
    z5 = (z10 + z12) * ((float) 1.847759065);
    tmp10 = ((float) 1.082392200) * z12 - z5;
    tmp12 = ((float) -2.613125930) * z10 + z5;
    tmp6 = tmp12 - tmp7;
    tmp5 = tmp11 - tmp6;
    tmp4 = tmp10 + tmp5;
    wsptr[8 * 0] = tmp0 + tmp7;
    wsptr[8 * 7] = tmp0 - tmp7;
    wsptr[8 * 1] = tmp1 + tmp6;
    wsptr[8 * 6] = tmp1 - tmp6;
    wsptr[8 * 2] = tmp2 + tmp5;
    wsptr[8 * 5] = tmp2 - tmp5;
    wsptr[8 * 4] = tmp3 + tmp4;
    wsptr[8 * 3] = tmp3 - tmp4;
    inptr++;
    quantptr++;
    wsptr++;
    }
  wsptr = workspace;
  for (ctr = 0; ctr < 8; ctr++)
   {
    outptr = output_buf[ctr] + output_col;
    tmp10 = wsptr[0] + wsptr[4];
    tmp11 = wsptr[0] - wsptr[4];
    tmp13 = wsptr[2] + wsptr[6];
    tmp12 = (wsptr[2] - wsptr[6]) * ((float) 1.414213562) - tmp13;
    tmp0 = tmp10 + tmp13;
    tmp3 = tmp10 - tmp13;
    tmp1 = tmp11 + tmp12;
    tmp2 = tmp11 - tmp12;
    z13 = wsptr[5] + wsptr[3];
    z10 = wsptr[5] - wsptr[3];
    z11 = wsptr[1] + wsptr[7];
    z12 = wsptr[1] - wsptr[7];
    tmp7 = z11 + z13;
    tmp11 = (z11 - z13) * ((float) 1.414213562);
    z5 = (z10 + z12) * ((float) 1.847759065);
    tmp10 = ((float) 1.082392200) * z12 - z5;
    tmp12 = ((float) -2.613125930) * z10 + z5;
    tmp6 = tmp12 - tmp7;
    tmp5 = tmp11 - tmp6;
    tmp4 = tmp10 + tmp5;
    outptr[0] = range_limit[(int) ((((int) (tmp0 + tmp7)) + (((int) 1) << ((3) - 1))) >> (3))
     & (255 * 4 + 3)];
    outptr[7] = range_limit[(int) ((((int) (tmp0 - tmp7)) + (((int) 1) << ((3) - 1))) >> (3))
     & (255 * 4 + 3)];
    outptr[1] = range_limit[(int) ((((int) (tmp1 + tmp6)) + (((int) 1) << ((3) - 1))) >> (3))
     & (255 * 4 + 3)];
    outptr[6] = range_limit[(int) ((((int) (tmp1 - tmp6)) + (((int) 1) << ((3) - 1))) >> (3))
     & (255 * 4 + 3)];
    outptr[2] = range_limit[(int) ((((int) (tmp2 + tmp5)) + (((int) 1) << ((3) - 1))) >> (3))
     & (255 * 4 + 3)];
    outptr[5] = range_limit[(int) ((((int) (tmp2 - tmp5)) + (((int) 1) << ((3) - 1))) >> (3))
     & (255 * 4 + 3)];
    outptr[4] = range_limit[(int) ((((int) (tmp3 + tmp4)) + (((int) 1) << ((3) - 1))) >> (3))
     & (255 * 4 + 3)];
    outptr[3] = range_limit[(int) ((((int) (tmp3 - tmp4)) + (((int) 1) << ((3) - 1))) >> (3))
     & (255 * 4 + 3)];
    wsptr += 8;
    }
  }
#endif
#define JPEG_INTERNALS
#ifdef DCT_IFAST_SUPPORTED
#if 8  != 8
#endif
#if 8  == 8
#undef CONST_BITS
#define CONST_BITS  8
#define PASS1_BITS  2
#else
#endif
#if 8  == 8
#define FIX_1_082392200 277
#define FIX_1_414213562 362
#undef FIX_1_847759065
#define FIX_1_847759065 473
#define FIX_2_613125930 669
#else
#endif
#ifndef USE_ACCURATE_ROUNDING
#undef DESCALE
#define DESCALE(x,n)  RIGHT_SHIFT(x, n)
#endif
#undef MULTIPLY
#define MULTIPLY(var,const)  ((DCTELEM) DESCALE((var) * (const), CONST_BITS))
#if 8  == 8
#undef DEQUANTIZE
#define DEQUANTIZE(coef,quantval)  (((IFAST_MULT_TYPE) (coef)) * (quantval))
#else
#endif
#ifdef RIGHT_SHIFT_IS_UNSIGNED
#else
#define ISHIFT_TEMPS
#define IRIGHT_SHIFT(x,shft) ((x) >> (shft))
#endif
#ifdef USE_ACCURATE_ROUNDING
#else
#define IDESCALE(x,n)  ((int) IRIGHT_SHIFT(x, n))
#endif
void
  jpeg_idct_ifast(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JCOEFPTR coef_block,
   JSAMPARRAY output_buf, JDIMENSION output_col)
 {
  DCTELEM tmp0,
    tmp1,
    tmp2,
    tmp3,
    tmp4,
    tmp5,
    tmp6,
    tmp7;
  DCTELEM tmp10,
    tmp11,
    tmp12,
    tmp13;
  DCTELEM z5,
    z10,
    z11,
    z12,
    z13;
  JCOEFPTR inptr;
  IFAST_MULT_TYPE *quantptr;
  int *wsptr;
  JSAMPROW outptr;
  JSAMPLE *range_limit = ((cinfo)->sample_range_limit + 128);
  int ctr;
  int workspace[64];
  inptr = coef_block;
  quantptr = (IFAST_MULT_TYPE *) compptr->dct_table;
  wsptr = workspace;
  for (ctr = 8; ctr > 0; ctr--)
   {
    if (inptr[8 * 1] == 0 && inptr[8 * 2] == 0 &&
     inptr[8 * 3] == 0 && inptr[8 * 4] == 0 &&
     inptr[8 * 5] == 0 && inptr[8 * 6] == 0 &&
     inptr[8 * 7] == 0)
     {
      int dcval = (int) (((IFAST_MULT_TYPE) (inptr[8 * 0])) * (quantptr[8 * 0]));
      wsptr[8 * 0] = dcval;
      wsptr[8 * 1] = dcval;
      wsptr[8 * 2] = dcval;
      wsptr[8 * 3] = dcval;
      wsptr[8 * 4] = dcval;
      wsptr[8 * 5] = dcval;
      wsptr[8 * 6] = dcval;
      wsptr[8 * 7] = dcval;
      inptr++;
      quantptr++;
      wsptr++;
      continue;
      }
    tmp0 = (((IFAST_MULT_TYPE) (inptr[8 * 0])) * (quantptr[8 * 0]));
    tmp1 = (((IFAST_MULT_TYPE) (inptr[8 * 2])) * (quantptr[8 * 2]));
    tmp2 = (((IFAST_MULT_TYPE) (inptr[8 * 4])) * (quantptr[8 * 4]));
    tmp3 = (((IFAST_MULT_TYPE) (inptr[8 * 6])) * (quantptr[8 * 6]));
    tmp10 = tmp0 + tmp2;
    tmp11 = tmp0 - tmp2;
    tmp13 = tmp1 + tmp3;
    tmp12 = ((DCTELEM) (((tmp1 - tmp3) * (((int) 362))) >> (8))) - tmp13;
    tmp0 = tmp10 + tmp13;
    tmp3 = tmp10 - tmp13;
    tmp1 = tmp11 + tmp12;
    tmp2 = tmp11 - tmp12;
    tmp4 = (((IFAST_MULT_TYPE) (inptr[8 * 1])) * (quantptr[8 * 1]));
    tmp5 = (((IFAST_MULT_TYPE) (inptr[8 * 3])) * (quantptr[8 * 3]));
    tmp6 = (((IFAST_MULT_TYPE) (inptr[8 * 5])) * (quantptr[8 * 5]));
    tmp7 = (((IFAST_MULT_TYPE) (inptr[8 * 7])) * (quantptr[8 * 7]));
    z13 = tmp6 + tmp5;
    z10 = tmp6 - tmp5;
    z11 = tmp4 + tmp7;
    z12 = tmp4 - tmp7;
    tmp7 = z11 + z13;
    tmp11 = ((DCTELEM) (((z11 - z13) * (((int) 362))) >> (8)));
    z5 = ((DCTELEM) (((z10 + z12) * (((int) 473))) >> (8)));
    tmp10 = ((DCTELEM) (((z12) * (((int) 277))) >> (8))) - z5;
    tmp12 = ((DCTELEM) (((z10) * (-((int) 669))) >> (8))) + z5;
    tmp6 = tmp12 - tmp7;
    tmp5 = tmp11 - tmp6;
    tmp4 = tmp10 + tmp5;
    wsptr[8 * 0] = (int) (tmp0 + tmp7);
    wsptr[8 * 7] = (int) (tmp0 - tmp7);
    wsptr[8 * 1] = (int) (tmp1 + tmp6);
    wsptr[8 * 6] = (int) (tmp1 - tmp6);
    wsptr[8 * 2] = (int) (tmp2 + tmp5);
    wsptr[8 * 5] = (int) (tmp2 - tmp5);
    wsptr[8 * 4] = (int) (tmp3 + tmp4);
    wsptr[8 * 3] = (int) (tmp3 - tmp4);
    inptr++;
    quantptr++;
    wsptr++;
    }
  wsptr = workspace;
  for (ctr = 0; ctr < 8; ctr++)
   {
    outptr = output_buf[ctr] + output_col;
#ifndef NO_ZERO_ROW_TEST
    if (wsptr[1] == 0 && wsptr[2] == 0 && wsptr[3] == 0 && wsptr[4] == 0 &&
     wsptr[5] == 0 && wsptr[6] == 0 && wsptr[7] == 0)
     {
      JSAMPLE dcval = range_limit[((int) ((wsptr[0]) >> (2 + 3)))
       & (255 * 4 + 3)];
      outptr[0] = dcval;
      outptr[1] = dcval;
      outptr[2] = dcval;
      outptr[3] = dcval;
      outptr[4] = dcval;
      outptr[5] = dcval;
      outptr[6] = dcval;
      outptr[7] = dcval;
      wsptr += 8;
      continue;
      }
#endif
    tmp10 = ((DCTELEM) wsptr[0] + (DCTELEM) wsptr[4]);
    tmp11 = ((DCTELEM) wsptr[0] - (DCTELEM) wsptr[4]);
    tmp13 = ((DCTELEM) wsptr[2] + (DCTELEM) wsptr[6]);
    tmp12 = ((DCTELEM) ((((DCTELEM) wsptr[2] - (DCTELEM) wsptr[6]) * (((int) 362))) >> (8)))
     - tmp13;
    tmp0 = tmp10 + tmp13;
    tmp3 = tmp10 - tmp13;
    tmp1 = tmp11 + tmp12;
    tmp2 = tmp11 - tmp12;
    z13 = (DCTELEM) wsptr[5] + (DCTELEM) wsptr[3];
    z10 = (DCTELEM) wsptr[5] - (DCTELEM) wsptr[3];
    z11 = (DCTELEM) wsptr[1] + (DCTELEM) wsptr[7];
    z12 = (DCTELEM) wsptr[1] - (DCTELEM) wsptr[7];
    tmp7 = z11 + z13;
    tmp11 = ((DCTELEM) (((z11 - z13) * (((int) 362))) >> (8)));
    z5 = ((DCTELEM) (((z10 + z12) * (((int) 473))) >> (8)));
    tmp10 = ((DCTELEM) (((z12) * (((int) 277))) >> (8))) - z5;
    tmp12 = ((DCTELEM) (((z10) * (-((int) 669))) >> (8))) + z5;
    tmp6 = tmp12 - tmp7;
    tmp5 = tmp11 - tmp6;
    tmp4 = tmp10 + tmp5;
    outptr[0] = range_limit[((int) ((tmp0 + tmp7) >> (2 + 3)))
     & (255 * 4 + 3)];
    outptr[7] = range_limit[((int) ((tmp0 - tmp7) >> (2 + 3)))
     & (255 * 4 + 3)];
    outptr[1] = range_limit[((int) ((tmp1 + tmp6) >> (2 + 3)))
     & (255 * 4 + 3)];
    outptr[6] = range_limit[((int) ((tmp1 - tmp6) >> (2 + 3)))
     & (255 * 4 + 3)];
    outptr[2] = range_limit[((int) ((tmp2 + tmp5) >> (2 + 3)))
     & (255 * 4 + 3)];
    outptr[5] = range_limit[((int) ((tmp2 - tmp5) >> (2 + 3)))
     & (255 * 4 + 3)];
    outptr[4] = range_limit[((int) ((tmp3 + tmp4) >> (2 + 3)))
     & (255 * 4 + 3)];
    outptr[3] = range_limit[((int) ((tmp3 - tmp4) >> (2 + 3)))
     & (255 * 4 + 3)];
    wsptr += 8;
    }
  }
#endif
#define JPEG_INTERNALS
#ifdef DCT_ISLOW_SUPPORTED
#if 8  != 8
#endif
#if 8  == 8
#undef CONST_BITS
#define CONST_BITS  13
#define PASS1_BITS  2
#else
#endif
#if 13  == 13
#undef FIX_0_298631336
#undef FIX_0_390180644
#undef FIX_0_541196100
#undef FIX_0_765366865
#undef FIX_0_899976223
#undef FIX_1_175875602
#undef FIX_1_501321110
#undef FIX_1_847759065
#undef FIX_1_961570560
#undef FIX_2_053119869
#undef FIX_2_562915447
#undef FIX_3_072711026

#define FIX_0_298631336 2446
#define FIX_0_390180644 3196
#define FIX_0_541196100 4433
#define FIX_0_765366865 6270
#define FIX_0_899976223 7373
#define FIX_1_175875602 9633
#define FIX_1_501321110 12299
#define FIX_1_847759065 15137
#define FIX_1_961570560 16069
#define FIX_2_053119869 16819
#define FIX_2_562915447 20995
#define FIX_3_072711026 25172
#else
#endif
#if 8  == 8
#undef MULTIPLY
#define MULTIPLY(var,const)  MULTIPLY16C16(var,const)
#else
#endif
#undef DEQUANTIZE
#define DEQUANTIZE(coef,quantval)  (((ISLOW_MULT_TYPE) (coef)) * (quantval))
void
  jpeg_idct_islow(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JCOEFPTR coef_block,
   JSAMPARRAY output_buf, JDIMENSION output_col)
 {
  int tmp0,
    tmp1,
    tmp2,
    tmp3;
  int tmp10,
    tmp11,
    tmp12,
    tmp13;
  int z1,
    z2,
    z3,
    z4,
    z5;
  JCOEFPTR inptr;
  ISLOW_MULT_TYPE *quantptr;
  int *wsptr;
  JSAMPROW outptr;
  JSAMPLE *range_limit = ((cinfo)->sample_range_limit + 128);
  int ctr;
  int workspace[64];
  inptr = coef_block;
  quantptr = (ISLOW_MULT_TYPE *) compptr->dct_table;
  wsptr = workspace;
  for (ctr = 8; ctr > 0; ctr--)
   {
    if (inptr[8 * 1] == 0 && inptr[8 * 2] == 0 &&
     inptr[8 * 3] == 0 && inptr[8 * 4] == 0 &&
     inptr[8 * 5] == 0 && inptr[8 * 6] == 0 &&
     inptr[8 * 7] == 0)
     {
      int dcval = (((ISLOW_MULT_TYPE) (inptr[8 * 0])) * (quantptr[8 * 0])) << 2;
      wsptr[8 * 0] = dcval;
      wsptr[8 * 1] = dcval;
      wsptr[8 * 2] = dcval;
      wsptr[8 * 3] = dcval;
      wsptr[8 * 4] = dcval;
      wsptr[8 * 5] = dcval;
      wsptr[8 * 6] = dcval;
      wsptr[8 * 7] = dcval;
      inptr++;
      quantptr++;
      wsptr++;
      continue;
      }
    z2 = (((ISLOW_MULT_TYPE) (inptr[8 * 2])) * (quantptr[8 * 2]));
    z3 = (((ISLOW_MULT_TYPE) (inptr[8 * 6])) * (quantptr[8 * 6]));
    z1 = ((z2 + z3) * (((int) 4433)));
    tmp2 = z1 + ((z3) * (-((int) 15137)));
    tmp3 = z1 + ((z2) * (((int) 6270)));
    z2 = (((ISLOW_MULT_TYPE) (inptr[8 * 0])) * (quantptr[8 * 0]));
    z3 = (((ISLOW_MULT_TYPE) (inptr[8 * 4])) * (quantptr[8 * 4]));
    tmp0 = (z2 + z3) << 13;
    tmp1 = (z2 - z3) << 13;
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    tmp0 = (((ISLOW_MULT_TYPE) (inptr[8 * 7])) * (quantptr[8 * 7]));
    tmp1 = (((ISLOW_MULT_TYPE) (inptr[8 * 5])) * (quantptr[8 * 5]));
    tmp2 = (((ISLOW_MULT_TYPE) (inptr[8 * 3])) * (quantptr[8 * 3]));
    tmp3 = (((ISLOW_MULT_TYPE) (inptr[8 * 1])) * (quantptr[8 * 1]));
    z1 = tmp0 + tmp3;
    z2 = tmp1 + tmp2;
    z3 = tmp0 + tmp2;
    z4 = tmp1 + tmp3;
    z5 = ((z3 + z4) * (((int) 9633)));
    tmp0 = ((tmp0) * (((int) 2446)));
    tmp1 = ((tmp1) * (((int) 16819)));
    tmp2 = ((tmp2) * (((int) 25172)));
    tmp3 = ((tmp3) * (((int) 12299)));
    z1 = ((z1) * (-((int) 7373)));
    z2 = ((z2) * (-((int) 20995)));
    z3 = ((z3) * (-((int) 16069)));
    z4 = ((z4) * (-((int) 3196)));
    z3 += z5;
    z4 += z5;
    tmp0 += z1 + z3;
    tmp1 += z2 + z4;
    tmp2 += z2 + z3;
    tmp3 += z1 + z4;
    wsptr[8 * 0] = (int) ((tmp10 + tmp3) >> (13 - 2));
    wsptr[8 * 7] = (int) ((tmp10 - tmp3) >> (13 - 2));
    wsptr[8 * 1] = (int) ((tmp11 + tmp2) >> (13 - 2));
    wsptr[8 * 6] = (int) ((tmp11 - tmp2) >> (13 - 2));
    wsptr[8 * 2] = (int) ((tmp12 + tmp1) >> (13 - 2));
    wsptr[8 * 5] = (int) ((tmp12 - tmp1) >> (13 - 2));
    wsptr[8 * 3] = (int) ((tmp13 + tmp0) >> (13 - 2));
    wsptr[8 * 4] = (int) ((tmp13 - tmp0) >> (13 - 2));
    inptr++;
    quantptr++;
    wsptr++;
    }
  wsptr = workspace;
  for (ctr = 0; ctr < 8; ctr++)
   {
    outptr = output_buf[ctr] + output_col;
#ifndef NO_ZERO_ROW_TEST
    if (wsptr[1] == 0 && wsptr[2] == 0 && wsptr[3] == 0 && wsptr[4] == 0 &&
     wsptr[5] == 0 && wsptr[6] == 0 && wsptr[7] == 0)
     {
      JSAMPLE dcval = range_limit[(int) (((int) wsptr[0]) >> (2 + 3))
       & (255 * 4 + 3)];
      outptr[0] = dcval;
      outptr[1] = dcval;
      outptr[2] = dcval;
      outptr[3] = dcval;
      outptr[4] = dcval;
      outptr[5] = dcval;
      outptr[6] = dcval;
      outptr[7] = dcval;
      wsptr += 8;
      continue;
      }
#endif
    z2 = (int) wsptr[2];
    z3 = (int) wsptr[6];
    z1 = ((z2 + z3) * (((int) 4433)));
    tmp2 = z1 + ((z3) * (-((int) 15137)));
    tmp3 = z1 + ((z2) * (((int) 6270)));
    tmp0 = ((int) wsptr[0] + (int) wsptr[4]) << 13;
    tmp1 = ((int) wsptr[0] - (int) wsptr[4]) << 13;
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    tmp0 = (int) wsptr[7];
    tmp1 = (int) wsptr[5];
    tmp2 = (int) wsptr[3];
    tmp3 = (int) wsptr[1];
    z1 = tmp0 + tmp3;
    z2 = tmp1 + tmp2;
    z3 = tmp0 + tmp2;
    z4 = tmp1 + tmp3;
    z5 = ((z3 + z4) * (((int) 9633)));
    tmp0 = ((tmp0) * (((int) 2446)));
    tmp1 = ((tmp1) * (((int) 16819)));
    tmp2 = ((tmp2) * (((int) 25172)));
    tmp3 = ((tmp3) * (((int) 12299)));
    z1 = ((z1) * (-((int) 7373)));
    z2 = ((z2) * (-((int) 20995)));
    z3 = ((z3) * (-((int) 16069)));
    z4 = ((z4) * (-((int) 3196)));
    z3 += z5;
    z4 += z5;
    tmp0 += z1 + z3;
    tmp1 += z2 + z4;
    tmp2 += z2 + z3;
    tmp3 += z1 + z4;
    outptr[0] = range_limit[(int) ((tmp10 + tmp3) >> (13 + 2 + 3))
     & (255 * 4 + 3)];
    outptr[7] = range_limit[(int) ((tmp10 - tmp3) >> (13 + 2 + 3))
     & (255 * 4 + 3)];
    outptr[1] = range_limit[(int) ((tmp11 + tmp2) >> (13 + 2 + 3))
     & (255 * 4 + 3)];
    outptr[6] = range_limit[(int) ((tmp11 - tmp2) >> (13 + 2 + 3))
     & (255 * 4 + 3)];
    outptr[2] = range_limit[(int) ((tmp12 + tmp1) >> (13 + 2 + 3))
     & (255 * 4 + 3)];
    outptr[5] = range_limit[(int) ((tmp12 - tmp1) >> (13 + 2 + 3))
     & (255 * 4 + 3)];
    outptr[3] = range_limit[(int) ((tmp13 + tmp0) >> (13 + 2 + 3))
     & (255 * 4 + 3)];
    outptr[4] = range_limit[(int) ((tmp13 - tmp0) >> (13 + 2 + 3))
     & (255 * 4 + 3)];
    wsptr += 8;
    }
  }
#endif
#define JPEG_INTERNALS
#ifdef UPSAMPLE_MERGING_SUPPORTED
typedef struct
 {
  struct jpeg_upsampler pub;
  void (*upmethod) (j_decompress_ptr cinfo, JSAMPIMAGE input_buf, JDIMENSION in_row_group_ctr, JSAMPARRAY output_buf);
  int *Cr_r_tab;
  int *Cb_b_tab;
  int *Cr_g_tab;
  int *Cb_g_tab;
  JSAMPROW spare_row;
  boolean spare_full;
  JDIMENSION out_row_width;
  JDIMENSION rows_to_go;
  } my_upsampler;
typedef my_upsampler *my_upsample_ptr;
#define SCALEBITS 16
#define ONE_HALF ((int) 1 << (SCALEBITS-1))
#undef FIX
#define FIX(x)  ((int) ((x) * (1L<<SCALEBITS) + 0.5))
static void
  build_ycc_rgb_table(j_decompress_ptr cinfo)
 {
  my_upsample_ptr upsample = (my_upsample_ptr) cinfo->upsample;
  int i;
  int x;
  upsample->Cr_r_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (255 + 1) * ((size_t) sizeof(int)));
  upsample->Cb_b_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (255 + 1) * ((size_t) sizeof(int)));
  upsample->Cr_g_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (255 + 1) * ((size_t) sizeof(int)));
  upsample->Cb_g_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (255 + 1) * ((size_t) sizeof(int)));
  for (i = 0, x = -128; i <= 255; i++, x++)
   {
    upsample->Cr_r_tab[i] = (int)
     ((((int) ((1.40200) * (1L << 16) + 0.5)) * x + ((int) 1 << (16 - 1))) >> (16));
    upsample->Cb_b_tab[i] = (int)
     ((((int) ((1.77200) * (1L << 16) + 0.5)) * x + ((int) 1 << (16 - 1))) >> (16));
    upsample->Cr_g_tab[i] = (-((int) ((0.71414) * (1L << 16) + 0.5))) * x;
    upsample->Cb_g_tab[i] = (-((int) ((0.34414) * (1L << 16) + 0.5))) * x + ((int) 1 << (16 - 1));
    }
  }
static void
  start_pass_merged_upsample(j_decompress_ptr cinfo)
 {
  my_upsample_ptr upsample = (my_upsample_ptr) cinfo->upsample;
  upsample->spare_full = 0;
  upsample->rows_to_go = cinfo->output_height;
  }
static void
  merged_2v_upsample(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr,
   JDIMENSION in_row_groups_avail,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  my_upsample_ptr upsample = (my_upsample_ptr) cinfo->upsample;
  JSAMPROW work_ptrs[2];
  JDIMENSION num_rows;
  if (upsample->spare_full)
   {
    jcopy_sample_rows(&upsample->spare_row, 0, output_buf + *out_row_ctr, 0,
     1, upsample->out_row_width);
    num_rows = 1;
    upsample->spare_full = 0;
    }
  else
   {
    num_rows = 2;
    if (num_rows > upsample->rows_to_go)
      num_rows = upsample->rows_to_go;
    out_rows_avail -= *out_row_ctr;
    if (num_rows > out_rows_avail)
      num_rows = out_rows_avail;
    work_ptrs[0] = output_buf[*out_row_ctr];
    if (num_rows > 1)
     {
      work_ptrs[1] = output_buf[*out_row_ctr + 1];
      }
    else
     {
      work_ptrs[1] = upsample->spare_row;
      upsample->spare_full = 1;
      }
    (*upsample->upmethod) (cinfo, input_buf, *in_row_group_ctr, work_ptrs);
    }
  *out_row_ctr += num_rows;
  upsample->rows_to_go -= num_rows;
  if (!upsample->spare_full)
    (*in_row_group_ctr)++;
  }
static void
  merged_1v_upsample(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr,
   JDIMENSION in_row_groups_avail,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  my_upsample_ptr upsample = (my_upsample_ptr) cinfo->upsample;
  (*upsample->upmethod) (cinfo, input_buf, *in_row_group_ctr,
   output_buf + *out_row_ctr);
  (*out_row_ctr)++;
  (*in_row_group_ctr)++;
  }
static void
  h2v1_merged_upsample(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION in_row_group_ctr,
   JSAMPARRAY output_buf)
 {
  my_upsample_ptr upsample = (my_upsample_ptr) cinfo->upsample;
  register int y,
    cred,
    cgreen,
    cblue;
  int cb,
    cr;
  register JSAMPROW outptr;
  JSAMPROW inptr0,
    inptr1,
    inptr2;
  JDIMENSION col;
  register JSAMPLE *range_limit = cinfo->sample_range_limit;
  int *Crrtab = upsample->Cr_r_tab;
  int *Cbbtab = upsample->Cb_b_tab;
  int *Crgtab = upsample->Cr_g_tab;
  int *Cbgtab = upsample->Cb_g_tab;
  inptr0 = input_buf[0][in_row_group_ctr];
  inptr1 = input_buf[1][in_row_group_ctr];
  inptr2 = input_buf[2][in_row_group_ctr];
  outptr = output_buf[0];
  for (col = cinfo->output_width >> 1; col > 0; col--)
   {
    cb = ((int) (*inptr1++));
    cr = ((int) (*inptr2++));
    cred = Crrtab[cr];
    cgreen = (int) ((Cbgtab[cb] + Crgtab[cr]) >> (16));
    cblue = Cbbtab[cb];
    y = ((int) (*inptr0++));
    outptr[0] = range_limit[y + cred];
    outptr[1] = range_limit[y + cgreen];
    outptr[2] = range_limit[y + cblue];
    outptr += 3;
    y = ((int) (*inptr0++));
    outptr[0] = range_limit[y + cred];
    outptr[1] = range_limit[y + cgreen];
    outptr[2] = range_limit[y + cblue];
    outptr += 3;
    }
  if (cinfo->output_width & 1)
   {
    cb = ((int) (*inptr1));
    cr = ((int) (*inptr2));
    cred = Crrtab[cr];
    cgreen = (int) ((Cbgtab[cb] + Crgtab[cr]) >> (16));
    cblue = Cbbtab[cb];
    y = ((int) (*inptr0));
    outptr[0] = range_limit[y + cred];
    outptr[1] = range_limit[y + cgreen];
    outptr[2] = range_limit[y + cblue];
    }
  }
static void
  h2v2_merged_upsample(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION in_row_group_ctr,
   JSAMPARRAY output_buf)
 {
  my_upsample_ptr upsample = (my_upsample_ptr) cinfo->upsample;
  register int y,
    cred,
    cgreen,
    cblue;
  int cb,
    cr;
  register JSAMPROW outptr0,
    outptr1;
  JSAMPROW inptr00,
    inptr01,
    inptr1,
    inptr2;
  JDIMENSION col;
  register JSAMPLE *range_limit = cinfo->sample_range_limit;
  int *Crrtab = upsample->Cr_r_tab;
  int *Cbbtab = upsample->Cb_b_tab;
  int *Crgtab = upsample->Cr_g_tab;
  int *Cbgtab = upsample->Cb_g_tab;
  inptr00 = input_buf[0][in_row_group_ctr * 2];
  inptr01 = input_buf[0][in_row_group_ctr * 2 + 1];
  inptr1 = input_buf[1][in_row_group_ctr];
  inptr2 = input_buf[2][in_row_group_ctr];
  outptr0 = output_buf[0];
  outptr1 = output_buf[1];
  for (col = cinfo->output_width >> 1; col > 0; col--)
   {
    cb = ((int) (*inptr1++));
    cr = ((int) (*inptr2++));
    cred = Crrtab[cr];
    cgreen = (int) ((Cbgtab[cb] + Crgtab[cr]) >> (16));
    cblue = Cbbtab[cb];
    y = ((int) (*inptr00++));
    outptr0[0] = range_limit[y + cred];
    outptr0[1] = range_limit[y + cgreen];
    outptr0[2] = range_limit[y + cblue];
    outptr0 += 3;
    y = ((int) (*inptr00++));
    outptr0[0] = range_limit[y + cred];
    outptr0[1] = range_limit[y + cgreen];
    outptr0[2] = range_limit[y + cblue];
    outptr0 += 3;
    y = ((int) (*inptr01++));
    outptr1[0] = range_limit[y + cred];
    outptr1[1] = range_limit[y + cgreen];
    outptr1[2] = range_limit[y + cblue];
    outptr1 += 3;
    y = ((int) (*inptr01++));
    outptr1[0] = range_limit[y + cred];
    outptr1[1] = range_limit[y + cgreen];
    outptr1[2] = range_limit[y + cblue];
    outptr1 += 3;
    }
  if (cinfo->output_width & 1)
   {
    cb = ((int) (*inptr1));
    cr = ((int) (*inptr2));
    cred = Crrtab[cr];
    cgreen = (int) ((Cbgtab[cb] + Crgtab[cr]) >> (16));
    cblue = Cbbtab[cb];
    y = ((int) (*inptr00));
    outptr0[0] = range_limit[y + cred];
    outptr0[1] = range_limit[y + cgreen];
    outptr0[2] = range_limit[y + cblue];
    y = ((int) (*inptr01));
    outptr1[0] = range_limit[y + cred];
    outptr1[1] = range_limit[y + cgreen];
    outptr1[2] = range_limit[y + cblue];
    }
  }
void
  jinit_merged_upsampler(j_decompress_ptr cinfo)
 {
  my_upsample_ptr upsample;
  upsample = (my_upsample_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_upsampler)));
  cinfo->upsample = (struct jpeg_upsampler *) upsample;
  upsample->pub.start_pass = start_pass_merged_upsample;
  upsample->pub.need_context_rows = 0;
  upsample->out_row_width = cinfo->output_width * cinfo->out_color_components;
  if (cinfo->max_v_samp_factor == 2)
   {
    upsample->pub.upsample = merged_2v_upsample;
    upsample->upmethod = h2v2_merged_upsample;
    upsample->spare_row = (JSAMPROW)
     (*cinfo->mem->alloc_large) ((j_common_ptr) cinfo, 1,
     (size_t) (upsample->out_row_width * ((size_t) sizeof(JSAMPLE))));
    }
  else
   {
    upsample->pub.upsample = merged_1v_upsample;
    upsample->upmethod = h2v1_merged_upsample;
    upsample->spare_row = ((void *) 0);
    }
  build_ycc_rgb_table(cinfo);
  }
#endif
#define JPEG_INTERNALS
typedef void (*upsample1_ptr) (j_decompress_ptr cinfo, jpeg_component_info * compptr, JSAMPARRAY input_data, JSAMPARRAY * output_data_ptr);
typedef struct
 {
  struct jpeg_upsampler pub;
  JSAMPARRAY color_buf[10];
  upsample1_ptr methods[10];
  int next_row_out;
  JDIMENSION rows_to_go;
  int rowgroup_height[10];
  UINT8 h_expand[10];
  UINT8 v_expand[10];
  } my_upsampler1;
typedef my_upsampler1 *my_upsample_ptr1;
static void
  start_pass_upsample(j_decompress_ptr cinfo)
 {
  my_upsample_ptr1 upsample = (my_upsample_ptr1) cinfo->upsample;
  upsample->next_row_out = cinfo->max_v_samp_factor;
  upsample->rows_to_go = cinfo->output_height;
  }
static void
  sep_upsample(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr,
   JDIMENSION in_row_groups_avail,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  my_upsample_ptr1 upsample = (my_upsample_ptr1) cinfo->upsample;
  int ci;
  jpeg_component_info *compptr;
  JDIMENSION num_rows;
  if (upsample->next_row_out >= cinfo->max_v_samp_factor)
   {
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
     ci++, compptr++)
     {
      (*upsample->methods[ci]) (cinfo, compptr,
       input_buf[ci] + (*in_row_group_ctr * upsample->rowgroup_height[ci]),
       upsample->color_buf + ci);
      }
    upsample->next_row_out = 0;
    }
  num_rows = (JDIMENSION) (cinfo->max_v_samp_factor - upsample->next_row_out);
  if (num_rows > upsample->rows_to_go)
    num_rows = upsample->rows_to_go;
  out_rows_avail -= *out_row_ctr;
  if (num_rows > out_rows_avail)
    num_rows = out_rows_avail;
  (*cinfo->cconvert->color_convert) (cinfo, upsample->color_buf,
   (JDIMENSION) upsample->next_row_out,
   output_buf + *out_row_ctr,
   (int) num_rows);
  *out_row_ctr += num_rows;
  upsample->rows_to_go -= num_rows;
  upsample->next_row_out += num_rows;
  if (upsample->next_row_out >= cinfo->max_v_samp_factor)
    (*in_row_group_ctr)++;
  }
static void
  fullsize_upsample(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY * output_data_ptr)
 {
  *output_data_ptr = input_data;
  }
static void
  noop_upsample(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY * output_data_ptr)
 {
  *output_data_ptr = ((void *) 0);
  }
static void
  int_upsample(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY * output_data_ptr)
 {
  my_upsample_ptr1 upsample = (my_upsample_ptr1) cinfo->upsample;
  JSAMPARRAY output_data = *output_data_ptr;
  register JSAMPROW inptr,
    outptr;
  register JSAMPLE invalue;
  register int h;
  JSAMPROW outend;
  int h_expand,
    v_expand;
  int inrow,
    outrow;
  h_expand = upsample->h_expand[compptr->component_index];
  v_expand = upsample->v_expand[compptr->component_index];
  inrow = outrow = 0;
  while (outrow < cinfo->max_v_samp_factor)
   {
    inptr = input_data[inrow];
    outptr = output_data[outrow];
    outend = outptr + cinfo->output_width;
    while (outptr < outend)
     {
      invalue = *inptr++;
      for (h = h_expand; h > 0; h--)
       {
        *outptr++ = invalue;
        }
      }
    if (v_expand > 1)
     {
      jcopy_sample_rows(output_data, outrow, output_data, outrow + 1,
       v_expand - 1, cinfo->output_width);
      }
    inrow++;
    outrow += v_expand;
    }
  }
static void
  h2v1_upsample(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY * output_data_ptr)
 {
  JSAMPARRAY output_data = *output_data_ptr;
  register JSAMPROW inptr,
    outptr;
  register JSAMPLE invalue;
  JSAMPROW outend;
  int inrow;
  for (inrow = 0; inrow < cinfo->max_v_samp_factor; inrow++)
   {
    inptr = input_data[inrow];
    outptr = output_data[inrow];
    outend = outptr + cinfo->output_width;
    while (outptr < outend)
     {
      invalue = *inptr++;
      *outptr++ = invalue;
      *outptr++ = invalue;
      }
    }
  }
static void
  h2v2_upsample(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY * output_data_ptr)
 {
  JSAMPARRAY output_data = *output_data_ptr;
  register JSAMPROW inptr,
    outptr;
  register JSAMPLE invalue;
  JSAMPROW outend;
  int inrow,
    outrow;
  inrow = outrow = 0;
  while (outrow < cinfo->max_v_samp_factor)
   {
    inptr = input_data[inrow];
    outptr = output_data[outrow];
    outend = outptr + cinfo->output_width;
    while (outptr < outend)
     {
      invalue = *inptr++;
      *outptr++ = invalue;
      *outptr++ = invalue;
      }
    jcopy_sample_rows(output_data, outrow, output_data, outrow + 1,
     1, cinfo->output_width);
    inrow++;
    outrow += 2;
    }
  }
static void
  h2v1_fancy_upsample(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY * output_data_ptr)
 {
  JSAMPARRAY output_data = *output_data_ptr;
  register JSAMPROW inptr,
    outptr;
  register int invalue;
  register JDIMENSION colctr;
  int inrow;
  for (inrow = 0; inrow < cinfo->max_v_samp_factor; inrow++)
   {
    inptr = input_data[inrow];
    outptr = output_data[inrow];
    invalue = ((int) (*inptr++));
    *outptr++ = (JSAMPLE) invalue;
    *outptr++ = (JSAMPLE) ((invalue * 3 + ((int) (*inptr)) + 2) >> 2);
    for (colctr = compptr->downsampled_width - 2; colctr > 0; colctr--)
     {
      invalue = ((int) (*inptr++)) * 3;
      *outptr++ = (JSAMPLE) ((invalue + ((int) (inptr[-2])) + 1) >> 2);
      *outptr++ = (JSAMPLE) ((invalue + ((int) (*inptr)) + 2) >> 2);
      }
    invalue = ((int) (*inptr));
    *outptr++ = (JSAMPLE) ((invalue * 3 + ((int) (inptr[-1])) + 1) >> 2);
    *outptr++ = (JSAMPLE) invalue;
    }
  }
static void
  h2v2_fancy_upsample(j_decompress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY * output_data_ptr)
 {
  JSAMPARRAY output_data = *output_data_ptr;
  register JSAMPROW inptr0,
    inptr1,
    outptr;
#if 8  == 8
  register int thiscolsum,
    lastcolsum,
    nextcolsum;
#else
#endif
  register JDIMENSION colctr;
  int inrow,
    outrow,
    v;
  inrow = outrow = 0;
  while (outrow < cinfo->max_v_samp_factor)
   {
    for (v = 0; v < 2; v++)
     {
      inptr0 = input_data[inrow];
      if (v == 0)
        inptr1 = input_data[inrow - 1];
      else
        inptr1 = input_data[inrow + 1];
      outptr = output_data[outrow++];
      thiscolsum = ((int) (*inptr0++)) * 3 + ((int) (*inptr1++));
      nextcolsum = ((int) (*inptr0++)) * 3 + ((int) (*inptr1++));
      *outptr++ = (JSAMPLE) ((thiscolsum * 4 + 8) >> 4);
      *outptr++ = (JSAMPLE) ((thiscolsum * 3 + nextcolsum + 7) >> 4);
      lastcolsum = thiscolsum;
      thiscolsum = nextcolsum;
      for (colctr = compptr->downsampled_width - 2; colctr > 0; colctr--)
       {
        nextcolsum = ((int) (*inptr0++)) * 3 + ((int) (*inptr1++));
        *outptr++ = (JSAMPLE) ((thiscolsum * 3 + lastcolsum + 8) >> 4);
        *outptr++ = (JSAMPLE) ((thiscolsum * 3 + nextcolsum + 7) >> 4);
        lastcolsum = thiscolsum;
        thiscolsum = nextcolsum;
        }
      *outptr++ = (JSAMPLE) ((thiscolsum * 3 + lastcolsum + 8) >> 4);
      *outptr++ = (JSAMPLE) ((thiscolsum * 4 + 7) >> 4);
      }
    inrow++;
    }
  }
void
  jinit_upsampler(j_decompress_ptr cinfo)
 {
  my_upsample_ptr1 upsample;
  int ci;
  jpeg_component_info *compptr;
  boolean need_buffer,
    do_fancy;
  int h_in_group,
    v_in_group,
    h_out_group,
    v_out_group;
  upsample = (my_upsample_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_upsampler1)));
  cinfo->upsample = (struct jpeg_upsampler *) upsample;
  upsample->pub.start_pass = start_pass_upsample;
  upsample->pub.upsample = sep_upsample;
  upsample->pub.need_context_rows = 0;
  if (cinfo->CCIR601_sampling)
    ((cinfo)->err->msg_code = (JERR_CCIR601_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  do_fancy = cinfo->do_fancy_upsampling && cinfo->min_DCT_scaled_size > 1;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    h_in_group = (compptr->h_samp_factor * compptr->DCT_scaled_size) /
     cinfo->min_DCT_scaled_size;
    v_in_group = (compptr->v_samp_factor * compptr->DCT_scaled_size) /
     cinfo->min_DCT_scaled_size;
    h_out_group = cinfo->max_h_samp_factor;
    v_out_group = cinfo->max_v_samp_factor;
    upsample->rowgroup_height[ci] = v_in_group;
    need_buffer = 1;
    if (!compptr->component_needed)
     {
      upsample->methods[ci] = noop_upsample;
      need_buffer = 0;
      }
    else if (h_in_group == h_out_group && v_in_group == v_out_group)
     {
      upsample->methods[ci] = fullsize_upsample;
      need_buffer = 0;
      }
    else if (h_in_group * 2 == h_out_group &&
     v_in_group == v_out_group)
     {
      if (do_fancy && compptr->downsampled_width > 2)
        upsample->methods[ci] = h2v1_fancy_upsample;
      else
        upsample->methods[ci] = h2v1_upsample;
      }
    else if (h_in_group * 2 == h_out_group &&
     v_in_group * 2 == v_out_group)
     {
      if (do_fancy && compptr->downsampled_width > 2)
       {
        upsample->methods[ci] = h2v2_fancy_upsample;
        upsample->pub.need_context_rows = 1;
        }
      else
        upsample->methods[ci] = h2v2_upsample;
      }
    else if ((h_out_group % h_in_group) == 0 &&
     (v_out_group % v_in_group) == 0)
     {
      upsample->methods[ci] = int_upsample;
      upsample->h_expand[ci] = (UINT8) (h_out_group / h_in_group);
      upsample->v_expand[ci] = (UINT8) (v_out_group / v_in_group);
      }
    else
      ((cinfo)->err->msg_code = (JERR_FRACT_SAMPLE_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (need_buffer)
     {
      upsample->color_buf[ci] = (*cinfo->mem->alloc_sarray)
       ((j_common_ptr) cinfo, 1,
       (JDIMENSION) jround_up((long) cinfo->output_width,
        (long) cinfo->max_h_samp_factor),
       (JDIMENSION) cinfo->max_v_samp_factor);
      }
    }
  }
#define JPEG_INTERNALS
typedef struct
 {
  struct jpeg_color_converter pub;
  int *rgb_ycc_tab;
  } my_color_converter;
typedef my_color_converter *my_cconvert_ptr;
#define SCALEBITS 16
#define CBCR_OFFSET ((int) CENTERJSAMPLE << SCALEBITS)
#define ONE_HALF ((int) 1 << (SCALEBITS-1))
#define FIX(x)  ((int) ((x) * (1L<<SCALEBITS) + 0.5))
#define R_Y_OFF  0
#define G_Y_OFF  (1*(MAXJSAMPLE+1))
#define B_Y_OFF  (2*(MAXJSAMPLE+1))
#define R_CB_OFF (3*(MAXJSAMPLE+1))
#define G_CB_OFF (4*(MAXJSAMPLE+1))
#define B_CB_OFF (5*(MAXJSAMPLE+1))
#define R_CR_OFF B_CB_OFF
#define G_CR_OFF (6*(MAXJSAMPLE+1))
#define B_CR_OFF (7*(MAXJSAMPLE+1))
#define TABLE_SIZE (8*(MAXJSAMPLE+1))
static void
  rgb_ycc_start(j_compress_ptr cinfo)
 {
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  int *rgb_ycc_tab;
  int i;
  cconvert->rgb_ycc_tab = rgb_ycc_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((8 * (255 + 1)) * ((size_t) sizeof(int))));
  for (i = 0; i <= 255; i++)
   {
    rgb_ycc_tab[i + 0] = ((int) ((0.29900) * (1L << 16) + 0.5)) * i;
    rgb_ycc_tab[i + (1 * (255 + 1))] = ((int) ((0.58700) * (1L << 16) + 0.5)) * i;
    rgb_ycc_tab[i + (2 * (255 + 1))] = ((int) ((0.11400) * (1L << 16) + 0.5)) * i + ((int) 1 << (16 - 1));
    rgb_ycc_tab[i + (3 * (255 + 1))] = (-((int) ((0.16874) * (1L << 16) + 0.5))) * i;
    rgb_ycc_tab[i + (4 * (255 + 1))] = (-((int) ((0.33126) * (1L << 16) + 0.5))) * i;
    rgb_ycc_tab[i + (5 * (255 + 1))] = ((int) ((0.50000) * (1L << 16) + 0.5)) * i + ((int) 128 << 16) + ((int) 1 << (16 - 1)) - 1;
    rgb_ycc_tab[i + (6 * (255 + 1))] = (-((int) ((0.41869) * (1L << 16) + 0.5))) * i;
    rgb_ycc_tab[i + (7 * (255 + 1))] = (-((int) ((0.08131) * (1L << 16) + 0.5))) * i;
    }
  }
static void
  rgb_ycc_convert(j_compress_ptr cinfo,
   JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
   JDIMENSION output_row, int num_rows)
 {
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r,
    g,
    b;
  register int *ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr0,
    outptr1,
    outptr2;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;
  while (--num_rows >= 0)
   {
    inptr = *input_buf++;
    outptr0 = output_buf[0][output_row];
    outptr1 = output_buf[1][output_row];
    outptr2 = output_buf[2][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++)
     {
      r = ((int) (inptr[0]));
      g = ((int) (inptr[1]));
      b = ((int) (inptr[2]));
      inptr += 3;
      outptr0[col] = (JSAMPLE)
       ((ctab[r + 0] + ctab[g + (1 * (255 + 1))] + ctab[b + (2 * (255 + 1))])
       >> 16);
      outptr1[col] = (JSAMPLE)
       ((ctab[r + (3 * (255 + 1))] + ctab[g + (4 * (255 + 1))] + ctab[b + (5 * (255 + 1))])
       >> 16);
      outptr2[col] = (JSAMPLE)
       ((ctab[r + (5 * (255 + 1))] + ctab[g + (6 * (255 + 1))] + ctab[b + (7 * (255 + 1))])
       >> 16);
      }
    }
  }
static void
  rgb_gray_convert(j_compress_ptr cinfo,
   JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
   JDIMENSION output_row, int num_rows)
 {
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r,
    g,
    b;
  register int *ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;
  while (--num_rows >= 0)
   {
    inptr = *input_buf++;
    outptr = output_buf[0][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++)
     {
      r = ((int) (inptr[0]));
      g = ((int) (inptr[1]));
      b = ((int) (inptr[2]));
      inptr += 3;
      outptr[col] = (JSAMPLE)
       ((ctab[r + 0] + ctab[g + (1 * (255 + 1))] + ctab[b + (2 * (255 + 1))])
       >> 16);
      }
    }
  }
static void
  cmyk_ycck_convert(j_compress_ptr cinfo,
   JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
   JDIMENSION output_row, int num_rows)
 {
  my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  register int r,
    g,
    b;
  register int *ctab = cconvert->rgb_ycc_tab;
  register JSAMPROW inptr;
  register JSAMPROW outptr0,
    outptr1,
    outptr2,
    outptr3;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;
  while (--num_rows >= 0)
   {
    inptr = *input_buf++;
    outptr0 = output_buf[0][output_row];
    outptr1 = output_buf[1][output_row];
    outptr2 = output_buf[2][output_row];
    outptr3 = output_buf[3][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++)
     {
      r = 255 - ((int) (inptr[0]));
      g = 255 - ((int) (inptr[1]));
      b = 255 - ((int) (inptr[2]));
      outptr3[col] = inptr[3];
      inptr += 4;
      outptr0[col] = (JSAMPLE)
       ((ctab[r + 0] + ctab[g + (1 * (255 + 1))] + ctab[b + (2 * (255 + 1))])
       >> 16);
      outptr1[col] = (JSAMPLE)
       ((ctab[r + (3 * (255 + 1))] + ctab[g + (4 * (255 + 1))] + ctab[b + (5 * (255 + 1))])
       >> 16);
      outptr2[col] = (JSAMPLE)
       ((ctab[r + (5 * (255 + 1))] + ctab[g + (6 * (255 + 1))] + ctab[b + (7 * (255 + 1))])
       >> 16);
      }
    }
  }
static void
  grayscale_convert(j_compress_ptr cinfo,
   JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
   JDIMENSION output_row, int num_rows)
 {
  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->image_width;
  int instride = cinfo->input_components;
  while (--num_rows >= 0)
   {
    inptr = *input_buf++;
    outptr = output_buf[0][output_row];
    output_row++;
    for (col = 0; col < num_cols; col++)
     {
      outptr[col] = inptr[0];
      inptr += instride;
      }
    }
  }
static void
  null_convert(j_compress_ptr cinfo,
   JSAMPARRAY input_buf, JSAMPIMAGE output_buf,
   JDIMENSION output_row, int num_rows)
 {
  register JSAMPROW inptr;
  register JSAMPROW outptr;
  register JDIMENSION col;
  register int ci;
  int nc = cinfo->num_components;
  JDIMENSION num_cols = cinfo->image_width;
  while (--num_rows >= 0)
   {
    for (ci = 0; ci < nc; ci++)
     {
      inptr = *input_buf;
      outptr = output_buf[ci][output_row];
      for (col = 0; col < num_cols; col++)
       {
        outptr[col] = inptr[ci];
        inptr += nc;
        }
      }
    input_buf++;
    output_row++;
    }
  }
static void
  null_method(j_compress_ptr cinfo)
 {
  }
void
  jinit_color_converter(j_compress_ptr cinfo)
 {
  my_cconvert_ptr cconvert;
  cconvert = (my_cconvert_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_color_converter)));
  cinfo->cconvert = (struct jpeg_color_converter *) cconvert;
  cconvert->pub.start_pass = null_method;
  switch (cinfo->in_color_space)
   {
   case JCS_GRAYSCALE:
    if (cinfo->input_components != 1)
      ((cinfo)->err->msg_code = (JERR_BAD_IN_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_RGB:
#if 3  != 3
#endif
   case JCS_YCbCr:
    if (cinfo->input_components != 3)
      ((cinfo)->err->msg_code = (JERR_BAD_IN_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_CMYK:
   case JCS_YCCK:
    if (cinfo->input_components != 4)
      ((cinfo)->err->msg_code = (JERR_BAD_IN_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   default:
    if (cinfo->input_components < 1)
      ((cinfo)->err->msg_code = (JERR_BAD_IN_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  switch (cinfo->jpeg_color_space)
   {
   case JCS_GRAYSCALE:
    if (cinfo->num_components != 1)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (cinfo->in_color_space == JCS_GRAYSCALE)
      cconvert->pub.color_convert = grayscale_convert;
    else if (cinfo->in_color_space == JCS_RGB)
     {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = rgb_gray_convert;
      }
    else if (cinfo->in_color_space == JCS_YCbCr)
      cconvert->pub.color_convert = grayscale_convert;
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_RGB:
    if (cinfo->num_components != 3)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (cinfo->in_color_space == JCS_RGB && 3 == 3)
      cconvert->pub.color_convert = null_convert;
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_YCbCr:
    if (cinfo->num_components != 3)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (cinfo->in_color_space == JCS_RGB)
     {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = rgb_ycc_convert;
      }
    else if (cinfo->in_color_space == JCS_YCbCr)
      cconvert->pub.color_convert = null_convert;
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_CMYK:
    if (cinfo->num_components != 4)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (cinfo->in_color_space == JCS_CMYK)
      cconvert->pub.color_convert = null_convert;
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_YCCK:
    if (cinfo->num_components != 4)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    if (cinfo->in_color_space == JCS_CMYK)
     {
      cconvert->pub.start_pass = rgb_ycc_start;
      cconvert->pub.color_convert = cmyk_ycck_convert;
      }
    else if (cinfo->in_color_space == JCS_YCCK)
      cconvert->pub.color_convert = null_convert;
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   default:
    if (cinfo->jpeg_color_space != cinfo->in_color_space ||
     cinfo->num_components != cinfo->input_components)
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    cconvert->pub.color_convert = null_convert;
    break;
    }
  }
#define JPEG_INTERNALS
typedef struct
 {
  struct jpeg_color_deconverter pub;
  int *Cr_r_tab;
  int *Cb_b_tab;
  int *Cr_g_tab;
  int *Cb_g_tab;
  } my_color_deconverter;
typedef my_color_deconverter *my_cconvert_ptr1;
#define SCALEBITS 16
#define ONE_HALF ((int) 1 << (SCALEBITS-1))
#define FIX(x)  ((int) ((x) * (1L<<SCALEBITS) + 0.5))
static void
  build_ycc_rgb_table1(j_decompress_ptr cinfo)
 {
  my_cconvert_ptr1 cconvert = (my_cconvert_ptr1) cinfo->cconvert;
  int i;
  int x;
  cconvert->Cr_r_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (255 + 1) * ((size_t) sizeof(int)));
  cconvert->Cb_b_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (255 + 1) * ((size_t) sizeof(int)));
  cconvert->Cr_g_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (255 + 1) * ((size_t) sizeof(int)));
  cconvert->Cb_g_tab = (int *)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (255 + 1) * ((size_t) sizeof(int)));
  for (i = 0, x = -128; i <= 255; i++, x++)
   {
    cconvert->Cr_r_tab[i] = (int)
     ((((int) ((1.40200) * (1L << 16) + 0.5)) * x + ((int) 1 << (16 - 1))) >> (16));
    cconvert->Cb_b_tab[i] = (int)
     ((((int) ((1.77200) * (1L << 16) + 0.5)) * x + ((int) 1 << (16 - 1))) >> (16));
    cconvert->Cr_g_tab[i] = (-((int) ((0.71414) * (1L << 16) + 0.5))) * x;
    cconvert->Cb_g_tab[i] = (-((int) ((0.34414) * (1L << 16) + 0.5))) * x + ((int) 1 << (16 - 1));
    }
  }
static void
  ycc_rgb_convert(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION input_row,
   JSAMPARRAY output_buf, int num_rows)
 {
  my_cconvert_ptr1 cconvert = (my_cconvert_ptr1) cinfo->cconvert;
  register int y,
    cb,
    cr;
  register JSAMPROW outptr;
  register JSAMPROW inptr0,
    inptr1,
    inptr2;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->output_width;
  register JSAMPLE *range_limit = cinfo->sample_range_limit;
  register int *Crrtab = cconvert->Cr_r_tab;
  register int *Cbbtab = cconvert->Cb_b_tab;
  register int *Crgtab = cconvert->Cr_g_tab;
  register int *Cbgtab = cconvert->Cb_g_tab;
  while (--num_rows >= 0)
   {
    inptr0 = input_buf[0][input_row];
    inptr1 = input_buf[1][input_row];
    inptr2 = input_buf[2][input_row];
    input_row++;
    outptr = *output_buf++;
    for (col = 0; col < num_cols; col++)
     {
      y = ((int) (inptr0[col]));
      cb = ((int) (inptr1[col]));
      cr = ((int) (inptr2[col]));
      outptr[0] = range_limit[y + Crrtab[cr]];
      outptr[1] = range_limit[y +
       ((int) ((Cbgtab[cb] + Crgtab[cr]) >> (16)))];
      outptr[2] = range_limit[y + Cbbtab[cb]];
      outptr += 3;
      }
    }
  }
static void
  null_convert1(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION input_row,
   JSAMPARRAY output_buf, int num_rows)
 {
  register JSAMPROW inptr,
    outptr;
  register JDIMENSION count;
  register int num_components = cinfo->num_components;
  JDIMENSION num_cols = cinfo->output_width;
  int ci;
  while (--num_rows >= 0)
   {
    for (ci = 0; ci < num_components; ci++)
     {
      inptr = input_buf[ci][input_row];
      outptr = output_buf[0] + ci;
      for (count = num_cols; count > 0; count--)
       {
        *outptr = *inptr++;
        outptr += num_components;
        }
      }
    input_row++;
    output_buf++;
    }
  }
static void
  grayscale_convert1(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION input_row,
   JSAMPARRAY output_buf, int num_rows)
 {
  jcopy_sample_rows(input_buf[0], (int) input_row, output_buf, 0,
   num_rows, cinfo->output_width);
  }
static void
  gray_rgb_convert(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION input_row,
   JSAMPARRAY output_buf, int num_rows)
 {
  register JSAMPROW inptr,
    outptr;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->output_width;
  while (--num_rows >= 0)
   {
    inptr = input_buf[0][input_row++];
    outptr = *output_buf++;
    for (col = 0; col < num_cols; col++)
     {
      outptr[0] = outptr[1] = outptr[2] = inptr[col];
      outptr += 3;
      }
    }
  }
static void
  ycck_cmyk_convert(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION input_row,
   JSAMPARRAY output_buf, int num_rows)
 {
  my_cconvert_ptr1 cconvert = (my_cconvert_ptr1) cinfo->cconvert;
  register int y,
    cb,
    cr;
  register JSAMPROW outptr;
  register JSAMPROW inptr0,
    inptr1,
    inptr2,
    inptr3;
  register JDIMENSION col;
  JDIMENSION num_cols = cinfo->output_width;
  register JSAMPLE *range_limit = cinfo->sample_range_limit;
  register int *Crrtab = cconvert->Cr_r_tab;
  register int *Cbbtab = cconvert->Cb_b_tab;
  register int *Crgtab = cconvert->Cr_g_tab;
  register int *Cbgtab = cconvert->Cb_g_tab;
  while (--num_rows >= 0)
   {
    inptr0 = input_buf[0][input_row];
    inptr1 = input_buf[1][input_row];
    inptr2 = input_buf[2][input_row];
    inptr3 = input_buf[3][input_row];
    input_row++;
    outptr = *output_buf++;
    for (col = 0; col < num_cols; col++)
     {
      y = ((int) (inptr0[col]));
      cb = ((int) (inptr1[col]));
      cr = ((int) (inptr2[col]));
      outptr[0] = range_limit[255 - (y + Crrtab[cr])];
      outptr[1] = range_limit[255 - (y +
        ((int) ((Cbgtab[cb] + Crgtab[cr]) >> (16))))];
      outptr[2] = range_limit[255 - (y + Cbbtab[cb])];
      outptr[3] = inptr3[col];
      outptr += 4;
      }
    }
  }
static void
  start_pass_dcolor(j_decompress_ptr cinfo)
 {
  }
void
  jinit_color_deconverter(j_decompress_ptr cinfo)
 {
  my_cconvert_ptr1 cconvert;
  int ci;
  cconvert = (my_cconvert_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_color_deconverter)));
  cinfo->cconvert = (struct jpeg_color_deconverter *) cconvert;
  cconvert->pub.start_pass = start_pass_dcolor;
  switch (cinfo->jpeg_color_space)
   {
   case JCS_GRAYSCALE:
    if (cinfo->num_components != 1)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_RGB:
   case JCS_YCbCr:
    if (cinfo->num_components != 3)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_CMYK:
   case JCS_YCCK:
    if (cinfo->num_components != 4)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   default:
    if (cinfo->num_components < 1)
      ((cinfo)->err->msg_code = (JERR_BAD_J_COLORSPACE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  switch (cinfo->out_color_space)
   {
   case JCS_GRAYSCALE:
    cinfo->out_color_components = 1;
    if (cinfo->jpeg_color_space == JCS_GRAYSCALE ||
     cinfo->jpeg_color_space == JCS_YCbCr)
     {
      cconvert->pub.color_convert = grayscale_convert1;
      for (ci = 1; ci < cinfo->num_components; ci++)
        cinfo->comp_info[ci].component_needed = 0;
      }
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_RGB:
    cinfo->out_color_components = 3;
    if (cinfo->jpeg_color_space == JCS_YCbCr)
     {
      cconvert->pub.color_convert = ycc_rgb_convert;
      build_ycc_rgb_table1(cinfo);
      }
    else if (cinfo->jpeg_color_space == JCS_GRAYSCALE)
     {
      cconvert->pub.color_convert = gray_rgb_convert;
      }
    else if (cinfo->jpeg_color_space == JCS_RGB && 3 == 3)
     {
      cconvert->pub.color_convert = null_convert1;
      }
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   case JCS_CMYK:
    cinfo->out_color_components = 4;
    if (cinfo->jpeg_color_space == JCS_YCCK)
     {
      cconvert->pub.color_convert = ycck_cmyk_convert;
      build_ycc_rgb_table1(cinfo);
      }
    else if (cinfo->jpeg_color_space == JCS_CMYK)
     {
      cconvert->pub.color_convert = null_convert1;
      }
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
   default:
    if (cinfo->out_color_space == cinfo->jpeg_color_space)
     {
      cinfo->out_color_components = cinfo->num_components;
      cconvert->pub.color_convert = null_convert1;
      }
    else
      ((cinfo)->err->msg_code = (JERR_CONVERSION_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  if (cinfo->quantize_colors)
    cinfo->output_components = 1;
  else
    cinfo->output_components = cinfo->out_color_components;
  }
#define JPEG_INTERNALS
#ifdef ENTROPY_OPT_SUPPORTED
#define FULL_COEF_BUFFER_SUPPORTED
#else
#endif
typedef struct
 {
  struct jpeg_c_coef_controller pub;
  JDIMENSION iMCU_row_num;
  JDIMENSION mcu_ctr;
  int MCU_vert_offset;
  int MCU_rows_per_iMCU_row;
  JBLOCKROW MCU_buffer[10];
  jvirt_barray_ptr whole_image[10];
  } my_coef_controller;
typedef my_coef_controller *my_coef_ptr;
static boolean compress_data
  (j_compress_ptr cinfo, JSAMPIMAGE input_buf);
#ifdef FULL_COEF_BUFFER_SUPPORTED
static boolean compress_first_pass
  (j_compress_ptr cinfo, JSAMPIMAGE input_buf);
static boolean compress_output
  (j_compress_ptr cinfo, JSAMPIMAGE input_buf);
#endif
static void
  start_iMCU_row(j_compress_ptr cinfo)
 {
  my_coef_ptr coef = (my_coef_ptr) cinfo->coef;
  if (cinfo->comps_in_scan > 1)
   {
    coef->MCU_rows_per_iMCU_row = 1;
    }
  else
   {
    if (coef->iMCU_row_num < (cinfo->total_iMCU_rows - 1))
      coef->MCU_rows_per_iMCU_row = cinfo->cur_comp_info[0]->v_samp_factor;
    else
      coef->MCU_rows_per_iMCU_row = cinfo->cur_comp_info[0]->last_row_height;
    }
  coef->mcu_ctr = 0;
  coef->MCU_vert_offset = 0;
  }
static void
  start_pass_coef(j_compress_ptr cinfo, J_BUF_MODE pass_mode)
 {
  my_coef_ptr coef = (my_coef_ptr) cinfo->coef;
  coef->iMCU_row_num = 0;
  start_iMCU_row(cinfo);
  switch (pass_mode)
   {
   case JBUF_PASS_THRU:
    if (coef->whole_image[0] != ((void *) 0))
      ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    coef->pub.compress_data = compress_data;
    break;
#ifdef FULL_COEF_BUFFER_SUPPORTED
   case JBUF_SAVE_AND_PASS:
    if (coef->whole_image[0] == ((void *) 0))
      ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    coef->pub.compress_data = compress_first_pass;
    break;
   case JBUF_CRANK_DEST:
    if (coef->whole_image[0] == ((void *) 0))
      ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    coef->pub.compress_data = compress_output;
    break;
#endif
   default:
    ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  }
static boolean
  compress_data(j_compress_ptr cinfo, JSAMPIMAGE input_buf)
 {
  my_coef_ptr coef = (my_coef_ptr) cinfo->coef;
  JDIMENSION MCU_col_num;
  JDIMENSION last_MCU_col = cinfo->MCUs_per_row - 1;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  int blkn,
    bi,
    ci,
    yindex,
    yoffset,
    blockcnt;
  JDIMENSION ypos,
    xpos;
  jpeg_component_info *compptr;
  for (yoffset = coef->MCU_vert_offset; yoffset < coef->MCU_rows_per_iMCU_row;
   yoffset++)
   {
    for (MCU_col_num = coef->mcu_ctr; MCU_col_num <= last_MCU_col;
     MCU_col_num++)
     {
      blkn = 0;
      for (ci = 0; ci < cinfo->comps_in_scan; ci++)
       {
        compptr = cinfo->cur_comp_info[ci];
        blockcnt = (MCU_col_num < last_MCU_col) ? compptr->MCU_width
         : compptr->last_col_width;
        xpos = MCU_col_num * compptr->MCU_sample_width;
        ypos = yoffset * 8;
        for (yindex = 0; yindex < compptr->MCU_height; yindex++)
         {
          if (coef->iMCU_row_num < last_iMCU_row ||
           yoffset + yindex < compptr->last_row_height)
           {
            (*cinfo->fdct->forward_DCT) (cinfo, compptr,
             input_buf[compptr->component_index],
             coef->MCU_buffer[blkn],
             ypos, xpos, (JDIMENSION) blockcnt);
            if (blockcnt < compptr->MCU_width)
             {
              jzero_far((void *) coef->MCU_buffer[blkn + blockcnt],
               (compptr->MCU_width - blockcnt) * ((size_t) sizeof(JBLOCK)));
              for (bi = blockcnt; bi < compptr->MCU_width; bi++)
               {
                coef->MCU_buffer[blkn + bi][0][0] = coef->MCU_buffer[blkn + bi - 1][0][0];
                }
              }
            }
          else
           {
            jzero_far((void *) coef->MCU_buffer[blkn],
             compptr->MCU_width * ((size_t) sizeof(JBLOCK)));
            for (bi = 0; bi < compptr->MCU_width; bi++)
             {
              coef->MCU_buffer[blkn + bi][0][0] = coef->MCU_buffer[blkn - 1][0][0];
              }
            }
          blkn += compptr->MCU_width;
          ypos += 8;
          }
        }
      if (!(*cinfo->entropy->encode_mcu) (cinfo, coef->MCU_buffer))
       {
        coef->MCU_vert_offset = yoffset;
        coef->mcu_ctr = MCU_col_num;
        return 0;
        }
      }
    coef->mcu_ctr = 0;
    }
  coef->iMCU_row_num++;
  start_iMCU_row(cinfo);
  return 1;
  }
#ifdef FULL_COEF_BUFFER_SUPPORTED
static boolean
  compress_first_pass(j_compress_ptr cinfo, JSAMPIMAGE input_buf)
 {
  my_coef_ptr coef = (my_coef_ptr) cinfo->coef;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  JDIMENSION blocks_across,
    MCUs_across,
    MCUindex;
  int bi,
    ci,
    h_samp_factor,
    block_row,
    block_rows,
    ndummy;
  JCOEF lastDC;
  jpeg_component_info *compptr;
  JBLOCKARRAY buffer;
  JBLOCKROW thisblockrow,
    lastblockrow;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    buffer = (*cinfo->mem->access_virt_barray)
     ((j_common_ptr) cinfo, coef->whole_image[ci],
     coef->iMCU_row_num * compptr->v_samp_factor,
     (JDIMENSION) compptr->v_samp_factor, 1);
    if (coef->iMCU_row_num < last_iMCU_row)
      block_rows = compptr->v_samp_factor;
    else
     {
      block_rows = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
      if (block_rows == 0)
        block_rows = compptr->v_samp_factor;
      }
    blocks_across = compptr->width_in_blocks;
    h_samp_factor = compptr->h_samp_factor;
    ndummy = (int) (blocks_across % h_samp_factor);
    if (ndummy > 0)
      ndummy = h_samp_factor - ndummy;
    for (block_row = 0; block_row < block_rows; block_row++)
     {
      thisblockrow = buffer[block_row];
      (*cinfo->fdct->forward_DCT) (cinfo, compptr,
       input_buf[ci], thisblockrow,
       (JDIMENSION) (block_row * 8),
       (JDIMENSION) 0, blocks_across);
      if (ndummy > 0)
       {
        thisblockrow += blocks_across;
        jzero_far((void *) thisblockrow, ndummy * ((size_t) sizeof(JBLOCK)));
        lastDC = thisblockrow[-1][0];
        for (bi = 0; bi < ndummy; bi++)
         {
          thisblockrow[bi][0] = lastDC;
          }
        }
      }
    if (coef->iMCU_row_num == last_iMCU_row)
     {
      blocks_across += ndummy;
      MCUs_across = blocks_across / h_samp_factor;
      for (block_row = block_rows; block_row < compptr->v_samp_factor;
       block_row++)
       {
        thisblockrow = buffer[block_row];
        lastblockrow = buffer[block_row - 1];
        jzero_far((void *) thisblockrow,
         (size_t) (blocks_across * ((size_t) sizeof(JBLOCK))));
        for (MCUindex = 0; MCUindex < MCUs_across; MCUindex++)
         {
          lastDC = lastblockrow[h_samp_factor - 1][0];
          for (bi = 0; bi < h_samp_factor; bi++)
           {
            thisblockrow[bi][0] = lastDC;
            }
          thisblockrow += h_samp_factor;
          lastblockrow += h_samp_factor;
          }
        }
      }
    }
  return compress_output(cinfo, input_buf);
  }
static boolean
  compress_output(j_compress_ptr cinfo, JSAMPIMAGE input_buf)
 {
  my_coef_ptr coef = (my_coef_ptr) cinfo->coef;
  JDIMENSION MCU_col_num;
  int blkn,
    ci,
    xindex,
    yindex,
    yoffset;
  JDIMENSION start_col;
  JBLOCKARRAY buffer[4];
  JBLOCKROW buffer_ptr;
  jpeg_component_info *compptr;
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    buffer[ci] = (*cinfo->mem->access_virt_barray)
     ((j_common_ptr) cinfo, coef->whole_image[compptr->component_index],
     coef->iMCU_row_num * compptr->v_samp_factor,
     (JDIMENSION) compptr->v_samp_factor, 0);
    }
  for (yoffset = coef->MCU_vert_offset; yoffset < coef->MCU_rows_per_iMCU_row;
   yoffset++)
   {
    for (MCU_col_num = coef->mcu_ctr; MCU_col_num < cinfo->MCUs_per_row;
     MCU_col_num++)
     {
      blkn = 0;
      for (ci = 0; ci < cinfo->comps_in_scan; ci++)
       {
        compptr = cinfo->cur_comp_info[ci];
        start_col = MCU_col_num * compptr->MCU_width;
        for (yindex = 0; yindex < compptr->MCU_height; yindex++)
         {
          buffer_ptr = buffer[ci][yindex + yoffset] + start_col;
          for (xindex = 0; xindex < compptr->MCU_width; xindex++)
           {
            coef->MCU_buffer[blkn++] = buffer_ptr++;
            }
          }
        }
      if (!(*cinfo->entropy->encode_mcu) (cinfo, coef->MCU_buffer))
       {
        coef->MCU_vert_offset = yoffset;
        coef->mcu_ctr = MCU_col_num;
        return 0;
        }
      }
    coef->mcu_ctr = 0;
    }
  coef->iMCU_row_num++;
  start_iMCU_row(cinfo);
  return 1;
  }
#endif
void
  jinit_c_coef_controller(j_compress_ptr cinfo, boolean need_full_buffer)
 {
  my_coef_ptr coef;
  coef = (my_coef_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_coef_controller)));
  cinfo->coef = (struct jpeg_c_coef_controller *) coef;
  coef->pub.start_pass = start_pass_coef;
  if (need_full_buffer)
   {
#ifdef FULL_COEF_BUFFER_SUPPORTED
    int ci;
    jpeg_component_info *compptr;
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
     ci++, compptr++)
     {
      coef->whole_image[ci] = (*cinfo->mem->request_virt_barray)
       ((j_common_ptr) cinfo, 1, 0,
       (JDIMENSION) jround_up((long) compptr->width_in_blocks,
        (long) compptr->h_samp_factor),
       (JDIMENSION) jround_up((long) compptr->height_in_blocks,
        (long) compptr->v_samp_factor),
       (JDIMENSION) compptr->v_samp_factor);
      }
#else
#endif
    }
  else
   {
    JBLOCKROW buffer;
    int i;
    buffer = (JBLOCKROW)
     (*cinfo->mem->alloc_large) ((j_common_ptr) cinfo, 1,
     10 * ((size_t) sizeof(JBLOCK)));
    for (i = 0; i < 10; i++)
     {
      coef->MCU_buffer[i] = buffer + i;
      }
    coef->whole_image[0] = ((void *) 0);
    }
  }
#define JPEG_INTERNALS
#ifdef C_PROGRESSIVE_SUPPORTED
typedef struct
 {
  struct jpeg_entropy_encoder pub;
  boolean gather_statistics;
  JOCTET *next_output_byte;
  size_t free_in_buffer;
  int put_buffer;
  int put_bits;
  j_compress_ptr cinfo;
  int last_dc_val[4];
  int ac_tbl_no;
  unsigned int EOBRUN;
  unsigned int BE;
  char *bit_buffer;
  unsigned int restarts_to_go;
  int next_restart_num;
  c_derived_tbl *derived_tbls[4];
  long *count_ptrs[4];
  } phuff_entropy_encoder;
typedef phuff_entropy_encoder *phuff_entropy_ptr;
#define MAX_CORR_BITS  1000
#ifdef RIGHT_SHIFT_IS_UNSIGNED
#else
#define ISHIFT_TEMPS
#define IRIGHT_SHIFT(x,shft) ((x) >> (shft))
#endif
static boolean encode_mcu_DC_first(j_compress_ptr cinfo, JBLOCKROW * MCU_data);
static boolean encode_mcu_AC_first(j_compress_ptr cinfo, JBLOCKROW * MCU_data);
static boolean encode_mcu_DC_refine(j_compress_ptr cinfo, JBLOCKROW * MCU_data);
static boolean encode_mcu_AC_refine(j_compress_ptr cinfo, JBLOCKROW * MCU_data);
static void finish_pass_phuff(j_compress_ptr cinfo);
static void finish_pass_gather_phuff(j_compress_ptr cinfo);
static void
  start_pass_phuff(j_compress_ptr cinfo, boolean gather_statistics)
 {
  phuff_entropy_ptr entropy = (phuff_entropy_ptr) cinfo->entropy;
  boolean is_DC_band;
  int ci,
    tbl;
  jpeg_component_info *compptr;
  entropy->cinfo = cinfo;
  entropy->gather_statistics = gather_statistics;
  is_DC_band = (cinfo->Ss == 0);
  if (cinfo->Ah == 0)
   {
    if (is_DC_band)
      entropy->pub.encode_mcu = encode_mcu_DC_first;
    else
      entropy->pub.encode_mcu = encode_mcu_AC_first;
    }
  else
   {
    if (is_DC_band)
      entropy->pub.encode_mcu = encode_mcu_DC_refine;
    else
     {
      entropy->pub.encode_mcu = encode_mcu_AC_refine;
      if (entropy->bit_buffer == ((void *) 0))
        entropy->bit_buffer = (char *)
         (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
         1000 * ((size_t) sizeof(char)));
      }
    }
  if (gather_statistics)
    entropy->pub.finish_pass = finish_pass_gather_phuff;
  else
    entropy->pub.finish_pass = finish_pass_phuff;
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    entropy->last_dc_val[ci] = 0;
    if (is_DC_band)
     {
      if (cinfo->Ah != 0)
        continue;
      tbl = compptr->dc_tbl_no;
      }
    else
     {
      entropy->ac_tbl_no = tbl = compptr->ac_tbl_no;
      }
    if (gather_statistics)
     {
      if (tbl < 0 || tbl >= 4)
        ((cinfo)->err->msg_code = (JERR_NO_HUFF_TABLE), (cinfo)->err->msg_parm.i[0] = (tbl), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      if (entropy->count_ptrs[tbl] == ((void *) 0))
        entropy->count_ptrs[tbl] = (long *)
         (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
         257 * ((size_t) sizeof(long)));
      memset((void *) (entropy->count_ptrs[tbl]), 0, (size_t) (257 * ((size_t) sizeof(long))));
      }
    else
     {
      jpeg_make_c_derived_tbl(cinfo, is_DC_band, tbl,
       &entropy->derived_tbls[tbl]);
      }
    }
  entropy->EOBRUN = 0;
  entropy->BE = 0;
  entropy->put_buffer = 0;
  entropy->put_bits = 0;
  entropy->restarts_to_go = cinfo->restart_interval;
  entropy->next_restart_num = 0;
  }
#undef emit_byte
#define emit_byte(entropy,val)   { *(entropy)->next_output_byte++ = (JOCTET) (val);     if (--(entropy)->free_in_buffer == 0)       dump_buffer1(entropy); }
static void
  dump_buffer1(phuff_entropy_ptr entropy)
 {
  struct jpeg_destination_mgr *dest = entropy->cinfo->dest;
  if (!(*dest->empty_output_buffer) (entropy->cinfo))
    ((entropy->cinfo)->err->msg_code = (JERR_CANT_SUSPEND), (*(entropy->cinfo)->err->error_exit) ((j_common_ptr) (entropy->cinfo)));
  entropy->next_output_byte = dest->next_output_byte;
  entropy->free_in_buffer = dest->free_in_buffer;
  }
static void
  emit_bits1(phuff_entropy_ptr entropy, unsigned int code, int size)
 {
  register int put_buffer = (int) code;
  register int put_bits = entropy->put_bits;
  if (size == 0)
    ((entropy->cinfo)->err->msg_code = (JERR_HUFF_MISSING_CODE), (*(entropy->cinfo)->err->error_exit) ((j_common_ptr) (entropy->cinfo)));
  if (entropy->gather_statistics)
    return;
  put_buffer &= (((int) 1) << size) - 1;
  put_bits += size;
  put_buffer <<= 24 - put_bits;
  put_buffer |= entropy->put_buffer;
  while (put_bits >= 8)
   {
    int c = (int) ((put_buffer >> 16) & 0xFF);
     {
      *(entropy)->next_output_byte++ = (JOCTET) (c);
      if (--(entropy)->free_in_buffer == 0)
        dump_buffer1(entropy);
      };
    if (c == 0xFF)
     {
       {
        *(entropy)->next_output_byte++ = (JOCTET) (0);
        if (--(entropy)->free_in_buffer == 0)
          dump_buffer1(entropy);
        };
      }
    put_buffer <<= 8;
    put_bits -= 8;
    }
  entropy->put_buffer = put_buffer;
  entropy->put_bits = put_bits;
  }
static void
  flush_bits1(phuff_entropy_ptr entropy)
 {
  emit_bits1(entropy, 0x7F, 7);
  entropy->put_buffer = 0;
  entropy->put_bits = 0;
  }
static void
  emit_symbol(phuff_entropy_ptr entropy, int tbl_no, int symbol)
 {
  if (entropy->gather_statistics)
    entropy->count_ptrs[tbl_no][symbol]++;
  else
   {
    c_derived_tbl *tbl = entropy->derived_tbls[tbl_no];
    emit_bits1(entropy, tbl->ehufco[symbol], tbl->ehufsi[symbol]);
    }
  }
static void
  emit_buffered_bits(phuff_entropy_ptr entropy, char *bufstart,
   unsigned int nbits)
 {
  if (entropy->gather_statistics)
    return;
  while (nbits > 0)
   {
    emit_bits1(entropy, (unsigned int) (*bufstart), 1);
    bufstart++;
    nbits--;
    }
  }
static void
  emit_eobrun(phuff_entropy_ptr entropy)
 {
  register int temp,
    nbits;
  if (entropy->EOBRUN > 0)
   {
    temp = entropy->EOBRUN;
    nbits = 0;
    while ((temp >>= 1))
      nbits++;
    if (nbits > 14)
      ((entropy->cinfo)->err->msg_code = (JERR_HUFF_MISSING_CODE), (*(entropy->cinfo)->err->error_exit) ((j_common_ptr) (entropy->cinfo)));
    emit_symbol(entropy, entropy->ac_tbl_no, nbits << 4);
    if (nbits)
      emit_bits1(entropy, entropy->EOBRUN, nbits);
    entropy->EOBRUN = 0;
    emit_buffered_bits(entropy, entropy->bit_buffer, entropy->BE);
    entropy->BE = 0;
    }
  }
static void
  emit_restart1(phuff_entropy_ptr entropy, int restart_num)
 {
  int ci;
  emit_eobrun(entropy);
  if (!entropy->gather_statistics)
   {
    flush_bits1(entropy);
     {
      *(entropy)->next_output_byte++ = (JOCTET) (0xFF);
      if (--(entropy)->free_in_buffer == 0)
        dump_buffer1(entropy);
      };
     {
      *(entropy)->next_output_byte++ = (JOCTET) (0xD0 + restart_num);
      if (--(entropy)->free_in_buffer == 0)
        dump_buffer1(entropy);
      };
    }
  if (entropy->cinfo->Ss == 0)
   {
    for (ci = 0; ci < entropy->cinfo->comps_in_scan; ci++)
      entropy->last_dc_val[ci] = 0;
    }
  else
   {
    entropy->EOBRUN = 0;
    entropy->BE = 0;
    }
  }
static boolean
  encode_mcu_DC_first(j_compress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  phuff_entropy_ptr entropy = (phuff_entropy_ptr) cinfo->entropy;
  register int temp,
    temp2;
  register int nbits;
  int blkn,
    ci;
  int Al = cinfo->Al;
  JBLOCKROW block;
  jpeg_component_info *compptr;
  entropy->next_output_byte = cinfo->dest->next_output_byte;
  entropy->free_in_buffer = cinfo->dest->free_in_buffer;
  if (cinfo->restart_interval)
    if (entropy->restarts_to_go == 0)
      emit_restart1(entropy, entropy->next_restart_num);
  for (blkn = 0; blkn < cinfo->blocks_in_MCU; blkn++)
   {
    block = MCU_data[blkn];
    ci = cinfo->MCU_membership[blkn];
    compptr = cinfo->cur_comp_info[ci];
    temp2 = (((int) ((*block)[0])) >> (Al));
    temp = temp2 - entropy->last_dc_val[ci];
    entropy->last_dc_val[ci] = temp2;
    temp2 = temp;
    if (temp < 0)
     {
      temp = -temp;
      temp2--;
      }
    nbits = 0;
    while (temp)
     {
      nbits++;
      temp >>= 1;
      }
    if (nbits > 10 + 1)
      ((cinfo)->err->msg_code = (JERR_BAD_DCT_COEF), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    emit_symbol(entropy, compptr->dc_tbl_no, nbits);
    if (nbits)
      emit_bits1(entropy, (unsigned int) temp2, nbits);
    }
  cinfo->dest->next_output_byte = entropy->next_output_byte;
  cinfo->dest->free_in_buffer = entropy->free_in_buffer;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
     {
      entropy->restarts_to_go = cinfo->restart_interval;
      entropy->next_restart_num++;
      entropy->next_restart_num &= 7;
      }
    entropy->restarts_to_go--;
    }
  return 1;
  }
static boolean
  encode_mcu_AC_first(j_compress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  phuff_entropy_ptr entropy = (phuff_entropy_ptr) cinfo->entropy;
  register int temp,
    temp2;
  register int nbits;
  register int r,
    k;
  int Se = cinfo->Se;
  int Al = cinfo->Al;
  JBLOCKROW block;
  entropy->next_output_byte = cinfo->dest->next_output_byte;
  entropy->free_in_buffer = cinfo->dest->free_in_buffer;
  if (cinfo->restart_interval)
    if (entropy->restarts_to_go == 0)
      emit_restart1(entropy, entropy->next_restart_num);
  block = MCU_data[0];
  r = 0;
  for (k = cinfo->Ss; k <= Se; k++)
   {
    if ((temp = (*block)[jpeg_natural_order[k]]) == 0)
     {
      r++;
      continue;
      }
    if (temp < 0)
     {
      temp = -temp;
      temp >>= Al;
      temp2 = ~temp;
      }
    else
     {
      temp >>= Al;
      temp2 = temp;
      }
    if (temp == 0)
     {
      r++;
      continue;
      }
    if (entropy->EOBRUN > 0)
      emit_eobrun(entropy);
    while (r > 15)
     {
      emit_symbol(entropy, entropy->ac_tbl_no, 0xF0);
      r -= 16;
      }
    nbits = 1;
    while ((temp >>= 1))
      nbits++;
    if (nbits > 10)
      ((cinfo)->err->msg_code = (JERR_BAD_DCT_COEF), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    emit_symbol(entropy, entropy->ac_tbl_no, (r << 4) + nbits);
    emit_bits1(entropy, (unsigned int) temp2, nbits);
    r = 0;
    }
  if (r > 0)
   {
    entropy->EOBRUN++;
    if (entropy->EOBRUN == 0x7FFF)
      emit_eobrun(entropy);
    }
  cinfo->dest->next_output_byte = entropy->next_output_byte;
  cinfo->dest->free_in_buffer = entropy->free_in_buffer;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
     {
      entropy->restarts_to_go = cinfo->restart_interval;
      entropy->next_restart_num++;
      entropy->next_restart_num &= 7;
      }
    entropy->restarts_to_go--;
    }
  return 1;
  }
static boolean
  encode_mcu_DC_refine(j_compress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  phuff_entropy_ptr entropy = (phuff_entropy_ptr) cinfo->entropy;
  register int temp;
  int blkn;
  int Al = cinfo->Al;
  JBLOCKROW block;
  entropy->next_output_byte = cinfo->dest->next_output_byte;
  entropy->free_in_buffer = cinfo->dest->free_in_buffer;
  if (cinfo->restart_interval)
    if (entropy->restarts_to_go == 0)
      emit_restart1(entropy, entropy->next_restart_num);
  for (blkn = 0; blkn < cinfo->blocks_in_MCU; blkn++)
   {
    block = MCU_data[blkn];
    temp = (*block)[0];
    emit_bits1(entropy, (unsigned int) (temp >> Al), 1);
    }
  cinfo->dest->next_output_byte = entropy->next_output_byte;
  cinfo->dest->free_in_buffer = entropy->free_in_buffer;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
     {
      entropy->restarts_to_go = cinfo->restart_interval;
      entropy->next_restart_num++;
      entropy->next_restart_num &= 7;
      }
    entropy->restarts_to_go--;
    }
  return 1;
  }
static boolean
  encode_mcu_AC_refine(j_compress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  phuff_entropy_ptr entropy = (phuff_entropy_ptr) cinfo->entropy;
  register int temp;
  register int r,
    k;
  int EOB;
  char *BR_buffer;
  unsigned int BR;
  int Se = cinfo->Se;
  int Al = cinfo->Al;
  JBLOCKROW block;
  int absvalues[64];
  entropy->next_output_byte = cinfo->dest->next_output_byte;
  entropy->free_in_buffer = cinfo->dest->free_in_buffer;
  if (cinfo->restart_interval)
    if (entropy->restarts_to_go == 0)
      emit_restart1(entropy, entropy->next_restart_num);
  block = MCU_data[0];
  EOB = 0;
  for (k = cinfo->Ss; k <= Se; k++)
   {
    temp = (*block)[jpeg_natural_order[k]];
    if (temp < 0)
      temp = -temp;
    temp >>= Al;
    absvalues[k] = temp;
    if (temp == 1)
      EOB = k;
    }
  r = 0;
  BR = 0;
  BR_buffer = entropy->bit_buffer + entropy->BE;
  for (k = cinfo->Ss; k <= Se; k++)
   {
    if ((temp = absvalues[k]) == 0)
     {
      r++;
      continue;
      }
    while (r > 15 && k <= EOB)
     {
      emit_eobrun(entropy);
      emit_symbol(entropy, entropy->ac_tbl_no, 0xF0);
      r -= 16;
      emit_buffered_bits(entropy, BR_buffer, BR);
      BR_buffer = entropy->bit_buffer;
      BR = 0;
      }
    if (temp > 1)
     {
      BR_buffer[BR++] = (char) (temp & 1);
      continue;
      }
    emit_eobrun(entropy);
    emit_symbol(entropy, entropy->ac_tbl_no, (r << 4) + 1);
    temp = ((*block)[jpeg_natural_order[k]] < 0) ? 0 : 1;
    emit_bits1(entropy, (unsigned int) temp, 1);
    emit_buffered_bits(entropy, BR_buffer, BR);
    BR_buffer = entropy->bit_buffer;
    BR = 0;
    r = 0;
    }
  if (r > 0 || BR > 0)
   {
    entropy->EOBRUN++;
    entropy->BE += BR;
    if (entropy->EOBRUN == 0x7FFF || entropy->BE > (1000 - 64 + 1))
      emit_eobrun(entropy);
    }
  cinfo->dest->next_output_byte = entropy->next_output_byte;
  cinfo->dest->free_in_buffer = entropy->free_in_buffer;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
     {
      entropy->restarts_to_go = cinfo->restart_interval;
      entropy->next_restart_num++;
      entropy->next_restart_num &= 7;
      }
    entropy->restarts_to_go--;
    }
  return 1;
  }
static void
  finish_pass_phuff(j_compress_ptr cinfo)
 {
  phuff_entropy_ptr entropy = (phuff_entropy_ptr) cinfo->entropy;
  entropy->next_output_byte = cinfo->dest->next_output_byte;
  entropy->free_in_buffer = cinfo->dest->free_in_buffer;
  emit_eobrun(entropy);
  flush_bits1(entropy);
  cinfo->dest->next_output_byte = entropy->next_output_byte;
  cinfo->dest->free_in_buffer = entropy->free_in_buffer;
  }
static void
  finish_pass_gather_phuff(j_compress_ptr cinfo)
 {
  phuff_entropy_ptr entropy = (phuff_entropy_ptr) cinfo->entropy;
  boolean is_DC_band;
  int ci,
    tbl;
  jpeg_component_info *compptr;
  JHUFF_TBL **htblptr;
  boolean did[4];
  emit_eobrun(entropy);
  is_DC_band = (cinfo->Ss == 0);
  memset((void *) (did), 0, (size_t) (((size_t) sizeof(did))));
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    if (is_DC_band)
     {
      if (cinfo->Ah != 0)
        continue;
      tbl = compptr->dc_tbl_no;
      }
    else
     {
      tbl = compptr->ac_tbl_no;
      }
    if (!did[tbl])
     {
      if (is_DC_band)
        htblptr = &cinfo->dc_huff_tbl_ptrs[tbl];
      else
        htblptr = &cinfo->ac_huff_tbl_ptrs[tbl];
      if (*htblptr == ((void *) 0))
        *htblptr = jpeg_alloc_huff_table((j_common_ptr) cinfo);
      jpeg_gen_optimal_table(cinfo, *htblptr, entropy->count_ptrs[tbl]);
      did[tbl] = 1;
      }
    }
  }
void
  jinit_phuff_encoder(j_compress_ptr cinfo)
 {
  phuff_entropy_ptr entropy;
  int i;
  entropy = (phuff_entropy_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(phuff_entropy_encoder)));
  cinfo->entropy = (struct jpeg_entropy_encoder *) entropy;
  entropy->pub.start_pass = start_pass_phuff;
  for (i = 0; i < 4; i++)
   {
    entropy->derived_tbls[i] = ((void *) 0);
    entropy->count_ptrs[i] = ((void *) 0);
    }
  entropy->bit_buffer = ((void *) 0);
  }
#endif
#define JPEG_INTERNALS
typedef struct
 {
  struct jpeg_forward_dct pub;
  forward_DCT_method_ptr do_dct;
  DCTELEM *divisors[4];
#ifdef DCT_FLOAT_SUPPORTED
  float_DCT_method_ptr do_float_dct;
  float *float_divisors[4];
#endif
  } my_fdct_controller;
typedef my_fdct_controller *my_fdct_ptr;
static void
  start_pass_fdctmgr(j_compress_ptr cinfo)
 {
  my_fdct_ptr fdct = (my_fdct_ptr) cinfo->fdct;
  int ci,
    qtblno,
    i;
  jpeg_component_info *compptr;
  JQUANT_TBL *qtbl;
  DCTELEM *dtbl;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    qtblno = compptr->quant_tbl_no;
    if (qtblno < 0 || qtblno >= 4 ||
     cinfo->quant_tbl_ptrs[qtblno] == ((void *) 0))
      ((cinfo)->err->msg_code = (JERR_NO_QUANT_TABLE), (cinfo)->err->msg_parm.i[0] = (qtblno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    qtbl = cinfo->quant_tbl_ptrs[qtblno];
    switch (cinfo->dct_method)
     {
#ifdef DCT_ISLOW_SUPPORTED
     case JDCT_ISLOW:
      if (fdct->divisors[qtblno] == ((void *) 0))
       {
        fdct->divisors[qtblno] = (DCTELEM *)
         (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
         64 * ((size_t) sizeof(DCTELEM)));
        }
      dtbl = fdct->divisors[qtblno];
      for (i = 0; i < 64; i++)
       {
        dtbl[i] = ((DCTELEM) qtbl->quantval[i]) << 3;
        }
      break;
#endif
#ifdef DCT_IFAST_SUPPORTED
     case JDCT_IFAST:
       {
#undef CONST_BITS
#define CONST_BITS 14
         static const short aanscales[64] = {
          16384, 22725, 21407, 19266, 16384, 12873, 8867, 4520,
          22725, 31521, 29692, 26722, 22725, 17855, 12299, 6270,
          21407, 29692, 27969, 25172, 21407, 16819, 11585, 5906,
          19266, 26722, 25172, 22654, 19266, 15137, 10426, 5315,
          16384, 22725, 21407, 19266, 16384, 12873, 8867, 4520,
          12873, 17855, 16819, 15137, 12873, 10114, 6967, 3552,
          8867, 12299, 11585, 10426, 8867, 6967, 4799, 2446,
          4520, 6270, 5906, 5315, 4520, 3552, 2446, 1247
          };
        if (fdct->divisors[qtblno] == ((void *) 0))
         {
          fdct->divisors[qtblno] = (DCTELEM *)
           (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
           64 * ((size_t) sizeof(DCTELEM)));
          }
        dtbl = fdct->divisors[qtblno];
        for (i = 0; i < 64; i++)
         {
          dtbl[i] = (DCTELEM)
           (((((int) qtbl->quantval[i]) * ((int) aanscales[i]))) >> (14 - 3));
          }
        }
      break;
#endif
#ifdef DCT_FLOAT_SUPPORTED
     case JDCT_FLOAT:
       {
        float *fdtbl;
        int row,
          col;
         static const double aanscalefactor[8] = {
          1.0, 1.387039845, 1.306562965, 1.175875602,
          1.0, 0.785694958, 0.541196100, 0.275899379
          };
        if (fdct->float_divisors[qtblno] == ((void *) 0))
         {
          fdct->float_divisors[qtblno] = (float *)
           (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
           64 * ((size_t) sizeof(float)));
          }
        fdtbl = fdct->float_divisors[qtblno];
        i = 0;
        for (row = 0; row < 8; row++)
         {
          for (col = 0; col < 8; col++)
           {
            fdtbl[i] = (float)
             (1.0 / (((double) qtbl->quantval[i] *
               aanscalefactor[row] * aanscalefactor[col] * 8.0)));
            i++;
            }
          }
        }
      break;
#endif
     default:
      ((cinfo)->err->msg_code = (JERR_NOT_COMPILED), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      break;
      }
    }
  }
static void
  forward_DCT(j_compress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY sample_data, JBLOCKROW coef_blocks,
   JDIMENSION start_row, JDIMENSION start_col,
   JDIMENSION num_blocks)
 {
  my_fdct_ptr fdct = (my_fdct_ptr) cinfo->fdct;
  forward_DCT_method_ptr do_dct = fdct->do_dct;
  DCTELEM *divisors = fdct->divisors[compptr->quant_tbl_no];
  DCTELEM workspace[64];
  JDIMENSION bi;
  sample_data += start_row;
  for (bi = 0; bi < num_blocks; bi++, start_col += 8)
   {
     {
      register DCTELEM *workspaceptr;
      register JSAMPROW elemptr;
      register int elemr;
      workspaceptr = workspace;
      for (elemr = 0; elemr < 8; elemr++)
       {
        elemptr = sample_data[elemr] + start_col;
#if 8  == 8
        *workspaceptr++ = ((int) (*elemptr++)) - 128;
        *workspaceptr++ = ((int) (*elemptr++)) - 128;
        *workspaceptr++ = ((int) (*elemptr++)) - 128;
        *workspaceptr++ = ((int) (*elemptr++)) - 128;
        *workspaceptr++ = ((int) (*elemptr++)) - 128;
        *workspaceptr++ = ((int) (*elemptr++)) - 128;
        *workspaceptr++ = ((int) (*elemptr++)) - 128;
        *workspaceptr++ = ((int) (*elemptr++)) - 128;
#else
#endif
        }
      }
    (*do_dct) (workspace);
     {
      register DCTELEM temp,
        qval;
      register int i;
      register JCOEFPTR output_ptr = coef_blocks[bi];
      for (i = 0; i < 64; i++)
       {
        qval = divisors[i];
        temp = workspace[i];
#ifdef FAST_DIVIDE
#else
#define DIVIDE_BY(a,b) if (a >= b) a /= b; else a = 0
#endif
        if (temp < 0)
         {
          temp = -temp;
          temp += qval >> 1;
          if (temp >= qval)
            temp /= qval;
          else
            temp = 0;
          temp = -temp;
          }
        else
         {
          temp += qval >> 1;
          if (temp >= qval)
            temp /= qval;
          else
            temp = 0;
          }
        output_ptr[i] = (JCOEF) temp;
        }
      }
    }
  }
#ifdef DCT_FLOAT_SUPPORTED
static void
  forward_DCT_float(j_compress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY sample_data, JBLOCKROW coef_blocks,
   JDIMENSION start_row, JDIMENSION start_col,
   JDIMENSION num_blocks)
 {
  my_fdct_ptr fdct = (my_fdct_ptr) cinfo->fdct;
  float_DCT_method_ptr do_dct = fdct->do_float_dct;
  float *divisors = fdct->float_divisors[compptr->quant_tbl_no];
  float workspace[64];
  JDIMENSION bi;
  sample_data += start_row;
  for (bi = 0; bi < num_blocks; bi++, start_col += 8)
   {
     {
      register float *workspaceptr;
      register JSAMPROW elemptr;
      register int elemr;
      workspaceptr = workspace;
      for (elemr = 0; elemr < 8; elemr++)
       {
        elemptr = sample_data[elemr] + start_col;
#if 8  == 8
        *workspaceptr++ = (float) (((int) (*elemptr++)) - 128);
        *workspaceptr++ = (float) (((int) (*elemptr++)) - 128);
        *workspaceptr++ = (float) (((int) (*elemptr++)) - 128);
        *workspaceptr++ = (float) (((int) (*elemptr++)) - 128);
        *workspaceptr++ = (float) (((int) (*elemptr++)) - 128);
        *workspaceptr++ = (float) (((int) (*elemptr++)) - 128);
        *workspaceptr++ = (float) (((int) (*elemptr++)) - 128);
        *workspaceptr++ = (float) (((int) (*elemptr++)) - 128);
#else
#endif
        }
      }
    (*do_dct) (workspace);
     {
      register float temp;
      register int i;
      register JCOEFPTR output_ptr = coef_blocks[bi];
      for (i = 0; i < 64; i++)
       {
        temp = workspace[i] * divisors[i];
        output_ptr[i] = (JCOEF) ((int) (temp + (float) 16384.5) - 16384);
        }
      }
    }
  }
#endif
void
  jinit_forward_dct(j_compress_ptr cinfo)
 {
  my_fdct_ptr fdct;
  int i;
  fdct = (my_fdct_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_fdct_controller)));
  cinfo->fdct = (struct jpeg_forward_dct *) fdct;
  fdct->pub.start_pass = start_pass_fdctmgr;
  switch (cinfo->dct_method)
   {
#ifdef DCT_ISLOW_SUPPORTED
   case JDCT_ISLOW:
    fdct->pub.forward_DCT = forward_DCT;
    fdct->do_dct = jpeg_fdct_islow;
    break;
#endif
#ifdef DCT_IFAST_SUPPORTED
   case JDCT_IFAST:
    fdct->pub.forward_DCT = forward_DCT;
    fdct->do_dct = jpeg_fdct_ifast;
    break;
#endif
#ifdef DCT_FLOAT_SUPPORTED
   case JDCT_FLOAT:
    fdct->pub.forward_DCT = forward_DCT_float;
    fdct->do_float_dct = jpeg_fdct_float;
    break;
#endif
   default:
    ((cinfo)->err->msg_code = (JERR_NOT_COMPILED), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  for (i = 0; i < 4; i++)
   {
    fdct->divisors[i] = ((void *) 0);
#ifdef DCT_FLOAT_SUPPORTED
    fdct->float_divisors[i] = ((void *) 0);
#endif
    }
  }
#define JPEG_INTERNALS
#ifdef INPUT_SMOOTHING_SUPPORTED
#define CONTEXT_ROWS_SUPPORTED
#endif
typedef struct
 {
  struct jpeg_c_prep_controller pub;
  JSAMPARRAY color_buf[10];
  JDIMENSION rows_to_go;
  int next_buf_row;
#ifdef CONTEXT_ROWS_SUPPORTED
  int this_row_group;
  int next_buf_stop;
#endif
  } my_prep_controller;
typedef my_prep_controller *my_prep_ptr;
static void
  start_pass_prep(j_compress_ptr cinfo, J_BUF_MODE pass_mode)
 {
  my_prep_ptr prep = (my_prep_ptr) cinfo->prep;
  if (pass_mode != JBUF_PASS_THRU)
    ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  prep->rows_to_go = cinfo->image_height;
  prep->next_buf_row = 0;
#ifdef CONTEXT_ROWS_SUPPORTED
  prep->this_row_group = 0;
  prep->next_buf_stop = 2 * cinfo->max_v_samp_factor;
#endif
  }
static void
  expand_bottom_edge(JSAMPARRAY image_data, JDIMENSION num_cols,
   int input_rows, int output_rows)
 {
  register int row;
  for (row = input_rows; row < output_rows; row++)
   {
    jcopy_sample_rows(image_data, input_rows - 1, image_data, row,
     1, num_cols);
    }
  }
static void
  pre_process_data(j_compress_ptr cinfo,
   JSAMPARRAY input_buf, JDIMENSION * in_row_ctr,
   JDIMENSION in_rows_avail,
   JSAMPIMAGE output_buf, JDIMENSION * out_row_group_ctr,
   JDIMENSION out_row_groups_avail)
 {
  my_prep_ptr prep = (my_prep_ptr) cinfo->prep;
  int numrows,
    ci;
  JDIMENSION inrows;
  jpeg_component_info *compptr;
  while (*in_row_ctr < in_rows_avail &&
   *out_row_group_ctr < out_row_groups_avail)
   {
    inrows = in_rows_avail - *in_row_ctr;
    numrows = cinfo->max_v_samp_factor - prep->next_buf_row;
    numrows = (int) (((JDIMENSION) numrows) < (inrows) ? ((JDIMENSION) numrows) : (inrows));
    (*cinfo->cconvert->color_convert) (cinfo, input_buf + *in_row_ctr,
     prep->color_buf,
     (JDIMENSION) prep->next_buf_row,
     numrows);
    *in_row_ctr += numrows;
    prep->next_buf_row += numrows;
    prep->rows_to_go -= numrows;
    if (prep->rows_to_go == 0 &&
     prep->next_buf_row < cinfo->max_v_samp_factor)
     {
      for (ci = 0; ci < cinfo->num_components; ci++)
       {
        expand_bottom_edge(prep->color_buf[ci], cinfo->image_width,
         prep->next_buf_row, cinfo->max_v_samp_factor);
        }
      prep->next_buf_row = cinfo->max_v_samp_factor;
      }
    if (prep->next_buf_row == cinfo->max_v_samp_factor)
     {
      (*cinfo->downsample->downsample) (cinfo,
       prep->color_buf, (JDIMENSION) 0,
       output_buf, *out_row_group_ctr);
      prep->next_buf_row = 0;
      (*out_row_group_ctr)++;
      }
    if (prep->rows_to_go == 0 &&
     *out_row_group_ctr < out_row_groups_avail)
     {
      for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
       ci++, compptr++)
       {
        expand_bottom_edge(output_buf[ci],
         compptr->width_in_blocks * 8,
         (int) (*out_row_group_ctr * compptr->v_samp_factor),
         (int) (out_row_groups_avail * compptr->v_samp_factor));
        }
      *out_row_group_ctr = out_row_groups_avail;
      break;
      }
    }
  }
#ifdef CONTEXT_ROWS_SUPPORTED
static void
  pre_process_context(j_compress_ptr cinfo,
   JSAMPARRAY input_buf, JDIMENSION * in_row_ctr,
   JDIMENSION in_rows_avail,
   JSAMPIMAGE output_buf, JDIMENSION * out_row_group_ctr,
   JDIMENSION out_row_groups_avail)
 {
  my_prep_ptr prep = (my_prep_ptr) cinfo->prep;
  int numrows,
    ci;
  int buf_height = cinfo->max_v_samp_factor * 3;
  JDIMENSION inrows;
  while (*out_row_group_ctr < out_row_groups_avail)
   {
    if (*in_row_ctr < in_rows_avail)
     {
      inrows = in_rows_avail - *in_row_ctr;
      numrows = prep->next_buf_stop - prep->next_buf_row;
      numrows = (int) (((JDIMENSION) numrows) < (inrows) ? ((JDIMENSION) numrows) : (inrows));
      (*cinfo->cconvert->color_convert) (cinfo, input_buf + *in_row_ctr,
       prep->color_buf,
       (JDIMENSION) prep->next_buf_row,
       numrows);
      if (prep->rows_to_go == cinfo->image_height)
       {
        for (ci = 0; ci < cinfo->num_components; ci++)
         {
          int row;
          for (row = 1; row <= cinfo->max_v_samp_factor; row++)
           {
            jcopy_sample_rows(prep->color_buf[ci], 0,
             prep->color_buf[ci], -row,
             1, cinfo->image_width);
            }
          }
        }
      *in_row_ctr += numrows;
      prep->next_buf_row += numrows;
      prep->rows_to_go -= numrows;
      }
    else
     {
      if (prep->rows_to_go != 0)
        break;
      if (prep->next_buf_row < prep->next_buf_stop)
       {
        for (ci = 0; ci < cinfo->num_components; ci++)
         {
          expand_bottom_edge(prep->color_buf[ci], cinfo->image_width,
           prep->next_buf_row, prep->next_buf_stop);
          }
        prep->next_buf_row = prep->next_buf_stop;
        }
      }
    if (prep->next_buf_row == prep->next_buf_stop)
     {
      (*cinfo->downsample->downsample) (cinfo,
       prep->color_buf,
       (JDIMENSION) prep->this_row_group,
       output_buf, *out_row_group_ctr);
      (*out_row_group_ctr)++;
      prep->this_row_group += cinfo->max_v_samp_factor;
      if (prep->this_row_group >= buf_height)
        prep->this_row_group = 0;
      if (prep->next_buf_row >= buf_height)
        prep->next_buf_row = 0;
      prep->next_buf_stop = prep->next_buf_row + cinfo->max_v_samp_factor;
      }
    }
  }
static void
  create_context_buffer(j_compress_ptr cinfo)
 {
  my_prep_ptr prep = (my_prep_ptr) cinfo->prep;
  int rgroup_height = cinfo->max_v_samp_factor;
  int ci,
    i;
  jpeg_component_info *compptr;
  JSAMPARRAY true_buffer,
    fake_buffer;
  fake_buffer = (JSAMPARRAY)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   (cinfo->num_components * 5 * rgroup_height) *
   ((size_t) sizeof(JSAMPROW)));
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    true_buffer = (*cinfo->mem->alloc_sarray)
     ((j_common_ptr) cinfo, 1,
     (JDIMENSION) (((long) compptr->width_in_blocks * 8 *
       cinfo->max_h_samp_factor) / compptr->h_samp_factor),
     (JDIMENSION) (3 * rgroup_height));
    memcpy((void *) (fake_buffer + rgroup_height), (const void *) (true_buffer), (size_t) (3 * rgroup_height * ((size_t) sizeof(JSAMPROW))));
    for (i = 0; i < rgroup_height; i++)
     {
      fake_buffer[i] = true_buffer[2 * rgroup_height + i];
      fake_buffer[4 * rgroup_height + i] = true_buffer[i];
      }
    prep->color_buf[ci] = fake_buffer + rgroup_height;
    fake_buffer += 5 * rgroup_height;
    }
  }
#endif
void
  jinit_c_prep_controller(j_compress_ptr cinfo, boolean need_full_buffer)
 {
  my_prep_ptr prep;
  int ci;
  jpeg_component_info *compptr;
  if (need_full_buffer)
    ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  prep = (my_prep_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_prep_controller)));
  cinfo->prep = (struct jpeg_c_prep_controller *) prep;
  prep->pub.start_pass = start_pass_prep;
  if (cinfo->downsample->need_context_rows)
   {
#ifdef CONTEXT_ROWS_SUPPORTED
    prep->pub.pre_process_data = pre_process_context;
    create_context_buffer(cinfo);
#else
#endif
    }
  else
   {
    prep->pub.pre_process_data = pre_process_data;
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
     ci++, compptr++)
     {
      prep->color_buf[ci] = (*cinfo->mem->alloc_sarray)
       ((j_common_ptr) cinfo, 1,
       (JDIMENSION) (((long) compptr->width_in_blocks * 8 *
         cinfo->max_h_samp_factor) / compptr->h_samp_factor),
       (JDIMENSION) cinfo->max_v_samp_factor);
      }
    }
  }
#define JPEG_INTERNALS
typedef void (*downsample1_ptr) (j_compress_ptr cinfo, jpeg_component_info * compptr, JSAMPARRAY input_data, JSAMPARRAY output_data);
typedef struct
 {
  struct jpeg_downsampler pub;
  downsample1_ptr methods[10];
  } my_downsampler;
typedef my_downsampler *my_downsample_ptr;
static void
  start_pass_downsample(j_compress_ptr cinfo)
 {
  }
static void
  expand_right_edge(JSAMPARRAY image_data, int num_rows,
   JDIMENSION input_cols, JDIMENSION output_cols)
 {
  register JSAMPROW ptr;
  register JSAMPLE pixval;
  register int count;
  int row;
  int numcols = (int) (output_cols - input_cols);
  if (numcols > 0)
   {
    for (row = 0; row < num_rows; row++)
     {
      ptr = image_data[row] + input_cols;
      pixval = ptr[-1];
      for (count = numcols; count > 0; count--)
        *ptr++ = pixval;
      }
    }
  }
static void
  sep_downsample(j_compress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION in_row_index,
   JSAMPIMAGE output_buf, JDIMENSION out_row_group_index)
 {
  my_downsample_ptr downsample = (my_downsample_ptr) cinfo->downsample;
  int ci;
  jpeg_component_info *compptr;
  JSAMPARRAY in_ptr,
    out_ptr;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    in_ptr = input_buf[ci] + in_row_index;
    out_ptr = output_buf[ci] + (out_row_group_index * compptr->v_samp_factor);
    (*downsample->methods[ci]) (cinfo, compptr, in_ptr, out_ptr);
    }
  }
static void
  int_downsample(j_compress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY output_data)
 {
  int inrow,
    outrow,
    h_expand,
    v_expand,
    numpix,
    numpix2,
    h,
    v;
  JDIMENSION outcol,
    outcol_h;
  JDIMENSION output_cols = compptr->width_in_blocks * 8;
  JSAMPROW inptr,
    outptr;
  int outvalue;
  h_expand = cinfo->max_h_samp_factor / compptr->h_samp_factor;
  v_expand = cinfo->max_v_samp_factor / compptr->v_samp_factor;
  numpix = h_expand * v_expand;
  numpix2 = numpix / 2;
  expand_right_edge(input_data, cinfo->max_v_samp_factor,
   cinfo->image_width, output_cols * h_expand);
  inrow = 0;
  for (outrow = 0; outrow < compptr->v_samp_factor; outrow++)
   {
    outptr = output_data[outrow];
    for (outcol = 0, outcol_h = 0; outcol < output_cols;
     outcol++, outcol_h += h_expand)
     {
      outvalue = 0;
      for (v = 0; v < v_expand; v++)
       {
        inptr = input_data[inrow + v] + outcol_h;
        for (h = 0; h < h_expand; h++)
         {
          outvalue += (int) ((int) (*inptr++));
          }
        }
      *outptr++ = (JSAMPLE) ((outvalue + numpix2) / numpix);
      }
    inrow += v_expand;
    }
  }
static void
  fullsize_downsample(j_compress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY output_data)
 {
  jcopy_sample_rows(input_data, 0, output_data, 0,
   cinfo->max_v_samp_factor, cinfo->image_width);
  expand_right_edge(output_data, cinfo->max_v_samp_factor,
   cinfo->image_width, compptr->width_in_blocks * 8);
  }
static void
  h2v1_downsample(j_compress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY output_data)
 {
  int outrow;
  JDIMENSION outcol;
  JDIMENSION output_cols = compptr->width_in_blocks * 8;
  register JSAMPROW inptr,
    outptr;
  register int bias;
  expand_right_edge(input_data, cinfo->max_v_samp_factor,
   cinfo->image_width, output_cols * 2);
  for (outrow = 0; outrow < compptr->v_samp_factor; outrow++)
   {
    outptr = output_data[outrow];
    inptr = input_data[outrow];
    bias = 0;
    for (outcol = 0; outcol < output_cols; outcol++)
     {
      *outptr++ = (JSAMPLE) ((((int) (*inptr)) + ((int) (inptr[1]))
        + bias) >> 1);
      bias ^= 1;
      inptr += 2;
      }
    }
  }
static void
  h2v2_downsample(j_compress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY output_data)
 {
  int inrow,
    outrow;
  JDIMENSION outcol;
  JDIMENSION output_cols = compptr->width_in_blocks * 8;
  register JSAMPROW inptr0,
    inptr1,
    outptr;
  register int bias;
  expand_right_edge(input_data, cinfo->max_v_samp_factor,
   cinfo->image_width, output_cols * 2);
  inrow = 0;
  for (outrow = 0; outrow < compptr->v_samp_factor; outrow++)
   {
    outptr = output_data[outrow];
    inptr0 = input_data[inrow];
    inptr1 = input_data[inrow + 1];
    bias = 1;
    for (outcol = 0; outcol < output_cols; outcol++)
     {
      *outptr++ = (JSAMPLE) ((((int) (*inptr0)) + ((int) (inptr0[1])) +
        ((int) (*inptr1)) + ((int) (inptr1[1]))
        + bias) >> 2);
      bias ^= 3;
      inptr0 += 2;
      inptr1 += 2;
      }
    inrow += 2;
    }
  }
#ifdef INPUT_SMOOTHING_SUPPORTED
static void
  h2v2_smooth_downsample(j_compress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY output_data)
 {
  int inrow,
    outrow;
  JDIMENSION colctr;
  JDIMENSION output_cols = compptr->width_in_blocks * 8;
  register JSAMPROW inptr0,
    inptr1,
    above_ptr,
    below_ptr,
    outptr;
  int membersum,
    neighsum,
    memberscale,
    neighscale;
  expand_right_edge(input_data - 1, cinfo->max_v_samp_factor + 2,
   cinfo->image_width, output_cols * 2);
  memberscale = 16384 - cinfo->smoothing_factor * 80;
  neighscale = cinfo->smoothing_factor * 16;
  inrow = 0;
  for (outrow = 0; outrow < compptr->v_samp_factor; outrow++)
   {
    outptr = output_data[outrow];
    inptr0 = input_data[inrow];
    inptr1 = input_data[inrow + 1];
    above_ptr = input_data[inrow - 1];
    below_ptr = input_data[inrow + 2];
    membersum = ((int) (*inptr0)) + ((int) (inptr0[1])) +
     ((int) (*inptr1)) + ((int) (inptr1[1]));
    neighsum = ((int) (*above_ptr)) + ((int) (above_ptr[1])) +
     ((int) (*below_ptr)) + ((int) (below_ptr[1])) +
     ((int) (*inptr0)) + ((int) (inptr0[2])) +
     ((int) (*inptr1)) + ((int) (inptr1[2]));
    neighsum += neighsum;
    neighsum += ((int) (*above_ptr)) + ((int) (above_ptr[2])) +
     ((int) (*below_ptr)) + ((int) (below_ptr[2]));
    membersum = membersum * memberscale + neighsum * neighscale;
    *outptr++ = (JSAMPLE) ((membersum + 32768) >> 16);
    inptr0 += 2;
    inptr1 += 2;
    above_ptr += 2;
    below_ptr += 2;
    for (colctr = output_cols - 2; colctr > 0; colctr--)
     {
      membersum = ((int) (*inptr0)) + ((int) (inptr0[1])) +
       ((int) (*inptr1)) + ((int) (inptr1[1]));
      neighsum = ((int) (*above_ptr)) + ((int) (above_ptr[1])) +
       ((int) (*below_ptr)) + ((int) (below_ptr[1])) +
       ((int) (inptr0[-1])) + ((int) (inptr0[2])) +
       ((int) (inptr1[-1])) + ((int) (inptr1[2]));
      neighsum += neighsum;
      neighsum += ((int) (above_ptr[-1])) + ((int) (above_ptr[2])) +
       ((int) (below_ptr[-1])) + ((int) (below_ptr[2]));
      membersum = membersum * memberscale + neighsum * neighscale;
      *outptr++ = (JSAMPLE) ((membersum + 32768) >> 16);
      inptr0 += 2;
      inptr1 += 2;
      above_ptr += 2;
      below_ptr += 2;
      }
    membersum = ((int) (*inptr0)) + ((int) (inptr0[1])) +
     ((int) (*inptr1)) + ((int) (inptr1[1]));
    neighsum = ((int) (*above_ptr)) + ((int) (above_ptr[1])) +
     ((int) (*below_ptr)) + ((int) (below_ptr[1])) +
     ((int) (inptr0[-1])) + ((int) (inptr0[1])) +
     ((int) (inptr1[-1])) + ((int) (inptr1[1]));
    neighsum += neighsum;
    neighsum += ((int) (above_ptr[-1])) + ((int) (above_ptr[1])) +
     ((int) (below_ptr[-1])) + ((int) (below_ptr[1]));
    membersum = membersum * memberscale + neighsum * neighscale;
    *outptr = (JSAMPLE) ((membersum + 32768) >> 16);
    inrow += 2;
    }
  }
static void
  fullsize_smooth_downsample(j_compress_ptr cinfo, jpeg_component_info * compptr,
   JSAMPARRAY input_data, JSAMPARRAY output_data)
 {
  int outrow;
  JDIMENSION colctr;
  JDIMENSION output_cols = compptr->width_in_blocks * 8;
  register JSAMPROW inptr,
    above_ptr,
    below_ptr,
    outptr;
  int membersum,
    neighsum,
    memberscale,
    neighscale;
  int colsum,
    lastcolsum,
    nextcolsum;
  expand_right_edge(input_data - 1, cinfo->max_v_samp_factor + 2,
   cinfo->image_width, output_cols);
  memberscale = 65536L - cinfo->smoothing_factor * 512L;
  neighscale = cinfo->smoothing_factor * 64;
  for (outrow = 0; outrow < compptr->v_samp_factor; outrow++)
   {
    outptr = output_data[outrow];
    inptr = input_data[outrow];
    above_ptr = input_data[outrow - 1];
    below_ptr = input_data[outrow + 1];
    colsum = ((int) (*above_ptr++)) + ((int) (*below_ptr++)) +
     ((int) (*inptr));
    membersum = ((int) (*inptr++));
    nextcolsum = ((int) (*above_ptr)) + ((int) (*below_ptr)) +
     ((int) (*inptr));
    neighsum = colsum + (colsum - membersum) + nextcolsum;
    membersum = membersum * memberscale + neighsum * neighscale;
    *outptr++ = (JSAMPLE) ((membersum + 32768) >> 16);
    lastcolsum = colsum;
    colsum = nextcolsum;
    for (colctr = output_cols - 2; colctr > 0; colctr--)
     {
      membersum = ((int) (*inptr++));
      above_ptr++;
      below_ptr++;
      nextcolsum = ((int) (*above_ptr)) + ((int) (*below_ptr)) +
       ((int) (*inptr));
      neighsum = lastcolsum + (colsum - membersum) + nextcolsum;
      membersum = membersum * memberscale + neighsum * neighscale;
      *outptr++ = (JSAMPLE) ((membersum + 32768) >> 16);
      lastcolsum = colsum;
      colsum = nextcolsum;
      }
    membersum = ((int) (*inptr));
    neighsum = lastcolsum + (colsum - membersum) + colsum;
    membersum = membersum * memberscale + neighsum * neighscale;
    *outptr = (JSAMPLE) ((membersum + 32768) >> 16);
    }
  }
#endif
void
  jinit_downsampler(j_compress_ptr cinfo)
 {
  my_downsample_ptr downsample;
  int ci;
  jpeg_component_info *compptr;
  boolean smoothok = 1;
  downsample = (my_downsample_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_downsampler)));
  cinfo->downsample = (struct jpeg_downsampler *) downsample;
  downsample->pub.start_pass = start_pass_downsample;
  downsample->pub.downsample = sep_downsample;
  downsample->pub.need_context_rows = 0;
  if (cinfo->CCIR601_sampling)
    ((cinfo)->err->msg_code = (JERR_CCIR601_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    if (compptr->h_samp_factor == cinfo->max_h_samp_factor &&
     compptr->v_samp_factor == cinfo->max_v_samp_factor)
     {
#ifdef INPUT_SMOOTHING_SUPPORTED
      if (cinfo->smoothing_factor)
       {
        downsample->methods[ci] = fullsize_smooth_downsample;
        downsample->pub.need_context_rows = 1;
        }
      else
#endif
        downsample->methods[ci] = fullsize_downsample;
      }
    else if (compptr->h_samp_factor * 2 == cinfo->max_h_samp_factor &&
     compptr->v_samp_factor == cinfo->max_v_samp_factor)
     {
      smoothok = 0;
      downsample->methods[ci] = h2v1_downsample;
      }
    else if (compptr->h_samp_factor * 2 == cinfo->max_h_samp_factor &&
     compptr->v_samp_factor * 2 == cinfo->max_v_samp_factor)
     {
#ifdef INPUT_SMOOTHING_SUPPORTED
      if (cinfo->smoothing_factor)
       {
        downsample->methods[ci] = h2v2_smooth_downsample;
        downsample->pub.need_context_rows = 1;
        }
      else
#endif
        downsample->methods[ci] = h2v2_downsample;
      }
    else if ((cinfo->max_h_samp_factor % compptr->h_samp_factor) == 0 &&
     (cinfo->max_v_samp_factor % compptr->v_samp_factor) == 0)
     {
      smoothok = 0;
      downsample->methods[ci] = int_downsample;
      }
    else
      ((cinfo)->err->msg_code = (JERR_FRACT_SAMPLE_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
#ifdef INPUT_SMOOTHING_SUPPORTED
  if (cinfo->smoothing_factor && !smoothok)
    ((cinfo)->err->msg_code = (JTRC_SMOOTH_NOTIMPL), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), (0)));
#endif
  }
#define JPEG_INTERNALS
typedef struct
 {
  struct jpeg_d_main_controller pub;
  JSAMPARRAY buffer[10];
  boolean buffer_full;
  JDIMENSION rowgroup_ctr;
  JSAMPIMAGE xbuffer[2];
  int whichptr;
  int context_state;
  JDIMENSION rowgroups_avail;
  JDIMENSION iMCU_row_ctr;
  } my_main_controller1;
typedef my_main_controller1 *my_main_ptr1;
#define CTX_PREPARE_FOR_IMCU 0
#define CTX_PROCESS_IMCU 1
#define CTX_POSTPONED_ROW 2
static void process_data_simple_main1
  (j_decompress_ptr cinfo, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
static void process_data_context_main
  (j_decompress_ptr cinfo, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
#ifdef QUANT_2PASS_SUPPORTED
static void process_data_crank_post
  (j_decompress_ptr cinfo, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
#endif
static void
  alloc_funny_pointers(j_decompress_ptr cinfo)
 {
  my_main_ptr1 main = (my_main_ptr1) cinfo->main;
  int ci,
    rgroup;
  int M = cinfo->min_DCT_scaled_size;
  jpeg_component_info *compptr;
  JSAMPARRAY xbuf;
  main->xbuffer[0] = (JSAMPIMAGE)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   cinfo->num_components * 2 * ((size_t) sizeof(JSAMPARRAY)));
  main->xbuffer[1] = main->xbuffer[0] + cinfo->num_components;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    rgroup = (compptr->v_samp_factor * compptr->DCT_scaled_size) /
     cinfo->min_DCT_scaled_size;
    xbuf = (JSAMPARRAY)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
     2 * (rgroup * (M + 4)) * ((size_t) sizeof(JSAMPROW)));
    xbuf += rgroup;
    main->xbuffer[0][ci] = xbuf;
    xbuf += rgroup * (M + 4);
    main->xbuffer[1][ci] = xbuf;
    }
  }
static void
  make_funny_pointers(j_decompress_ptr cinfo)
 {
  my_main_ptr1 main = (my_main_ptr1) cinfo->main;
  int ci,
    i,
    rgroup;
  int M = cinfo->min_DCT_scaled_size;
  jpeg_component_info *compptr;
  JSAMPARRAY buf,
    xbuf0,
    xbuf1;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    rgroup = (compptr->v_samp_factor * compptr->DCT_scaled_size) /
     cinfo->min_DCT_scaled_size;
    xbuf0 = main->xbuffer[0][ci];
    xbuf1 = main->xbuffer[1][ci];
    buf = main->buffer[ci];
    for (i = 0; i < rgroup * (M + 2); i++)
     {
      xbuf0[i] = xbuf1[i] = buf[i];
      }
    for (i = 0; i < rgroup * 2; i++)
     {
      xbuf1[rgroup * (M - 2) + i] = buf[rgroup * M + i];
      xbuf1[rgroup * M + i] = buf[rgroup * (M - 2) + i];
      }
    for (i = 0; i < rgroup; i++)
     {
      xbuf0[i - rgroup] = xbuf0[0];
      }
    }
  }
static void
  set_wraparound_pointers(j_decompress_ptr cinfo)
 {
  my_main_ptr1 main = (my_main_ptr1) cinfo->main;
  int ci,
    i,
    rgroup;
  int M = cinfo->min_DCT_scaled_size;
  jpeg_component_info *compptr;
  JSAMPARRAY xbuf0,
    xbuf1;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    rgroup = (compptr->v_samp_factor * compptr->DCT_scaled_size) /
     cinfo->min_DCT_scaled_size;
    xbuf0 = main->xbuffer[0][ci];
    xbuf1 = main->xbuffer[1][ci];
    for (i = 0; i < rgroup; i++)
     {
      xbuf0[i - rgroup] = xbuf0[rgroup * (M + 1) + i];
      xbuf1[i - rgroup] = xbuf1[rgroup * (M + 1) + i];
      xbuf0[rgroup * (M + 2) + i] = xbuf0[i];
      xbuf1[rgroup * (M + 2) + i] = xbuf1[i];
      }
    }
  }
static void
  set_bottom_pointers(j_decompress_ptr cinfo)
 {
  my_main_ptr1 main = (my_main_ptr1) cinfo->main;
  int ci,
    i,
    rgroup,
    iMCUheight,
    rows_left;
  jpeg_component_info *compptr;
  JSAMPARRAY xbuf;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    iMCUheight = compptr->v_samp_factor * compptr->DCT_scaled_size;
    rgroup = iMCUheight / cinfo->min_DCT_scaled_size;
    rows_left = (int) (compptr->downsampled_height % (JDIMENSION) iMCUheight);
    if (rows_left == 0)
      rows_left = iMCUheight;
    if (ci == 0)
     {
      main->rowgroups_avail = (JDIMENSION) ((rows_left - 1) / rgroup + 1);
      }
    xbuf = main->xbuffer[main->whichptr][ci];
    for (i = 0; i < rgroup * 2; i++)
     {
      xbuf[rows_left + i] = xbuf[rows_left - 1];
      }
    }
  }
static void
  start_pass_main1(j_decompress_ptr cinfo, J_BUF_MODE pass_mode)
 {
  my_main_ptr1 main = (my_main_ptr1) cinfo->main;
  switch (pass_mode)
   {
   case JBUF_PASS_THRU:
    if (cinfo->upsample->need_context_rows)
     {
      main->pub.process_data = process_data_context_main;
      make_funny_pointers(cinfo);
      main->whichptr = 0;
      main->context_state = 0;
      main->iMCU_row_ctr = 0;
      }
    else
     {
      main->pub.process_data = process_data_simple_main1;
      }
    main->buffer_full = 0;
    main->rowgroup_ctr = 0;
    break;
#ifdef QUANT_2PASS_SUPPORTED
   case JBUF_CRANK_DEST:
    main->pub.process_data = process_data_crank_post;
    break;
#endif
   default:
    ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  }
static void
  process_data_simple_main1(j_decompress_ptr cinfo,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  my_main_ptr1 main = (my_main_ptr1) cinfo->main;
  JDIMENSION rowgroups_avail;
  if (!main->buffer_full)
   {
    if (!(*cinfo->coef->decompress_data) (cinfo, main->buffer))
      return;
    main->buffer_full = 1;
    }
  rowgroups_avail = (JDIMENSION) cinfo->min_DCT_scaled_size;
  (*cinfo->post->post_process_data) (cinfo, main->buffer,
   &main->rowgroup_ctr, rowgroups_avail,
   output_buf, out_row_ctr, out_rows_avail);
  if (main->rowgroup_ctr >= rowgroups_avail)
   {
    main->buffer_full = 0;
    main->rowgroup_ctr = 0;
    }
  }
static void
  process_data_context_main(j_decompress_ptr cinfo,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  my_main_ptr1 main = (my_main_ptr1) cinfo->main;
  if (!main->buffer_full)
   {
    if (!(*cinfo->coef->decompress_data) (cinfo,
      main->xbuffer[main->whichptr]))
      return;
    main->buffer_full = 1;
    main->iMCU_row_ctr++;
    }
  switch (main->context_state)
   {
   case 2:
    (*cinfo->post->post_process_data) (cinfo, main->xbuffer[main->whichptr],
     &main->rowgroup_ctr, main->rowgroups_avail,
     output_buf, out_row_ctr, out_rows_avail);
    if (main->rowgroup_ctr < main->rowgroups_avail)
      return;
    main->context_state = 0;
    if (*out_row_ctr >= out_rows_avail)
      return;
   case 0:
    main->rowgroup_ctr = 0;
    main->rowgroups_avail = (JDIMENSION) (cinfo->min_DCT_scaled_size - 1);
    if (main->iMCU_row_ctr == cinfo->total_iMCU_rows)
      set_bottom_pointers(cinfo);
    main->context_state = 1;
   case 1:
    (*cinfo->post->post_process_data) (cinfo, main->xbuffer[main->whichptr],
     &main->rowgroup_ctr, main->rowgroups_avail,
     output_buf, out_row_ctr, out_rows_avail);
    if (main->rowgroup_ctr < main->rowgroups_avail)
      return;
    if (main->iMCU_row_ctr == 1)
      set_wraparound_pointers(cinfo);
    main->whichptr ^= 1;
    main->buffer_full = 0;
    main->rowgroup_ctr = (JDIMENSION) (cinfo->min_DCT_scaled_size + 1);
    main->rowgroups_avail = (JDIMENSION) (cinfo->min_DCT_scaled_size + 2);
    main->context_state = 2;
    }
  }
#ifdef QUANT_2PASS_SUPPORTED
static void
  process_data_crank_post(j_decompress_ptr cinfo,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  (*cinfo->post->post_process_data) (cinfo, (JSAMPIMAGE) ((void *) 0),
   (JDIMENSION *) ((void *) 0), (JDIMENSION) 0,
   output_buf, out_row_ctr, out_rows_avail);
  }
#endif
void
  jinit_d_main_controller(j_decompress_ptr cinfo, boolean need_full_buffer)
 {
  my_main_ptr1 main;
  int ci,
    rgroup,
    ngroups;
  jpeg_component_info *compptr;
  main = (my_main_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_main_controller1)));
  cinfo->main = (struct jpeg_d_main_controller *) main;
  main->pub.start_pass = start_pass_main1;
  if (need_full_buffer)
    ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->upsample->need_context_rows)
   {
    if (cinfo->min_DCT_scaled_size < 2)
      ((cinfo)->err->msg_code = (JERR_NOTIMPL), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    alloc_funny_pointers(cinfo);
    ngroups = cinfo->min_DCT_scaled_size + 2;
    }
  else
   {
    ngroups = cinfo->min_DCT_scaled_size;
    }
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    rgroup = (compptr->v_samp_factor * compptr->DCT_scaled_size) /
     cinfo->min_DCT_scaled_size;
    main->buffer[ci] = (*cinfo->mem->alloc_sarray)
     ((j_common_ptr) cinfo, 1,
     compptr->width_in_blocks * compptr->DCT_scaled_size,
     (JDIMENSION) (rgroup * ngroups));
    }
  }
#define JPEG_INTERNALS
typedef enum
 {
  main_pass,
  huff_opt_pass,
  output_pass
  } c_pass_type;
typedef struct
 {
  struct jpeg_comp_master pub;
  c_pass_type pass_type;
  int pass_number;
  int total_passes;
  int scan_number;
  } my_comp_master;
typedef my_comp_master *my_master_ptr;
static void
  initial_setup1(j_compress_ptr cinfo)
 {
  int ci;
  jpeg_component_info *compptr;
  long samplesperrow;
  JDIMENSION jd_samplesperrow;
  if (cinfo->image_height <= 0 || cinfo->image_width <= 0
   || cinfo->num_components <= 0 || cinfo->input_components <= 0)
    ((cinfo)->err->msg_code = (JERR_EMPTY_IMAGE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if ((long) cinfo->image_height > (long) 65500L ||
   (long) cinfo->image_width > (long) 65500L)
    ((cinfo)->err->msg_code = (JERR_IMAGE_TOO_BIG), (cinfo)->err->msg_parm.i[0] = ((unsigned int) 65500L), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  samplesperrow = (long) cinfo->image_width * (long) cinfo->input_components;
  jd_samplesperrow = (JDIMENSION) samplesperrow;
  if ((long) jd_samplesperrow != samplesperrow)
    ((cinfo)->err->msg_code = (JERR_WIDTH_OVERFLOW), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->data_precision != 8)
    ((cinfo)->err->msg_code = (JERR_BAD_PRECISION), (cinfo)->err->msg_parm.i[0] = (cinfo->data_precision), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  if (cinfo->num_components > 10)
    ((cinfo)->err->msg_code = (JERR_COMPONENT_COUNT), (cinfo)->err->msg_parm.i[0] = (cinfo->num_components), (cinfo)->err->msg_parm.i[1] = (10), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  cinfo->max_h_samp_factor = 1;
  cinfo->max_v_samp_factor = 1;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    if (compptr->h_samp_factor <= 0 || compptr->h_samp_factor > 4 ||
     compptr->v_samp_factor <= 0 || compptr->v_samp_factor > 4)
      ((cinfo)->err->msg_code = (JERR_BAD_SAMPLING), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    cinfo->max_h_samp_factor = ((cinfo->max_h_samp_factor) > (compptr->h_samp_factor) ? (cinfo->max_h_samp_factor) : (compptr->h_samp_factor));
    cinfo->max_v_samp_factor = ((cinfo->max_v_samp_factor) > (compptr->v_samp_factor) ? (cinfo->max_v_samp_factor) : (compptr->v_samp_factor));
    }
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    compptr->component_index = ci;
    compptr->DCT_scaled_size = 8;
    compptr->width_in_blocks = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width * (long) compptr->h_samp_factor,
     (long) (cinfo->max_h_samp_factor * 8));
    compptr->height_in_blocks = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height * (long) compptr->v_samp_factor,
     (long) (cinfo->max_v_samp_factor * 8));
    compptr->downsampled_width = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width * (long) compptr->h_samp_factor,
     (long) cinfo->max_h_samp_factor);
    compptr->downsampled_height = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height * (long) compptr->v_samp_factor,
     (long) cinfo->max_v_samp_factor);
    compptr->component_needed = 1;
    }
  cinfo->total_iMCU_rows = (JDIMENSION)
   jdiv_round_up((long) cinfo->image_height,
   (long) (cinfo->max_v_samp_factor * 8));
  }
#ifdef C_MULTISCAN_FILES_SUPPORTED
static void
  validate_script(j_compress_ptr cinfo)
 {
  const jpeg_scan_info *scanptr;
  int scanno,
    ncomps,
    ci,
    coefi,
    thisi;
  int Ss,
    Se,
    Ah,
    Al;
  boolean component_sent[10];
#ifdef C_PROGRESSIVE_SUPPORTED
  int *last_bitpos_ptr;
  int last_bitpos[10][64];
#endif
  if (cinfo->num_scans <= 0)
    ((cinfo)->err->msg_code = (JERR_BAD_SCAN_SCRIPT), (cinfo)->err->msg_parm.i[0] = (0), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  scanptr = cinfo->scan_info;
  if (scanptr->Ss != 0 || scanptr->Se != 64 - 1)
   {
#ifdef C_PROGRESSIVE_SUPPORTED
    cinfo->progressive_mode = 1;
    last_bitpos_ptr = &last_bitpos[0][0];
    for (ci = 0; ci < cinfo->num_components; ci++)
      for (coefi = 0; coefi < 64; coefi++)
        *last_bitpos_ptr++ = -1;
#else
#endif
    }
  else
   {
    cinfo->progressive_mode = 0;
    for (ci = 0; ci < cinfo->num_components; ci++)
      component_sent[ci] = 0;
    }
  for (scanno = 1; scanno <= cinfo->num_scans; scanptr++, scanno++)
   {
    ncomps = scanptr->comps_in_scan;
    if (ncomps <= 0 || ncomps > 4)
      ((cinfo)->err->msg_code = (JERR_COMPONENT_COUNT), (cinfo)->err->msg_parm.i[0] = (ncomps), (cinfo)->err->msg_parm.i[1] = (4), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    for (ci = 0; ci < ncomps; ci++)
     {
      thisi = scanptr->component_index[ci];
      if (thisi < 0 || thisi >= cinfo->num_components)
        ((cinfo)->err->msg_code = (JERR_BAD_SCAN_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      if (ci > 0 && thisi <= scanptr->component_index[ci - 1])
        ((cinfo)->err->msg_code = (JERR_BAD_SCAN_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      }
    Ss = scanptr->Ss;
    Se = scanptr->Se;
    Ah = scanptr->Ah;
    Al = scanptr->Al;
    if (cinfo->progressive_mode)
     {
#ifdef C_PROGRESSIVE_SUPPORTED
#if 8  == 8
#define MAX_AH_AL 10
#else
#endif
      if (Ss < 0 || Ss >= 64 || Se < Ss || Se >= 64 ||
       Ah < 0 || Ah > 10 || Al < 0 || Al > 10)
        ((cinfo)->err->msg_code = (JERR_BAD_PROG_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      if (Ss == 0)
       {
        if (Se != 0)
          ((cinfo)->err->msg_code = (JERR_BAD_PROG_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
        }
      else
       {
        if (ncomps != 1)
          ((cinfo)->err->msg_code = (JERR_BAD_PROG_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
        }
      for (ci = 0; ci < ncomps; ci++)
       {
        last_bitpos_ptr = &last_bitpos[scanptr->component_index[ci]][0];
        if (Ss != 0 && last_bitpos_ptr[0] < 0)
          ((cinfo)->err->msg_code = (JERR_BAD_PROG_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
        for (coefi = Ss; coefi <= Se; coefi++)
         {
          if (last_bitpos_ptr[coefi] < 0)
           {
            if (Ah != 0)
              ((cinfo)->err->msg_code = (JERR_BAD_PROG_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
            }
          else
           {
            if (Ah != last_bitpos_ptr[coefi] || Al != Ah - 1)
              ((cinfo)->err->msg_code = (JERR_BAD_PROG_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
            }
          last_bitpos_ptr[coefi] = Al;
          }
        }
#endif
      }
    else
     {
      if (Ss != 0 || Se != 64 - 1 || Ah != 0 || Al != 0)
        ((cinfo)->err->msg_code = (JERR_BAD_PROG_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      for (ci = 0; ci < ncomps; ci++)
       {
        thisi = scanptr->component_index[ci];
        if (component_sent[thisi])
          ((cinfo)->err->msg_code = (JERR_BAD_SCAN_SCRIPT), (cinfo)->err->msg_parm.i[0] = (scanno), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
        component_sent[thisi] = 1;
        }
      }
    }
  if (cinfo->progressive_mode)
   {
#ifdef C_PROGRESSIVE_SUPPORTED
    for (ci = 0; ci < cinfo->num_components; ci++)
     {
      if (last_bitpos[ci][0] < 0)
        ((cinfo)->err->msg_code = (JERR_MISSING_DATA), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      }
#endif
    }
  else
   {
    for (ci = 0; ci < cinfo->num_components; ci++)
     {
      if (!component_sent[ci])
        ((cinfo)->err->msg_code = (JERR_MISSING_DATA), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      }
    }
  }
#endif
static void
  select_scan_parameters(j_compress_ptr cinfo)
 {
  int ci;
#ifdef C_MULTISCAN_FILES_SUPPORTED
  if (cinfo->scan_info != ((void *) 0))
   {
    my_master_ptr master = (my_master_ptr) cinfo->master;
    const jpeg_scan_info *scanptr = cinfo->scan_info + master->scan_number;
    cinfo->comps_in_scan = scanptr->comps_in_scan;
    for (ci = 0; ci < scanptr->comps_in_scan; ci++)
     {
      cinfo->cur_comp_info[ci] =
       &cinfo->comp_info[scanptr->component_index[ci]];
      }
    cinfo->Ss = scanptr->Ss;
    cinfo->Se = scanptr->Se;
    cinfo->Ah = scanptr->Ah;
    cinfo->Al = scanptr->Al;
    }
  else
#endif
   {
    if (cinfo->num_components > 4)
      ((cinfo)->err->msg_code = (JERR_COMPONENT_COUNT), (cinfo)->err->msg_parm.i[0] = (cinfo->num_components), (cinfo)->err->msg_parm.i[1] = (4), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    cinfo->comps_in_scan = cinfo->num_components;
    for (ci = 0; ci < cinfo->num_components; ci++)
     {
      cinfo->cur_comp_info[ci] = &cinfo->comp_info[ci];
      }
    cinfo->Ss = 0;
    cinfo->Se = 64 - 1;
    cinfo->Ah = 0;
    cinfo->Al = 0;
    }
  }
static void
  per_scan_setup1(j_compress_ptr cinfo)
 {
  int ci,
    mcublks,
    tmp;
  jpeg_component_info *compptr;
  if (cinfo->comps_in_scan == 1)
   {
    compptr = cinfo->cur_comp_info[0];
    cinfo->MCUs_per_row = compptr->width_in_blocks;
    cinfo->MCU_rows_in_scan = compptr->height_in_blocks;
    compptr->MCU_width = 1;
    compptr->MCU_height = 1;
    compptr->MCU_blocks = 1;
    compptr->MCU_sample_width = 8;
    compptr->last_col_width = 1;
    tmp = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
    if (tmp == 0)
      tmp = compptr->v_samp_factor;
    compptr->last_row_height = tmp;
    cinfo->blocks_in_MCU = 1;
    cinfo->MCU_membership[0] = 0;
    }
  else
   {
    if (cinfo->comps_in_scan <= 0 || cinfo->comps_in_scan > 4)
      ((cinfo)->err->msg_code = (JERR_COMPONENT_COUNT), (cinfo)->err->msg_parm.i[0] = (cinfo->comps_in_scan), (cinfo)->err->msg_parm.i[1] = (4), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    cinfo->MCUs_per_row = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_width,
     (long) (cinfo->max_h_samp_factor * 8));
    cinfo->MCU_rows_in_scan = (JDIMENSION)
     jdiv_round_up((long) cinfo->image_height,
     (long) (cinfo->max_v_samp_factor * 8));
    cinfo->blocks_in_MCU = 0;
    for (ci = 0; ci < cinfo->comps_in_scan; ci++)
     {
      compptr = cinfo->cur_comp_info[ci];
      compptr->MCU_width = compptr->h_samp_factor;
      compptr->MCU_height = compptr->v_samp_factor;
      compptr->MCU_blocks = compptr->MCU_width * compptr->MCU_height;
      compptr->MCU_sample_width = compptr->MCU_width * 8;
      tmp = (int) (compptr->width_in_blocks % compptr->MCU_width);
      if (tmp == 0)
        tmp = compptr->MCU_width;
      compptr->last_col_width = tmp;
      tmp = (int) (compptr->height_in_blocks % compptr->MCU_height);
      if (tmp == 0)
        tmp = compptr->MCU_height;
      compptr->last_row_height = tmp;
      mcublks = compptr->MCU_blocks;
      if (cinfo->blocks_in_MCU + mcublks > 10)
        ((cinfo)->err->msg_code = (JERR_BAD_MCU_SIZE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
      while (mcublks-- > 0)
       {
        cinfo->MCU_membership[cinfo->blocks_in_MCU++] = ci;
        }
      }
    }
  if (cinfo->restart_in_rows > 0)
   {
    long nominal = (long) cinfo->restart_in_rows * (long) cinfo->MCUs_per_row;
    cinfo->restart_interval = (unsigned int) ((nominal) < (65535L) ? (nominal) : (65535L));
    }
  }
static void
  prepare_for_pass(j_compress_ptr cinfo)
 {
  my_master_ptr master = (my_master_ptr) cinfo->master;
  switch (master->pass_type)
   {
   case main_pass:
    select_scan_parameters(cinfo);
    per_scan_setup1(cinfo);
    if (!cinfo->raw_data_in)
     {
      (*cinfo->cconvert->start_pass) (cinfo);
      (*cinfo->downsample->start_pass) (cinfo);
      (*cinfo->prep->start_pass) (cinfo, JBUF_PASS_THRU);
      }
    (*cinfo->fdct->start_pass) (cinfo);
    (*cinfo->entropy->start_pass) (cinfo, cinfo->optimize_coding);
    (*cinfo->coef->start_pass) (cinfo,
     (master->total_passes > 1 ?
      JBUF_SAVE_AND_PASS : JBUF_PASS_THRU));
    (*cinfo->main->start_pass) (cinfo, JBUF_PASS_THRU);
    if (cinfo->optimize_coding)
     {
      master->pub.call_pass_startup = 0;
      }
    else
     {
      master->pub.call_pass_startup = 1;
      }
    break;
#ifdef ENTROPY_OPT_SUPPORTED
   case huff_opt_pass:
    select_scan_parameters(cinfo);
    per_scan_setup1(cinfo);
    if (cinfo->Ss != 0 || cinfo->Ah == 0 || cinfo->arith_code)
     {
      (*cinfo->entropy->start_pass) (cinfo, 1);
      (*cinfo->coef->start_pass) (cinfo, JBUF_CRANK_DEST);
      master->pub.call_pass_startup = 0;
      break;
      }
    master->pass_type = output_pass;
    master->pass_number++;
#endif
   case output_pass:
    if (!cinfo->optimize_coding)
     {
      select_scan_parameters(cinfo);
      per_scan_setup1(cinfo);
      }
    (*cinfo->entropy->start_pass) (cinfo, 0);
    (*cinfo->coef->start_pass) (cinfo, JBUF_CRANK_DEST);
    if (master->scan_number == 0)
      (*cinfo->marker->write_frame_header) (cinfo);
    (*cinfo->marker->write_scan_header) (cinfo);
    master->pub.call_pass_startup = 0;
    break;
   default:
    ((cinfo)->err->msg_code = (JERR_NOT_COMPILED), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    }
  master->pub.is_last_pass = (master->pass_number == master->total_passes - 1);
  if (cinfo->progress != ((void *) 0))
   {
    cinfo->progress->completed_passes = master->pass_number;
    cinfo->progress->total_passes = master->total_passes;
    }
  }
static void
  pass_startup(j_compress_ptr cinfo)
 {
  cinfo->master->call_pass_startup = 0;
  (*cinfo->marker->write_frame_header) (cinfo);
  (*cinfo->marker->write_scan_header) (cinfo);
  }
static void
  finish_pass_master(j_compress_ptr cinfo)
 {
  my_master_ptr master = (my_master_ptr) cinfo->master;
  (*cinfo->entropy->finish_pass) (cinfo);
  switch (master->pass_type)
   {
   case main_pass:
    master->pass_type = output_pass;
    if (!cinfo->optimize_coding)
      master->scan_number++;
    break;
   case huff_opt_pass:
    master->pass_type = output_pass;
    break;
   case output_pass:
    if (cinfo->optimize_coding)
      master->pass_type = huff_opt_pass;
    master->scan_number++;
    break;
    }
  master->pass_number++;
  }
void
  jinit_c_master_control(j_compress_ptr cinfo, boolean transcode_only)
 {
  my_master_ptr master;
  master = (my_master_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_comp_master)));
  cinfo->master = (struct jpeg_comp_master *) master;
  master->pub.prepare_for_pass = prepare_for_pass;
  master->pub.pass_startup = pass_startup;
  master->pub.finish_pass = finish_pass_master;
  master->pub.is_last_pass = 0;
  initial_setup1(cinfo);
  if (cinfo->scan_info != ((void *) 0))
   {
#ifdef C_MULTISCAN_FILES_SUPPORTED
    validate_script(cinfo);
#else
#endif
    }
  else
   {
    cinfo->progressive_mode = 0;
    cinfo->num_scans = 1;
    }
  if (cinfo->progressive_mode)
    cinfo->optimize_coding = 1;
  if (transcode_only)
   {
    if (cinfo->optimize_coding)
      master->pass_type = huff_opt_pass;
    else
      master->pass_type = output_pass;
    }
  else
   {
    master->pass_type = main_pass;
    }
  master->scan_number = 0;
  master->pass_number = 0;
  if (cinfo->optimize_coding)
    master->total_passes = cinfo->num_scans * 2;
  else
    master->total_passes = cinfo->num_scans;
  }
#define JPEG_INTERNALS
#ifndef D_PROGRESSIVE_SUPPORTED
#endif
typedef struct
 {
  struct jpeg_d_coef_controller pub;
  JDIMENSION MCU_ctr;
  int MCU_vert_offset;
  int MCU_rows_per_iMCU_row;
  JBLOCKROW MCU_buffer[10];
#ifdef D_MULTISCAN_FILES_SUPPORTED
  jvirt_barray_ptr whole_image[10];
#endif
#ifdef BLOCK_SMOOTHING_SUPPORTED
  int *coef_bits_latch;
#define SAVED_COEFS  6
#endif
  } my_coef_controller1;
typedef my_coef_controller1 *my_coef_ptr1;
static int decompress_onepass
  (j_decompress_ptr cinfo, JSAMPIMAGE output_buf);
#ifdef D_MULTISCAN_FILES_SUPPORTED
static int decompress_data
  (j_decompress_ptr cinfo, JSAMPIMAGE output_buf);
#endif
#ifdef BLOCK_SMOOTHING_SUPPORTED
static boolean smoothing_ok(j_decompress_ptr cinfo);
static int decompress_smooth_data
  (j_decompress_ptr cinfo, JSAMPIMAGE output_buf);
#endif
static void
  start_iMCU_row1(j_decompress_ptr cinfo)
 {
  my_coef_ptr1 coef = (my_coef_ptr1) cinfo->coef;
  if (cinfo->comps_in_scan > 1)
   {
    coef->MCU_rows_per_iMCU_row = 1;
    }
  else
   {
    if (cinfo->input_iMCU_row < (cinfo->total_iMCU_rows - 1))
      coef->MCU_rows_per_iMCU_row = cinfo->cur_comp_info[0]->v_samp_factor;
    else
      coef->MCU_rows_per_iMCU_row = cinfo->cur_comp_info[0]->last_row_height;
    }
  coef->MCU_ctr = 0;
  coef->MCU_vert_offset = 0;
  }
static void
  start_input_pass1(j_decompress_ptr cinfo)
 {
  cinfo->input_iMCU_row = 0;
  start_iMCU_row1(cinfo);
  }
static void
  start_output_pass(j_decompress_ptr cinfo)
 {
#ifdef BLOCK_SMOOTHING_SUPPORTED
  my_coef_ptr1 coef = (my_coef_ptr1) cinfo->coef;
  if (coef->pub.coef_arrays != ((void *) 0))
   {
    if (cinfo->do_block_smoothing && smoothing_ok(cinfo))
      coef->pub.decompress_data = decompress_smooth_data;
    else
      coef->pub.decompress_data = decompress_data;
    }
#endif
  cinfo->output_iMCU_row = 0;
  }
static int
  decompress_onepass(j_decompress_ptr cinfo, JSAMPIMAGE output_buf)
 {
  my_coef_ptr1 coef = (my_coef_ptr1) cinfo->coef;
  JDIMENSION MCU_col_num;
  JDIMENSION last_MCU_col = cinfo->MCUs_per_row - 1;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  int blkn,
    ci,
    xindex,
    yindex,
    yoffset,
    useful_width;
  JSAMPARRAY output_ptr;
  JDIMENSION start_col,
    output_col;
  jpeg_component_info *compptr;
  inverse_DCT_method_ptr inverse_DCT;
  for (yoffset = coef->MCU_vert_offset; yoffset < coef->MCU_rows_per_iMCU_row;
   yoffset++)
   {
    for (MCU_col_num = coef->MCU_ctr; MCU_col_num <= last_MCU_col;
     MCU_col_num++)
     {
      jzero_far((void *) coef->MCU_buffer[0],
       (size_t) (cinfo->blocks_in_MCU * ((size_t) sizeof(JBLOCK))));
      if (!(*cinfo->entropy->decode_mcu) (cinfo, coef->MCU_buffer))
       {
        coef->MCU_vert_offset = yoffset;
        coef->MCU_ctr = MCU_col_num;
        return 0;
        }
      blkn = 0;
      for (ci = 0; ci < cinfo->comps_in_scan; ci++)
       {
        compptr = cinfo->cur_comp_info[ci];
        if (!compptr->component_needed)
         {
          blkn += compptr->MCU_blocks;
          continue;
          }
        inverse_DCT = cinfo->idct->inverse_DCT[compptr->component_index];
        useful_width = (MCU_col_num < last_MCU_col) ? compptr->MCU_width
         : compptr->last_col_width;
        output_ptr = output_buf[compptr->component_index] +
         yoffset * compptr->DCT_scaled_size;
        start_col = MCU_col_num * compptr->MCU_sample_width;
        for (yindex = 0; yindex < compptr->MCU_height; yindex++)
         {
          if (cinfo->input_iMCU_row < last_iMCU_row ||
           yoffset + yindex < compptr->last_row_height)
           {
            output_col = start_col;
            for (xindex = 0; xindex < useful_width; xindex++)
             {
              (*inverse_DCT) (cinfo, compptr,
               (JCOEFPTR) coef->MCU_buffer[blkn + xindex],
               output_ptr, output_col);
              output_col += compptr->DCT_scaled_size;
              }
            }
          blkn += compptr->MCU_width;
          output_ptr += compptr->DCT_scaled_size;
          }
        }
      }
    coef->MCU_ctr = 0;
    }
  cinfo->output_iMCU_row++;
  if (++(cinfo->input_iMCU_row) < cinfo->total_iMCU_rows)
   {
    start_iMCU_row1(cinfo);
    return 3;
    }
  (*cinfo->inputctl->finish_input_pass) (cinfo);
  return 4;
  }
static int
  dummy_consume_data(j_decompress_ptr cinfo)
 {
  return 0;
  }
#ifdef D_MULTISCAN_FILES_SUPPORTED
static int
  consume_data(j_decompress_ptr cinfo)
 {
  my_coef_ptr1 coef = (my_coef_ptr1) cinfo->coef;
  JDIMENSION MCU_col_num;
  int blkn,
    ci,
    xindex,
    yindex,
    yoffset;
  JDIMENSION start_col;
  JBLOCKARRAY buffer[4];
  JBLOCKROW buffer_ptr;
  jpeg_component_info *compptr;
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    buffer[ci] = (*cinfo->mem->access_virt_barray)
     ((j_common_ptr) cinfo, coef->whole_image[compptr->component_index],
     cinfo->input_iMCU_row * compptr->v_samp_factor,
     (JDIMENSION) compptr->v_samp_factor, 1);
    }
  for (yoffset = coef->MCU_vert_offset; yoffset < coef->MCU_rows_per_iMCU_row;
   yoffset++)
   {
    for (MCU_col_num = coef->MCU_ctr; MCU_col_num < cinfo->MCUs_per_row;
     MCU_col_num++)
     {
      blkn = 0;
      for (ci = 0; ci < cinfo->comps_in_scan; ci++)
       {
        compptr = cinfo->cur_comp_info[ci];
        start_col = MCU_col_num * compptr->MCU_width;
        for (yindex = 0; yindex < compptr->MCU_height; yindex++)
         {
          buffer_ptr = buffer[ci][yindex + yoffset] + start_col;
          for (xindex = 0; xindex < compptr->MCU_width; xindex++)
           {
            coef->MCU_buffer[blkn++] = buffer_ptr++;
            }
          }
        }
      if (!(*cinfo->entropy->decode_mcu) (cinfo, coef->MCU_buffer))
       {
        coef->MCU_vert_offset = yoffset;
        coef->MCU_ctr = MCU_col_num;
        return 0;
        }
      }
    coef->MCU_ctr = 0;
    }
  if (++(cinfo->input_iMCU_row) < cinfo->total_iMCU_rows)
   {
    start_iMCU_row1(cinfo);
    return 3;
    }
  (*cinfo->inputctl->finish_input_pass) (cinfo);
  return 4;
  }
static int
  decompress_data(j_decompress_ptr cinfo, JSAMPIMAGE output_buf)
 {
  my_coef_ptr1 coef = (my_coef_ptr1) cinfo->coef;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  JDIMENSION block_num;
  int ci,
    block_row,
    block_rows;
  JBLOCKARRAY buffer;
  JBLOCKROW buffer_ptr;
  JSAMPARRAY output_ptr;
  JDIMENSION output_col;
  jpeg_component_info *compptr;
  inverse_DCT_method_ptr inverse_DCT;
  while (cinfo->input_scan_number < cinfo->output_scan_number ||
   (cinfo->input_scan_number == cinfo->output_scan_number &&
    cinfo->input_iMCU_row <= cinfo->output_iMCU_row))
   {
    if ((*cinfo->inputctl->consume_input) (cinfo) == 0)
      return 0;
    }
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    if (!compptr->component_needed)
      continue;
    buffer = (*cinfo->mem->access_virt_barray)
     ((j_common_ptr) cinfo, coef->whole_image[ci],
     cinfo->output_iMCU_row * compptr->v_samp_factor,
     (JDIMENSION) compptr->v_samp_factor, 0);
    if (cinfo->output_iMCU_row < last_iMCU_row)
      block_rows = compptr->v_samp_factor;
    else
     {
      block_rows = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
      if (block_rows == 0)
        block_rows = compptr->v_samp_factor;
      }
    inverse_DCT = cinfo->idct->inverse_DCT[ci];
    output_ptr = output_buf[ci];
    for (block_row = 0; block_row < block_rows; block_row++)
     {
      buffer_ptr = buffer[block_row];
      output_col = 0;
      for (block_num = 0; block_num < compptr->width_in_blocks; block_num++)
       {
        (*inverse_DCT) (cinfo, compptr, (JCOEFPTR) buffer_ptr,
         output_ptr, output_col);
        buffer_ptr++;
        output_col += compptr->DCT_scaled_size;
        }
      output_ptr += compptr->DCT_scaled_size;
      }
    }
  if (++(cinfo->output_iMCU_row) < cinfo->total_iMCU_rows)
    return 3;
  return 4;
  }
#endif
#ifdef BLOCK_SMOOTHING_SUPPORTED
#define Q01_POS  1
#define Q10_POS  8
#define Q20_POS  16
#define Q11_POS  9
#define Q02_POS  2
static boolean
  smoothing_ok(j_decompress_ptr cinfo)
 {
  my_coef_ptr1 coef = (my_coef_ptr1) cinfo->coef;
  boolean smoothing_useful = 0;
  int ci,
    coefi;
  jpeg_component_info *compptr;
  JQUANT_TBL *qtable;
  int *coef_bits;
  int *coef_bits_latch;
  if (!cinfo->progressive_mode || cinfo->coef_bits == ((void *) 0))
    return 0;
  if (coef->coef_bits_latch == ((void *) 0))
    coef->coef_bits_latch = (int *)
     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
     cinfo->num_components *
     (6 * ((size_t) sizeof(int))));
  coef_bits_latch = coef->coef_bits_latch;
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    if ((qtable = compptr->quant_table) == ((void *) 0))
      return 0;
    if (qtable->quantval[0] == 0 ||
     qtable->quantval[1] == 0 ||
     qtable->quantval[8] == 0 ||
     qtable->quantval[16] == 0 ||
     qtable->quantval[9] == 0 ||
     qtable->quantval[2] == 0)
      return 0;
    coef_bits = cinfo->coef_bits[ci];
    if (coef_bits[0] < 0)
      return 0;
    for (coefi = 1; coefi <= 5; coefi++)
     {
      coef_bits_latch[coefi] = coef_bits[coefi];
      if (coef_bits[coefi] != 0)
        smoothing_useful = 1;
      }
    coef_bits_latch += 6;
    }
  return smoothing_useful;
  }
static int
  decompress_smooth_data(j_decompress_ptr cinfo, JSAMPIMAGE output_buf)
 {
  my_coef_ptr1 coef = (my_coef_ptr1) cinfo->coef;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  JDIMENSION block_num,
    last_block_column;
  int ci,
    block_row,
    block_rows,
    access_rows;
  JBLOCKARRAY buffer;
  JBLOCKROW buffer_ptr,
    prev_block_row,
    next_block_row;
  JSAMPARRAY output_ptr;
  JDIMENSION output_col;
  jpeg_component_info *compptr;
  inverse_DCT_method_ptr inverse_DCT;
  boolean first_row,
    last_row;
  JBLOCK workspace;
  int *coef_bits;
  JQUANT_TBL *quanttbl;
  int Q00,
    Q01,
    Q02,
    Q10,
    Q11,
    Q20,
    num;
  int DC1,
    DC2,
    DC3,
    DC4,
    DC5,
    DC6,
    DC7,
    DC8,
    DC9;
  int Al,
    pred;
  while (cinfo->input_scan_number <= cinfo->output_scan_number &&
   !cinfo->inputctl->eoi_reached)
   {
    if (cinfo->input_scan_number == cinfo->output_scan_number)
     {
      JDIMENSION delta = (cinfo->Ss == 0) ? 1 : 0;
      if (cinfo->input_iMCU_row > cinfo->output_iMCU_row + delta)
        break;
      }
    if ((*cinfo->inputctl->consume_input) (cinfo) == 0)
      return 0;
    }
  for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
   ci++, compptr++)
   {
    if (!compptr->component_needed)
      continue;
    if (cinfo->output_iMCU_row < last_iMCU_row)
     {
      block_rows = compptr->v_samp_factor;
      access_rows = block_rows * 2;
      last_row = 0;
      }
    else
     {
      block_rows = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
      if (block_rows == 0)
        block_rows = compptr->v_samp_factor;
      access_rows = block_rows;
      last_row = 1;
      }
    if (cinfo->output_iMCU_row > 0)
     {
      access_rows += compptr->v_samp_factor;
      buffer = (*cinfo->mem->access_virt_barray)
       ((j_common_ptr) cinfo, coef->whole_image[ci],
       (cinfo->output_iMCU_row - 1) * compptr->v_samp_factor,
       (JDIMENSION) access_rows, 0);
      buffer += compptr->v_samp_factor;
      first_row = 0;
      }
    else
     {
      buffer = (*cinfo->mem->access_virt_barray)
       ((j_common_ptr) cinfo, coef->whole_image[ci],
       (JDIMENSION) 0, (JDIMENSION) access_rows, 0);
      first_row = 1;
      }
    coef_bits = coef->coef_bits_latch + (ci * 6);
    quanttbl = compptr->quant_table;
    Q00 = quanttbl->quantval[0];
    Q01 = quanttbl->quantval[1];
    Q10 = quanttbl->quantval[8];
    Q20 = quanttbl->quantval[16];
    Q11 = quanttbl->quantval[9];
    Q02 = quanttbl->quantval[2];
    inverse_DCT = cinfo->idct->inverse_DCT[ci];
    output_ptr = output_buf[ci];
    for (block_row = 0; block_row < block_rows; block_row++)
     {
      buffer_ptr = buffer[block_row];
      if (first_row && block_row == 0)
        prev_block_row = buffer_ptr;
      else
        prev_block_row = buffer[block_row - 1];
      if (last_row && block_row == block_rows - 1)
        next_block_row = buffer_ptr;
      else
        next_block_row = buffer[block_row + 1];
      DC1 = DC2 = DC3 = (int) prev_block_row[0][0];
      DC4 = DC5 = DC6 = (int) buffer_ptr[0][0];
      DC7 = DC8 = DC9 = (int) next_block_row[0][0];
      output_col = 0;
      last_block_column = compptr->width_in_blocks - 1;
      for (block_num = 0; block_num <= last_block_column; block_num++)
       {
        jcopy_block_row(buffer_ptr, (JBLOCKROW) workspace, (JDIMENSION) 1);
        if (block_num < last_block_column)
         {
          DC3 = (int) prev_block_row[1][0];
          DC6 = (int) buffer_ptr[1][0];
          DC9 = (int) next_block_row[1][0];
          }
        if ((Al = coef_bits[1]) != 0 && workspace[1] == 0)
         {
          num = 36 * Q00 * (DC4 - DC6);
          if (num >= 0)
           {
            pred = (int) (((Q01 << 7) + num) / (Q01 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            }
          else
           {
            pred = (int) (((Q01 << 7) - num) / (Q01 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            pred = -pred;
            }
          workspace[1] = (JCOEF) pred;
          }
        if ((Al = coef_bits[2]) != 0 && workspace[8] == 0)
         {
          num = 36 * Q00 * (DC2 - DC8);
          if (num >= 0)
           {
            pred = (int) (((Q10 << 7) + num) / (Q10 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            }
          else
           {
            pred = (int) (((Q10 << 7) - num) / (Q10 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            pred = -pred;
            }
          workspace[8] = (JCOEF) pred;
          }
        if ((Al = coef_bits[3]) != 0 && workspace[16] == 0)
         {
          num = 9 * Q00 * (DC2 + DC8 - 2 * DC5);
          if (num >= 0)
           {
            pred = (int) (((Q20 << 7) + num) / (Q20 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            }
          else
           {
            pred = (int) (((Q20 << 7) - num) / (Q20 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            pred = -pred;
            }
          workspace[16] = (JCOEF) pred;
          }
        if ((Al = coef_bits[4]) != 0 && workspace[9] == 0)
         {
          num = 5 * Q00 * (DC1 - DC3 - DC7 + DC9);
          if (num >= 0)
           {
            pred = (int) (((Q11 << 7) + num) / (Q11 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            }
          else
           {
            pred = (int) (((Q11 << 7) - num) / (Q11 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            pred = -pred;
            }
          workspace[9] = (JCOEF) pred;
          }
        if ((Al = coef_bits[5]) != 0 && workspace[2] == 0)
         {
          num = 9 * Q00 * (DC4 + DC6 - 2 * DC5);
          if (num >= 0)
           {
            pred = (int) (((Q02 << 7) + num) / (Q02 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            }
          else
           {
            pred = (int) (((Q02 << 7) - num) / (Q02 << 8));
            if (Al > 0 && pred >= (1 << Al))
              pred = (1 << Al) - 1;
            pred = -pred;
            }
          workspace[2] = (JCOEF) pred;
          }
        (*inverse_DCT) (cinfo, compptr, (JCOEFPTR) workspace,
         output_ptr, output_col);
        DC1 = DC2;
        DC2 = DC3;
        DC4 = DC5;
        DC5 = DC6;
        DC7 = DC8;
        DC8 = DC9;
        buffer_ptr++, prev_block_row++, next_block_row++;
        output_col += compptr->DCT_scaled_size;
        }
      output_ptr += compptr->DCT_scaled_size;
      }
    }
  if (++(cinfo->output_iMCU_row) < cinfo->total_iMCU_rows)
    return 3;
  return 4;
  }
#endif
void
  jinit_d_coef_controller(j_decompress_ptr cinfo, boolean need_full_buffer)
 {
  my_coef_ptr1 coef;
  coef = (my_coef_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_coef_controller1)));
  cinfo->coef = (struct jpeg_d_coef_controller *) coef;
  coef->pub.start_input_pass = start_input_pass1;
  coef->pub.start_output_pass = start_output_pass;
#ifdef BLOCK_SMOOTHING_SUPPORTED
  coef->coef_bits_latch = ((void *) 0);
#endif
  if (need_full_buffer)
   {
#ifdef D_MULTISCAN_FILES_SUPPORTED
    int ci,
      access_rows;
    jpeg_component_info *compptr;
    for (ci = 0, compptr = cinfo->comp_info; ci < cinfo->num_components;
     ci++, compptr++)
     {
      access_rows = compptr->v_samp_factor;
#ifdef BLOCK_SMOOTHING_SUPPORTED
      if (cinfo->progressive_mode)
        access_rows *= 3;
#endif
      coef->whole_image[ci] = (*cinfo->mem->request_virt_barray)
       ((j_common_ptr) cinfo, 1, 1,
       (JDIMENSION) jround_up((long) compptr->width_in_blocks,
        (long) compptr->h_samp_factor),
       (JDIMENSION) jround_up((long) compptr->height_in_blocks,
        (long) compptr->v_samp_factor),
       (JDIMENSION) access_rows);
      }
    coef->pub.consume_data = consume_data;
    coef->pub.decompress_data = decompress_data;
    coef->pub.coef_arrays = coef->whole_image;
#else
#endif
    }
  else
   {
    JBLOCKROW buffer;
    int i;
    buffer = (JBLOCKROW)
     (*cinfo->mem->alloc_large) ((j_common_ptr) cinfo, 1,
     10 * ((size_t) sizeof(JBLOCK)));
    for (i = 0; i < 10; i++)
     {
      coef->MCU_buffer[i] = buffer + i;
      }
    coef->pub.consume_data = dummy_consume_data;
    coef->pub.decompress_data = decompress_onepass;
    coef->pub.coef_arrays = ((void *) 0);
    }
  }
#define JPEG_INTERNALS
#ifdef D_PROGRESSIVE_SUPPORTED
typedef struct
 {
  unsigned int EOBRUN;
  int last_dc_val[4];
  } savable_state2;
#ifndef NO_STRUCT_ASSIGN
#define ASSIGN_STATE(dest,src)  ((dest) = (src))
#else
#endif
typedef struct
 {
  struct jpeg_entropy_decoder pub;
  bitread_perm_state bitstate;
  savable_state2 saved;
  unsigned int restarts_to_go;
  d_derived_tbl *derived_tbls[4];
  d_derived_tbl *ac_derived_tbl;
  } phuff_entropy_decoder;
typedef phuff_entropy_decoder *phuff_entropy_ptr1;
static boolean decode_mcu_DC_first(j_decompress_ptr cinfo, JBLOCKROW * MCU_data);
static boolean decode_mcu_AC_first(j_decompress_ptr cinfo, JBLOCKROW * MCU_data);
static boolean decode_mcu_DC_refine(j_decompress_ptr cinfo, JBLOCKROW * MCU_data);
static boolean decode_mcu_AC_refine(j_decompress_ptr cinfo, JBLOCKROW * MCU_data);
static void
  start_pass_phuff_decoder(j_decompress_ptr cinfo)
 {
  phuff_entropy_ptr1 entropy = (phuff_entropy_ptr1) cinfo->entropy;
  boolean is_DC_band,
    bad;
  int ci,
    coefi,
    tbl;
  int *coef_bit_ptr;
  jpeg_component_info *compptr;
  is_DC_band = (cinfo->Ss == 0);
  bad = 0;
  if (is_DC_band)
   {
    if (cinfo->Se != 0)
      bad = 1;
    }
  else
   {
    if (cinfo->Ss > cinfo->Se || cinfo->Se >= 64)
      bad = 1;
    if (cinfo->comps_in_scan != 1)
      bad = 1;
    }
  if (cinfo->Ah != 0)
   {
    if (cinfo->Al != cinfo->Ah - 1)
      bad = 1;
    }
  if (cinfo->Al > 13)
    bad = 1;
  if (bad)
    ((cinfo)->err->msg_code = (JERR_BAD_PROGRESSION), (cinfo)->err->msg_parm.i[0] = (cinfo->Ss), (cinfo)->err->msg_parm.i[1] = (cinfo->Se), (cinfo)->err->msg_parm.i[2] = (cinfo->Ah), (cinfo)->err->msg_parm.i[3] = (cinfo->Al), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    int cindex = cinfo->cur_comp_info[ci]->component_index;
    coef_bit_ptr = &cinfo->coef_bits[cindex][0];
    if (!is_DC_band && coef_bit_ptr[0] < 0)
      ((cinfo)->err->msg_code = (JWRN_BOGUS_PROGRESSION), (cinfo)->err->msg_parm.i[0] = (cindex), (cinfo)->err->msg_parm.i[1] = (0), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
    for (coefi = cinfo->Ss; coefi <= cinfo->Se; coefi++)
     {
      int expected = (coef_bit_ptr[coefi] < 0) ? 0 : coef_bit_ptr[coefi];
      if (cinfo->Ah != expected)
        ((cinfo)->err->msg_code = (JWRN_BOGUS_PROGRESSION), (cinfo)->err->msg_parm.i[0] = (cindex), (cinfo)->err->msg_parm.i[1] = (coefi), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
      coef_bit_ptr[coefi] = cinfo->Al;
      }
    }
  if (cinfo->Ah == 0)
   {
    if (is_DC_band)
      entropy->pub.decode_mcu = decode_mcu_DC_first;
    else
      entropy->pub.decode_mcu = decode_mcu_AC_first;
    }
  else
   {
    if (is_DC_band)
      entropy->pub.decode_mcu = decode_mcu_DC_refine;
    else
      entropy->pub.decode_mcu = decode_mcu_AC_refine;
    }
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
   {
    compptr = cinfo->cur_comp_info[ci];
    if (is_DC_band)
     {
      if (cinfo->Ah == 0)
       {
        tbl = compptr->dc_tbl_no;
        jpeg_make_d_derived_tbl(cinfo, 1, tbl,
         &entropy->derived_tbls[tbl]);
        }
      }
    else
     {
      tbl = compptr->ac_tbl_no;
      jpeg_make_d_derived_tbl(cinfo, 0, tbl,
       &entropy->derived_tbls[tbl]);
      entropy->ac_derived_tbl = entropy->derived_tbls[tbl];
      }
    entropy->saved.last_dc_val[ci] = 0;
    }
  entropy->bitstate.bits_left = 0;
  entropy->bitstate.get_buffer = 0;
  entropy->pub.insufficient_data = 0;
  entropy->saved.EOBRUN = 0;
  entropy->restarts_to_go = cinfo->restart_interval;
  }
#ifdef AVOID_TABLES
#else
#undef HUFF_EXTEND
#define HUFF_EXTEND(x,s)  ((x) < extend_test1[s] ? (x) + extend_offset1[s] : (x))
static const int extend_test1[16] =
 {0, 0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080,
  0x0100, 0x0200, 0x0400, 0x0800, 0x1000, 0x2000, 0x4000};
static const int extend_offset1[16] =
 {0, ((-1) << 1) + 1, ((-1) << 2) + 1, ((-1) << 3) + 1, ((-1) << 4) + 1,
  ((-1) << 5) + 1, ((-1) << 6) + 1, ((-1) << 7) + 1, ((-1) << 8) + 1,
  ((-1) << 9) + 1, ((-1) << 10) + 1, ((-1) << 11) + 1, ((-1) << 12) + 1,
  ((-1) << 13) + 1, ((-1) << 14) + 1, ((-1) << 15) + 1};
#endif
static boolean
  process_restart1(j_decompress_ptr cinfo)
 {
  phuff_entropy_ptr1 entropy = (phuff_entropy_ptr1) cinfo->entropy;
  int ci;
  cinfo->marker->discarded_bytes += entropy->bitstate.bits_left / 8;
  entropy->bitstate.bits_left = 0;
  if (!(*cinfo->marker->read_restart_marker) (cinfo))
    return 0;
  for (ci = 0; ci < cinfo->comps_in_scan; ci++)
    entropy->saved.last_dc_val[ci] = 0;
  entropy->saved.EOBRUN = 0;
  entropy->restarts_to_go = cinfo->restart_interval;
  if (cinfo->unread_marker == 0)
    entropy->pub.insufficient_data = 0;
  return 1;
  }
static boolean
  decode_mcu_DC_first(j_decompress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  phuff_entropy_ptr1 entropy = (phuff_entropy_ptr1) cinfo->entropy;
  int Al = cinfo->Al;
  register int s,
    r;
  int blkn,
    ci;
  JBLOCKROW block;
  register bit_buf_type get_buffer;
  register int bits_left;
  bitread_working_state br_state;
  savable_state2 state;
  d_derived_tbl *tbl;
  jpeg_component_info *compptr;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
      if (!process_restart1(cinfo))
        return 0;
    }
  if (!entropy->pub.insufficient_data)
   {
    br_state.cinfo = cinfo;
    br_state.next_input_byte = cinfo->src->next_input_byte;
    br_state.bytes_in_buffer = cinfo->src->bytes_in_buffer;
    get_buffer = entropy->bitstate.get_buffer;
    bits_left = entropy->bitstate.bits_left;;
    ((state) = (entropy->saved));
    for (blkn = 0; blkn < cinfo->blocks_in_MCU; blkn++)
     {
      block = MCU_data[blkn];
      ci = cinfo->MCU_membership[blkn];
      compptr = cinfo->cur_comp_info[ci];
      tbl = entropy->derived_tbls[compptr->dc_tbl_no];
       {
        register int nb,
          look;
        if (bits_left < 8)
         {
          if (!jpeg_fill_bit_buffer(&br_state, get_buffer, bits_left, 0))
           {
            return 0;
            } get_buffer = br_state.get_buffer;
          bits_left = br_state.bits_left;
          if (bits_left < 8)
           {
            nb = 1;
            goto label1;
            }
          } look = (((int) (get_buffer >> (bits_left - (8)))) & ((1 << (8)) - 1));
        if ((nb = tbl->look_nbits[look]) != 0)
         {
          (bits_left -= (nb));
          s = tbl->look_sym[look];
          }
        else
         {
          nb = 8 + 1;
      label1:if ((s = jpeg_huff_decode(&br_state, get_buffer, bits_left, tbl, nb)) < 0)
           {
            return 0;
            } get_buffer = br_state.get_buffer;
          bits_left = br_state.bits_left;
          }
        };
      if (s)
       {
         {
          if (bits_left < (s))
           {
            if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, s))
             {
              return 0;
              } get_buffer = (br_state).get_buffer;
            bits_left = (br_state).bits_left;
            }
          };
        r = (((int) (get_buffer >> (bits_left -= (s)))) & ((1 << (s)) - 1));
        s = ((r) < extend_test1[s] ? (r) + extend_offset1[s] : (r));
        }
      s += state.last_dc_val[ci];
      state.last_dc_val[ci] = s;
      (*block)[0] = (JCOEF) (s << Al);
      }
    cinfo->src->next_input_byte = br_state.next_input_byte;
    cinfo->src->bytes_in_buffer = br_state.bytes_in_buffer;
    entropy->bitstate.get_buffer = get_buffer;
    entropy->bitstate.bits_left = bits_left;
    ((entropy->saved) = (state));
    }
  entropy->restarts_to_go--;
  return 1;
  }
static boolean
  decode_mcu_AC_first(j_decompress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  phuff_entropy_ptr1 entropy = (phuff_entropy_ptr1) cinfo->entropy;
  int Se = cinfo->Se;
  int Al = cinfo->Al;
  register int s,
    k,
    r;
  unsigned int EOBRUN;
  JBLOCKROW block;
  register bit_buf_type get_buffer;
  register int bits_left;
  bitread_working_state br_state;
  d_derived_tbl *tbl;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
      if (!process_restart1(cinfo))
        return 0;
    }
  if (!entropy->pub.insufficient_data)
   {
    EOBRUN = entropy->saved.EOBRUN;
    if (EOBRUN > 0)
      EOBRUN--;
    else
     {
      br_state.cinfo = cinfo;
      br_state.next_input_byte = cinfo->src->next_input_byte;
      br_state.bytes_in_buffer = cinfo->src->bytes_in_buffer;
      get_buffer = entropy->bitstate.get_buffer;
      bits_left = entropy->bitstate.bits_left;;
      block = MCU_data[0];
      tbl = entropy->ac_derived_tbl;
      for (k = cinfo->Ss; k <= Se; k++)
       {
         {
          register int nb,
            look;
          if (bits_left < 8)
           {
            if (!jpeg_fill_bit_buffer(&br_state, get_buffer, bits_left, 0))
             {
              return 0;
              } get_buffer = br_state.get_buffer;
            bits_left = br_state.bits_left;
            if (bits_left < 8)
             {
              nb = 1;
              goto label2;
              }
            } look = (((int) (get_buffer >> (bits_left - (8)))) & ((1 << (8)) - 1));
          if ((nb = tbl->look_nbits[look]) != 0)
           {
            (bits_left -= (nb));
            s = tbl->look_sym[look];
            }
          else
           {
            nb = 8 + 1;
        label2:if ((s = jpeg_huff_decode(&br_state, get_buffer, bits_left, tbl, nb)) < 0)
             {
              return 0;
              } get_buffer = br_state.get_buffer;
            bits_left = br_state.bits_left;
            }
          };
        r = s >> 4;
        s &= 15;
        if (s)
         {
          k += r;
           {
            if (bits_left < (s))
             {
              if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, s))
               {
                return 0;
                } get_buffer = (br_state).get_buffer;
              bits_left = (br_state).bits_left;
              }
            };
          r = (((int) (get_buffer >> (bits_left -= (s)))) & ((1 << (s)) - 1));
          s = ((r) < extend_test1[s] ? (r) + extend_offset1[s] : (r));
          (*block)[jpeg_natural_order[k]] = (JCOEF) (s << Al);
          }
        else
         {
          if (r == 15)
           {
            k += 15;
            }
          else
           {
            EOBRUN = 1 << r;
            if (r)
             {
               {
                if (bits_left < (r))
                 {
                  if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, r))
                   {
                    return 0;
                    } get_buffer = (br_state).get_buffer;
                  bits_left = (br_state).bits_left;
                  }
                };
              r = (((int) (get_buffer >> (bits_left -= (r)))) & ((1 << (r)) - 1));
              EOBRUN += r;
              }
            EOBRUN--;
            break;
            }
          }
        }
      cinfo->src->next_input_byte = br_state.next_input_byte;
      cinfo->src->bytes_in_buffer = br_state.bytes_in_buffer;
      entropy->bitstate.get_buffer = get_buffer;
      entropy->bitstate.bits_left = bits_left;
      }
    entropy->saved.EOBRUN = EOBRUN;
    }
  entropy->restarts_to_go--;
  return 1;
  }
static boolean
  decode_mcu_DC_refine(j_decompress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  phuff_entropy_ptr1 entropy = (phuff_entropy_ptr1) cinfo->entropy;
  int p1 = 1 << cinfo->Al;
  int blkn;
  JBLOCKROW block;
  register bit_buf_type get_buffer;
  register int bits_left;
  bitread_working_state br_state;
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
      if (!process_restart1(cinfo))
        return 0;
    }
  br_state.cinfo = cinfo;
  br_state.next_input_byte = cinfo->src->next_input_byte;
  br_state.bytes_in_buffer = cinfo->src->bytes_in_buffer;
  get_buffer = entropy->bitstate.get_buffer;
  bits_left = entropy->bitstate.bits_left;;
  for (blkn = 0; blkn < cinfo->blocks_in_MCU; blkn++)
   {
    block = MCU_data[blkn];
     {
      if (bits_left < (1))
       {
        if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, 1))
         {
          return 0;
          } get_buffer = (br_state).get_buffer;
        bits_left = (br_state).bits_left;
        }
      };
    if ((((int) (get_buffer >> (bits_left -= (1)))) & ((1 << (1)) - 1)))
      (*block)[0] |= p1;
    }
  cinfo->src->next_input_byte = br_state.next_input_byte;
  cinfo->src->bytes_in_buffer = br_state.bytes_in_buffer;
  entropy->bitstate.get_buffer = get_buffer;
  entropy->bitstate.bits_left = bits_left;
  entropy->restarts_to_go--;
  return 1;
  }
static boolean
  decode_mcu_AC_refine(j_decompress_ptr cinfo, JBLOCKROW * MCU_data)
 {
  phuff_entropy_ptr1 entropy = (phuff_entropy_ptr1) cinfo->entropy;
  int Se = cinfo->Se;
  int p1 = 1 << cinfo->Al;
  int m1 = (-1) << cinfo->Al;
  register int s,
    k,
    r;
  unsigned int EOBRUN;
  JBLOCKROW block;
  JCOEFPTR thiscoef;
  register bit_buf_type get_buffer;
  register int bits_left;
  bitread_working_state br_state;
  d_derived_tbl *tbl;
  int num_newnz;
  int newnz_pos[64];
  if (cinfo->restart_interval)
   {
    if (entropy->restarts_to_go == 0)
      if (!process_restart1(cinfo))
        return 0;
    }
  if (!entropy->pub.insufficient_data)
   {
    br_state.cinfo = cinfo;
    br_state.next_input_byte = cinfo->src->next_input_byte;
    br_state.bytes_in_buffer = cinfo->src->bytes_in_buffer;
    get_buffer = entropy->bitstate.get_buffer;
    bits_left = entropy->bitstate.bits_left;;
    EOBRUN = entropy->saved.EOBRUN;
    block = MCU_data[0];
    tbl = entropy->ac_derived_tbl;
    num_newnz = 0;
    k = cinfo->Ss;
    if (EOBRUN == 0)
     {
      for (; k <= Se; k++)
       {
         {
          register int nb,
            look;
          if (bits_left < 8)
           {
            if (!jpeg_fill_bit_buffer(&br_state, get_buffer, bits_left, 0))
             {
              goto undoit;
              } get_buffer = br_state.get_buffer;
            bits_left = br_state.bits_left;
            if (bits_left < 8)
             {
              nb = 1;
              goto label3;
              }
            } look = (((int) (get_buffer >> (bits_left - (8)))) & ((1 << (8)) - 1));
          if ((nb = tbl->look_nbits[look]) != 0)
           {
            (bits_left -= (nb));
            s = tbl->look_sym[look];
            }
          else
           {
            nb = 8 + 1;
        label3:if ((s = jpeg_huff_decode(&br_state, get_buffer, bits_left, tbl, nb)) < 0)
             {
              goto undoit;
              } get_buffer = br_state.get_buffer;
            bits_left = br_state.bits_left;
            }
          };
        r = s >> 4;
        s &= 15;
        if (s)
         {
          if (s != 1)
            ((cinfo)->err->msg_code = (JWRN_HUFF_BAD_CODE), (*(cinfo)->err->emit_message) ((j_common_ptr) (cinfo), -1));
           {
            if (bits_left < (1))
             {
              if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, 1))
               {
                goto undoit;
                } get_buffer = (br_state).get_buffer;
              bits_left = (br_state).bits_left;
              }
            };
          if ((((int) (get_buffer >> (bits_left -= (1)))) & ((1 << (1)) - 1)))
            s = p1;
          else
            s = m1;
          }
        else
         {
          if (r != 15)
           {
            EOBRUN = 1 << r;
            if (r)
             {
               {
                if (bits_left < (r))
                 {
                  if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, r))
                   {
                    goto undoit;
                    } get_buffer = (br_state).get_buffer;
                  bits_left = (br_state).bits_left;
                  }
                };
              r = (((int) (get_buffer >> (bits_left -= (r)))) & ((1 << (r)) - 1));
              EOBRUN += r;
              }
            break;
            }
          }
        do
         {
          thiscoef = *block + jpeg_natural_order[k];
          if (*thiscoef != 0)
           {
             {
              if (bits_left < (1))
               {
                if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, 1))
                 {
                  goto undoit;
                  } get_buffer = (br_state).get_buffer;
                bits_left = (br_state).bits_left;
                }
              };
            if ((((int) (get_buffer >> (bits_left -= (1)))) & ((1 << (1)) - 1)))
             {
              if ((*thiscoef & p1) == 0)
               {
                if (*thiscoef >= 0)
                  *thiscoef += p1;
                else
                  *thiscoef += m1;
                }
              }
            }
          else
           {
            if (--r < 0)
              break;
            }
          k++;
          }
        while (k <= Se);
        if (s)
         {
          int pos = jpeg_natural_order[k];
          (*block)[pos] = (JCOEF) s;
          newnz_pos[num_newnz++] = pos;
          }
        }
      }
    if (EOBRUN > 0)
     {
      for (; k <= Se; k++)
       {
        thiscoef = *block + jpeg_natural_order[k];
        if (*thiscoef != 0)
         {
           {
            if (bits_left < (1))
             {
              if (!jpeg_fill_bit_buffer(&(br_state), get_buffer, bits_left, 1))
               {
                goto undoit;
                } get_buffer = (br_state).get_buffer;
              bits_left = (br_state).bits_left;
              }
            };
          if ((((int) (get_buffer >> (bits_left -= (1)))) & ((1 << (1)) - 1)))
           {
            if ((*thiscoef & p1) == 0)
             {
              if (*thiscoef >= 0)
                *thiscoef += p1;
              else
                *thiscoef += m1;
              }
            }
          }
        }
      EOBRUN--;
      }
    cinfo->src->next_input_byte = br_state.next_input_byte;
    cinfo->src->bytes_in_buffer = br_state.bytes_in_buffer;
    entropy->bitstate.get_buffer = get_buffer;
    entropy->bitstate.bits_left = bits_left;
    entropy->saved.EOBRUN = EOBRUN;
    }
  entropy->restarts_to_go--;
  return 1;
undoit:
  while (num_newnz > 0)
    (*block)[newnz_pos[--num_newnz]] = 0;
  return 0;
  }
void
  jinit_phuff_decoder(j_decompress_ptr cinfo)
 {
  phuff_entropy_ptr1 entropy;
  int *coef_bit_ptr;
  int ci,
    i;
  entropy = (phuff_entropy_ptr1)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(phuff_entropy_decoder)));
  cinfo->entropy = (struct jpeg_entropy_decoder *) entropy;
  entropy->pub.start_pass = start_pass_phuff_decoder;
  for (i = 0; i < 4; i++)
   {
    entropy->derived_tbls[i] = ((void *) 0);
    }
  cinfo->coef_bits = (int (*)[64])
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   cinfo->num_components * 64 * ((size_t) sizeof(int)));
  coef_bit_ptr = &cinfo->coef_bits[0][0];
  for (ci = 0; ci < cinfo->num_components; ci++)
    for (i = 0; i < 64; i++)
      *coef_bit_ptr++ = -1;
  }
#endif
#define JPEG_INTERNALS
typedef struct
 {
  struct jpeg_d_post_controller pub;
  jvirt_sarray_ptr whole_image;
  JSAMPARRAY buffer;
  JDIMENSION strip_height;
  JDIMENSION starting_row;
  JDIMENSION next_row;
  } my_post_controller;
typedef my_post_controller *my_post_ptr;
static void post_process_1pass
  (j_decompress_ptr cinfo, JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr, JDIMENSION in_row_groups_avail, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
#ifdef QUANT_2PASS_SUPPORTED
static void post_process_prepass
  (j_decompress_ptr cinfo, JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr, JDIMENSION in_row_groups_avail, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
static void post_process_2pass
  (j_decompress_ptr cinfo, JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr, JDIMENSION in_row_groups_avail, JSAMPARRAY output_buf, JDIMENSION * out_row_ctr, JDIMENSION out_rows_avail);
#endif
static void
  start_pass_dpost(j_decompress_ptr cinfo, J_BUF_MODE pass_mode)
 {
  my_post_ptr post = (my_post_ptr) cinfo->post;
  switch (pass_mode)
   {
   case JBUF_PASS_THRU:
    if (cinfo->quantize_colors)
     {
      post->pub.post_process_data = post_process_1pass;
      if (post->buffer == ((void *) 0))
       {
        post->buffer = (*cinfo->mem->access_virt_sarray)
         ((j_common_ptr) cinfo, post->whole_image,
         (JDIMENSION) 0, post->strip_height, 1);
        }
      }
    else
     {
      post->pub.post_process_data = cinfo->upsample->upsample;
      }
    break;
#ifdef QUANT_2PASS_SUPPORTED
   case JBUF_SAVE_AND_PASS:
    if (post->whole_image == ((void *) 0))
      ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    post->pub.post_process_data = post_process_prepass;
    break;
   case JBUF_CRANK_DEST:
    if (post->whole_image == ((void *) 0))
      ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    post->pub.post_process_data = post_process_2pass;
    break;
#endif
   default:
    ((cinfo)->err->msg_code = (JERR_BAD_BUFFER_MODE), (*(cinfo)->err->error_exit) ((j_common_ptr) (cinfo)));
    break;
    }
  post->starting_row = post->next_row = 0;
  }
static void
  post_process_1pass(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr,
   JDIMENSION in_row_groups_avail,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  my_post_ptr post = (my_post_ptr) cinfo->post;
  JDIMENSION num_rows,
    max_rows;
  max_rows = out_rows_avail - *out_row_ctr;
  if (max_rows > post->strip_height)
    max_rows = post->strip_height;
  num_rows = 0;
  (*cinfo->upsample->upsample) (cinfo,
   input_buf, in_row_group_ctr, in_row_groups_avail,
   post->buffer, &num_rows, max_rows);
  (*cinfo->cquantize->color_quantize) (cinfo,
   post->buffer, output_buf + *out_row_ctr, (int) num_rows);
  *out_row_ctr += num_rows;
  }
#ifdef QUANT_2PASS_SUPPORTED
static void
  post_process_prepass(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr,
   JDIMENSION in_row_groups_avail,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  my_post_ptr post = (my_post_ptr) cinfo->post;
  JDIMENSION old_next_row,
    num_rows;
  if (post->next_row == 0)
   {
    post->buffer = (*cinfo->mem->access_virt_sarray)
     ((j_common_ptr) cinfo, post->whole_image,
     post->starting_row, post->strip_height, 1);
    }
  old_next_row = post->next_row;
  (*cinfo->upsample->upsample) (cinfo,
   input_buf, in_row_group_ctr, in_row_groups_avail,
   post->buffer, &post->next_row, post->strip_height);
  if (post->next_row > old_next_row)
   {
    num_rows = post->next_row - old_next_row;
    (*cinfo->cquantize->color_quantize) (cinfo, post->buffer + old_next_row,
     (JSAMPARRAY) ((void *) 0), (int) num_rows);
    *out_row_ctr += num_rows;
    }
  if (post->next_row >= post->strip_height)
   {
    post->starting_row += post->strip_height;
    post->next_row = 0;
    }
  }
static void
  post_process_2pass(j_decompress_ptr cinfo,
   JSAMPIMAGE input_buf, JDIMENSION * in_row_group_ctr,
   JDIMENSION in_row_groups_avail,
   JSAMPARRAY output_buf, JDIMENSION * out_row_ctr,
   JDIMENSION out_rows_avail)
 {
  my_post_ptr post = (my_post_ptr) cinfo->post;
  JDIMENSION num_rows,
    max_rows;
  if (post->next_row == 0)
   {
    post->buffer = (*cinfo->mem->access_virt_sarray)
     ((j_common_ptr) cinfo, post->whole_image,
     post->starting_row, post->strip_height, 0);
    }
  num_rows = post->strip_height - post->next_row;
  max_rows = out_rows_avail - *out_row_ctr;
  if (num_rows > max_rows)
    num_rows = max_rows;
  max_rows = cinfo->output_height - post->starting_row;
  if (num_rows > max_rows)
    num_rows = max_rows;
  (*cinfo->cquantize->color_quantize) (cinfo,
   post->buffer + post->next_row, output_buf + *out_row_ctr,
   (int) num_rows);
  *out_row_ctr += num_rows;
  post->next_row += num_rows;
  if (post->next_row >= post->strip_height)
   {
    post->starting_row += post->strip_height;
    post->next_row = 0;
    }
  }
#endif
void
  jinit_d_post_controller(j_decompress_ptr cinfo, boolean need_full_buffer)
 {
  my_post_ptr post;
  post = (my_post_ptr)
   (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 1,
   ((size_t) sizeof(my_post_controller)));
  cinfo->post = (struct jpeg_d_post_controller *) post;
  post->pub.start_pass = start_pass_dpost;
  post->whole_image = ((void *) 0);
  post->buffer = ((void *) 0);
  if (cinfo->quantize_colors)
   {
    post->strip_height = (JDIMENSION) cinfo->max_v_samp_factor;
    if (need_full_buffer)
     {
#ifdef QUANT_2PASS_SUPPORTED
      post->whole_image = (*cinfo->mem->request_virt_sarray)
       ((j_common_ptr) cinfo, 1, 0,
       cinfo->output_width * cinfo->out_color_components,
       (JDIMENSION) jround_up((long) cinfo->output_height,
        (long) post->strip_height),
       post->strip_height);
#else
#endif
      }
    else
     {
      post->buffer = (*cinfo->mem->alloc_sarray)
       ((j_common_ptr) cinfo, 1,
       cinfo->output_width * cinfo->out_color_components,
       post->strip_height);
      }
    }
  }
#define JPEG_INTERNALS
#ifdef DCT_FLOAT_SUPPORTED
#if 8  != 8
#endif
void
  jpeg_fdct_float(float *data)
 {
  float tmp0,
    tmp1,
    tmp2,
    tmp3,
    tmp4,
    tmp5,
    tmp6,
    tmp7;
  float tmp10,
    tmp11,
    tmp12,
    tmp13;
  float z1,
    z2,
    z3,
    z4,
    z5,
    z11,
    z13;
  float *dataptr;
  int ctr;
  dataptr = data;
  for (ctr = 8 - 1; ctr >= 0; ctr--)
   {
    tmp0 = dataptr[0] + dataptr[7];
    tmp7 = dataptr[0] - dataptr[7];
    tmp1 = dataptr[1] + dataptr[6];
    tmp6 = dataptr[1] - dataptr[6];
    tmp2 = dataptr[2] + dataptr[5];
    tmp5 = dataptr[2] - dataptr[5];
    tmp3 = dataptr[3] + dataptr[4];
    tmp4 = dataptr[3] - dataptr[4];
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    dataptr[0] = tmp10 + tmp11;
    dataptr[4] = tmp10 - tmp11;
    z1 = (tmp12 + tmp13) * ((float) 0.707106781);
    dataptr[2] = tmp13 + z1;
    dataptr[6] = tmp13 - z1;
    tmp10 = tmp4 + tmp5;
    tmp11 = tmp5 + tmp6;
    tmp12 = tmp6 + tmp7;
    z5 = (tmp10 - tmp12) * ((float) 0.382683433);
    z2 = ((float) 0.541196100) * tmp10 + z5;
    z4 = ((float) 1.306562965) * tmp12 + z5;
    z3 = tmp11 * ((float) 0.707106781);
    z11 = tmp7 + z3;
    z13 = tmp7 - z3;
    dataptr[5] = z13 + z2;
    dataptr[3] = z13 - z2;
    dataptr[1] = z11 + z4;
    dataptr[7] = z11 - z4;
    dataptr += 8;
    }
  dataptr = data;
  for (ctr = 8 - 1; ctr >= 0; ctr--)
   {
    tmp0 = dataptr[8 * 0] + dataptr[8 * 7];
    tmp7 = dataptr[8 * 0] - dataptr[8 * 7];
    tmp1 = dataptr[8 * 1] + dataptr[8 * 6];
    tmp6 = dataptr[8 * 1] - dataptr[8 * 6];
    tmp2 = dataptr[8 * 2] + dataptr[8 * 5];
    tmp5 = dataptr[8 * 2] - dataptr[8 * 5];
    tmp3 = dataptr[8 * 3] + dataptr[8 * 4];
    tmp4 = dataptr[8 * 3] - dataptr[8 * 4];
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    dataptr[8 * 0] = tmp10 + tmp11;
    dataptr[8 * 4] = tmp10 - tmp11;
    z1 = (tmp12 + tmp13) * ((float) 0.707106781);
    dataptr[8 * 2] = tmp13 + z1;
    dataptr[8 * 6] = tmp13 - z1;
    tmp10 = tmp4 + tmp5;
    tmp11 = tmp5 + tmp6;
    tmp12 = tmp6 + tmp7;
    z5 = (tmp10 - tmp12) * ((float) 0.382683433);
    z2 = ((float) 0.541196100) * tmp10 + z5;
    z4 = ((float) 1.306562965) * tmp12 + z5;
    z3 = tmp11 * ((float) 0.707106781);
    z11 = tmp7 + z3;
    z13 = tmp7 - z3;
    dataptr[8 * 5] = z13 + z2;
    dataptr[8 * 3] = z13 - z2;
    dataptr[8 * 1] = z11 + z4;
    dataptr[8 * 7] = z11 - z4;
    dataptr++;
    }
  }
#endif
#define JPEG_INTERNALS
#ifdef DCT_IFAST_SUPPORTED
#if 8  != 8
#endif
#undef CONST_BITS
#define CONST_BITS  8
#if 8  == 8
#undef FIX_0_382683433
#undef FIX_0_541196100
#undef FIX_0_707106781
#undef FIX_1_306562965

#define FIX_0_382683433 98
#define FIX_0_541196100 139
#define FIX_0_707106781 181
#define FIX_1_306562965 334
#else
#endif
#ifndef USE_ACCURATE_ROUNDING
#undef DESCALE
#define DESCALE(x,n)  RIGHT_SHIFT(x, n)
#endif
#undef MULTIPLY
#define MULTIPLY(var,const)  ((DCTELEM) DESCALE((var) * (const), CONST_BITS))
void
  jpeg_fdct_ifast(DCTELEM * data)
 {
  DCTELEM tmp0,
    tmp1,
    tmp2,
    tmp3,
    tmp4,
    tmp5,
    tmp6,
    tmp7;
  DCTELEM tmp10,
    tmp11,
    tmp12,
    tmp13;
  DCTELEM z1,
    z2,
    z3,
    z4,
    z5,
    z11,
    z13;
  DCTELEM *dataptr;
  int ctr;
  dataptr = data;
  for (ctr = 8 - 1; ctr >= 0; ctr--)
   {
    tmp0 = dataptr[0] + dataptr[7];
    tmp7 = dataptr[0] - dataptr[7];
    tmp1 = dataptr[1] + dataptr[6];
    tmp6 = dataptr[1] - dataptr[6];
    tmp2 = dataptr[2] + dataptr[5];
    tmp5 = dataptr[2] - dataptr[5];
    tmp3 = dataptr[3] + dataptr[4];
    tmp4 = dataptr[3] - dataptr[4];
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    dataptr[0] = tmp10 + tmp11;
    dataptr[4] = tmp10 - tmp11;
    z1 = ((DCTELEM) (((tmp12 + tmp13) * (((int) 181))) >> (8)));
    dataptr[2] = tmp13 + z1;
    dataptr[6] = tmp13 - z1;
    tmp10 = tmp4 + tmp5;
    tmp11 = tmp5 + tmp6;
    tmp12 = tmp6 + tmp7;
    z5 = ((DCTELEM) (((tmp10 - tmp12) * (((int) 98))) >> (8)));
    z2 = ((DCTELEM) (((tmp10) * (((int) 139))) >> (8))) + z5;
    z4 = ((DCTELEM) (((tmp12) * (((int) 334))) >> (8))) + z5;
    z3 = ((DCTELEM) (((tmp11) * (((int) 181))) >> (8)));
    z11 = tmp7 + z3;
    z13 = tmp7 - z3;
    dataptr[5] = z13 + z2;
    dataptr[3] = z13 - z2;
    dataptr[1] = z11 + z4;
    dataptr[7] = z11 - z4;
    dataptr += 8;
    }
  dataptr = data;
  for (ctr = 8 - 1; ctr >= 0; ctr--)
   {
    tmp0 = dataptr[8 * 0] + dataptr[8 * 7];
    tmp7 = dataptr[8 * 0] - dataptr[8 * 7];
    tmp1 = dataptr[8 * 1] + dataptr[8 * 6];
    tmp6 = dataptr[8 * 1] - dataptr[8 * 6];
    tmp2 = dataptr[8 * 2] + dataptr[8 * 5];
    tmp5 = dataptr[8 * 2] - dataptr[8 * 5];
    tmp3 = dataptr[8 * 3] + dataptr[8 * 4];
    tmp4 = dataptr[8 * 3] - dataptr[8 * 4];
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    dataptr[8 * 0] = tmp10 + tmp11;
    dataptr[8 * 4] = tmp10 - tmp11;
    z1 = ((DCTELEM) (((tmp12 + tmp13) * (((int) 181))) >> (8)));
    dataptr[8 * 2] = tmp13 + z1;
    dataptr[8 * 6] = tmp13 - z1;
    tmp10 = tmp4 + tmp5;
    tmp11 = tmp5 + tmp6;
    tmp12 = tmp6 + tmp7;
    z5 = ((DCTELEM) (((tmp10 - tmp12) * (((int) 98))) >> (8)));
    z2 = ((DCTELEM) (((tmp10) * (((int) 139))) >> (8))) + z5;
    z4 = ((DCTELEM) (((tmp12) * (((int) 334))) >> (8))) + z5;
    z3 = ((DCTELEM) (((tmp11) * (((int) 181))) >> (8)));
    z11 = tmp7 + z3;
    z13 = tmp7 - z3;
    dataptr[8 * 5] = z13 + z2;
    dataptr[8 * 3] = z13 - z2;
    dataptr[8 * 1] = z11 + z4;
    dataptr[8 * 7] = z11 - z4;
    dataptr++;
    }
  }
#endif
#define JPEG_INTERNALS
#ifdef DCT_ISLOW_SUPPORTED
#if 8  != 8
#endif
#if 8  == 8
#undef CONST_BITS
#define CONST_BITS  13
#define PASS1_BITS  2
#else
#endif
#if 13  == 13
#undef FIX_0_298631336
#undef FIX_0_390180644
#undef FIX_0_541196100
#undef FIX_0_765366865
#undef FIX_0_899976223
#undef FIX_1_175875602
#undef FIX_1_501321110
#undef FIX_1_847759065
#undef FIX_1_961570560
#undef FIX_2_053119869
#undef FIX_2_562915447
#undef FIX_3_072711026

#define FIX_0_298631336 2446
#define FIX_0_390180644 3196
#define FIX_0_541196100 4433
#define FIX_0_765366865 6270
#define FIX_0_899976223 7373
#define FIX_1_175875602 9633
#define FIX_1_501321110 12299
#define FIX_1_847759065 15137
#define FIX_1_961570560 16069
#define FIX_2_053119869 16819
#define FIX_2_562915447 20995
#define FIX_3_072711026 25172
#else
#endif
#if 8  == 8
#undef MULTIPLY
#define MULTIPLY(var,const)  MULTIPLY16C16(var,const)
#else
#endif
void
  jpeg_fdct_islow(DCTELEM * data)
 {
  int tmp0,
    tmp1,
    tmp2,
    tmp3,
    tmp4,
    tmp5,
    tmp6,
    tmp7;
  int tmp10,
    tmp11,
    tmp12,
    tmp13;
  int z1,
    z2,
    z3,
    z4,
    z5;
  DCTELEM *dataptr;
  int ctr;
  dataptr = data;
  for (ctr = 8 - 1; ctr >= 0; ctr--)
   {
    tmp0 = dataptr[0] + dataptr[7];
    tmp7 = dataptr[0] - dataptr[7];
    tmp1 = dataptr[1] + dataptr[6];
    tmp6 = dataptr[1] - dataptr[6];
    tmp2 = dataptr[2] + dataptr[5];
    tmp5 = dataptr[2] - dataptr[5];
    tmp3 = dataptr[3] + dataptr[4];
    tmp4 = dataptr[3] - dataptr[4];
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    dataptr[0] = (DCTELEM) ((tmp10 + tmp11) << 2);
    dataptr[4] = (DCTELEM) ((tmp10 - tmp11) << 2);
    z1 = ((tmp12 + tmp13) * (((int) 4433)));
    dataptr[2] = (DCTELEM) ((z1 + ((tmp13) * (((int) 6270)))) >> (13 - 2));
    dataptr[6] = (DCTELEM) ((z1 + ((tmp12) * (-((int) 15137)))) >> (13 - 2));
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;
    z5 = ((z3 + z4) * (((int) 9633)));
    tmp4 = ((tmp4) * (((int) 2446)));
    tmp5 = ((tmp5) * (((int) 16819)));
    tmp6 = ((tmp6) * (((int) 25172)));
    tmp7 = ((tmp7) * (((int) 12299)));
    z1 = ((z1) * (-((int) 7373)));
    z2 = ((z2) * (-((int) 20995)));
    z3 = ((z3) * (-((int) 16069)));
    z4 = ((z4) * (-((int) 3196)));
    z3 += z5;
    z4 += z5;
    dataptr[7] = (DCTELEM) ((tmp4 + z1 + z3) >> (13 - 2));
    dataptr[5] = (DCTELEM) ((tmp5 + z2 + z4) >> (13 - 2));
    dataptr[3] = (DCTELEM) ((tmp6 + z2 + z3) >> (13 - 2));
    dataptr[1] = (DCTELEM) ((tmp7 + z1 + z4) >> (13 - 2));
    dataptr += 8;
    }
  dataptr = data;
  for (ctr = 8 - 1; ctr >= 0; ctr--)
   {
    tmp0 = dataptr[8 * 0] + dataptr[8 * 7];
    tmp7 = dataptr[8 * 0] - dataptr[8 * 7];
    tmp1 = dataptr[8 * 1] + dataptr[8 * 6];
    tmp6 = dataptr[8 * 1] - dataptr[8 * 6];
    tmp2 = dataptr[8 * 2] + dataptr[8 * 5];
    tmp5 = dataptr[8 * 2] - dataptr[8 * 5];
    tmp3 = dataptr[8 * 3] + dataptr[8 * 4];
    tmp4 = dataptr[8 * 3] - dataptr[8 * 4];
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    dataptr[8 * 0] = (DCTELEM) ((tmp10 + tmp11) >> (2));
    dataptr[8 * 4] = (DCTELEM) ((tmp10 - tmp11) >> (2));
    z1 = ((tmp12 + tmp13) * (((int) 4433)));
    dataptr[8 * 2] = (DCTELEM) ((z1 + ((tmp13) * (((int) 6270)))) >> (13 + 2));
    dataptr[8 * 6] = (DCTELEM) ((z1 + ((tmp12) * (-((int) 15137)))) >> (13 + 2));
    z1 = tmp4 + tmp7;
    z2 = tmp5 + tmp6;
    z3 = tmp4 + tmp6;
    z4 = tmp5 + tmp7;
    z5 = ((z3 + z4) * (((int) 9633)));
    tmp4 = ((tmp4) * (((int) 2446)));
    tmp5 = ((tmp5) * (((int) 16819)));
    tmp6 = ((tmp6) * (((int) 25172)));
    tmp7 = ((tmp7) * (((int) 12299)));
    z1 = ((z1) * (-((int) 7373)));
    z2 = ((z2) * (-((int) 20995)));
    z3 = ((z3) * (-((int) 16069)));
    z4 = ((z4) * (-((int) 3196)));
    z3 += z5;
    z4 += z5;
    dataptr[8 * 7] = (DCTELEM) ((tmp4 + z1 + z3) >> (13 + 2));
    dataptr[8 * 5] = (DCTELEM) ((tmp5 + z2 + z4) >> (13 + 2));
    dataptr[8 * 3] = (DCTELEM) ((tmp6 + z2 + z3) >> (13 + 2));
    dataptr[8 * 1] = (DCTELEM) ((tmp7 + z1 + z4) >> (13 + 2));
    dataptr++;
    }
  }
#endif
#if 0x720  || __RCC__
#pragma once
#endif
#ifndef __SETJMP_H
#define __SETJMP_H 1
#if __cplusplus
#endif
#ifndef _CRTAPI1
#endif
#ifndef _CRTAPI2
#endif
#ifndef _CRTIMP
#endif
#if defined(_WIN32)
#define _JBLEN 16
#elif __INTSIZE == 4
#endif
//typedef int jmp_buf[16];
#if __OS2__ && 4  == 4
#else
#define __CLIB __cdecl
#endif
#if defined(_WINDOWS) && 4  == 2
#else
#if M_UNIX || M_XENIX
#endif
#endif
#if __cplusplus
#endif
#endif

#define JPEG_CJPEG_DJPEG
#define JPEG_INTERNAL_OPTIONS

typedef struct cjpeg_source_struct *cjpeg_source_ptr;
struct cjpeg_source_struct
 {
  void (*start_input) (j_compress_ptr cinfo, cjpeg_source_ptr sinfo);
    JDIMENSION(*get_pixel_rows) (j_compress_ptr cinfo, cjpeg_source_ptr sinfo);
  void (*finish_input) (j_compress_ptr cinfo, cjpeg_source_ptr sinfo);
  FILE *input_file;
  JSAMPARRAY buffer;
  JDIMENSION buffer_height;
  };
typedef struct djpeg_dest_struct *djpeg_dest_ptr;
struct djpeg_dest_struct
 {
  void (*start_output) (j_decompress_ptr cinfo, djpeg_dest_ptr dinfo);
  void (*put_pixel_rows) (j_decompress_ptr cinfo, djpeg_dest_ptr dinfo, JDIMENSION rows_supplied);
  void (*finish_output) (j_decompress_ptr cinfo, djpeg_dest_ptr dinfo);
  FILE *output_file;
  JSAMPARRAY buffer;
  JDIMENSION buffer_height;
  };
struct cdjpeg_progress_mgr
 {
  struct jpeg_progress_mgr pub;
  int completed_extra_passes;
  int total_extra_passes;
  int percent_done;
  };
typedef struct cdjpeg_progress_mgr *cd_progress_ptr;
#ifdef NEED_SHORT_EXTERNAL_NAMES
#endif
extern cjpeg_source_ptr jinit_read_bmp(j_compress_ptr cinfo);
extern djpeg_dest_ptr jinit_write_bmp(j_decompress_ptr cinfo, boolean is_os2);
extern cjpeg_source_ptr jinit_read_gif(j_compress_ptr cinfo);
extern djpeg_dest_ptr jinit_write_gif(j_decompress_ptr cinfo);
extern cjpeg_source_ptr jinit_read_ppm(j_compress_ptr cinfo);
extern djpeg_dest_ptr jinit_write_ppm(j_decompress_ptr cinfo);
extern cjpeg_source_ptr jinit_read_rle(j_compress_ptr cinfo);
extern djpeg_dest_ptr jinit_write_rle(j_decompress_ptr cinfo);
extern cjpeg_source_ptr jinit_read_targa(j_compress_ptr cinfo);
extern djpeg_dest_ptr jinit_write_targa(j_decompress_ptr cinfo);
extern boolean read_quant_tables(j_compress_ptr cinfo, char *filename, int scale_factor, boolean force_baseline);
extern boolean read_scan_script(j_compress_ptr cinfo, char *filename);
extern boolean set_quant_slots(j_compress_ptr cinfo, char *arg);
extern boolean set_sample_factors(j_compress_ptr cinfo, char *arg);
extern void read_color_map(j_decompress_ptr cinfo, FILE * infile);
extern void enable_signal_catcher(j_common_ptr cinfo);
extern void start_progress_monitor(j_common_ptr cinfo, cd_progress_ptr progress);
extern void end_progress_monitor(j_common_ptr cinfo);
extern boolean keymatch(char *arg, const char *keyword, int minchars);
extern FILE *read_stdin(void);
extern FILE *write_stdout(void);
#ifdef DONT_USE_B_MODE
#else
#ifdef VMS
#else
#define READ_BINARY "rb"
#define WRITE_BINARY "wb"
#endif
#endif
#ifndef EXIT_FAILURE
#endif
#ifndef EXIT_SUCCESS
#endif
#ifndef EXIT_WARNING
#ifdef VMS
#else
#define EXIT_WARNING  2
#endif
#endif
int WriteBlackAndWhiteJPEG(BITMAPINFOHEADER * bM, char *fname)
 {
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  FILE *outfile;
  JSAMPROW row_pointer[1];
  int row_stride;
  BYTE *image;
  DWORD *palette;
  if (bM->biSize != sizeof(BITMAPINFOHEADER))
    return (38);
  if (bM->biPlanes != 1)
    return (40);
  if (bM->biCompression != 0)
    return (42);
  if (bM->biBitCount != 8)
    return (44);
  if (bM->biWidth < 8 || bM->biWidth > 16384)
    return (46);
  if (bM->biHeight < 8 || bM->biHeight > 16384)
    return (48);
  palette = (DWORD *) (((BYTE *) bM) + sizeof(BITMAPINFOHEADER));
  image = (BYTE *) (palette + bM->biClrUsed);
  if ((outfile = fopen(fname, "wb")) == ((void *) 0))
    return (53);
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_CreateCompress((&cinfo), 62, (size_t) sizeof(struct jpeg_compress_struct));
  jpeg_stdio_dest(&cinfo, outfile);
  cinfo.image_width = bM->biWidth;
  cinfo.image_height = bM->biHeight;
  cinfo.input_components = 1;
  cinfo.in_color_space = JCS_GRAYSCALE;
  jpeg_set_defaults(&cinfo);
  jpeg_start_compress(&cinfo, 1);
  row_stride = BMPwidth(bM->biWidth, 8);
  while (cinfo.next_scanline < cinfo.image_height)
   {
    row_pointer[0] = image + cinfo.next_scanline * row_stride;
    jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }
  jpeg_finish_compress(&cinfo);
  fclose(outfile);
  jpeg_destroy_compress(&cinfo);
  return (0);
  }
int WriteColorJPEG(BITMAPINFOHEADER * bM, char *fname)
 {
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  FILE *outfile;
  JSAMPROW row_pointer[1];
  int row_stride;
  int i,
    j,
    k;
  BYTE b,
    g,
   *image,
    r,
   *scanline;
  if (bM->biSize != sizeof(BITMAPINFOHEADER))
    return (157);
  if (bM->biPlanes != 1)
    return (159);
  if (bM->biCompression != 0)
    return (161);
  if (bM->biBitCount != 24)
    return (163);
  if (bM->biWidth < 8 || bM->biWidth > 16384)
    return (165);
  if (bM->biHeight < 8 || bM->biHeight > 16384)
    return (167);
  image = ((BYTE *) bM) + sizeof(BITMAPINFOHEADER);
  if ((outfile = fopen(fname, "wb")) == ((void *) 0))
    return (171);
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_CreateCompress((&cinfo), 62, (size_t) sizeof(struct jpeg_compress_struct));
  jpeg_stdio_dest(&cinfo, outfile);
  cinfo.image_width = bM->biWidth;
  cinfo.image_height = bM->biHeight;
  cinfo.input_components = 3;
  cinfo.in_color_space = JCS_RGB;
  jpeg_set_defaults(&cinfo);
  jpeg_start_compress(&cinfo, 1);
  row_stride = BMPwidth(bM->biWidth, 24);
  while (cinfo.next_scanline < cinfo.image_height)
   {
    scanline = image + cinfo.next_scanline * row_stride;
    for (i = j = k = 0; i < bM->biWidth; i++)
     {
      b = scanline[j++];
      g = scanline[j++];
      r = scanline[j++];
      scanline[k++] = r;
      scanline[k++] = g;
      scanline[k++] = b;
      }
    row_pointer[0] = scanline;
    jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }
  jpeg_finish_compress(&cinfo);
  fclose(outfile);
  jpeg_destroy_compress(&cinfo);
  return (0);
  }
int WriteJPEG1(BITMAPINFOHEADER * bM1, char *fname)
 {
  int h,
    i,
    j,
    w,
    wide1,
    wide8;
  BITMAPINFOHEADER *bM8;
  BYTE *image1,
   *image8,
    mask;
  DWORD *palette1,
   *palette8;
  if (bM1->biSize != sizeof(BITMAPINFOHEADER))
    return (267);
  if (bM1->biPlanes != 1)
    return (269);
  if (bM1->biCompression != 0)
    return (271);
  if (bM1->biBitCount != 1)
    return (273);
  if (bM1->biWidth < 8 || bM1->biWidth > 16384)
    return (275);
  if (bM1->biHeight < 8 || bM1->biHeight > 16384)
    return (277);
  palette1 = (DWORD *) (((BYTE *) bM1) + sizeof(BITMAPINFOHEADER));
  image1 = (BYTE *) (palette1 + bM1->biClrUsed);
  wide1 = BMPwidth(bM1->biWidth, 1);
  wide8 = BMPwidth(bM1->biWidth, 8);
  bM8 = calloc(sizeof(BITMAPINFOHEADER) + 256 * sizeof(DWORD) + wide8 * bM1->biHeight, 1);
  bM8->biSize = sizeof(BITMAPINFOHEADER);
  bM8->biPlanes = 1;
  bM8->biBitCount = 8;
  bM8->biWidth = bM1->biWidth;
  bM8->biHeight = bM1->biHeight;
  bM8->biClrUsed = 256;
  bM8->biSizeImage = wide8 * bM1->biHeight;
  palette8 = (DWORD *) (((BYTE *) bM8) + sizeof(BITMAPINFOHEADER));
  for (i = 0; i < 256; i++)
    palette8[i] = (((i << 8) | i) << 8) | i;
  image8 = (BYTE *) (palette8 + bM8->biClrUsed);
  for (h = 0; h < bM1->biHeight; h++)
   {
    mask = 0x80;
    i = wide1 * h;
    j = wide8 * h;
    for (w = 0; w < bM1->biWidth; w++)
     {
      if (mask & image1[i])
       {
        if (palette1[1])
          image8[j++] = 255;
        else
          image8[j++] = 0;
        }
      else
       {
        if (palette1[0])
          image8[j++] = 255;
        else
          image8[j++] = 0;
        }
      mask >>= 1;
      if (mask == 0)
       {
        mask = 0x80;
        i++;
        }
      }
    }
  i = WriteBlackAndWhiteJPEG(bM8, fname);
  free(bM8);
  return (i);
  }
int WriteJPEG4(BITMAPINFOHEADER * bM4, char *fname)
 {
  int color,
    h,
    i,
    j,
    w,
    wide4,
    wide8,
    wide24;
  BITMAPINFOHEADER *bM8,
   *bM24;
  BYTE b,
    g,
    r;
  BYTE *image4,
   *image8,
   *image24;
  DWORD *palette4,
   *palette8;
  if (bM4->biSize != sizeof(BITMAPINFOHEADER))
    return (336);
  if (bM4->biPlanes != 1)
    return (338);
  if (bM4->biCompression != 0)
    return (340);
  if (bM4->biBitCount != 4)
    return (342);
  if (bM4->biWidth < 8 || bM4->biWidth > 16384)
    return (344);
  if (bM4->biHeight < 8 || bM4->biHeight > 16384)
    return (346);
  palette4 = (DWORD *) (((BYTE *) bM4) + sizeof(BITMAPINFOHEADER));
  image4 = (BYTE *) (palette4 + bM4->biClrUsed);
  wide4 = BMPwidth(bM4->biWidth, 4);
  for (color = i = 0; i < 16; i++)
   {
    r = RedRGB(palette4[i]);
    g = GreenRGB(palette4[i]);
    b = BlueRGB(palette4[i]);
    if (r != g || r != b)
      color = 1;
    }
  if (color)
   {
    wide24 = BMPwidth(bM4->biWidth, 24);
    bM24 = calloc(sizeof(BITMAPINFOHEADER) + wide24 * bM4->biHeight, 1);
    bM24->biSize = sizeof(BITMAPINFOHEADER);
    bM24->biPlanes = 1;
    bM24->biBitCount = 24;
    bM24->biWidth = bM4->biWidth;
    bM24->biHeight = bM4->biHeight;
    bM24->biSizeImage = wide24 * bM4->biHeight;
    image24 = ((BYTE *) bM24) + sizeof(BITMAPINFOHEADER);
    for (h = 0; h < bM4->biHeight; h++)
     {
      i = wide4 * h;
      j = wide24 * h;
      for (w = 0; w < bM4->biWidth; w++)
       {
        if (w & 1)
         {
          b = image4[i] & 0xF;
          i++;
          }
        else
          b = image4[i] >> 4;
        image24[j++] = BlueRGB(palette4[b]);
        image24[j++] = GreenRGB(palette4[b]);
        image24[j++] = RedRGB(palette4[b]);
        }
      }
    i = WriteColorJPEG(bM24, fname);
    free(bM24);
    return (i);
    }
  wide8 = BMPwidth(bM4->biWidth, 8);
  bM8 = calloc(sizeof(BITMAPINFOHEADER) + 256 * sizeof(DWORD) + wide8 * bM4->biHeight, 1);
  bM8->biSize = sizeof(BITMAPINFOHEADER);
  bM8->biPlanes = 1;
  bM8->biBitCount = 8;
  bM8->biWidth = bM4->biWidth;
  bM8->biHeight = bM4->biHeight;
  bM8->biClrUsed = 256;
  bM8->biSizeImage = wide8 * bM4->biHeight;
  palette8 = (DWORD *) (((BYTE *) bM8) + sizeof(BITMAPINFOHEADER));
  for (i = 0; i < 256; i++)
    palette8[i] = (((i << 8) | i) << 8) | i;
  image8 = (BYTE *) (palette8 + bM8->biClrUsed);
  for (h = 0; h < bM4->biHeight; h++)
   {
    i = wide4 * h;
    j = wide8 * h;
    for (w = 0; w < bM4->biWidth; w++)
     {
      if (w & 1)
       {
        b = image4[i] & 0xF;
        i++;
        }
      else
        b = image4[i] >> 4;
      image8[j++] = (BYTE) (palette4[b] & 0xFF);
      }
    }
  i = WriteBlackAndWhiteJPEG(bM8, fname);
  free(bM8);
  return (i);
  }
int WriteJPEG8(BITMAPINFOHEADER * bM8, char *fname)
 {
  int color,
    h,
    i,
    j,
    w,
    wide8,
    wide24;
  BITMAPINFOHEADER *bM24;
  BYTE b,
    g,
    r;
  BYTE *image8,
   *image24;
  DWORD *palette8;
  if (bM8->biSize != sizeof(BITMAPINFOHEADER))
    return (433);
  if (bM8->biPlanes != 1)
    return (435);
  if (bM8->biCompression != 0)
    return (437);
  if (bM8->biBitCount != 8)
    return (439);
  if (bM8->biWidth < 8 || bM8->biWidth > 16384)
    return (441);
  if (bM8->biHeight < 8 || bM8->biHeight > 16384)
    return (443);
  palette8 = (DWORD *) (((BYTE *) bM8) + sizeof(BITMAPINFOHEADER));
  image8 = (BYTE *) (palette8 + bM8->biClrUsed);
  for (color = i = 0; i < 256; i++)
   {
    r = RedRGB(palette8[i]);
    g = GreenRGB(palette8[i]);
    b = BlueRGB(palette8[i]);
    if (r != g || r != b)
      color = 1;
    }
  if (!color)
    return (WriteBlackAndWhiteJPEG(bM8, fname));
  wide8 = BMPwidth(bM8->biWidth, 8);
  wide24 = BMPwidth(bM8->biWidth, 24);
  bM24 = calloc(sizeof(BITMAPINFOHEADER) + wide24 * bM8->biHeight, 1);
  bM24->biSize = sizeof(BITMAPINFOHEADER);
  bM24->biPlanes = 1;
  bM24->biBitCount = 24;
  bM24->biWidth = bM8->biWidth;
  bM24->biHeight = bM8->biHeight;
  bM24->biSizeImage = wide24 * bM8->biHeight;
  image24 = ((BYTE *) bM24) + sizeof(BITMAPINFOHEADER);
  for (h = 0; h < bM8->biHeight; h++)
   {
    i = wide8 * h;
    j = wide24 * h;
    for (w = 0; w < bM8->biWidth; w++)
     {
      b = image8[i++];
      image24[j++] = BlueRGB(palette8[b]);
      image24[j++] = GreenRGB(palette8[b]);
      image24[j++] = RedRGB(palette8[b]);
      }
    }
  i = WriteColorJPEG(bM24, fname);
  free(bM24);
  return (i);
  }
int WriteJPEG(BITMAPINFOHEADER * bM, char *fname)
 {
  if (bM->biWidth < 8 || bM->biWidth > 16384)
    return (487);
  if (bM->biHeight < 8 || bM->biHeight > 16384)
    return (489);
  if (bM->biPlanes != 1)
    return (491);
  if (bM->biCompression != 0)
    return (493);
  if (bM->biBitCount == 1)
    return (WriteJPEG1(bM, fname));
  if (bM->biBitCount == 4)
    return (WriteJPEG4(bM, fname));
  if (bM->biBitCount == 8)
    return (WriteJPEG8(bM, fname));
  if (bM->biBitCount == 24)
    return (WriteColorJPEG(bM, fname));
  return (502);
  }

int BMPwidth(int width, int bits_per_pixel);

struct my_error_mgr
 {
  struct jpeg_error_mgr pub;
  jmp_buf setjmp_buffer;
  };

typedef struct my_error_mgr *my_error_ptr;

static void my_error_exit(j_common_ptr cinfo)
 {
  my_error_ptr myerr = (my_error_ptr) cinfo->err;
  longjmp(myerr->setjmp_buffer, 1);
  }

BITMAPINFOHEADER *ReadJPEG(char *fname)
 {
  BYTE b,
   *buf,
    g,
   *img,
   *image,
    r;
  int high,
    i,
    size,
    wide;
  FILE *input_file;
  BITMAPINFOHEADER *bM;
  struct jpeg_decompress_struct cinfo;
  struct my_error_mgr jerr;
  JSAMPARRAY buffer;
  if ((input_file = fopen(fname, "rb")) == ((void *) 0))
    return (((void *) 0));
  bM = ((void *) 0);
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  if (_setjmp(jerr.setjmp_buffer))
   {
clean_up_and_exit:
    jpeg_destroy_decompress(&cinfo);
    if (input_file)
      fclose(input_file);
    if (bM)
      free(bM);
    return (((void *) 0));
    }
  jpeg_CreateDecompress((&cinfo), 62, (size_t) sizeof(struct jpeg_decompress_struct));
  jpeg_stdio_src(&cinfo, input_file);
  jpeg_read_header(&cinfo, 1);
  if (cinfo.image_width < 8 || cinfo.image_width > 16384 ||
   cinfo.image_height < 8 || cinfo.image_height > 16384)
    goto clean_up_and_exit;
  if (cinfo.num_components == 1)
   {
    DWORD *palette;
    wide = BMPwidth(cinfo.image_width, 8);
    high = cinfo.image_height;
    size = sizeof(BITMAPINFOHEADER) + 256 * sizeof(DWORD) + wide * high;
    if ((bM = calloc(size, 1)) == ((void *) 0))
      goto clean_up_and_exit;
    palette = (DWORD *) (((BYTE *) bM) + sizeof(BITMAPINFOHEADER));
    image = (BYTE *) (((BYTE *) palette) + 256 * sizeof(DWORD));
    bM->biSize = sizeof(BITMAPINFOHEADER);
    bM->biWidth = cinfo.image_width;
    bM->biHeight = cinfo.image_height;
    bM->biPlanes = 1;
    bM->biBitCount = 8;
    bM->biSizeImage = wide * high;
    bM->biClrUsed = 256;
    bM->biClrImportant = 256;
    for (i = 0; i < 256; i++)
      palette[i] = (((i << 8) | i) << 8) | i;
    }
  else if (cinfo.num_components == 3)
   {
    wide = BMPwidth(cinfo.image_width, 24);
    high = cinfo.image_height;
    size = sizeof(BITMAPINFOHEADER) + wide * high;
    if ((bM = calloc(size, 1)) == ((void *) 0))
      goto clean_up_and_exit;
    image = ((BYTE *) bM) + sizeof(BITMAPINFOHEADER);
    bM->biSize = sizeof(BITMAPINFOHEADER);
    bM->biWidth = cinfo.image_width;
    bM->biHeight = cinfo.image_height;
    bM->biPlanes = 1;
    bM->biBitCount = 24;
    bM->biSizeImage = wide * high;
    }
  else
    goto clean_up_and_exit;
  jpeg_start_decompress(&cinfo);
  buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) & cinfo, 1, wide, 1);
  while (cinfo.output_scanline < cinfo.image_height)
   {
    jpeg_read_scanlines(&cinfo, buffer, 1);
    buf = (BYTE *) buffer[0];
    img = image + wide * (cinfo.image_height - cinfo.output_scanline);
    if (cinfo.num_components == 3)
     {
      for (i = 0; i < (int)cinfo.image_width; i++)
       {
        r = *buf++;
        g = *buf++;
        b = *buf++;
        *img++ = b;
        *img++ = g;
        *img++ = r;
        }
      }
    else
      memcpy(img, buf, cinfo.image_width);
    }
  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  fclose(input_file);
  return (bM);
  }

BITMAPINFOHEADER*LoadJPEG(void*data,size_t length)
  {
  BITMAPINFOHEADER*bm;
  FILE*fp;
  fp=fopen("bobo.jpg","wb");
  fwrite(data,1,length,fp);
  fclose(fp);
  bm=ReadJPEG("bobo.jpg");
  remove("bobo.jpg");
  return(bm);
  }
